 /*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdarg.h> /* for variable argument */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "via_driver.h"
#include "via_dri.h"
#include "via_rotate.h" /* for rotate feature.*/
#include "via_Hw3DScaling.h"
#include "via_eng_regs.h"

void hw3d_CursorPositionUpdate(VIABIOSInfoPtr pBIOSInfo)
{
    int x = 0, y = 0;
    ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr      pVia = VIAPTR(pScrn);
    
    if (pBIOSInfo->IGA1SettingInfo.IsDISP3DScaling)
    {
        x = pVia->pBIOSInfo->IGA1SettingInfo.DISP3DSCALCTRLInfo.CursorX;
        y = pVia->pBIOSInfo->IGA1SettingInfo.DISP3DSCALCTRLInfo.CursorY;    
    }
    if(pBIOSInfo->IGA2SettingInfo.IsDISP3DScaling)
    {
        x = pVia->pBIOSInfo->IGA2SettingInfo.DISP3DSCALCTRLInfo.CursorX;
        y = pVia->pBIOSInfo->IGA2SettingInfo.DISP3DSCALCTRLInfo.CursorY;    
    }
    VIASetCursorPosition(pScrn, x, y);
}

/*3D DISPLAY SCaling supportting*/
void Reserve3DScalingMemory(ScrnInfoPtr pScrn)
{
    int needBufferSet;
    VIAPtr pVia = VIAPTR(pScrn);

    VIAPtr    pVia1 = NULL;
    unsigned int size = 0;
    unsigned int addr = 0;


    if(pVia->pVIAEnt->pPrimaryScrn != NULL)
    {
        pVia1 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);
    }
    
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Reserve3DScalingMemory begin\n"));

    /*now add the double buffle mechiem to get good qualtity display when enable 3D display function*/
    if(!pVia->IsSecondary)
    {
        needBufferSet =  pVia->NumOfDevNeed3dScl;
        size = DISP_3D_SCALING_BUFFER_SIZE*DISP_3D_SCALING_BUFNUM_ONESET*(needBufferSet);
        addr = viaFBAlloctor(pScrn, size, INVERT_SEQ, "3D Scaling Buf");
        if(addr) {
            pVia->DISP3DScalingBufferStart = addr;
        }
        else {
            return ;
        }
        if ( 1 == needBufferSet)
        {
            pVia->iga1DISP3DScalingBufferStart = pVia->iga2DISP3DScalingBufferStart = pVia->DISP3DScalingBufferStart;
        }

        if (needBufferSet > 1)
        {
            pVia->iga1DISP3DScalingBufferStart = pVia->DISP3DScalingBufferStart;
            pVia->iga2DISP3DScalingBufferStart = pVia->DISP3DScalingBufferStart + DISP_3D_SCALING_BUFFER_SIZE*DISP_3D_SCALING_BUFNUM_ONESET;
        }
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Reserve3DScalingMemory , need buffer set =%d\n", needBufferSet));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Reserve3DScalingMemory pVia->iga1DISP3DScalingBufferStart=%x, pVia->iga2DISP3DScalingBufferStart=%x\n",pVia->iga1DISP3DScalingBufferStart,pVia->iga2DISP3DScalingBufferStart));
        size = MAX_CURS * MAX_CURS * 4;
        addr = viaFBAlloctor(pScrn, size, INVERT_SEQ, "3D Scaling Cursor Buf");
        if(addr) {
            pVia->Cursor3DScalingBufferStart = addr;
        } 
        else {
            return ;
        }
    }
    /*  Needn't to reserve 3D buffer in second scrn because the first scrn has reserved all necessary buffer.
        Here Just need to save the 3D buffer address in second scrn.
    */
    else
    {
        pVia->iga1DISP3DScalingBufferStart = pVia1->iga1DISP3DScalingBufferStart;
        pVia->iga2DISP3DScalingBufferStart = pVia1->iga2DISP3DScalingBufferStart;
        pVia->DISP3DScalingBufferStart = pVia1->DISP3DScalingBufferStart;
        pVia->Cursor3DScalingBufferStart = pVia1->Cursor3DScalingBufferStart;
        
    }
    
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Reserve3DScalingMemory exit\n"));
}


void VIABlankRectangle( VIAPtr pVia, CARD32 RectAddr, CARD32 RectDimension, CARD32 RectPitch)
{
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];    
    via2DBltSrcRec src;
    via2DBltDstRec dst;
    via2DBltConfigRec config;

    memset(&src, 0x00, sizeof(via2DBltSrcRec));
    memset(&dst, 0x00, sizeof(via2DBltDstRec));
    memset(&config, 0x00, sizeof(via2DBltConfigRec));

    src.bpp = pBIOSInfo->bitsPerPixel;;
    src.color = 0x00;
    
    dst.bpp = pBIOSInfo->bitsPerPixel;
    dst.offset = RectAddr;
    dst.pitch = RectPitch;
    dst.x = 0;
    dst.y = 0;
    dst.w = RectDimension & 0xFFFF;
    dst.h = (RectDimension>>16) & 0xFFFF;

    config.solidfill = TRUE;

    viaAccel2DBlt(pScrn, &src, &dst, &config);
}  /* VIABlankRectangle()*/

Bool BlackSclScrnBoundary(VIAPtr pVia, VIADISP3DSCALCTRLPtr p3DSCALCTRLInfo, RECTL rDest, int index)
{
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    CARD32 RectAddr = p3DSCALCTRLInfo->DISP3DScalingBufAddr[index];
    CARD32 ulTopLeftOffset;
    long dwWidth, dwHeigght;
    CARD32 Demension;
    int BytePP;
    long dwShrinkMemWidth, dwShrinkMemHeight;
    CARD32 MemPitch;

    BytePP = pBIOSInfo->bitsPerPixel >> 3;
    dwShrinkMemWidth = p3DSCALCTRLInfo->RealHActive;
    dwShrinkMemHeight = p3DSCALCTRLInfo->RealVActive;

    MemPitch = dwShrinkMemWidth * BytePP;
    MemPitch = (MemPitch+31)&~31;

    /*black left side rectangle  */
    /*we clear a little bit more to avoid align problem*/
    ulTopLeftOffset = 0; 
    dwWidth = rDest.left; 
    dwHeigght = dwShrinkMemHeight;
    Demension = (dwHeigght<<16) | dwWidth;
    if( (dwWidth > 0) && (dwHeigght > 0) )
        VIABlankRectangle( pVia, RectAddr, Demension, MemPitch);
        
    /*black right side rectangle*/
    ulTopLeftOffset = (rDest.right ) * BytePP ;
    dwWidth = dwShrinkMemWidth - rDest.right; 
    dwHeigght = dwShrinkMemHeight;
    Demension = (dwHeigght<<16) | dwWidth;
    if( (dwWidth > 0) && (dwHeigght > 0) )
            VIABlankRectangle( pVia, RectAddr+ulTopLeftOffset, Demension, MemPitch);

    /*black up rectangle*/
    ulTopLeftOffset = 0;
    dwWidth = dwShrinkMemWidth;
    dwHeigght = rDest.top;
    Demension = (dwHeigght<<16) | dwWidth;    
    if( (dwWidth > 0) && (dwHeigght > 0) )
        VIABlankRectangle( pVia, RectAddr+ulTopLeftOffset, Demension, MemPitch);

    /*draw bottom rectangle*/
    ulTopLeftOffset =  (rDest.bottom) * MemPitch;
    dwWidth = dwShrinkMemWidth;
    dwHeigght = dwShrinkMemHeight - rDest.bottom;
    Demension = (dwHeigght<<16) | dwWidth;
    if( (dwWidth > 0) && (dwHeigght > 0) )
        VIABlankRectangle( pVia, RectAddr+ulTopLeftOffset, Demension, MemPitch);

    return TRUE;
}


void patchForVidColorKey(VIAPtr pVia,CARD32 pSrc,CARD32 pDest,CARD32 ColoyKey,CARD32 pitch,
                            RECTL* prclDst,RECTL* prclClip)
{
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];    
    via2DBltSrcRec src;
    via2DBltDstRec dst;
    via2DBltConfigRec config;

    memset(&src, 0x00, sizeof(via2DBltSrcRec));
    memset(&dst, 0x00, sizeof(via2DBltDstRec));
    memset(&config, 0x00, sizeof(via2DBltConfigRec));

    src.bpp = pBIOSInfo->bitsPerPixel;;
    src.offset = pSrc;
    src.pitch = pitch;
    src.x = prclClip->left;
    src.y = prclClip->top;
    
    dst.bpp = pBIOSInfo->bitsPerPixel;;
    dst.offset = pDest;
    dst.pitch = pitch;
    dst.colorKey = ColoyKey;
    dst.x = prclClip->left;
    dst.y = prclClip->top;
    dst.w = prclClip->right - prclClip->left;
    dst.h = prclClip->bottom - prclClip->top;

    config.srckey = TRUE;

    viaAccel2DBlt(pScrn, &src, &dst, &config);
}

Bool DISP3DScalWindowAdjust(int scrnIndex, RECTLPtr prSrc, RECTLPtr prDest, VIA3DSclSCNParasPtr pDISP3DScalScreenInfo)
{
    CARD32 dwShrinkMemWidth;
    CARD32 dwShrinkMemHeight;
    CARD32 dwSrcMemWidth, dwSrcMemHeight;
    CARD32 dwDesMemWidth, dwDesMemHeight;

    dwShrinkMemWidth = pDISP3DScalScreenInfo->XSIZEMax;
    dwShrinkMemHeight = pDISP3DScalScreenInfo->YSIZEMax;

    dwSrcMemWidth = prSrc->right - prSrc->left;
    dwSrcMemHeight = prSrc->bottom - prSrc->top;

    /*Size debug information*/
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo SIZE: XSIZEMax=%ld, YSIZEMax=%ld, XSIZEMaxRange=%ld, YSIZEMaxRange=%ld\n", \
        pDISP3DScalScreenInfo->XSIZEMax, pDISP3DScalScreenInfo->YSIZEMax, pDISP3DScalScreenInfo->XSIZEMaxRange,pDISP3DScalScreenInfo->YSIZEMaxRange));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo SIZE: XScaleStep=%ld, YScaleStep=%ld, XScaleMaximum=%ld, YScaleMaximum=%ld\n", \
        pDISP3DScalScreenInfo->XScaleStep, pDISP3DScalScreenInfo->YScaleStep, pDISP3DScalScreenInfo->XScaleMaximum,pDISP3DScalScreenInfo->YScaleMaximum));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo SIZE: XScaleLevel=%ld, YScaleLevel=%ld, XSIZEOffset=%ld, YSIZEOffset=%ld\n", \
        pDISP3DScalScreenInfo->XScaleLevel, pDISP3DScalScreenInfo->YScaleLevel, pDISP3DScalScreenInfo->XSIZEOffset,pDISP3DScalScreenInfo->YSIZEOffset));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo SIZE: XSIZENow=%ld, YSIZENow=%ld\n", \
        pDISP3DScalScreenInfo->XSIZENow, pDISP3DScalScreenInfo->YSIZENow));
        
    /*Position debug information*/
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONMaxRange=%ld, YPOSITIONMaxRange=%ld\n", \
    pDISP3DScalScreenInfo->XPOSITIONMaxRange,pDISP3DScalScreenInfo->YPOSITIONMaxRange));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONStep=%ld, YPOSITIONStep=%ld, XPOSITIONMaximum=%ld, YPOSITIONMaximum=%ld\n", \
        pDISP3DScalScreenInfo->XPOSITIONStep, pDISP3DScalScreenInfo->YPOSITIONStep, pDISP3DScalScreenInfo->XPOSITIONMaximum,pDISP3DScalScreenInfo->YPOSITIONMaximum));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONLevel=%ld, YPOSITIONLevel=%ld, XPOSITIONOffset=%ld, YPOSITIONOffset=%ld\n", \
        pDISP3DScalScreenInfo->XPOSITIONLevel, pDISP3DScalScreenInfo->YPOSITIONLevel, pDISP3DScalScreenInfo->XPOSITIONOffset,pDISP3DScalScreenInfo->YPOSITIONOffset));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONNow=%ld, YPOSITIONNow=%ld\n", \
        pDISP3DScalScreenInfo->XPOSITIONNow, pDISP3DScalScreenInfo->YPOSITIONNow));
    
    /*software shrink adjust windows size*/
    prDest->left += pDISP3DScalScreenInfo->XSIZEOffset;
    prDest->top += pDISP3DScalScreenInfo->YSIZEOffset;
    prDest->right -= pDISP3DScalScreenInfo->XSIZEOffset; 
    prDest->bottom -= pDISP3DScalScreenInfo->YSIZEOffset;
  
    prDest->left += pDISP3DScalScreenInfo->XPOSITIONOffset;
    prDest->right += pDISP3DScalScreenInfo->XPOSITIONOffset;
    prDest->top -= pDISP3DScalScreenInfo->YPOSITIONOffset;
    prDest->bottom -= pDISP3DScalScreenInfo->YPOSITIONOffset;

    dwDesMemWidth = prDest->right - prDest->left;
    dwDesMemHeight = prDest->bottom - prDest->top;
  
    if(prDest->left < 0)
    {   
        prSrc->left += -prDest->left * dwSrcMemWidth /dwDesMemWidth;
        prDest->left = 0;
    }
  
    if(prDest->top < 0)
    {
        prSrc->top += -prDest->top * dwSrcMemHeight / dwDesMemHeight;
        prDest->top = 0;
    }
  
    if(prDest->right > dwShrinkMemWidth)
    {
        prSrc->right -= (prDest->right -dwShrinkMemWidth) * dwSrcMemWidth /dwDesMemWidth;
        prDest->right = dwShrinkMemWidth;
    }
  
    if(prDest->bottom > dwShrinkMemHeight)
    {
        prSrc->bottom -= (prDest->bottom - dwShrinkMemHeight) * dwSrcMemHeight / dwDesMemHeight;
        prDest->bottom = dwShrinkMemHeight;
    }
    return TRUE;
}

Bool VIASetTv3DScalScreenParas(VIABIOSInfoPtr  pBIOSInfo)
{
       VIA3DSclSCNParasPtr pTV3DScalInfo = &(pBIOSInfo->TVSettingInfo.TV3DScalInfo);
       /*Get the  physical size*/
       pTV3DScalInfo->XSIZEMax = pBIOSInfo->TVSettingInfo.HActive;
       pTV3DScalInfo->YSIZEMax = pBIOSInfo->TVSettingInfo.VActive;
       
       /*make adjust SIZE range 50% of physical mode, 25% up and down*/
       pTV3DScalInfo->XSIZEMaxRange= (pBIOSInfo->TVSettingInfo.HActive >> 1);
       pTV3DScalInfo->YSIZEMaxRange= (pBIOSInfo->TVSettingInfo.VActive >> 1);    


       pTV3DScalInfo->XScaleStep = 3;
       pTV3DScalInfo->YScaleStep = 3;

       pTV3DScalInfo->XScaleMaximum = (pTV3DScalInfo->XSIZEMaxRange / pTV3DScalInfo->XScaleStep);
       pTV3DScalInfo->YScaleMaximum = (pTV3DScalInfo->YSIZEMaxRange / pTV3DScalInfo->YScaleStep);
       
       pTV3DScalInfo->XScaleLevel = pTV3DScalInfo->XScaleDefault = pTV3DScalInfo->XScaleMaximum >>1;
       pTV3DScalInfo->YScaleLevel = pTV3DScalInfo->YScaleDefault = pTV3DScalInfo->YScaleMaximum >>1;
       
       pTV3DScalInfo->XSIZEOffset = (pTV3DScalInfo->XScaleDefault - pTV3DScalInfo->XScaleLevel)*pTV3DScalInfo->XScaleStep;
       pTV3DScalInfo->YSIZEOffset = (pTV3DScalInfo->YScaleDefault - pTV3DScalInfo->YScaleLevel)*pTV3DScalInfo->YScaleStep;

       pTV3DScalInfo->XSIZENow = pTV3DScalInfo->XSIZEMax - (pTV3DScalInfo->XScaleLevel*pTV3DScalInfo->XScaleStep);
       pTV3DScalInfo->YSIZENow = pTV3DScalInfo->YSIZEMax - (pTV3DScalInfo->YScaleLevel*pTV3DScalInfo->YScaleStep);

       /*make adjust POSITION range 50% of physical mode, 25% up and down       */
       pTV3DScalInfo->XPOSITIONMaxRange = (pBIOSInfo->TVSettingInfo.HActive >> 1);
       pTV3DScalInfo->YPOSITIONMaxRange = (pBIOSInfo->TVSettingInfo.VActive >> 1);
       
       pTV3DScalInfo->XPOSITIONStep = 3;
       pTV3DScalInfo->YPOSITIONStep = 3;

       pTV3DScalInfo->XPOSITIONMaximum = (pTV3DScalInfo->XPOSITIONMaxRange / pTV3DScalInfo->XPOSITIONStep);
       pTV3DScalInfo->YPOSITIONMaximum = (pTV3DScalInfo->YPOSITIONMaxRange / pTV3DScalInfo->YPOSITIONStep);
       
       pTV3DScalInfo->XPOSITIONLevel = pTV3DScalInfo->XPOSITIONDefault = pTV3DScalInfo->XPOSITIONMaximum >>1;
       pTV3DScalInfo->YPOSITIONLevel = pTV3DScalInfo->YPOSITIONDefault = pTV3DScalInfo->YPOSITIONMaximum >>1;

       pTV3DScalInfo->XPOSITIONOffset= (pTV3DScalInfo->XPOSITIONLevel - pTV3DScalInfo->XPOSITIONDefault)*pTV3DScalInfo->XPOSITIONStep;
       pTV3DScalInfo->YPOSITIONOffset= (pTV3DScalInfo->YPOSITIONLevel - pTV3DScalInfo->YPOSITIONDefault)*pTV3DScalInfo->YPOSITIONStep;

       pTV3DScalInfo->XPOSITIONNow =( pTV3DScalInfo->XPOSITIONLevel - pTV3DScalInfo->XPOSITIONDefault)*pTV3DScalInfo->XPOSITIONStep;
       pTV3DScalInfo->YPOSITIONNow =( pTV3DScalInfo->YPOSITIONLevel - pTV3DScalInfo->YPOSITIONDefault)*pTV3DScalInfo->YPOSITIONStep;

       pTV3DScalInfo->bNeedBlack3dScalingBorders = FALSE;

       if((pBIOSInfo->TVSettingInfo.IsPreSetting)&&(pBIOSInfo->TVSettingInfo.TVInfo.HasUserSetting))
       {
            pTV3DScalInfo->XScaleLevel = pBIOSInfo->TVSettingInfo.TVInfo.XScalingLevel;
            pTV3DScalInfo->YScaleLevel = pBIOSInfo->TVSettingInfo.TVInfo.YScalingLevel;
            pTV3DScalInfo->XSIZEOffset = (pTV3DScalInfo->XScaleDefault - pTV3DScalInfo->XScaleLevel)*pTV3DScalInfo->XScaleStep;
            pTV3DScalInfo->YSIZEOffset = (pTV3DScalInfo->YScaleDefault - pTV3DScalInfo->YScaleLevel)*pTV3DScalInfo->YScaleStep;
            
            pTV3DScalInfo->XPOSITIONLevel = pBIOSInfo->TVSettingInfo.TVInfo.XScalPosLevel;
            pTV3DScalInfo->YPOSITIONLevel = pBIOSInfo->TVSettingInfo.TVInfo.YScalPosLevel;
            pTV3DScalInfo->XPOSITIONOffset= (pTV3DScalInfo->XPOSITIONLevel - pTV3DScalInfo->XPOSITIONDefault)*pTV3DScalInfo->XPOSITIONStep;
            pTV3DScalInfo->YPOSITIONOffset= (pTV3DScalInfo->YPOSITIONLevel - pTV3DScalInfo->YPOSITIONDefault)*pTV3DScalInfo->YPOSITIONStep;
       }
       return TRUE;
       
}

Bool VIASetLCD3DScalScreenParas(VIABIOSInfoPtr  pBIOSInfo)
{
       VIA3DSclSCNParasPtr pLCD3DScalInfo = &(pBIOSInfo->LVDSSettingInfo.LCD3DScalInfo);
       /*Get the  physical size*/
       pLCD3DScalInfo->XSIZEMax = pBIOSInfo->LVDSSettingInfo.HActive;
       pLCD3DScalInfo->YSIZEMax = pBIOSInfo->LVDSSettingInfo.VActive;
       
       /*make adjust SIZE range 50% of physical mode, 25% up and down*/
       pLCD3DScalInfo->XSIZEMaxRange= (pBIOSInfo->LVDSSettingInfo.HActive >> 1);
       pLCD3DScalInfo->YSIZEMaxRange= (pBIOSInfo->LVDSSettingInfo.VActive >> 1);    


       pLCD3DScalInfo->XScaleStep = 3;
       pLCD3DScalInfo->YScaleStep = 3;

       pLCD3DScalInfo->XScaleMaximum = (pLCD3DScalInfo->XSIZEMaxRange / pLCD3DScalInfo->XScaleStep);
       pLCD3DScalInfo->YScaleMaximum = (pLCD3DScalInfo->YSIZEMaxRange / pLCD3DScalInfo->YScaleStep);
       
       pLCD3DScalInfo->XScaleLevel = pLCD3DScalInfo->XScaleDefault = pLCD3DScalInfo->XScaleMaximum >>1;
       pLCD3DScalInfo->YScaleLevel = pLCD3DScalInfo->YScaleDefault = pLCD3DScalInfo->YScaleMaximum >>1;
       
       pLCD3DScalInfo->XSIZEOffset = (pLCD3DScalInfo->XScaleDefault - pLCD3DScalInfo->XScaleLevel)*pLCD3DScalInfo->XScaleStep;
       pLCD3DScalInfo->YSIZEOffset = (pLCD3DScalInfo->YScaleDefault - pLCD3DScalInfo->YScaleLevel)*pLCD3DScalInfo->YScaleStep;

       pLCD3DScalInfo->XSIZENow = pLCD3DScalInfo->XSIZEMax - (pLCD3DScalInfo->XScaleLevel*pLCD3DScalInfo->XScaleStep);
       pLCD3DScalInfo->YSIZENow = pLCD3DScalInfo->YSIZEMax - (pLCD3DScalInfo->YScaleLevel*pLCD3DScalInfo->YScaleStep);

       /*make adjust POSITION range 50% of physical mode, 25% up and down       */
       pLCD3DScalInfo->XPOSITIONMaxRange = (pBIOSInfo->LVDSSettingInfo.HActive >> 1);
       pLCD3DScalInfo->YPOSITIONMaxRange = (pBIOSInfo->LVDSSettingInfo.VActive >> 1);
       
       pLCD3DScalInfo->XPOSITIONStep = 3;
       pLCD3DScalInfo->YPOSITIONStep = 3;

       pLCD3DScalInfo->XPOSITIONMaximum = (pLCD3DScalInfo->XPOSITIONMaxRange / pLCD3DScalInfo->XPOSITIONStep);
       pLCD3DScalInfo->YPOSITIONMaximum = (pLCD3DScalInfo->YPOSITIONMaxRange / pLCD3DScalInfo->YPOSITIONStep);
       
       pLCD3DScalInfo->XPOSITIONLevel = pLCD3DScalInfo->XPOSITIONDefault = pLCD3DScalInfo->XPOSITIONMaximum >>1;
       pLCD3DScalInfo->YPOSITIONLevel = pLCD3DScalInfo->YPOSITIONDefault = pLCD3DScalInfo->YPOSITIONMaximum >>1;

       pLCD3DScalInfo->XPOSITIONOffset= (pLCD3DScalInfo->XPOSITIONLevel - pLCD3DScalInfo->XPOSITIONDefault)*pLCD3DScalInfo->XPOSITIONStep;
       pLCD3DScalInfo->YPOSITIONOffset= (pLCD3DScalInfo->YPOSITIONLevel - pLCD3DScalInfo->YPOSITIONDefault)*pLCD3DScalInfo->YPOSITIONStep;

       pLCD3DScalInfo->XPOSITIONNow =( pLCD3DScalInfo->XPOSITIONLevel - pLCD3DScalInfo->XPOSITIONDefault)*pLCD3DScalInfo->XPOSITIONStep;
       pLCD3DScalInfo->YPOSITIONNow =( pLCD3DScalInfo->YPOSITIONLevel - pLCD3DScalInfo->YPOSITIONDefault)*pLCD3DScalInfo->YPOSITIONStep;

       pLCD3DScalInfo->bNeedBlack3dScalingBorders = FALSE;

       if((pBIOSInfo->LVDSSettingInfo.IsPreSetting))
       {
            pLCD3DScalInfo->XScaleLevel = pBIOSInfo->LVDSSettingInfo.XScalingLevel;
            pLCD3DScalInfo->YScaleLevel = pBIOSInfo->LVDSSettingInfo.YScalingLevel; 
            pLCD3DScalInfo->XSIZEOffset = (pLCD3DScalInfo->XScaleDefault - pLCD3DScalInfo->XScaleLevel)*pLCD3DScalInfo->XScaleStep;
            pLCD3DScalInfo->YSIZEOffset = (pLCD3DScalInfo->YScaleDefault - pLCD3DScalInfo->YScaleLevel)*pLCD3DScalInfo->YScaleStep;
            
            pLCD3DScalInfo->XPOSITIONLevel = pBIOSInfo->LVDSSettingInfo.XScalPosLevel;
            pLCD3DScalInfo->YPOSITIONLevel = pBIOSInfo->LVDSSettingInfo.YScalPosLevel;
            pLCD3DScalInfo->XPOSITIONOffset= (pLCD3DScalInfo->XPOSITIONLevel - pLCD3DScalInfo->XPOSITIONDefault)*pLCD3DScalInfo->XPOSITIONStep;
            pLCD3DScalInfo->YPOSITIONOffset= (pLCD3DScalInfo->YPOSITIONLevel - pLCD3DScalInfo->YPOSITIONDefault)*pLCD3DScalInfo->YPOSITIONStep;
       }
       return TRUE;
       
}

Bool    VIASet3DScalCtrlParas(ScrnInfoPtr pScrn, VIADISP3DSCALCTRLPtr p3DSCALCTRLInfo)
{
    int i;
    CARD32  Pitch;
    CARD32  Dimension;
    CARD32  ScalingAddress;
    VIAPtr pVia = VIAPTR(pScrn);    
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;

    /*Initialize the 3d scaling control for display */
    p3DSCALCTRLInfo->VirtualHActive = pBIOSInfo->displayWidth;
    p3DSCALCTRLInfo->BitsPerPixel = pBIOSInfo->bitsPerPixel;
 /*   p3DSCALCTRLInfo->DISP3DScalIGAPath = pBIOSInfo->TVSettingInfo.IGAPath;*/
    
    Pitch = p3DSCALCTRLInfo->RealHActive*(p3DSCALCTRLInfo->BitsPerPixel>>3);
    /* 3D engine of VT3353/3409 require that the src and dst pitch must be aligned to 32 */
    Pitch = ALIGN_TO(Pitch, 32);

    p3DSCALCTRLInfo->DISP3DScalingBufPitch= Pitch;
    if (pVia->IsSecondary)
    {
        p3DSCALCTRLInfo->OrigFBAddress = pScrn->fbOffset;
    }
    else
    {
        p3DSCALCTRLInfo->OrigFBAddress=0;
    }

    p3DSCALCTRLInfo->PanningX=0;
    p3DSCALCTRLInfo->PanningY=0;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "pVia->videoRambytes=%lud\n",pVia->videoRambytes));        

    /*Re-allocate the scaling buffers for 3d scaling*/
    ScalingAddress = pVia->DISP3DScalingBufferStart;    
    if(ScalingAddress)
    {       
        Dimension = (p3DSCALCTRLInfo->RealVActive<<16) | p3DSCALCTRLInfo->RealHActive;
        if (p3DSCALCTRLInfo->DISP3DScalIGAPath ==IGA1)
        {
            ScalingAddress = pVia->iga1DISP3DScalingBufferStart;
        }
        else
        {
            ScalingAddress = pVia->iga2DISP3DScalingBufferStart;
        }
        for(i=0; i < DISP_3D_SCALING_BUFNUM_ONESET;i++)
        {
            p3DSCALCTRLInfo->DISP3DScalingBufAddr[i]= ScalingAddress+ DISP_3D_SCALING_BUFFER_SIZE*i;
            /*256byte alignment for 3D scaling function*/
            p3DSCALCTRLInfo->DISP3DScalingBufAddr[i] = ((p3DSCALCTRLInfo->DISP3DScalingBufAddr[i]+ 255) & ~255);
//                VIABlankRectangle( pVia, p3DSCALCTRLInfo->DISP3DScalingBufAddr[i], Dimension, Pitch);
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "p3DSCALCTRLInfo->DISP3DScalingBufAddr=%x\n",p3DSCALCTRLInfo->DISP3DScalingBufAddr[i]));        
        }
    }
    return TRUE;
}

void vid3DScalColorKeyWindowAdjust(RECTL* lpSrcRectVideoWin, RECTL* lpDstRectVideoWin, HW3D_Scaling_INFO   *lpgfx3DScal_info)
{

    int OrigHActive,OrigVActive,RealHActive,RealVActive;
    int tmpLeft, tmpRight,tmpTop, tmpBottom;

    int screenX, screenY;
    OrigHActive = lpgfx3DScal_info->OrigHActive;
    OrigVActive = lpgfx3DScal_info->OrigVActive;
    RealHActive = lpgfx3DScal_info->RealHActive;
    RealVActive = lpgfx3DScal_info->RealVActive;
    
    tmpLeft = (lpSrcRectVideoWin->left)*RealHActive/OrigHActive;
    tmpRight = (lpSrcRectVideoWin->right)*RealHActive/OrigHActive;
    tmpTop = (lpSrcRectVideoWin->top)*RealVActive/OrigVActive;
    tmpBottom = (lpSrcRectVideoWin->bottom)*RealVActive/OrigVActive;
    
    screenX = RealHActive - lpgfx3DScal_info->XSIZEOffset * 2;
    screenY = RealVActive - lpgfx3DScal_info->YSIZEOffset * 2;

    tmpLeft = (screenX /2) - (RealHActive /2 - tmpLeft) * screenX /RealHActive;
    tmpRight = (screenX/2 ) - (RealHActive/2 - tmpRight) * screenX /RealHActive;
    tmpTop = (screenY/2) - (RealVActive /2 - tmpTop) * screenY /RealVActive;
    tmpBottom = (screenY/2) - (RealVActive/2  - tmpBottom) * screenY /RealVActive;
      
    /*software shrink adjust windows size*/
    tmpLeft += (lpgfx3DScal_info->XSIZEOffset + lpgfx3DScal_info->XPOSITIONOffset);
    tmpRight += (lpgfx3DScal_info->XSIZEOffset + lpgfx3DScal_info->XPOSITIONOffset);
    tmpTop +=  (lpgfx3DScal_info->YSIZEOffset - lpgfx3DScal_info->YPOSITIONOffset);
    tmpBottom +=  (lpgfx3DScal_info->YSIZEOffset - lpgfx3DScal_info->YPOSITIONOffset);

    lpDstRectVideoWin->left =(unsigned long) tmpLeft;
    lpDstRectVideoWin->right =(unsigned long) tmpRight;
    lpDstRectVideoWin->top =  (unsigned long)tmpTop;
    lpDstRectVideoWin->bottom = (unsigned long) tmpBottom;

}


void VIADISP3DScalingParasSetting(ScrnInfoPtr pScrn)
{
    CARD32 OffsetWidthByQWord, CountWidthByQWord;
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    IGASETTINGINFO     *pTargetIGASettingInfo = &(pBIOSInfo->IGA1SettingInfo);
    VIADISP3DSCALCTRLPtr p3DSCALCTRLInfo = &(pTargetIGASettingInfo->DISP3DSCALCTRLInfo);
    int panelWidth=0, panelHeight=0;

    /*Reset the setting*/
    pBIOSInfo->IGA1SettingInfo.IsDISP3DScaling = FALSE;
    pBIOSInfo->IGA2SettingInfo.IsDISP3DScaling = FALSE;
  
    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_TV)
    {
        TVSETTINGINFOPTR  pTVSettingInfo=&(pBIOSInfo->TVSettingInfo);
        switch (pBIOSInfo->TVSettingInfo.IGAPath )
        {
            case IGA2:
                pTargetIGASettingInfo = &(pBIOSInfo->IGA2SettingInfo);
                p3DSCALCTRLInfo =& (pTargetIGASettingInfo->DISP3DSCALCTRLInfo);
                break;
            case IGA1:
            default:
                break;
        }
         p3DSCALCTRLInfo->DISP3DScalIGAPath = pBIOSInfo->TVSettingInfo.IGAPath;
         pTargetIGASettingInfo->IsDISP3DScaling = TRUE;
         DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIADISP3DScalingParasSetting begin\n"));
         DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTVSettingInfo:HActive =%d, VActive=%d, IGAPath=%d,TVType =%d\n", \
            pTVSettingInfo->HActive, pTVSettingInfo->VActive, pTVSettingInfo->IGAPath,pTVSettingInfo->TVType));
         DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTVSettingInfo: ModeIndex=%d\n", pTVSettingInfo->ModeIndex));
         
        /*1. first to get the physical size, and store the current mode size*/
        /*Replace the TV setting mode to native mode*/
        if ((TVTYPE_NTSC==pBIOSInfo->TVSettingInfo.TVType)|| (TVTYPE_480P==pBIOSInfo->TVSettingInfo.TVType) )
        {
            pTVSettingInfo->ModeIndex = M720x480;
            p3DSCALCTRLInfo->OrigHActive = pTVSettingInfo->HActive;
            p3DSCALCTRLInfo->OrigVActive = pTVSettingInfo->VActive;
            p3DSCALCTRLInfo->RealHActive = 720;
            p3DSCALCTRLInfo->RealVActive = 480;            
            pTVSettingInfo->HActive = 720;
            pTVSettingInfo->VActive = 480;
        }
        
        if((TVTYPE_PAL==pBIOSInfo->TVSettingInfo.TVType)|| (TVTYPE_576P==pBIOSInfo->TVSettingInfo.TVType))
        {
            pTVSettingInfo->ModeIndex = M720x576;
            p3DSCALCTRLInfo->OrigHActive = pTVSettingInfo->HActive;
            p3DSCALCTRLInfo->OrigVActive = pTVSettingInfo->VActive;
            p3DSCALCTRLInfo->RealHActive = 720;
            p3DSCALCTRLInfo->RealVActive = 576;            
            pTVSettingInfo->HActive = 720;
            pTVSettingInfo->VActive = 576;            
        }

        if (TVTYPE_720P==pBIOSInfo->TVSettingInfo.TVType )
        {
            pTVSettingInfo->ModeIndex = M1280x720;
            p3DSCALCTRLInfo->OrigHActive = pTVSettingInfo->HActive;
            p3DSCALCTRLInfo->OrigVActive = pTVSettingInfo->VActive;
            p3DSCALCTRLInfo->RealHActive = 1280;
            p3DSCALCTRLInfo->RealVActive = 720;            
            pTVSettingInfo->HActive = 1280;
            pTVSettingInfo->VActive = 720;            

        }   

        if (TVTYPE_1080I==pBIOSInfo->TVSettingInfo.TVType )
        {
            pTVSettingInfo->ModeIndex = M1920x1080;
            p3DSCALCTRLInfo->OrigHActive = pTVSettingInfo->HActive;
            p3DSCALCTRLInfo->OrigVActive = pTVSettingInfo->VActive;
            p3DSCALCTRLInfo->RealHActive = 1920;
            p3DSCALCTRLInfo->RealVActive = 1080;            
            pTVSettingInfo->HActive = 1920;
            pTVSettingInfo->VActive = 1080;            

        }     
       
        /*2.initial the scaling screen paras */
        VIASetTv3DScalScreenParas(pBIOSInfo);
        #if 1       
         /*Size debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo SIZE: XSIZEMax=%ld, YSIZEMax=%ld, XSIZEMaxRange=%ld, YSIZEMaxRange=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XSIZEMax, pTVSettingInfo->TV3DScalInfo.YSIZEMax, pTVSettingInfo->TV3DScalInfo.XSIZEMaxRange,pTVSettingInfo->TV3DScalInfo.YSIZEMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo SIZE: XScaleStep=%ld, YScaleStep=%ld, XScaleMaximum=%ld, YScaleMaximum=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XScaleStep, pTVSettingInfo->TV3DScalInfo.YScaleStep, pTVSettingInfo->TV3DScalInfo.XScaleMaximum,pTVSettingInfo->TV3DScalInfo.YScaleMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo SIZE: XScaleLevel=%ld, YScaleLevel=%ld, XSIZEOffset=%ld, YSIZEOffset=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XScaleLevel, pTVSettingInfo->TV3DScalInfo.YScaleLevel, pTVSettingInfo->TV3DScalInfo.XSIZEOffset,pTVSettingInfo->TV3DScalInfo.YSIZEOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo SIZE: XSIZENow=%ld, YSIZENow=%ld\n", pTVSettingInfo->TV3DScalInfo.XSIZENow, pTVSettingInfo->TV3DScalInfo.YSIZENow));
        
        /*Position debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo POSITION: XPOSITIONMaxRange=%ld, YPOSITIONMaxRange=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XPOSITIONMaxRange,pTVSettingInfo->TV3DScalInfo.YPOSITIONMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo POSITION: XPOSITIONStep=%ld, YPOSITIONStep=%ld, XPOSITIONMaximum=%ld, YPOSITIONMaximum=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XPOSITIONStep, pTVSettingInfo->TV3DScalInfo.YPOSITIONStep, pTVSettingInfo->TV3DScalInfo.XPOSITIONMaximum,pTVSettingInfo->TV3DScalInfo.YPOSITIONMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo POSITION: XPOSITIONLevel=%ld, YPOSITIONLevel=%ld, XPOSITIONOffset=%ld, YPOSITIONOffset=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XPOSITIONLevel, pTVSettingInfo->TV3DScalInfo.YPOSITIONLevel, pTVSettingInfo->TV3DScalInfo.XPOSITIONOffset,pTVSettingInfo->TV3DScalInfo.YPOSITIONOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pTV3DScalInfo POSITION: XPOSITIONNow=%ld, YPOSITIONNow=%ld\n", \
            pTVSettingInfo->TV3DScalInfo.XPOSITIONNow, pTVSettingInfo->TV3DScalInfo.YPOSITIONNow));
                  
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIASet3DScalScreenParas exit\n")); 
        #endif     
        /*3. initial the scaling control paras*/
        VIASet3DScalCtrlParas(pScrn, p3DSCALCTRLInfo);
       
        /*4. Update the timing factor to IGA */
        /* We get the offset and fetch count from the DISP3DScalingBufPitch which has been 
        aligned to 16 bytes when 3D scaling */ 
        CountWidthByQWord = OffsetWidthByQWord = p3DSCALCTRLInfo->DISP3DScalingBufPitch >> 3;

        pTargetIGASettingInfo->IsPanning = FALSE;
        pTargetIGASettingInfo->HActive = p3DSCALCTRLInfo->RealHActive;
        pTargetIGASettingInfo->VActive = p3DSCALCTRLInfo->RealVActive;
        pTargetIGASettingInfo->OffsetWidthByQWord = OffsetWidthByQWord;
        pTargetIGASettingInfo->CountWidthByQWord = CountWidthByQWord;
        pTargetIGASettingInfo->SetStartingAddrFirstTime = TRUE;
        pTargetIGASettingInfo->IsActive = TRUE;

        /*5. Enable the 3D scaling */
        pBIOSInfo->Is3DScalingEnable |= DISP_3D_SCALING_ENABLE;
    }   

    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_HDMI)
    {
        HDMISETTINGINFOPTR    pHDMISettingInfo = &(pBIOSInfo->HDMISettingInfo);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
              "into VIADISP3DScalingParasSetting for HDMI  :NativeSize=%d\n", 
              pBIOSInfo->HDMISettingInfo.NativeSizeID));       
                 
        switch (pBIOSInfo->HDMISettingInfo.IGAPath )
        {
            case IGA2:
                pTargetIGASettingInfo = &(pBIOSInfo->IGA2SettingInfo);
                p3DSCALCTRLInfo = & (pTargetIGASettingInfo->DISP3DSCALCTRLInfo);
                break;
                
            case IGA1:
            default:
                pTargetIGASettingInfo = &(pBIOSInfo->IGA1SettingInfo);
                p3DSCALCTRLInfo = &(pTargetIGASettingInfo->DISP3DSCALCTRLInfo);            
                break;
        }

        p3DSCALCTRLInfo->DISP3DScalIGAPath = pBIOSInfo->HDMISettingInfo.IGAPath;
        pTargetIGASettingInfo->IsDISP3DScaling = TRUE;

        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
              "VIADISP3DScalingParasSetting begin\n"));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "pHDMISettingInfo:HActive =%d, VActive=%d, IGAPath=%d,ModeIndex =%d\n", 
              pHDMISettingInfo->HActive, 
              pHDMISettingInfo->VActive, 
              pHDMISettingInfo->IGAPath,
              pHDMISettingInfo->ModeIndex));
        
        /*1. first to get the physical size, and store the current mode size*/
        /* DVI just use native size as the destination buffer size.
        * HDMI maybe has several native mode, so we just support 720P/1080P as 
        * the destination buffer size.
        */
        p3DSCALCTRLInfo->OrigHActive = pHDMISettingInfo->HActive;
        p3DSCALCTRLInfo->OrigVActive = pHDMISettingInfo->VActive;        
        
   
        /*Ignore the ReduceBlanking(+0x100) and Interlace mode(+0x200)
          gene to get the true size.*/
        VIAGetModeSizeByIndex((pHDMISettingInfo->ModeIndex)&0xFF,
                              (CARD32 *)&p3DSCALCTRLInfo->RealHActive,
                              (CARD32 *)&p3DSCALCTRLInfo->RealVActive);
        
        /*Ignore the HDMI native mode? Maybe need refine...*/
        if (p3DSCALCTRLInfo->RealHActive < 1920)
        {
            pHDMISettingInfo->ModeIndex = M1280x720;
            p3DSCALCTRLInfo->RealHActive = 1280;
            p3DSCALCTRLInfo->RealVActive = 720;
        }
        else
        {
            pHDMISettingInfo->ModeIndex = M1920x1080;
            p3DSCALCTRLInfo->RealHActive = 1920;
            p3DSCALCTRLInfo->RealVActive = 1080;
        }
        
        pHDMISettingInfo->HActive = p3DSCALCTRLInfo->RealHActive;
        pHDMISettingInfo->VActive = p3DSCALCTRLInfo->RealVActive; 
       
        /*2.initial the scaling screen paras */
        VIASetHDMI3DScalScreenParas(pBIOSInfo);
#if 1       
         /*Size debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo SIZE: XSIZEMax=%ld, YSIZEMax=%ld, XSIZEMaxRange=%ld,"
              "YSIZEMaxRange=%ld\n",
              pHDMISettingInfo->HDMI3DScalInfo.XSIZEMax,
              pHDMISettingInfo->HDMI3DScalInfo.YSIZEMax,
              pHDMISettingInfo->HDMI3DScalInfo.XSIZEMaxRange,
              pHDMISettingInfo->HDMI3DScalInfo.YSIZEMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo SIZE: XScaleStep=%ld, YScaleStep=%ld, XScaleMaximum=%ld,"
              "YScaleMaximum=%ld\n",
              pHDMISettingInfo->HDMI3DScalInfo.XScaleStep,
              pHDMISettingInfo->HDMI3DScalInfo.YScaleStep,
              pHDMISettingInfo->HDMI3DScalInfo.XScaleMaximum,
              pHDMISettingInfo->HDMI3DScalInfo.YScaleMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo SIZE: XScaleLevel=%ld, YScaleLevel=%ld, XSIZEOffset=%ld,"
              "YSIZEOffset=%ld\n", 
              pHDMISettingInfo->HDMI3DScalInfo.XScaleLevel,
              pHDMISettingInfo->HDMI3DScalInfo.YScaleLevel,
              pHDMISettingInfo->HDMI3DScalInfo.XSIZEOffset,
              pHDMISettingInfo->HDMI3DScalInfo.YSIZEOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo SIZE: XSIZENow=%ld, YSIZENow=%ld\n", 
              pHDMISettingInfo->HDMI3DScalInfo.XSIZENow, 
              pHDMISettingInfo->HDMI3DScalInfo.YSIZENow));
        
        /*Position debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo POSITION: XPOSITIONMaxRange=%ld, YPOSITIONMaxRange=%ld\n",
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONMaxRange,
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo POSITION: XPOSITIONStep=%ld, YPOSITIONStep=%ld,"
              "XPOSITIONMaximum=%ld, YPOSITIONMaximum=%ld\n",
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONStep,
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONStep,
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONMaximum,
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
             "3DScalInfo POSITION: XPOSITIONLevel=%ld, YPOSITIONLevel=%ld,"
              "XPOSITIONOffset=%ld, YPOSITIONOffset=%ld\n", 
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONLevel, 
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONLevel, 
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONOffset,
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "3DScalInfo POSITION: XPOSITIONNow=%ld, YPOSITIONNow=%ld\n", 
              pHDMISettingInfo->HDMI3DScalInfo.XPOSITIONNow, 
              pHDMISettingInfo->HDMI3DScalInfo.YPOSITIONNow));
                  
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "VIASet3DScalScreenParas exit\n")); 
#endif     
       
        /*3. initial the scaling control paras*/
        VIASet3DScalCtrlParas(pScrn, p3DSCALCTRLInfo);
       
        /*4. Update the timing factor to IGA */
        CountWidthByQWord = OffsetWidthByQWord = p3DSCALCTRLInfo->DISP3DScalingBufPitch >> 3;
        
        pTargetIGASettingInfo->IsPanning = FALSE;
        pTargetIGASettingInfo->HActive = p3DSCALCTRLInfo->RealHActive;
        pTargetIGASettingInfo->VActive = p3DSCALCTRLInfo->RealVActive;
        pTargetIGASettingInfo->OffsetWidthByQWord = OffsetWidthByQWord;
        pTargetIGASettingInfo->CountWidthByQWord = CountWidthByQWord;
        pTargetIGASettingInfo->SetStartingAddrFirstTime = TRUE;
        pTargetIGASettingInfo->IsActive = TRUE;

        /*5. Enable the 3D scaling */
        pBIOSInfo->Is3DScalingEnable |= DISP_3D_SCALING_ENABLE;
    }
    

    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_LCD)
    {
        LVDSSETTINGINFOPTR    pLVDSSettingInfo = &(pBIOSInfo->LVDSSettingInfo);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "into VIADISP3DScalingParasSetting for LCD  :PanelID=%d, PanelSizeX=%d, PanelSizeY=%d\n", \
            pBIOSInfo->LVDSSettingInfo.PanelSizeID, pBIOSInfo->LVDSSettingInfo.PanelSizeX, pBIOSInfo->LVDSSettingInfo.PanelSizeY));       
                 
        switch (pBIOSInfo->LVDSSettingInfo.IGAPath )
        {
            case IGA2:
                pTargetIGASettingInfo = &(pBIOSInfo->IGA2SettingInfo);
                p3DSCALCTRLInfo = & (pTargetIGASettingInfo->DISP3DSCALCTRLInfo);
                break;
            case IGA1:
            default:
                pTargetIGASettingInfo = &(pBIOSInfo->IGA1SettingInfo);
                p3DSCALCTRLInfo = &(pTargetIGASettingInfo->DISP3DSCALCTRLInfo);            
                break;
        }
        p3DSCALCTRLInfo->DISP3DScalIGAPath = pBIOSInfo->LVDSSettingInfo.IGAPath;
        pTargetIGASettingInfo->IsDISP3DScaling = TRUE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIADISP3DScalingParasSetting begin\n"));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pLVDSSettingInfo:HActive =%d, VActive=%d, IGAPath=%d,ModeIndex =%d\n", \
            pLVDSSettingInfo->HActive, pLVDSSettingInfo->VActive, pLVDSSettingInfo->IGAPath,pLVDSSettingInfo->ModeIndex));
         
        /*1. first to get the physical size, and store the current mode size*/
        /*Replace the LCD setting mode to native mode*/
        
        p3DSCALCTRLInfo->OrigHActive = pLVDSSettingInfo->HActive;
        p3DSCALCTRLInfo->OrigVActive = pLVDSSettingInfo->VActive;

        panelWidth = VIA_PANEL_WIDTH(pLVDSSettingInfo->PanelSizeID);
        panelHeight = VIA_PANEL_HEIGHT(pLVDSSettingInfo->PanelSizeID);

        p3DSCALCTRLInfo->RealHActive = panelWidth;
        p3DSCALCTRLInfo->RealVActive = panelHeight;

        pLVDSSettingInfo->HActive = panelWidth;
        pLVDSSettingInfo->VActive = panelHeight;

        pLVDSSettingInfo->ModeIndex = VIAGetModeIndex(panelWidth, panelHeight);

        if ((p3DSCALCTRLInfo->OrigHActive == p3DSCALCTRLInfo->RealHActive) &&
            (p3DSCALCTRLInfo->OrigVActive == p3DSCALCTRLInfo->RealVActive)) {
            pBIOSInfo->Is3DScalingEnable &= ~VIA_DEVICE_LCD;
            return;
        }    

        /*2.initial the scaling screen paras */
        VIASetLCD3DScalScreenParas(pBIOSInfo);
#if 1       
         /*Size debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo SIZE: XSIZEMax=%ld, YSIZEMax=%ld, XSIZEMaxRange=%ld, YSIZEMaxRange=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XSIZEMax, pLVDSSettingInfo->LCD3DScalInfo.YSIZEMax, pLVDSSettingInfo->LCD3DScalInfo.XSIZEMaxRange,pLVDSSettingInfo->LCD3DScalInfo.YSIZEMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo SIZE: XScaleStep=%ld, YScaleStep=%ld, XScaleMaximum=%ld, YScaleMaximum=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XScaleStep, pLVDSSettingInfo->LCD3DScalInfo.YScaleStep, pLVDSSettingInfo->LCD3DScalInfo.XScaleMaximum,pLVDSSettingInfo->LCD3DScalInfo.YScaleMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo SIZE: XScaleLevel=%ld, YScaleLevel=%ld, XSIZEOffset=%ld, YSIZEOffset=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XScaleLevel, pLVDSSettingInfo->LCD3DScalInfo.YScaleLevel, pLVDSSettingInfo->LCD3DScalInfo.XSIZEOffset,pLVDSSettingInfo->LCD3DScalInfo.YSIZEOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo SIZE: XSIZENow=%ld, YSIZENow=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XSIZENow, pLVDSSettingInfo->LCD3DScalInfo.YSIZENow));
        
        /*Position debug information*/
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONMaxRange=%ld, YPOSITIONMaxRange=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONMaxRange,pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONMaxRange));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONStep=%ld, YPOSITIONStep=%ld, XPOSITIONMaximum=%ld, YPOSITIONMaximum=%ld\n", 
            pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONStep, pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONStep, pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONMaximum,pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONMaximum));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONLevel=%ld, YPOSITIONLevel=%ld, XPOSITIONOffset=%ld, YPOSITIONOffset=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONLevel, pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONLevel, pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONOffset,pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONOffset));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "3DScalInfo POSITION: XPOSITIONNow=%ld, YPOSITIONNow=%ld\n", \
            pLVDSSettingInfo->LCD3DScalInfo.XPOSITIONNow, pLVDSSettingInfo->LCD3DScalInfo.YPOSITIONNow));
                  
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIASet3DScalScreenParas exit\n")); 
#endif     
       
        /*3. initial the scaling control paras*/
        VIASet3DScalCtrlParas(pScrn, p3DSCALCTRLInfo);
       
        /*4. Update the timing factor to IGA */
        CountWidthByQWord = OffsetWidthByQWord = p3DSCALCTRLInfo->DISP3DScalingBufPitch >> 3;
        
        pTargetIGASettingInfo->IsPanning = FALSE;
        pTargetIGASettingInfo->HActive = p3DSCALCTRLInfo->RealHActive;
        pTargetIGASettingInfo->VActive = p3DSCALCTRLInfo->RealVActive;
        pTargetIGASettingInfo->OffsetWidthByQWord = OffsetWidthByQWord;
        pTargetIGASettingInfo->CountWidthByQWord = CountWidthByQWord;
        pTargetIGASettingInfo->SetStartingAddrFirstTime = TRUE;
        pTargetIGASettingInfo->IsActive = TRUE;

        /*5. Enable the 3D scaling */
        pBIOSInfo->Is3DScalingEnable |= DISP_3D_SCALING_ENABLE;  
    }
}

Bool doVIADISP3DScalingImageProcessing(VIAPtr pVia, 
    VIADISP3DSCALCTRLPtr p3DSCALCTRLInfo, VIA3DSclSCNParasPtr pDISP3DScalScreenInfo)
{
    ScrnInfoPtr pScrn = xf86Screens[pVia->pBIOSInfo->scrnIndex];
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;
    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;
    HW3D_Scaling_INFO *    lpGFX3DScal_info;
    viaTexture3DBltSrcRec texBltSrc;
    viaTexture3DBltDstRec texBltDst;
    viaTexture3DBltRotationRec texBltRotation;

    CARD32 SrcScalingAddr, DestScalingAddr, colorKeyScalingAddr;
    CARD32 dwSrcWidth, dwSrcHeight, lSrcPitch;
    CARD32 dwPanningOffset;
    RECTL rSrc, rDest, rDestCkPatch;

    CARD32 LinearScaleMode = TRUE;
    Bool bDuoviewDualVidEng = FALSE;
    Bool bPatchColorKey = FALSE;
    int i, currentBufIndex;
    
    int numOfActiveVideo=0;
    int videoOnIGA=0;
    RECTL rVideoSrc, rVideoDest;

    unsigned long curIGAIndex = 0;
    unsigned long VideoFlag[2];
    Bool vidColorKeyOn[2];
    unsigned long vidColorKey[2];
    unsigned long patchedColorKey;
    ViaXvRectRec dwOriDestRect[2];

    Bool isH2 = (H5_UMA_CHIPID||H6_UMA_CHIPID) ? \
        FALSE: TRUE;

    /*Get the 3D scaling source address */
    dwPanningOffset =( (p3DSCALCTRLInfo->VirtualHActive)*(p3DSCALCTRLInfo->PanningY)+ \
        (p3DSCALCTRLInfo->PanningX))*(p3DSCALCTRLInfo->BitsPerPixel>> 3);
    SrcScalingAddr = (p3DSCALCTRLInfo->OrigFBAddress) + dwPanningOffset;
    
    /* Get the 3D scaling source pitch: 
        3D engine of VT3353/3409 require that the src and dst pitch must be aligned to 32
    */
    lSrcPitch = ALIGN_TO(p3DSCALCTRLInfo->VirtualHActive*(p3DSCALCTRLInfo->BitsPerPixel>> 3), 32);

    rSrc.left = rSrc.top = 0;
    rSrc.right = pBIOSInfo->CrtcHDisplay;
    rSrc.bottom = pBIOSInfo->CrtcVDisplay;
    dwSrcWidth = pBIOSInfo->CrtcHDisplay;
    dwSrcHeight = pBIOSInfo->CrtcVDisplay;

    /*Get the current buffer index to be processed and updated to screen */
    currentBufIndex=(p3DSCALCTRLInfo->DISP3DScalingBufIndex+1)%(DISP_3D_SCALING_BUFNUM_ONESET-1);
    /* Get the 3D scaling destination address */
    DestScalingAddr = p3DSCALCTRLInfo->DISP3DScalingBufAddr[currentBufIndex];

    rDest.left = rDest.top = 0;
    rDest.right = p3DSCALCTRLInfo->RealHActive;
    rDest.bottom = p3DSCALCTRLInfo->RealVActive;

    DISP3DScalWindowAdjust(pScrn->scrnIndex, &rSrc , &rDest, pDISP3DScalScreenInfo);

    vidColorKeyOn[0]=FALSE;
    vidColorKeyOn[1]=FALSE;
    vidColorKeyOn[2]=FALSE;

    /* NOTE:
        XV Overlay and VIA Video Acceration can not be running at the same time when 3D Scaling is enabled.
        The following 2 scetion of code will make conflicts actually.
    */

    for(i=0; i<2; i++)
    {
        if(pVia->pPriv[i] != NULL){
            if((pVia->pPriv[i]->videoFlag & VIDEO_ACTIVE) && !pVia->pPriv[i]->redirected)
            {
                VideoFlag[i] =  pVia->pPriv[i]->videoFlag;
                curIGAIndex = pVia->pPriv[i]->curIGA - 1;
                
                vidColorKeyOn[i] = pVia->pPriv[i]->ovlInfo[curIGAIndex].dwColorKeyOn;
                vidColorKey[i] = pVia->pPriv[i]->ovlInfo[curIGAIndex].dwKeyLow;
                dwOriDestRect[i] = pVia->pPriv[i]->ovlInfo[curIGAIndex].dwOriDestRect;               
                numOfActiveVideo++;
                DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "video stream is SW %d \n",i));
                DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "NO. %d \n",numOfActiveVideo));
            }
        }
    }

    /* First step, do the scaling using 3D engine linear mode*/
    /* If video overlay is actived with "GoodPerformance" option setting, we only do no linear 3D scaling once*/
    if(numOfActiveVideo && pVia->GoodPerformance)
    {
        LinearScaleMode = FALSE;
    }

    memset(&texBltSrc, 0x00, sizeof(texBltSrc));
    memset(&texBltDst, 0x00, sizeof(texBltDst));
    memset(&texBltRotation, 0x00, sizeof(texBltRotation));

    switch(pBIOSInfo->bitsPerPixel)
    {
        case 15:
            texBltSrc.format = PICT_x1r5g5b5;
            texBltDst.format = PICT_x1r5g5b5;
            break;
        case 16:
            texBltSrc.format = PICT_r5g6b5;
            texBltDst.format = PICT_r5g6b5;
            break;
            
        case 24:
            texBltSrc.format = PICT_x8r8g8b8;
            texBltDst.format = PICT_x8r8g8b8;
            break;
            
        case 32:
        default:
            texBltSrc.format = PICT_a8r8g8b8;
            texBltDst.format = PICT_a8r8g8b8;
            break;
    }

    /* NOTE: the first time, Blt destination address: DestScalingAddr */
    texBltDst.offset = DestScalingAddr;
    texBltDst.pitch = p3DSCALCTRLInfo->DISP3DScalingBufPitch;
    texBltDst.x = rDest.left;
    texBltDst.y = rDest.top;
    texBltDst.w = rDest.right-rDest.left;
    texBltDst.h = rDest.bottom-rDest.top;

    texBltSrc.offset = SrcScalingAddr;
    texBltSrc.pitch = lSrcPitch;
    texBltSrc.x = rSrc.left;
    texBltSrc.y = rSrc.top;
    texBltSrc.w = rSrc.right-rSrc.left;
    texBltSrc.h = rSrc.bottom-rSrc.top;
    texBltSrc.surfwidth = isH2 ? dwSrcWidth : pVia->pBIOSInfo->SaveVirtualX;
    texBltSrc.surfheight = isH2 ? dwSrcHeight: pVia->pBIOSInfo->SaveVirtualY;
    texBltSrc.filter = LinearScaleMode ? via_FilterBilinear : via_FilterNearest;
    texBltSrc.memLoc = LOC_SL;

    texBltRotation.rotate = VIA_TO_RR(0);
    texBltRotation.width = 0;
    texBltRotation.height = 0;

    /* Must acquire the DRM LOCK before accessing HW 3D engine */
#ifdef XF86DRI 
    if(pVia->directRenderingEnabled)
    {
        DRILock(pScrn->pScreen, 0);
    }
#endif   
    viaAccelTexture3DBlt(pScrn, &texBltSrc, &texBltDst, &texBltRotation, NULL);
#ifdef XF86DRI
    if(pVia->directRenderingEnabled)
    {
        DRIUnlock(pScrn->pScreen);
    }
#endif  

    /************* patch video color key in 3D scaling mode **************/
    if (!pVia->GoodPerformance && numOfActiveVideo)
    {
        for (i=0; i < 2; i++)
        {
            bPatchColorKey = FALSE;
            if(vidColorKeyOn[i])  /* if video color key enabled ? */
            {
                patchedColorKey = vidColorKey[i];
                DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Check the NO. %d video stream \n",i));

                /* if duoview mode 2 video engine case ? */
                if(bDuoviewDualVidEng){
                    bPatchColorKey = TRUE;
                } else { 
                    /* if the video on Iga is the 3D scaling target Iga ?*/
                    videoOnIGA = (VideoFlag[i] & VIDEO_ON_IGA2) ?IGA2 : IGA1;
                     if(videoOnIGA == p3DSCALCTRLInfo->DISP3DScalIGAPath) {
                        bPatchColorKey = TRUE;
                     }
                }
            }
            
            if(bPatchColorKey)  /* if need patch video color key ? */
            {
                colorKeyScalingAddr = p3DSCALCTRLInfo->DISP3DScalingBufAddr[2];
                
                memset(&texBltSrc, 0x00, sizeof(texBltSrc));
                memset(&texBltDst, 0x00, sizeof(texBltDst));
                memset(&texBltRotation, 0x00, sizeof(texBltRotation));

                switch(pBIOSInfo->bitsPerPixel)
                {
                    case 15:
                        texBltSrc.format = PICT_x1r5g5b5;
                        texBltDst.format = PICT_x1r5g5b5;
                        break;
                    case 16:
                        texBltSrc.format = PICT_r5g6b5;
                        texBltDst.format = PICT_r5g6b5;
                        break;
                        
                    case 24:
                        texBltSrc.format = PICT_x8r8g8b8;
                        texBltDst.format = PICT_x8r8g8b8;
                        break;
                        
                    case 32:
                    default:
                        texBltSrc.format = PICT_a8r8g8b8;
                        texBltDst.format = PICT_a8r8g8b8;
                        break;
                }

                /* NOTE: the second time, Blt destination address: colorKeyScalingAddr */
                texBltDst.offset = colorKeyScalingAddr;
                texBltDst.pitch = p3DSCALCTRLInfo->DISP3DScalingBufPitch;
                texBltDst.x = 0;
                texBltDst.y = 0;
                texBltDst.w = rDest.right;
                texBltDst.h = rDest.bottom;

                texBltSrc.offset = SrcScalingAddr;
                texBltSrc.pitch = lSrcPitch;
                texBltSrc.x = 0;
                texBltSrc.y = 0;
                texBltSrc.w = rSrc.right;
                texBltSrc.h = rSrc.bottom;
                texBltSrc.surfwidth = isH2 ? dwSrcWidth : pVia->pBIOSInfo->SaveVirtualX;
                texBltSrc.surfheight = isH2 ? dwSrcHeight: pVia->pBIOSInfo->SaveVirtualY;
                texBltSrc.filter = via_FilterNearest;
                texBltSrc.memLoc = LOC_SL;

                texBltRotation.rotate = VIA_TO_RR(0);
                texBltRotation.width = 0;
                texBltRotation.height = 0;
                
                /* Must acquire the DRM LOCK before accessing HW 3D engine */
#ifdef XF86DRI 
                if(pVia->directRenderingEnabled)
                {
                    DRILock(pScrn->pScreen, 0);
                }
#endif  
                viaAccelTexture3DBlt(pScrn, &texBltSrc, &texBltDst, &texBltRotation, NULL);
#ifdef XF86DRI
                if(pVia->directRenderingEnabled)
                {
                    DRIUnlock(pScrn->pScreen);
                }
#endif                   
                rVideoSrc.left   = max(0, dwOriDestRect[i].left);
                rVideoSrc.top    = max(0, dwOriDestRect[i].top);
                rVideoSrc.right  = min(p3DSCALCTRLInfo->OrigHActive, dwOriDestRect[i].right);
                rVideoSrc.bottom = min(p3DSCALCTRLInfo->OrigVActive, dwOriDestRect[i].bottom);

                /*get the desternation video window position parameters*/
                switch(p3DSCALCTRLInfo->DISP3DScalIGAPath)
                {
                    case IGA2:
                        lpGFX3DScal_info = 
                            &(pVidData->viaGfxInfo->igaInfo[IGA2-1].igagfx3DScaling_info);
                        break;
                        
                    case IGA1:
                      default:
                        lpGFX3DScal_info = 
                            &(pVidData->viaGfxInfo->igaInfo[IGA1-1].igagfx3DScaling_info);
                        break;
                }
                vid3DScalColorKeyWindowAdjust(&rVideoSrc , &rVideoDest, lpGFX3DScal_info);

                rVideoDest.left    = range_limit(rVideoDest.left, 0, p3DSCALCTRLInfo->RealHActive);
                rVideoDest.top     = range_limit(rVideoDest.top, 0, p3DSCALCTRLInfo->RealVActive);
                rVideoDest.right   = range_limit(rVideoDest.right, 0, p3DSCALCTRLInfo->RealHActive);
                rVideoDest.bottom  = range_limit(rVideoDest.bottom, 0, p3DSCALCTRLInfo->RealVActive);

                rDestCkPatch.left = rDestCkPatch.top =0;
                rDestCkPatch.right = p3DSCALCTRLInfo->RealHActive;
                rDestCkPatch.bottom = p3DSCALCTRLInfo->RealVActive;

                /* NOTE: Blt Color key from VidDest Addr to DestScalingAddr with 2D engine */
                patchForVidColorKey(pVia,colorKeyScalingAddr,DestScalingAddr, patchedColorKey,\
                    p3DSCALCTRLInfo->DISP3DScalingBufPitch,&rDestCkPatch,&rVideoDest);

            }
        }
    }

    /*clean the shrink buffer if adjusted the size or position*/
    if(pDISP3DScalScreenInfo->bNeedBlack3dScalingBorders ==  TRUE)
    {
        for(i=0; i<DISP_3D_SCALING_BUFNUM_ONESET-1;i++)
        {
            BlackSclScrnBoundary(pVia, p3DSCALCTRLInfo, rDest,i);
        }
        pDISP3DScalScreenInfo->bNeedBlack3dScalingBorders = FALSE;
    }

    /*update the current display buffer index*/
    p3DSCALCTRLInfo->DISP3DScalingBufIndex = currentBufIndex;

    /* Unlock Extended Regs */
    VGAOUT8(0x3c4, 0x10);
    VGAOUT8(0x3c5, VGAIN8(0x3c5) | 0x01);
    /* Unlock CRTC register protect */
    VGAOUT8(0x3d4, 0x47);
    VGAOUT8(0x3d5, VGAIN8(0x3d5) & ~0x01);

    switch(p3DSCALCTRLInfo->DISP3DScalIGAPath)
    {
        case IGA1:
            VIASETREG(0x214, DestScalingAddr>>1);
            break;
        case IGA2:
            VIASETREG(0x218, DestScalingAddr);
            break;
    }

    return TRUE;
}

Bool VIADISP3DScalingImageProcessing(VIAPtr pVia)
{
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    VIADISP3DSCALCTRLPtr p3DSCALCTRLInfo = NULL;
    VIA3DSclSCNParasPtr pDISP3DScalScreenInfo = NULL;
	
    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_TV)
    {
        switch (pBIOSInfo->TVSettingInfo.IGAPath )
        {
            case IGA2:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA2SettingInfo.DISP3DSCALCTRLInfo);
                break;
            case IGA1:
            default:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA1SettingInfo.DISP3DSCALCTRLInfo);
                break;
        }
        pDISP3DScalScreenInfo = &(pBIOSInfo->TVSettingInfo.TV3DScalInfo);

        doVIADISP3DScalingImageProcessing(pVia, p3DSCALCTRLInfo, pDISP3DScalScreenInfo);
    } 
    
    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_HDMI)
    {
        switch (pBIOSInfo->HDMISettingInfo.IGAPath )
        {
            case IGA2:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA2SettingInfo.DISP3DSCALCTRLInfo);
                break;
            case IGA1:
            default:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA1SettingInfo.DISP3DSCALCTRLInfo);
                break;
        }
        pDISP3DScalScreenInfo = &(pBIOSInfo->HDMISettingInfo.HDMI3DScalInfo);
        doVIADISP3DScalingImageProcessing(pVia, p3DSCALCTRLInfo, pDISP3DScalScreenInfo);        
    } 

   
    if (pBIOSInfo->Is3DScalingEnable & VIA_DEVICE_LCD)
    {
        switch (pBIOSInfo->LVDSSettingInfo.IGAPath )
        {
            case IGA2:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA2SettingInfo.DISP3DSCALCTRLInfo);
                break;
            case IGA1:
            default:
                p3DSCALCTRLInfo =& (pBIOSInfo->IGA1SettingInfo.DISP3DSCALCTRLInfo);
                break;
        }
        pDISP3DScalScreenInfo = &(pBIOSInfo->LVDSSettingInfo.LCD3DScalInfo);
        doVIADISP3DScalingImageProcessing(pVia, p3DSCALCTRLInfo, pDISP3DScalScreenInfo);        
    } 

    return TRUE;
} 

