/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*************************************************************************
 *
 *  File:       via_bios.c
 *  Content:    Get the mode table from VGA BIOS and set mode by this table
 *
 ************************************************************************/

#include "via_driver.h"
#include "xf86Priv.h"
#include "via_rotate.h"
#include "via_serial.h"
#include "via_hdmi.h"

/* Functions Prototype Declaration */
Bool IsValidMode(CARD16 HDisplay, CARD16 VDisplay);

/* Extern Functions Prototypes Declaration */
extern __inline__ void write_reg(CARD8 index, CARD16 io_port, CARD8 data);
extern void write_reg_mask(CARD8 index, int io_port, CARD8 data, CARD8 mask);
extern __inline__ CARD8 read_reg(int io_port, CARD8 index);
extern int SearchModeSetting(int ModeInfoIndex);
extern void write_regx(struct io_reg RegTable[],int ItemNum);
extern void unlock_crt(void);
extern void lock_crt(void);
extern CARD32 get_clk_value(VIABIOSInfoPtr pBIOSInfo, int clk);
extern void SetVCLK(VIABIOSInfoPtr pBIOSInfo, CARD32 CLK, int set_iga);
extern void EnableSecondDisplayChannel();
extern void DisableSecondDisplayChannel();
extern void load_offset_reg(int h_addr, int bpp_byte, int set_iga);
extern void set_color_depth(VIABIOSInfoPtr pBIOSInfo, int bpp_byte, int set_iga);
extern void DoTVSWReset(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo);

extern Bool vt1625_module_loaded;
extern Bool ad9389_module_loaded;

/* Support Mode, the order must follow the rule "small to big". */
struct _VIASUPPORTMODE VIASupportMode[] = {
/*  Index,         HActive, VActive*/
    {"480x640",     M480x640,     480,     640},
    {"640x480",     M640x480,     640,     480},
    {"720x480",     M720x480,     720,     480},
    {"720x576",     M720x576,     720,     576},
    {"800x480",     M800x480,     800,     480},
    {"800x600",     M800x600,     800,     600},
    {"848x480",     M848x480,     848,     480},
    {"852x480",     M852x480,     856,     480},
    {"960x600",     M960x600,     960,     600}, 
    {"1000x600",    M1000x600,    1000,    600}, 
    {"1024x512",    M1024x512,    1024,    512},  
    {"1024x600",    M1024x600,    1024,    600},     
    {"1024x576",    M1024x576,    1024,    576},        
    {"1024x768",    M1024x768,    1024,    768},
    {"1088x612",    M1088x612,    1088,    612}, 
    {"1152x720",    M1152x720,    1152,    720},
    {"1152x864",    M1152x864,    1152,    864},
    {"1200x720",    M1200x720,    1200,    720},
    {"1280x600",    M1280x600,    1280,    600},
    {"1280x720",    M1280x720,    1280,    720},
    {"1280x768",    M1280x768,    1280,    768},
    {"1280x800",    M1280x800,    1280,    800},
    {"1280x960",    M1280x960,    1280,    960},
    {"1280x1024",   M1280x1024,   1280,    1024},
    {"1360x768",    M1360x768,    1360,    768},
    {"1366x768",    M1366x768,    1366,    768},
    {"1368x768",	M1368x768,	  1368,    768},
    {"1400x1050",   M1400x1050,   1400,    1050},
    {"1440x900",    M1440x900,    1440,    900},
    {"1440x1050",   M1440x1050,   1440,    1050},
    {"1600x900",    M1600x900,    1600,    900},
    {"1600x1024",   M1600x1024,   1600,    1024},
    {"1600x1200",   M1600x1200,   1600,    1200},
    {"1680x1050",   M1680x1050,   1680,    1050},
    {"1792x1344",   M1792x1344,   1792,    1344},
    {"1856x1392",   M1856x1392,   1856,    1392},
    {"1920x1080",   M1920x1080,   1920,    1080},
    {"1920x1200",   M1920x1200,   1920,    1200},
    {"1920x1440",   M1920x1440,   1920,    1440},   
    {"2048x1536",   M2048x1536,   2048,    1536}, 
    {"720x400",     M720x400,     720,     400},
    {"832x624",     M832x624,     832,     624},
    {"",            -1,    0,       0}
};



/* Search Valid mode */
Bool IsValidMode(CARD16 HDisplay, CARD16 VDisplay)
{
    int i = 0;
    
    while (VIASupportMode[i].ModeIndex > -1)
    {
        if ((HDisplay==VIASupportMode[i].HActive)&&(VDisplay==VIASupportMode[i].VActive))
        {
			return 1;
        }
        i++;
    }
    return 0;
}

/*=*
 *
 * unsigned char VIAGetActiveDisplay(VIABIOSInfoPtr, unsigned char)
 *
 *     - Get Active Display Device
 *
 * Return Type:    unsigned char
 *
 * The Definition of Input Value:
 *
 *                 VIABIOSInfoPtr
 *
 * The Definition of Return Value:
 *
 *                 Bit[7] 2nd Path
 *                 Bit[6] 1/0 MHS Enable/Disable
 *                 Bit[5] 0 = Bypass Callback, 1 = Enable Callback
 *                 Bit[4] 0 = Hot-Key Sequence Control (OEM Specific)
 *                 Bit[3] LCD
 *                 Bit[2] TV
 *                 Bit[1] CRT
 *                 Bit[0] DVI
 *=*/

unsigned char VIAGetActiveDisplay(VIABIOSInfoPtr pBIOSInfo)
{
    VIABIOSInfoPtr  pVia = pBIOSInfo;
    unsigned char tmp;

    tmp = (read_reg(VIACR, CR3E) & 0xF0) >> 4;
    tmp |= (read_reg(VIACR, CR3B) & 0x18) << 3;

    return tmp;
}

/*=*
 *
 * unsigned char VIAGetDeviceDetect(VIABIOSInfoPtr)
 *
 *     - Get Display Device Attched
 *
 * Return Type:    unsigned char
 *
 * The Definition of Input Value:
 *
 *                 VIABIOSInfoPtr
 *
 * The Definition of Return Value:
 *
 *                 Bit[7] Reserved ------------ 2nd TV Connector
 *                 Bit[6] Reserved ------------ 1st TV Connector
 *                 Bit[5] Reserved
 *                 Bit[4] CRT2
 *                 Bit[3] DFP
 *                 Bit[2] TV
 *                 Bit[1] LCD
 *                 Bit[0] CRT
 *=*/

CARD32 VIAGetDeviceDetect(VIABIOSInfoPtr pBIOSInfo)
{
    CARD32   ret;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAGetDeviceDetect\n"));

    ret = VIA_DEVICE_CRT1; /* Default assume color CRT attached */

    ret |= VIAGetDeviceDetectTV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo));
    ret |= VIAGetDeviceDetectLCD(pBIOSInfo);
    ret |= VIAGetDeviceDetectDVI(pBIOSInfo,&(pBIOSInfo->TMDSSettingInfo),VIA_DEVICE_DFP);
    ret |= viaGetDeviceDetectHDMI(pBIOSInfo);
    ret |= VIAGetDeviceDetectDP(pBIOSInfo); 

    if (pBIOSInfo->TVSettingInfo2.TVEncoder)
    {
        if(VIAGetDeviceDetectTV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo2)))
        {
            ret |= VIA_DEVICE_TV2;
        }
    }

    if (pBIOSInfo->TMDSSettingInfo2.ChipID)
    {
        ret |= VIAGetDeviceDetectDVI(pBIOSInfo,&(pBIOSInfo->TMDSSettingInfo2),VIA_DEVICE_DFP2);
    }
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
                    "Connect Device is 0x%x\n", ret));
    
    return ret;
}

Bool VIAFindRefreshRate(VIABIOSInfoPtr pBIOSInfo, int ModeIndex)
{
    struct VideoModeTable   *VideoMode;
    int     i;
    int     needRefresh = pBIOSInfo->Refresh;

    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        switch (ModeIndex)
        {
            case M1024x768:
            case M1152x864:
            case M1280x1024:
            case M1600x1200:
            case M1920x1080:
                ModeIndex |= 0x200;
                break;
        }
    }
    
    VideoMode = &VIARoutineModes[SearchModeSetting(ModeIndex)];

    for (i = 0; i < VideoMode->mode_array; i++)
    {
        pBIOSInfo->FoundRefresh = VideoMode->crtc[i].refresh_rate;
        
        if (VideoMode->crtc[i].refresh_rate == needRefresh)
        {            
            return TRUE;
        }
    }

    xf86DrvMsg(pBIOSInfo->scrnIndex, X_WARNING, "Can't find refresh rate %d Hz.\n", needRefresh);

    return FALSE;
}

void VIAGetPanelSizeByIndex(int PanelSizeIndex, int *pSizeX, int *pSizeY)
{
	*pSizeX = VIA_PANEL_WIDTH(PanelSizeIndex);
	*pSizeY = VIA_PANEL_HEIGHT(PanelSizeIndex);
}


int VIAGetModeIndex(int HorResolution, int VerResolution)
{
    int ModeIndex = M1024x768;
    int i = 0;    
    
    while (VIASupportMode[i].ModeIndex > -1)
    {        
        if ((HorResolution == VIASupportMode[i].HActive) &&
            (VerResolution == VIASupportMode[i].VActive))
        {
            ModeIndex = VIASupportMode[i].ModeIndex;
            break;
        }

        i++;
    }    
    
    return ModeIndex;
}


void VIAGetModeSizeByName(char *pName, int *pHSize, int *pVSize)
{
    int i = 0;

    if (!pName)
    {
        return;
    }

    while (VIASupportMode[i].ModeIndex > -1)
    {
        if (xf86NameCmp(pName, VIASupportMode[i].Name) == 0)
        {
            *pHSize = VIASupportMode[i].HActive;
            *pVSize = VIASupportMode[i].VActive;
            break;
        }

        i++;
    }
}

/*
*   Get mode size by mode index.
*/
void VIAGetModeSizeByIndex(int ModeIndex, CARD32 *pHSize, CARD32 *pVSize)
{
    int i = 0;
    
    while (VIASupportMode[i].ModeIndex > -1)
    {
        if (VIASupportMode[i].ModeIndex == ModeIndex)
        {
            *pHSize = VIASupportMode[i].HActive;
            *pVSize = VIASupportMode[i].VActive;
            break;
        }

        i++;
    }
}

CARD32 VIAGetPixelClockByMode(CARD32 ModeIndex)
{        
    CARD32 i = 0;
    CARD32 DesireCLK = 0;

    for (i=0; i<NUM_TOTAL_MODETABLE; i++)
    {
        if (ModeIndex == VIARoutineModes[i].ModeIndex)
        {
             DesireCLK = VIARoutineModes[i].crtc->clk;
        }
    }    

    return DesireCLK;
}

/*
    Set LCD Panel Info according to Panel ID.
*/
static Bool
VIASetLCDPanelSizeFromID(VIABIOSInfoPtr pBIOSInfo, int panel_id)
{
    switch (panel_id)
    {
        case 0x0:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_640X480;
            break;
        case 0x1:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_800X600;
            break;
        case 0x2:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1024X768;
            break;
        case 0x3:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X768;
            break;
        case 0x4:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X1024;
            break;
        case 0x5:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1400X1050;
            break;
        case 0x6:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1440X900;
            break;
        case 0x7:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X800;
            break;
        case 0x8:       /*For Quanta*/
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_800X480;
            break;
        case 0x9:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1024X600;
            break;
        case 0xA:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1366X768;
            break;
        case 0xB:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1600X1200;
            break;
        case 0xC: /* Need to be fixed to 1680x1050 */
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X768;
            break;
        case 0xD: /* Need to be fixed to 1920x1200 */
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X1024;
            break;
        case 0xE: /* Need to be fixed to 640x240 */
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1600X1200;
            break;
         case 0xF:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_480X640;
            break;
        case 0x10:  /*For Panasonic 1280x768 18bit Dual-Channel Panel */
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1280X768;
            break;
        case 0x11:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1360X768;
            break;
        case 0x12:
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_1024X768;
            break;
        case 0x13:  /*General 8x4 panel use this setting*/
            pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_800X480_GEN;
            break;
        default:
            return FALSE;
    }
    return TRUE;

}


static void 
VIAEnabledPrimaryExtendedFIFO(VIABIOSInfoPtr pBIOSInfo)
{
    VIABIOSInfoPtr  pVia = pBIOSInfo;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAEnabledPrimaryExtendedFIFO\n"));
    switch (pBIOSInfo->Chipset) {
    case VIA_K8M890:
        VGAOUT8(0x3c4, 0x16);
        VGAOUT8(0x3c5, 0x92);
        VGAOUT8(0x3c4, 0x17);
        VGAOUT8(0x3c5, 0xB3);
        VGAOUT8(0x3c4, 0x18);
        VGAOUT8(0x3c5, 0x8A);
        break;
    default:
        break;
    }
}


/* Determine whether to use reduce blanking mode. */
void VIAReduceBlanking(VIABIOSInfoPtr pBIOSInfo, CARD32 ModeIndex, CARD32 DeviceType)
{        
    CARD32 DesiredModePixelClock;    

    /* if NoDDCValue is enable, we don't know the max pixel clock of monitor. */
    if (pBIOSInfo->NoDDCValue)
        return;

    /* Only 60Hz can reduce blanking */
    if (pBIOSInfo->FoundRefresh != 60)
        return;    
    
    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAReduceBlanking!!\n");
    
    switch (ModeIndex)
    {   
        /* Following modes have reduce blanking mode. */
        case M1360x768:
        case M1400x1050:
        case M1440x900:
        case M1600x900:
        case M1680x1050:
        case M1920x1080:
        case M1920x1200:                                           
            break;
            
        default:              
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "Reduce Blanking Mode doesn't exist!\n"));
            return; 
            break;
    }
    
    /* Start to determine whether to use reduce blanking mode. */
    DesiredModePixelClock = VIAGetPixelClockByMode(ModeIndex)/1000000; /* MHz */
    
    switch (DeviceType)
    {
        case VIA_DEVICE_CRT1:
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                    "Max Pixel Clock of CRT = %d\n", pBIOSInfo->CRTSettingInfo.MaxPixelClock));
            if ((DesiredModePixelClock > pBIOSInfo->CRTSettingInfo.MaxPixelClock) &&
                 pBIOSInfo->CRTSettingInfo.IsFlatMonitor)       
            {             
                 pBIOSInfo->CRTSettingInfo.ModeIndex |= 0x100;       
                 DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                    "CRT Use CVT Reduce Blaning Mode, Mode Index = %x\n", pBIOSInfo->CRTSettingInfo.ModeIndex));
            }
            break;

        case VIA_DEVICE_DFP:
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                    "Max Pixel Clock of DVI = %d\n!", pBIOSInfo->TMDSSettingInfo.MaxPixelClock));
            if (DesiredModePixelClock > pBIOSInfo->TMDSSettingInfo.MaxPixelClock)
            {
                pBIOSInfo->TMDSSettingInfo.ModeIndex |= 0x100;   
                DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                    "DVI Use CVT Reduce Blaning Mode, Mode Index = %x\n", pBIOSInfo->TMDSSettingInfo.ModeIndex));
            }
            break;
    }
    
}

void VIAInitSettingInfo_IGA(VIABIOSInfoPtr pBIOSInfo, int DisplayPath, int IsPanning, int IsDownScaling, int HActive, int VActive, CARD16 activeDevice)

{
    int OffsetWidthByQWord, CountWidthByQWord;
    OffsetWidthByQWord = (pBIOSInfo->displayWidth * (pBIOSInfo->bitsPerPixel >> 3)) >> 3;
    CountWidthByQWord = ((HActive * (pBIOSInfo->bitsPerPixel >> 3) + 15 ) & ~15) >> 3; /* Do 16 bytes alignment */
    
    if (DisplayPath == IGA1)
    {
        pBIOSInfo->IGA1SettingInfo.IsPanning = IsPanning;
        pBIOSInfo->IGA1SettingInfo.IsDownScaling = IsDownScaling;		
        pBIOSInfo->IGA1SettingInfo.HActive = HActive;
        pBIOSInfo->IGA1SettingInfo.VActive = VActive;
        pBIOSInfo->IGA1SettingInfo.OffsetWidthByQWord = OffsetWidthByQWord;
        pBIOSInfo->IGA1SettingInfo.CountWidthByQWord = CountWidthByQWord;
        pBIOSInfo->IGA1SettingInfo.SetStartingAddrFirstTime = TRUE;
        pBIOSInfo->IGA1SettingInfo.IsActive = TRUE;
        pBIOSInfo->IGA1SettingInfo.deviceActived = activeDevice;
    }
    else if (DisplayPath == IGA2)
    {
        pBIOSInfo->IGA2SettingInfo.IsPanning = IsPanning;
		pBIOSInfo->IGA2SettingInfo.IsDownScaling = IsDownScaling;
        pBIOSInfo->IGA2SettingInfo.HActive = HActive;
        pBIOSInfo->IGA2SettingInfo.VActive = VActive;
        pBIOSInfo->IGA2SettingInfo.OffsetWidthByQWord = OffsetWidthByQWord;
        pBIOSInfo->IGA2SettingInfo.CountWidthByQWord = CountWidthByQWord;
        pBIOSInfo->IGA2SettingInfo.SetStartingAddrFirstTime = TRUE;
        pBIOSInfo->IGA2SettingInfo.IsActive = TRUE;
        pBIOSInfo->IGA2SettingInfo.deviceActived = activeDevice;        
    }
}




void VIAAdjustActiveSize_LCD(VIABIOSInfoPtr pBIOSInfo, LVDSSETTINGINFOPTR pLVDSSettingInfo, int *pHActive, int *pVActive,
                          unsigned long whichLCD)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;

    /*When LCD scaling function is enable, we neednt to take care of Panning. */
    if (pBIOSInfo->Is3DScalingEnable & whichLCD)
    {
         pLVDSSettingInfo->IsPanning = FALSE; 
    }
    else
    {
        if ((HResToSet > pLVDSSettingInfo->PanelSizeX) || 
            (VResToSet > pLVDSSettingInfo->PanelSizeY))
        {
            if ((*pHActive > pLVDSSettingInfo->PanelSizeX) || 
                (*pVActive > pLVDSSettingInfo->PanelSizeY))
            {
                if ((pBIOSInfo->Chipset == VIA_VX900)&&
                    (*pHActive<=pLVDSSettingInfo->PanelSizeX*1.33)&&
                    (*pVActive<=pLVDSSettingInfo->PanelSizeY*1.33)) {
                    pLVDSSettingInfo->IsPanning = FALSE;
                    pLVDSSettingInfo->IsDownScaling = TRUE;
                    pBIOSInfo->HorDownScaleSize = *pHActive;
                    pBIOSInfo->VerDownScaleSize = *pVActive;
                } else {
                    /* Need to adjust active size: (Panning)*/
                    if (*pHActive > pLVDSSettingInfo->PanelSizeX)
                        *pHActive = pLVDSSettingInfo->PanelSizeX;
                    *pVActive = pLVDSSettingInfo->PanelSizeY;
                    pLVDSSettingInfo->IsPanning = TRUE;
                    pLVDSSettingInfo->IsDownScaling = FALSE;
                }
            }                            
        }
        else if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
        {
            /* If use Virtual Desktop, and the size is larger than the resolution, */
            /* we need to do panning, too. */
            pLVDSSettingInfo->IsPanning = TRUE;
            pLVDSSettingInfo->IsDownScaling = FALSE;
        }
        else 
        {
            pLVDSSettingInfo->IsPanning = FALSE;
            pLVDSSettingInfo->IsDownScaling = FALSE;  
        }
    }
}


void VIAAdjustActiveSize_DP(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo, int *pHActive, int *pVActive)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;

    if ((pDPSettingInfo->PanelSizeH == 0) &&
        (pDPSettingInfo->PanelSizeV == 0) &&
        (pDPSettingInfo->PanelSizeH != 0xFFF))
    {
        VIAGetModeSizeByIndex(pDPSettingInfo->DPSize,
                             &pDPSettingInfo->PanelSizeH,
                             &pDPSettingInfo->PanelSizeV);
    }
    
    if ((HResToSet > pDPSettingInfo->PanelSizeH) ||
        (VResToSet > pDPSettingInfo->PanelSizeV))
    {        
        if ((*pHActive > pDPSettingInfo->PanelSizeH) ||
            (*pVActive > pDPSettingInfo->PanelSizeV))
        {            
            /* Panning */
            *pHActive = pDPSettingInfo->PanelSizeH;
            *pVActive = pDPSettingInfo->PanelSizeV;
            pDPSettingInfo->IsPanning = TRUE;
            pDPSettingInfo->IsDownScaling = FALSE;
        }            
    }
    else if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
    {
        /* If use Virtual Desktop, and the size is larger than the resolution, */
        /* we need to do panning, too. */
        pDPSettingInfo->IsPanning = TRUE;
    }
    else
    {
        pDPSettingInfo->IsPanning = FALSE;
    }
}

void VIAAdjustActiveSize_CRT(VIABIOSInfoPtr pBIOSInfo, CRTSETTINGINFOPTR pCRTSettingInfo, int* pHActive, int* pVActive)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;
	

    
    if ((HResToSet > pCRTSettingInfo->MonitorSizeH) ||
        (VResToSet > pCRTSettingInfo->MonitorSizeV))
    {
        if ((*pHActive > pCRTSettingInfo->MonitorSizeH) ||
            (*pVActive > pCRTSettingInfo->MonitorSizeV))
        {
            *pHActive = pCRTSettingInfo->MonitorSizeH;
            *pVActive = pCRTSettingInfo->MonitorSizeV;
        }

        pCRTSettingInfo->IsPanning = TRUE;
    }
    else if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
    {
        /* If use Virtual Desktop, and the size is larger than the resolution, */
        /* we need to do panning, too. */
        pCRTSettingInfo->IsPanning = TRUE;
    }
    else
    {
        pCRTSettingInfo->IsPanning = FALSE;
    }
    
}


void VIAAdjustActiveSize_DVI(VIABIOSInfoPtr pBIOSInfo, TMDSSETTINGINFOPTR pTMDSSettingInfo, int *pHActive, int *pVActive)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;

    if ((pTMDSSettingInfo->PanelSizeH == 0) &&
        (pTMDSSettingInfo->PanelSizeV == 0) &&
        (pTMDSSettingInfo->PanelSizeH != 0xFFF))
    {
        VIAGetModeSizeByIndex(pTMDSSettingInfo->DFPSize,
                             &(pTMDSSettingInfo->PanelSizeH),
                             &(pTMDSSettingInfo->PanelSizeV));
    }
    
    if ((HResToSet > pTMDSSettingInfo->PanelSizeH) ||
        (VResToSet > pTMDSSettingInfo->PanelSizeV))
    {        
        if ((*pHActive > pTMDSSettingInfo->PanelSizeH) ||
            (*pVActive > pTMDSSettingInfo->PanelSizeV))
        {            
            /* Panning */
            *pHActive = pTMDSSettingInfo->PanelSizeH;
            *pVActive = pTMDSSettingInfo->PanelSizeV;
            pTMDSSettingInfo->IsPanning = TRUE;
            pTMDSSettingInfo->IsDownScaling = FALSE;
        }            
    }
    else if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
    {
        /* If use Virtual Desktop, and the size is larger than the resolution, */
        /* we need to do panning, too. */
        pTMDSSettingInfo->IsPanning = TRUE;
    }
    else
    {
        pTMDSSettingInfo->IsPanning = FALSE;
    }

}

void VIAAdjustActiveSize(VIABIOSInfoPtr pBIOSInfo)
{
    int     HActive = pBIOSInfo->CrtcHDisplay;
    int     VActive = pBIOSInfo->CrtcVDisplay;    

    /* Try to find a suitable view size, we should use the minimum one if 
       more than one device */
    /* need to panning */
    
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV)
        VIAAdjustActiveSize_TV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo),
                               &HActive, &VActive, VIA_DEVICE_TV);

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2)
        VIAAdjustActiveSize_TV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo2),
                               &HActive, &VActive, VIA_DEVICE_TV2);

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT1)    
        VIAAdjustActiveSize_CRT(pBIOSInfo, &(pBIOSInfo->CRTSettingInfo),
                                &HActive, &VActive);

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT2)    
        VIAAdjustActiveSize_CRT(pBIOSInfo, &(pBIOSInfo->CRTSettingInfo2),
                                &HActive, &VActive);
    
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DFP)
        VIAAdjustActiveSize_DVI(pBIOSInfo, &(pBIOSInfo->TMDSSettingInfo),
                                &HActive, &VActive);    

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DFP2)
        VIAAdjustActiveSize_DVI(pBIOSInfo, &(pBIOSInfo->TMDSSettingInfo2),
                                &HActive, &VActive); 
    
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD)
        VIAAdjustActiveSize_LCD(pBIOSInfo, &(pBIOSInfo->LVDSSettingInfo),
                                &HActive, &VActive, VIA_DEVICE_LCD);

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD2)
        VIAAdjustActiveSize_LCD(pBIOSInfo, &(pBIOSInfo->LVDSSettingInfo2),
                                &HActive, &VActive, VIA_DEVICE_LCD2);

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI)
        VIAAdjustActiveSize_HDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo),
                                 &HActive, &VActive, VIA_DEVICE_HDMI);    

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI2)
        VIAAdjustActiveSize_HDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo2),
                                 &HActive, &VActive, VIA_DEVICE_HDMI2);
	
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP)
        VIAAdjustActiveSize_DP(pBIOSInfo, &(pBIOSInfo->DPSettingInfo),
                                &HActive, &VActive); 

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP2)
        VIAAdjustActiveSize_DP(pBIOSInfo, &(pBIOSInfo->DPSettingInfo2),
                                &HActive, &VActive); 

    /* pBIOSInfo->frameX1 pBIOSInfo->frameY1 always remain the actual view size
       whenever rotation or not */
    pBIOSInfo->frameX1 = HActive - 1;
    pBIOSInfo->frameY1 = VActive - 1;
    
    pBIOSInfo->HDisplay = HActive;
    pBIOSInfo->VDisplay = VActive;
}

void VIAInitSettingInfo_CRT(VIABIOSInfoPtr pBIOSInfo, CRTSETTINGINFOPTR pCRTSettingInfo, CARD16 activeDevice)
{
    int     HActive, VActive;
    Bool    IsPanning = pCRTSettingInfo->IsPanning;
	Bool    IsDownScaling = FALSE;

    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;

        xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "CRT Max Resolution %dx%d, set virtual desktop!!\n", 
            pCRTSettingInfo->MonitorSizeH, pCRTSettingInfo->MonitorSizeV);
    }
    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }

    /* support two crt, so we need update the relative iga path rather than the first crt's iga path */
    VIAInitSettingInfo_IGA(pBIOSInfo, pCRTSettingInfo->IGAPath, IsPanning, IsDownScaling, HActive, VActive, activeDevice);

    pCRTSettingInfo->ModeIndex = VIAGetModeIndex(HActive, VActive);
    pCRTSettingInfo->HActive = HActive;
    pCRTSettingInfo->VActive = VActive;    

    /* switch to interlace mode index. */
    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        switch (pCRTSettingInfo->ModeIndex)
        {
            case M1024x768:
            case M1152x864:
            case M1280x1024:
            case M1600x1200:
            case M1920x1080:
                pCRTSettingInfo->ModeIndex |= 0x200;
                break;
        }
    }

}


void VIAInitSettingInfo_DVI(VIABIOSInfoPtr pBIOSInfo, TMDSSETTINGINFOPTR pTMDSSettingInfo, CARD16 activeDevice)
{
    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAInitSettingInfo_DVI!!\n");
    
    int     HActive, VActive;
    Bool    IsPanning = pTMDSSettingInfo->IsPanning;
    Bool    IsDownScaling = pTMDSSettingInfo->IsDownScaling;
    
    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;

        xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DFP Max Resolution %lux%lu, set virtual desktop!!\n", 
            pTMDSSettingInfo->PanelSizeH, pTMDSSettingInfo->PanelSizeV);
    }
    else if (IsDownScaling)
    {
        HActive = pBIOSInfo->HorDownScaleSize;
        VActive = pBIOSInfo->VerDownScaleSize;
    }
    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }

    
 	VIAInitSettingInfo_IGA (pBIOSInfo, pTMDSSettingInfo->IGAPath, IsPanning, IsDownScaling, HActive, VActive, activeDevice);

    pTMDSSettingInfo->ModeIndex = VIAGetModeIndex(HActive, VActive);
    pTMDSSettingInfo->HActive = HActive;
    pTMDSSettingInfo->VActive = VActive;        
    pBIOSInfo->IsDownScaleEnable = IsDownScaling;        

    /* switch to interlace mode index. */
    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        switch (pTMDSSettingInfo->ModeIndex)
        {
            case M1024x768:
            case M1152x864:
            case M1280x1024:
            case M1600x1200:
            case M1920x1080:
                pTMDSSettingInfo->ModeIndex |= 0x200;
                break;
        }
    }
    else
    {
        /* Determine whether to use reduce blanking mode. */
        VIAReduceBlanking(pBIOSInfo, pTMDSSettingInfo->ModeIndex, activeDevice);
    }
}


void VIAInitSettingInfo_LCD(VIABIOSInfoPtr pBIOSInfo, LVDSSETTINGINFOPTR pLVDSSettingInfo,  CARD16 activeDevice)
{
    int     HActive, VActive;
    Bool    IsPanning = pLVDSSettingInfo->IsPanning;
    Bool    IsDownScaling = pLVDSSettingInfo->IsDownScaling;

    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;

        xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "LCD Max Resolution %dx%d, set virtual desktop!!\n", 
            pLVDSSettingInfo->PanelSizeX, pLVDSSettingInfo->PanelSizeY);
    }
    else if (IsDownScaling)
    {
        HActive = pBIOSInfo->HorDownScaleSize;
        VActive = pBIOSInfo->VerDownScaleSize;
    }

    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }

    VIAInitSettingInfo_IGA(pBIOSInfo, pLVDSSettingInfo->IGAPath, IsPanning, IsDownScaling, HActive, VActive, activeDevice); 
    

    pLVDSSettingInfo->ModeIndex = VIAGetModeIndex(HActive, VActive);
    pLVDSSettingInfo->HActive = HActive;
    pLVDSSettingInfo->VActive = VActive;        
    pBIOSInfo->IsDownScaleEnable = IsDownScaling;  
}

void VIAInitSettingInfo_DP(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo,  CARD16 activeDevice)
{
    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAInitSettingInfo_DP!!\n");
    
    int     HActive, VActive;
    Bool    IsPanning = pDPSettingInfo->IsPanning;
    Bool    IsDownScaling = pDPSettingInfo->IsDownScaling;
    
    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;

        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                   "HDMI Max Resolution %lux%lu, set virtual desktop!!\n", 
                   pDPSettingInfo->PanelSizeH,
                   pDPSettingInfo->PanelSizeV));
    }
    else if (IsDownScaling)
    {
        HActive = pBIOSInfo->HorDownScaleSize;
        VActive = pBIOSInfo->VerDownScaleSize;
    }
    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }
    
    VIAInitSettingInfo_IGA (pBIOSInfo, pDPSettingInfo->IGAPath,
                            IsPanning, IsDownScaling, HActive, VActive,
                            activeDevice);

    pDPSettingInfo->ModeIndex = VIAGetModeIndex(HActive, VActive);
    pDPSettingInfo->HActive = HActive;
    pDPSettingInfo->VActive = VActive;        
    pBIOSInfo->IsDownScaleEnable = IsDownScaling;        

    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        switch (pDPSettingInfo->ModeIndex)
        {
            case M1024x768:
            case M1152x864:
            case M1280x1024:
            case M1600x1200:
            case M1920x1080:              
                pDPSettingInfo->ModeIndex |= 0x200;
                break;
        }
    }
}

Bool VIAFindModeUseBIOSTable(ScrnInfoPtr pScrn)
{
    VIAPtr          pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    int             needRefresh;
    int             ModeIndex;
    Bool            IsFoundRefresh = FALSE;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAFindModeUseBIOSTable\n"));

    /* Initial down scaling variable. */
    pBIOSInfo->IsDownScaleEnable = FALSE;

    if (pBIOSInfo->VirtualX > 0)
    {
        /* If set Virtual Desktop Size in xorg.conf: */
        pBIOSInfo->ActualDesktopSizeX = pBIOSInfo->VirtualX;
        pBIOSInfo->ActualDesktopSizeY = pBIOSInfo->VirtualY;
    }
    else
    {
        pBIOSInfo->ActualDesktopSizeX = pBIOSInfo->CrtcHDisplay;
        pBIOSInfo->ActualDesktopSizeY = pBIOSInfo->CrtcVDisplay;
    }

    /* Initial device's setting information: */
    VIAAdjustActiveSize(pBIOSInfo);
                        
    /* Find Resolution index: */
    ModeIndex = VIAGetModeIndex(pBIOSInfo->CrtcHDisplay, pBIOSInfo->CrtcVDisplay);

    /* Find suitable refresh rate: */
    /* Set the refresh rate to 60MHz for DuoView. */
    /* Note: We should support other refresh rate later. */
    if (/*pBIOSInfo->DuoView ||*/ 
        !(pBIOSInfo->ActiveDevice & (VIA_DEVICE_LCD | VIA_DEVICE_LCD2 | VIA_DEVICE_TV)))
    {
        if (pBIOSInfo->OptRefresh)
        {
            pBIOSInfo->Refresh = pBIOSInfo->OptRefresh;           
            IsFoundRefresh = VIAFindRefreshRate(pBIOSInfo, ModeIndex);            
        }
        else if (pBIOSInfo->ActiveDevice & (VIA_DEVICE_DFP |VIA_DEVICE_DFP2
                            | VIA_DEVICE_HDMI | VIA_DEVICE_HDMI2 | VIA_DEVICE_DP | VIA_DEVICE_DP2))
        {
            /* DVI will reference the CRT's capability to 
               determine the refresh rate. But the most DVI cannot support the
               higher refresh rate. I consider that we should to use option to 
               set refresh rate if we want to set the higher refresh rate for DVI.
               Otherwise the default refresh rate of DVI is 60Hz.
            */
            pBIOSInfo->FoundRefresh = 60;
            IsFoundRefresh = TRUE;
        }
        
        if (!IsFoundRefresh)
        {
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_WARNING, "Refresh rate setting in xorg.conf is not supported!!\n");
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_WARNING, "Driver will try to find another refresh rate instead.\n");
                
            /* use the monitor information */
            /* needRefresh = (pBIOSInfo->Clock * 1000) / (pBIOSInfo->HTotal * pBIOSInfo->VTotal); */
            /* Do rounding */
            needRefresh = ((pBIOSInfo->Clock * 10000) / (pBIOSInfo->HTotal * pBIOSInfo->VTotal) + 5) / 10;
            pBIOSInfo->Refresh = needRefresh;            
            IsFoundRefresh = VIAFindRefreshRate(pBIOSInfo, ModeIndex);            
        }
        
        if (!IsFoundRefresh)
        {
            pBIOSInfo->FoundRefresh = 60;
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_WARNING, "Can't find suitable refresh rate, use 60Hz as default.\n");
        }
    }
    else
    {
        pBIOSInfo->FoundRefresh = 60;
    }
    
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT1)
    {        
        VIAInitSettingInfo_CRT(pBIOSInfo, &(pBIOSInfo->CRTSettingInfo), VIA_DEVICE_CRT1 );
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT2)
    {        
        VIAInitSettingInfo_CRT(pBIOSInfo, &(pBIOSInfo->CRTSettingInfo2), VIA_DEVICE_CRT2 );
    }
    
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DFP)
    {
        VIAInitSettingInfo_DVI(pBIOSInfo, &(pBIOSInfo->TMDSSettingInfo), VIA_DEVICE_DFP);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DFP2)
    {
        VIAInitSettingInfo_DVI(pBIOSInfo, &(pBIOSInfo->TMDSSettingInfo2), VIA_DEVICE_DFP2);
    }
        
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD)
    {
        VIAInitSettingInfo_LCD(pBIOSInfo, &(pBIOSInfo->LVDSSettingInfo), VIA_DEVICE_LCD);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI)
    {
        VIAInitSettingInfo_HDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo), VIA_DEVICE_HDMI);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI2)
    {
        VIAInitSettingInfo_HDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo2), VIA_DEVICE_HDMI2);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP)
    {
        VIAInitSettingInfo_DP(pBIOSInfo, &(pBIOSInfo->DPSettingInfo), VIA_DEVICE_DP);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP2)
    {
        VIAInitSettingInfo_DP(pBIOSInfo, &(pBIOSInfo->DPSettingInfo2), VIA_DEVICE_DP2);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD2)
    {
        VIAInitSettingInfo_LCD(pBIOSInfo, &(pBIOSInfo->LVDSSettingInfo2), VIA_DEVICE_LCD2);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV)
    {
        VIAInitSettingInfo_TV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo), VIA_DEVICE_TV);
    }

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2)
    {
        VIAInitSettingInfo_TV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo2), VIA_DEVICE_TV2);
    }
    
    /* pBIOSInfo->widthByQWord = (pBIOSInfo->displayWidth * (pBIOSInfo->bitsPerPixel >> 3)) >> 3; */
    pBIOSInfo->offsetWidthByQWord = (pBIOSInfo->displayWidth * (pBIOSInfo->bitsPerPixel >> 3)) >> 3;
    pBIOSInfo->countWidthByQWord = (pBIOSInfo->CrtcHDisplay * (pBIOSInfo->bitsPerPixel >> 3)) >> 3;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "pBIOSInfo->FoundRefresh: %d\n", pBIOSInfo->FoundRefresh));
    return TRUE;
}


Bool VIARestoreGamma(ScrnInfoPtr pScrn, LOCO *colors) {
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    int i, sr1a, cr6a, sr1b, sr2d;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIARestorePalette\n"));
    /* 8 bpp mode can't adjust gamma */
    if (pScrn->bitsPerPixel == 8)
        return FALSE;

    /* Note:
       VT3409 has a hardware issue, when we enable IGA1 gamma, the display color
       will become bad if playing video.
       Solution: disable IGA1 gamma on VT3409 (16/32bpp)
    */
    if (pBIOSInfo->Chipset != VIA_VX855)
    {
        /* Enable IGA1 Gamma */
        write_reg_mask(CR33, VIACR, 0x80, BIT7);
    }
    
    /* 3C6/3C7/3C8/3C9 map to IGA1 gamma */
    sr1a = read_reg(VIASR, SR1A);
    write_reg_mask(SR1A, VIASR, 0x0, BIT0);                    

    sr1b = read_reg(VIASR, SR1B);
    write_reg_mask(SR1B, VIASR, 0xC0, BIT6+BIT7);  

    sr2d = read_reg(VIASR, SR2D);
    write_reg_mask(SR2D, VIASR, 0x0C, BIT2+BIT3);
            
    VGAOUT8(0x3c8, 0);
    for (i = 0; i < 256; i++)
    {
        VGAOUT8(0x3c9, colors[i].red);
        VGAOUT8(0x3c9, colors[i].green);
        VGAOUT8(0x3c9, colors[i].blue);
    }    

    /* 3C6/3C7/3C8/3C9 map to IGA2 gamma */
    write_reg_mask(SR1A, VIASR, 0x01, BIT0);

    /* Enable IGA2 Gamma */   
    write_reg_mask(CR6A, VIACR, 0x02, BIT1);

    /* We should set the format of IGA2's LUT to 8-BITs, because we use 8-BITs LUT except VT3123Ax. */
    /* This register added in VT3314 or later chipsets. */
    write_reg_mask(CR6A, VIACR, BIT5, BIT5);

    /* 
     Situation:
     On VT3324, when we switch to two devices, the second LUT maybe fault.
     Solution:
     Before we fill the second LUT, we have to enable second display channel.
    */
    /* Save */
    VGAOUT8(0x3D4, CR6A);
    cr6a = VGAIN8(0x3D5);
        
    VGAOUT8(0x3D4, CR6A);           
    if(!(VGAIN8(0x3D5) & 0x80))
    {                
        EnableSecondDisplayChannel();
    }

    /* Fill Second LUT */
    VGAOUT8(0x3c8, 0);
    for (i = 0; i < 256; i++)
    {
        VGAOUT8(0x3c9, colors[i].red);
        VGAOUT8(0x3c9, colors[i].green);
        VGAOUT8(0x3c9, colors[i].blue);
    }

    /* After filled second LUT, we should restore to previous value. */
    VGAOUT8(0x3D4, CR6A);
    VGAOUT8(0x3D5, cr6a);   

    write_reg(SR1A, VIASR, sr1a);
    write_reg(SR1B, VIASR, sr1b);
    write_reg(SR2D, VIASR, sr2d);
    return TRUE;
}


/* query NB to get frame buffer size */
int VIAGetFBSize(ScrnInfoPtr pScrn)
{
    unsigned long   configid,deviceid,FBSize = 0;
    int   VideoMemSize;
    int   DeviceFound = FALSE;
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    
    for ( configid = 0x80000000; configid < 0x80010800; configid += 0x100 )
    {
        outl((unsigned long)0xCF8,configid);
        deviceid = (inl((unsigned long)0xCFC)>>16) & 0xffff;

        /* Add KM800 case*/
        switch (deviceid)
        {
            case CN700_FUNCTION3:
            case CX700_FUNCTION3:
            case KM890_FUNCTION3:
            case P4M890_FUNCTION3:
            case P4M900_FUNCTION3:
            case VX800_FUNCTION3:
            case VX855_FUNCTION3:
            case VX900_FUNCTION3:
                outl((unsigned long)0xCF8,configid+0xA0);
                FBSize = inl((unsigned long)0xCFC);
                DeviceFound = TRUE; /* Found device id */
                break;
                
            /* Can't set default, otherwise FBSize will be modified incorrectly */    
            default:    
                break;
         }

         if (DeviceFound)
         {
            break; /* Found device id, so exit for loop */
         }
    }

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Device ID = %lx\n", deviceid));
    
    FBSize = FBSize & 0x00007000;
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "FB Size = %lx\n", FBSize));
    
    if (pBIOSInfo->Chipset < VIA_CX700)
    {
        switch(FBSize)
        {
            case 0x00004000:
                VideoMemSize = (16 << 10);  /*16M*/              
                break;

            case 0x00005000:
                VideoMemSize = (32 << 10);  /*32M*/                
                break;

            case 0x00006000:
                VideoMemSize = (64 << 10);  /*64M*/                
                break;

            default:
                VideoMemSize = (32 << 10);  /*32M*/
                break;     
        }
    }
    else
    {
        switch(FBSize)
        {
            case 0x00001000:
                VideoMemSize = (8 << 10);   /*8M*/                
                break;

            case 0x00002000:
                VideoMemSize = (16 << 10);  /*16M*/               
                break;

            case 0x00003000:
                VideoMemSize = (32 << 10);  /*32M*/                
                break;

            case 0x00004000:
                VideoMemSize = (64 << 10);  /*64M*/               
                break;

            case 0x00005000:
                VideoMemSize = (128 << 10); /*128M*/
                break;

            case 0x00006000:
                VideoMemSize = (256 << 10); /*256M*/
                break;

            case 0x00007000:
                VideoMemSize = (512 << 10); /*512M*/
                break;

            default:
                VideoMemSize = (32 << 10);  /*32M*/                
                break;
        }
    }

    pScrn->videoRam = VideoMemSize;

    return VideoMemSize;
}


void VIADelayIn_usec(VIABIOSInfoPtr pBIOSInfo, int usec)
{
    VIABIOSInfoPtr  pVia = pBIOSInfo;
    volatile long   i;
    unsigned char   tmp;

    if (usec > 0)
    {
        for (i = usec * 4; i > 0; i--)
        {
            tmp = VGAIN8(0x3c3);
        }
    }
}


/* CRT DAC Sense */
Bool VIASenseCRT(ScrnInfoPtr pScrn)
{    
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;    
    
    CARD8   sr2d, sr1b, sr01, cr36, sr40;
    Bool    IsAttch = FALSE;

    /* Save Value */
    sr2d = read_reg(VIASR, SR2D);
    sr1b = read_reg(VIASR, SR1B);
    sr01 = read_reg(VIASR, SR01);
    cr36 = read_reg(VIACR, CR36);
    sr40 = read_reg(VIASR, SR40);
    
    write_reg_mask(SR2D, VIASR, BIT4+BIT5, BIT4+BIT5);      /* Turn on PLL of VCK */
    write_reg_mask(SR1B, VIASR, BIT4+BIT5, BIT4+BIT5);      /* Turn on VCK */
    write_reg_mask(SR01, VIASR, 0x0, BIT5);                 /* Screen On */
    write_reg_mask(CR36, VIACR, 0x0, BIT4+BIT5+BIT6+BIT7);  /* Power On DPMS */                        

    /* This method could be fail. So I add delay. */
    while(!(INREG(0x200) & BIT1));                          /* Delay for V-sync */
    while(INREG(0x200) & BIT1);                             /* 1: Display, 0:VerticalBlank*/

    VIADelay_Nmsec(pBIOSInfo, 16);
    
    write_reg_mask(SR40, VIASR, BIT7, BIT7);                /* Enable CRT Sense */

    /* Read Status Bit */
    if (VGAIN8(0x3C2) & BIT4)
    {
        IsAttch = TRUE;        
    }
    else
    {
        IsAttch = FALSE;        
    }

    /* Restore */
    write_reg(SR40, VIASR, sr40);
    write_reg(CR36, VIASR, cr36);
    write_reg(SR01, VIASR, sr01);
    write_reg(SR1B, VIASR, sr1b);
    write_reg(SR2D, VIASR, sr2d);
    
    return IsAttch;
}


void VIASetOutputPath(VIABIOSInfoPtr pBIOSInfo, int device, int set_iga, int output_interface)
{
    switch(device)
    {
        case VIA_DEVICE_CRT1:
            VIASetCRTOutputPath(pBIOSInfo, set_iga, output_interface);
            break;
        case VIA_DEVICE_TV:
            VIASetTVOutputPath(pBIOSInfo, set_iga, output_interface);
            break;
        case VIA_DEVICE_DFP:
        case VIA_DEVICE_DFP2:
            VIASetDVIOutputPath(pBIOSInfo, set_iga, output_interface);
            break;
        case VIA_DEVICE_LCD:
            VIASetLCDOutputPath(pBIOSInfo, set_iga, output_interface);
            break;
        case VIA_DEVICE_HDMI:
            viaSetHDMIOutputPath(pBIOSInfo, set_iga, output_interface);
            break;        
    }
}


void VIASetCRTOutputPath(VIABIOSInfoPtr pBIOSInfo, int set_iga, int output_interface)
{
    switch(set_iga)
    {
        case IGA1:
            /* normal crt */
            if(!output_interface)
            {
                write_reg_mask(SR16, VIASR, 0x00, BIT6);
            }            
            break;
            
        case IGA2:
            /* Modify for sequence issue. */
            EnableSecondDisplayChannel();
            /* normal crt */
            if(!output_interface)
            {
                write_reg_mask(SR16, VIASR, 0x40, BIT6);
            }           
            break;
    }

    if(!output_interface)
    {
        return ;
    }

    /* We use VT1625 to output CRT2 signal, so we have to power on the digital port. */
    /*Set Iga source and turn on DI port clk*/
    switch(output_interface)
    {
        case VIA_DI_DVP0:
            if (set_iga == IGA1)
            {
                /* DVP0 Data Source Selection.*/
                write_reg_mask(CR96, VIACR, 0x00, BIT4);
            }
            else
            {
                /* DVP0 Data Source Selection.*/
                write_reg_mask(CR96, VIACR, 0x10, BIT4);
            }	
            /* enable dvp0 under CX700 */
            if (pBIOSInfo->Chipset == VIA_CX700)
            {
                write_reg_mask(CR91, VIACR, 0x00, BIT5);
            }
            
            /*Turn on DVP0 clk*/
            write_reg_mask(SR1E, VIASR, 0xc0, BIT6+BIT7);
            break;

        case VIA_DI_DVP1:
            if (set_iga == IGA1)
            {
                write_reg_mask(CR9B, VIACR, 0x00, BIT4);
            }
            else
            {
                write_reg_mask(CR9B, VIACR, 0x10, BIT4);
            }
            /* enable dvp1 under this chipset */
            if (pBIOSInfo->Chipset == VIA_CX700 ||
                pBIOSInfo->Chipset == VIA_VX800 ||
                pBIOSInfo->Chipset == VIA_VX855 ||
                pBIOSInfo->Chipset == VIA_VX900)
            {
                write_reg_mask(CRD3, VIACR, 0x00, BIT5);
            }
            /*Turn on DVP1 clk*/
            write_reg_mask(SR1E, VIASR, 0x30, BIT4+BIT5);
            break;
            
        case VIA_DI_DFPHIGH:
            if (set_iga == IGA1)
            {
                /*3324,3353 CR96[4] control DVP0*/
                if ((pBIOSInfo->Chipset == VIA_CX700 ) ||
                    (pBIOSInfo->Chipset == VIA_VX800 ) || 
                    (pBIOSInfo->Chipset == VIA_VX855) ||
                    (pBIOSInfo->Chipset == VIA_VX900))
                {
                    write_reg_mask(CR96, VIACR, 0x00, BIT4);
                }
                write_reg_mask(CR97, VIACR, 0x00, BIT4);
            }
            else
            {
                /*3324,3353 CR96[4] control DVP0*/
                if ((pBIOSInfo->Chipset == VIA_CX700 ) ||
                    (pBIOSInfo->Chipset == VIA_VX800 ) ||
                    (pBIOSInfo->Chipset == VIA_VX855) ||
                    (pBIOSInfo->Chipset == VIA_VX900))
                {
                    write_reg_mask(CR96, VIACR, 0x10, BIT4);
                }
                write_reg_mask(CR97, VIACR, 0x10, BIT4);
            }

            /*Turn on DFPH clk*/
            write_reg_mask(SR2A, VIASR, 0x0C, BIT2+BIT3);
            break;
            
        case VIA_DI_DFPLOW:
            if (set_iga == IGA1)
            {
                /*3324,3353 CR9B[4] control DVP1*/
                if ((pBIOSInfo->Chipset == VIA_CX700 ) ||
                    (pBIOSInfo->Chipset == VIA_VX800 ) ||
                    (pBIOSInfo->Chipset == VIA_VX855) ||
                    (pBIOSInfo->Chipset == VIA_VX900))
                {
                    write_reg_mask(CR9B, VIACR, 0x00, BIT4);
                }
                write_reg_mask(CR99, VIACR, 0x00, BIT4);
            }
            else
            {
                /*3324,3353 CR9B[4] control DVP1*/
                if ((pBIOSInfo->Chipset == VIA_CX700 ) ||
                    (pBIOSInfo->Chipset == VIA_VX800 ) ||
                    (pBIOSInfo->Chipset == VIA_VX855) ||
                    (pBIOSInfo->Chipset == VIA_VX900))
                {
                    write_reg_mask(CR9B, VIACR, 0x10, BIT4);
                }
                write_reg_mask(CR99, VIACR, 0x10, BIT4);
            }

            /*Turn on DFPL clk*/
            write_reg_mask(SR2A, VIASR, 0x03, BIT1+BIT0);
            break;
    }

    /*If Iga2 is used, enable Iga2 output for DI port. In order to enable Iga2, 3324,3353,3293 use CR91 and CRD3 to control
    DVP0 and DVP1 seperately, but other chip only use CR91 to control all Di port*/
    if(IGA2 == set_iga && pBIOSInfo->Chipset != VIA_CX700)
    {
        write_reg_mask(CR91, VIACR, 0x00, BIT5);
    }    

}

void VIASetDisplayChannel(VIABIOSInfoPtr pBIOSInfo)
{
	ScrnInfoPtr     pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr          pVia = VIAPTR(pScrn);
    BOOL            lcd_is_internal = FALSE;
    BOOL            lcd2_is_internal = FALSE;
    BOOL            tmds_is_internal = FALSE;
    BOOL            lcd_is_dualedge = FALSE;
    BOOL            lcd2_is_dualedge = FALSE;
    
    /* SAMM */
    if (pVia->pVIAEnt->HasSecondary)
    {
        ScrnInfoPtr pScrn0 = xf86Screens[0];
        VIAPtr pVia0 = VIAPTR(pScrn0);
        VIABIOSInfoPtr  pBIOSInfo0 = pVia0->pBIOSInfo;

        ScrnInfoPtr pScrn1 = xf86Screens[1];
        VIAPtr pVia1 = VIAPTR(pScrn1);
        VIABIOSInfoPtr  pBIOSInfo1 = pVia1->pBIOSInfo;

        lcd_is_internal = ((pBIOSInfo0->ActualActiveDevice & VIA_DEVICE_LCD) && 
                              (pBIOSInfo0->LVDSSettingInfo.ChipID == VIA_INTEGRATED_LVDS)) ||
                          ((pBIOSInfo1->ActualActiveDevice & VIA_DEVICE_LCD) && 
                              (pBIOSInfo1->LVDSSettingInfo.ChipID == VIA_INTEGRATED_LVDS));
        lcd2_is_internal = ((pBIOSInfo0->ActualActiveDevice & VIA_DEVICE_LCD2) &&
                              (pBIOSInfo0->LVDSSettingInfo2.ChipID == VIA_INTEGRATED_LVDS)) ||
                           ((pBIOSInfo1->ActualActiveDevice & VIA_DEVICE_LCD2) &&
                              (pBIOSInfo1->LVDSSettingInfo2.ChipID == VIA_INTEGRATED_LVDS));
        tmds_is_internal = ((pBIOSInfo0->ActualActiveDevice & VIA_DEVICE_DFP2) &&
                              (pBIOSInfo0->TMDSSettingInfo2.ChipID == VIA_INTEGRATED_TMDS)) ||
                           ((pBIOSInfo1->ActualActiveDevice & VIA_DEVICE_DFP2) &&
                              (pBIOSInfo1->TMDSSettingInfo2.ChipID == VIA_INTEGRATED_TMDS));
        lcd_is_dualedge = pBIOSInfo0->LVDSSettingInfo.IsDualEdge ||
                          pBIOSInfo1->LVDSSettingInfo.IsDualEdge;
        lcd2_is_dualedge = pBIOSInfo0->LVDSSettingInfo2.IsDualEdge ||
                           pBIOSInfo1->LVDSSettingInfo2.IsDualEdge;                   

    }
    /* Duoview or Single */
    else
    {
        lcd_is_internal = (pBIOSInfo->ActualActiveDevice & VIA_DEVICE_LCD) && 
                           (pBIOSInfo->LVDSSettingInfo.ChipID == VIA_INTEGRATED_LVDS);
        lcd2_is_internal = (pBIOSInfo->ActualActiveDevice & VIA_DEVICE_LCD2) &&
                            (pBIOSInfo->LVDSSettingInfo2.ChipID == VIA_INTEGRATED_LVDS);
        tmds_is_internal = (pBIOSInfo->ActualActiveDevice & VIA_DEVICE_DFP2) &&
                            (pBIOSInfo->TMDSSettingInfo2.ChipID == VIA_INTEGRATED_TMDS);
        lcd_is_dualedge = pBIOSInfo->LVDSSettingInfo.IsDualEdge;
        lcd2_is_dualedge = pBIOSInfo->LVDSSettingInfo2.IsDualEdge;
    }

    if ((lcd_is_internal && lcd_is_dualedge) ||
        (lcd2_is_internal && lcd2_is_dualedge))
    {
        /* For dual channel LCD: */
        /* Set to Dual LVDS channel. */
        write_reg_mask(CRD2, VIACR, 0x20, BIT4+BIT5);
    }
    else if (tmds_is_internal)
    {
        if (lcd_is_internal || lcd2_is_internal)
        {
            /* For Internal LCD + Internal DFP: */
            /* Set to LVDS1 + TMDS channel. */
            write_reg_mask(CRD2, VIACR, 0x10, BIT4+BIT5);
        }
        else
        {
            /* Set to single TMDS channel. */
            write_reg_mask(CRD2, VIACR, 0x30, BIT4+BIT5);   
        }
    }
    else
    {
        /* Set to LVDS0 + LVDS1 channel. */
        write_reg_mask(CRD2, VIACR, 0x00, BIT4+BIT5);
    }   
}

/*
    If monitor's max support resolution is a strange mode that
    we don't support, we have to correct it. Find a new mode that we support to replace it.

*/
Bool VIACorrectMonitorSize(int* MonitorH, int* MonitorV)
{   
    int i = 0, MaxH = 0, MaxV = 0;;

    while (VIASupportMode[i].ModeIndex > -1)
    {
        if (*MonitorH < VIASupportMode[i].HActive)
        {
            i--;
            break;
        }
        else if ((*MonitorH == VIASupportMode[i].HActive)
              && (*MonitorV < VIASupportMode[i].VActive))
        {
            i--;
            break;
        }
        else if ((*MonitorH == VIASupportMode[i].HActive)
              && (*MonitorV == VIASupportMode[i].VActive))
        {
            break;
        }
        else
        {
            if (MaxH < VIASupportMode[i].HActive)
            {
                MaxH = VIASupportMode[i].HActive;
                MaxV = VIASupportMode[i].VActive;
            }
            else if (MaxH == VIASupportMode[i].HActive)
            {
                if (MaxV < VIASupportMode[i].VActive)
                {
                    MaxV = VIASupportMode[i].VActive;
                }
            }
            i++;
        }
    }
    
    if (i < 0)    /* For too small size */
    {
        return FALSE;
    }
    else if (VIASupportMode[i].ModeIndex < 0)    /* For size larger than resolutions supported in table */
    {
        *MonitorH = MaxH;
        *MonitorV = MaxV;
    }
    else
    {
        *MonitorH = VIASupportMode[i].HActive;
        *MonitorV = VIASupportMode[i].VActive;        
        return TRUE;
    }
}


/*
*   Attach a new mode to the modeline.
*   So we don't need to add modeline in the config file.
*/
void ViaModesAttachHelper(ScrnInfoPtr pScrn, MonPtr monitorp, DisplayModePtr Modes)
{
    DisplayModePtr mode;
    DisplayModePtr last = monitorp->Last;
    int i;

    for (i = 0; Modes[i].name; i++)
    {        
        mode = xnfalloc(sizeof(DisplayModeRec));
        memcpy(mode, &Modes[i], sizeof(DisplayModeRec));
        mode->name = xnfstrdup(Modes[i].name);
        
        if (last)
        {
            mode->prev = last;
            last->next = mode;
        }
        else
        { 
            /* this is the first mode */
            monitorp->Modes = mode;
            mode->prev = NULL;
        }
        last = mode;
    }
    
    monitorp->Last = last;
}


Bool VIAEnumModes(VIABIOSInfoPtr pBIOSInfo, int IndexOfTable, VIASUPPORTMODE *pVIASupportMode, Bool *pIsSupported)
{
    int ItemNum;
    
    ItemNum = sizeof(VIASupportMode) / sizeof(VIASUPPORTMODE);

    *pIsSupported = FALSE;

    if (IndexOfTable < ItemNum)
    {
        memcpy(pVIASupportMode, &VIASupportMode[IndexOfTable], sizeof(VIASUPPORTMODE));

        /* Validate for TV: */
        if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV)
        {
            if (IsTVMode(pBIOSInfo, pVIASupportMode->ModeIndex, &(pBIOSInfo->TVSettingInfo)))
            {
                if ((pVIASupportMode->ModeIndex == M720x480) &&
                    ((pBIOSInfo->TVSettingInfo.TVType == TVTYPE_PAL) || (pBIOSInfo->TVSettingInfo.TVType == TVTYPE_576P)))                    
                {
                    *pIsSupported = FALSE;
                }
                else if ((pVIASupportMode->ModeIndex == M720x576) &&
                         ((pBIOSInfo->TVSettingInfo.TVType == TVTYPE_NTSC) || (pBIOSInfo->TVSettingInfo.TVType == TVTYPE_480P)))
                {
                    *pIsSupported = FALSE;
                }
                else if ((pVIASupportMode->ModeIndex == M1280x720) && (pBIOSInfo->TVSettingInfo.TVType != TVTYPE_720P))
                {
                    *pIsSupported = FALSE;
                }
                else if ((pVIASupportMode->ModeIndex == M1920x1080) && (pBIOSInfo->TVSettingInfo.TVType != TVTYPE_1080I))
                {
                    *pIsSupported = FALSE;
                }
                else
                {
                    *pIsSupported = TRUE;
                }
            }
        }
        
        /* Validate for CRT: */
        if ((pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT1) || (pBIOSInfo->ActiveDevice & VIA_DEVICE_CRT2))
        {
            if ((pBIOSInfo->CRTSettingInfo.MonitorSizeH >= pVIASupportMode->HActive) &&
                (pBIOSInfo->CRTSettingInfo.MonitorSizeV >= pVIASupportMode->VActive))
            {
                *pIsSupported = TRUE;
            }
        }

        /* Validate for DVI: */
        if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DFP)
        {
            if ((pBIOSInfo->TMDSSettingInfo.PanelSizeH >= pVIASupportMode->HActive) &&
                (pBIOSInfo->TMDSSettingInfo.PanelSizeV >= pVIASupportMode->VActive))
            {
                *pIsSupported = TRUE;
            }
        }

        /* Validate for LCD: */
        if (pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD)
        {
            if ((pBIOSInfo->LVDSSettingInfo.PanelSizeX >= pVIASupportMode->HActive) &&
                (pBIOSInfo->LVDSSettingInfo.PanelSizeY >= pVIASupportMode->VActive))
            {
                *pIsSupported = TRUE;
            }
        }

        /* Don't support if mode size is larger than virtual size. */
        if (pVIASupportMode->HActive >= pBIOSInfo->SaveVirtualX)
        {
            if (pVIASupportMode->VActive > pBIOSInfo->SaveVirtualY)
            {
                *pIsSupported = FALSE;
            }
        }
        
        return TRUE;
    }

    return FALSE;
}


void VIAEnumRefreshRate(int ModeIndex, int RefreshRateNum, int *pRefreshRate)
{
    struct VideoModeTable   *VideoMode;
    
    VideoMode = &VIARoutineModes[SearchModeSetting(ModeIndex)];

    if (RefreshRateNum < VideoMode->mode_array)
    {
        *pRefreshRate = VideoMode->crtc[RefreshRateNum].refresh_rate;
    }
    else
    {
        *pRefreshRate = 0;
    }
}

/*
*   Purpose: Set DPA value to register.
*/
void VIASetDPA_Gfx(int DIPort, GFX_DPA_VALUE_PTR pGfxDPASetting)
{
    DEBUG(ErrorF("VIASetDPA_Gfx, Use Port %x.\n", DIPort));
    
    switch (DIPort)
    {
        case VIA_DI_DVP0:
        {
            /* DVP0 Clock Polarity and Adjust: */
            write_reg_mask(CR96, VIACR, pGfxDPASetting->DVP0, 0x0F);

            /* DVP0 Clock and Data Pads Driving: */
            write_reg_mask(SR1E, VIASR, pGfxDPASetting->DVP0ClockDri_S, BIT2);
            write_reg_mask(SR2A, VIASR, pGfxDPASetting->DVP0ClockDri_S1, BIT4);
            write_reg_mask(SR1B, VIASR, pGfxDPASetting->DVP0DataDri_S, BIT1);
            write_reg_mask(SR2A, VIASR, pGfxDPASetting->DVP0DataDri_S1, BIT5);

            DEBUG(ErrorF("DVP0 = %x.\n", pGfxDPASetting->DVP0));
            DEBUG(ErrorF("DVP0ClockDri_S = %x.\n", pGfxDPASetting->DVP0ClockDri_S));
            DEBUG(ErrorF("DVP0ClockDri_S1 = %x.\n", pGfxDPASetting->DVP0ClockDri_S1));
            DEBUG(ErrorF("DVP0DataDri_S = %x.\n", pGfxDPASetting->DVP0DataDri_S));
            DEBUG(ErrorF("DVP0DataDri_S1 = %x.\n", pGfxDPASetting->DVP0DataDri_S1));
            break;
        }

        case VIA_DI_DVP1:
        {
            /* DVP1 Clock Polarity and Adjust: */
            write_reg_mask(CR9B, VIACR, pGfxDPASetting->DVP1, 0x0F);

            /* DVP1 Clock and Data Pads Driving: */
            write_reg_mask(SR65, VIASR, pGfxDPASetting->DVP1Driving, 0x0F);
            
            DEBUG(ErrorF("DVP1 = %x.\n", pGfxDPASetting->DVP1));
            DEBUG(ErrorF("DVP1Driving = %x.\n", pGfxDPASetting->DVP1Driving));
            break;
        }

        case VIA_DI_DFPHIGH:
        {
            write_reg_mask(CR97, VIACR, pGfxDPASetting->DFPHigh, 0x0F);
            DEBUG(ErrorF("DFPHigh = %x.\n", pGfxDPASetting->DFPHigh));
            break;
        }

        case VIA_DI_DFPLOW:
        {
            write_reg_mask(CR99, VIACR, pGfxDPASetting->DFPLow, 0x0F);
            DEBUG(ErrorF("DFPLow = %x.\n", pGfxDPASetting->DFPLow));
            break;
        }

        case VIA_DI_DFPHIGHLOW:
        {
            write_reg_mask(CR97, VIACR, pGfxDPASetting->DFPHigh, 0x0F);
            write_reg_mask(CR99, VIACR, pGfxDPASetting->DFPLow, 0x0F);
            DEBUG(ErrorF("DFPHigh = %x.\n", pGfxDPASetting->DFPHigh));
            DEBUG(ErrorF("DFPLow = %x.\n", pGfxDPASetting->DFPLow));
            break;
        }
    }
}


Bool
VIACollectBIOSInfo(ScrnInfoPtr pScrn)
{
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIACollectBIOSInfo!\n"));

    VIAPtr          pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    int tmp;

    /* Read PCI configuration space to get frame buffer size. */
    VIAGetFBSize(pScrn);

    /* Check if spread spectrum is enable */
    VGAOUT8(0x3c4, SR1E);
    if (VGAIN8(0x3c5) & BIT3) {
        pVia->pChipInfo->biosIsSpreadSpectrum = TRUE;
    } else {
        pVia->pChipInfo->biosIsSpreadSpectrum = FALSE;
    }

    if (!pVia->IsMAMMEnable) {

        /* get default TV standard from scratch pad */
        if (pBIOSInfo->TVSettingInfo.TVType == TVTYPE_NONE) {
            VGAOUT8(0x3D4, 0x4E);
            tmp = VGAIN8(0x3D5) & 0x0F;

            if (tmp) {
                pVia->pChipInfo->biosTVType = TVTYPE_PAL;
                pBIOSInfo->TVSettingInfo.TVType = TVTYPE_PAL;
            } else {
                pVia->pChipInfo->biosTVType = TVTYPE_NTSC;
                pBIOSInfo->TVSettingInfo.TVType = TVTYPE_NTSC;
            }
        }

        /* get default LCD panel ID from scratch pad. */
        if (pBIOSInfo->LVDSSettingInfo.PanelSizeID == VIA_INVALID) {
            VGAOUT8(0x3D4, 0x3F);
            tmp = VGAIN8(0x3D5) & 0x0F;

            pVia->pChipInfo->biosPanelID = tmp;
            if(VIASetLCDPanelSizeFromID(pBIOSInfo,
                                        pVia->pChipInfo->biosPanelID)) {
                DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                      "Get PanelID From Scratch Pad is 0x%x\n",
                      pVia->pChipInfo->biosPanelID));
            } else {
                DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_ERROR,
                      "No Panel ID\n"));
            }
        }
    }

    return TRUE;
}
