/*
* Copyright 1998-2009 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2009 S3 Graphics, Inc. All Rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sub license,
* and/or sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice (including the
* next paragraph) shall be included in all copies or substantial portions
* of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
* THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
* DEALINGS IN THE SOFTWARE.
*/
#include "via_driver.h"
#include "via_output.h"
#include "via_disphdmi.h"
#include "via_hdmi.h"
#include "via_common.h"
#include "via_dp.h"



/*****************************************************/
/*                 HDMI                              */
/*****************************************************/
extern _X_EXPORT Bool ad9389_module_loaded;

Bool 
viaGetStdEDIDBlock(CARD32 serialPort, CARD8* pEDIDData)
{
    CARD8   EDIDHeader[3];
    int     i;
    
    /* First Read EDID Block 0 */
    for (i = 0; i < 3; i++) {    
        viaSerialReadByte(serialPort, 0xA0, i, EDIDHeader+i);        
    }

    /* EDID Header is 00 FF FF FF FF FF FF 00 */
    if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
        (EDIDHeader[2] == 0xFF)) {
        /* Read Block 0 Raw Data. */        
        for (i = 0; i < 128; i++) {
            viaSerialReadByte(serialPort, 0xA0, i, pEDIDData+i);        
        }    
        return TRUE;
    }
    else
        return FALSE;
}

Bool 
viaGetStdEDIDBlockByHDCP(CARD8* pEDIDData)
{
    CARD8   EDIDHeader[3];
    int     i;
    
    /* First Read EDID Block 0 */
    for (i = 0; i < 3; i++) {    
        viaSerialReadByteByHDCP(0xA0, i, EDIDHeader+i);        
    }

    /* EDID Header is 00 FF FF FF FF FF FF 00 */
    if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
        (EDIDHeader[2] == 0xFF)) {
        /* Read Block 0 Raw Data. */        
        viaSerialReadBytesByHDCP(0xA0, 0x0, pEDIDData, 128);
        
        return TRUE;
    }
    else
        return FALSE;
}

/* Return EDID Status*/
Bool 
viaGetExtEDIDBlock(CARD32 serialPort, CARD8* pEDIDData, CARD8 extFlag)
{
    CARD8   ExtensionTag;   /* Extension Type */
    CARD8   RevisionNum;
    int     i;
    int     ret = FALSE;    
    
    DEBUG(ErrorF("viaGetExtEDIDBlock\n"));                

    if (extFlag == 1) {
        DEBUG(ErrorF("Only one extension block\n"));
    
        viaSerialReadByte(serialPort, 0xA0, 128, &ExtensionTag);

        /* 0x02 is timing extension block */
        if (ExtensionTag == 0x02) {
            viaSerialReadByte(serialPort, 0xA0, 129, &RevisionNum);
            /* Currently, we only support rev.3 */
            if (RevisionNum == 0x03) {                
                /* This have determined as HDMI EDID. */
                DEBUG(ErrorF("Found CEA EDID\n"));

                /* Read Block 2 Raw Data. */
                for (i=0; i<128; i++) {
                    viaSerialReadByte(serialPort, 0xA0, i+128, pEDIDData+i);                        
                }                
                ret = TRUE;
            }
        }
    } else if (extFlag > 1) {
        viaSerialReadByte(serialPort, 0xA0, 128, &ExtensionTag);
        /* 0xF0 is block map */
        if (ExtensionTag == 0xF0) {
            DEBUG(ErrorF("Found Extension Map\n"));

            /* Currently, we only can support 2 segments. (segment0,1) */
            viaSerialWriteByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0xC4, 1);
            
            /* Read Block 2 0x0 */
            viaSerialReadByte(serialPort, 0xA0, 0, &ExtensionTag);                
            /* 0x02 is timing extension block */
            if (ExtensionTag == 0x02) {
                viaSerialReadByte(serialPort, 0xA0, 1, &RevisionNum);
                /* Currently, we only support rev.3 */
                if (RevisionNum == 0x03) {
                    DEBUG(ErrorF("Found CEA EDID\n"));

                    /* Read Block 2 Raw Data. */
                    for (i=0; i<128; i++) {
                        viaSerialReadByte(serialPort, 0xA0, i, pEDIDData+i);
                    }                        
                    ret = TRUE;
                }                 
            }

            /* Read Block 3 0x0 */
            viaSerialReadByte(serialPort, 0xA0, 128, &ExtensionTag);                
            /* 0x02 is timing extension block */
            if (ExtensionTag == 0x02) {
                viaSerialReadByte(serialPort, 0xA0, 129, &RevisionNum);                
                /* We only support rev.3 now */
                if (RevisionNum == 0x03) {                
                    DEBUG(ErrorF("Found CEA EDID\n"));
                
                    /* Read Block 3 Raw Data. */
                    for (i=0; i<128; i++) {
                        viaSerialReadByte(serialPort, 0xA0, i+128, pEDIDData+i);                
                    }
                    ret = TRUE;                    
                }
            }
            /* Restore to segment 0. */
            viaSerialWriteByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0xC4, 0);
        }
    }   

    return ret;
}

/* Return EDID Status*/
Bool 
viaGetExtEDIDBlockByHDCP(CARD8* pEDIDData, CARD8 extFlag)
{
    CARD8   ExtensionTag;   /* Extension Type */
    CARD8   RevisionNum;
    int     i;
    int     ret = FALSE;    
    
    DEBUG(ErrorF("viaGetExtEDIDBlockByHDCP\n"));                

    if (extFlag == 1) {
        DEBUG(ErrorF("Only one extension block\n"));
    
        viaSerialReadByteByHDCP(0xA0, 128, &ExtensionTag);

        /* 0x02 is timing extension block */
        if (ExtensionTag == 0x02) {
            viaSerialReadByteByHDCP(0xA0, 129, &RevisionNum);
            /* Currently, we only support rev.3 */
            if (RevisionNum == 0x03) {                
                /* This have determined as HDMI EDID. */
                DEBUG(ErrorF("Found CEA EDID\n"));

                /* Read Block 2 Raw Data. */
                viaSerialReadBytesByHDCP(0xA0, 128, pEDIDData, 128);                
                ret = TRUE;
            }
        }
    }  

    return ret;
}


CARD32
viaDetectConnectType(CARD8* extEdid)
{
    CARD8  VendorBlockIndex = 0;      
    CARD8  tmp = 0;
    CARD32 ret = DISP_DEV_NONE;
    
    /* Check Data Block amount. If no data block exists, return DVI connect. 
       Byte 2: byte number offset d where detail timing data begin. */
    /* if 0, No Data Block, No Detailed Timing Descriptor.*/
    /* if 4, No Data Block.*/
    if(!(extEdid[2] == 0 ||    
         extEdid[2] == 4)) {
        /* Byte 04 is where the Data Block Collection begins */
        VendorBlockIndex = 4; 
        tmp = extEdid[VendorBlockIndex]; 
        
        /* Locate Vendor Description Block to identify if connect to HDMI.*/
        /* Bit[7:5]: Block Type Tag: value 3 is vendor specific*/
        while (((tmp & 0xE0) != 0x60) 
             && (VendorBlockIndex < (extEdid[2]))) {
            /* Bit[4:0]: Total number of bytes in this block following 
               this byte */
            tmp &= 0x1F;  

            /* The next byte is assumed to be the beginning of the next 
               data block. */
            VendorBlockIndex += (tmp+1); 

            /* Locate the next data block */
            tmp = extEdid[VendorBlockIndex]; 
        }
        
        if((tmp & 0xE0) == 0x60) {
            /* Read 24-bit IEEE registration.*/            
            /* HDMI identifier = 0x000C03.*/
            if( (extEdid[VendorBlockIndex+1] == 0x03)
             && (extEdid[VendorBlockIndex+2] == 0x0C)
             && (extEdid[VendorBlockIndex+3] == 0x00)) {                        
                ret =  DISP_DEV_HDMI;
            } else
                ret = DISP_DEV_DVI;  
        }        
    }
    return ret;
}

void
viaFindMaxSupportMode(ScrnInfoPtr pScrn, CARD8* stdEdid, CARD8* extEDID,
                             CARD32* maxWidth, CARD32* maxHeight)
{
    xf86MonPtr          edid_mon;      

    int    TmpHSize = 0;
    int    TmpVSize = 0;
    int    MaxHSize = 0;
    int    MaxVSize = 0;
    int    i = 0;
    CARD8  tmp = 0;
    CARD8  VideoBlockIndex = 0;      
    CARD32 DescriptorNum = 0;
    CARD8  HDMIModeIndex;

    if (stdEdid[0x7E] == 0) {
        /* Parse the EDID block */
        edid_mon = xf86InterpretEDID(pScrn->scrnIndex, stdEdid);
        if (edid_mon) {          
            /*Find max resolution from Established timing*/
            VIAFindMaxResFromEstablishedTiming(edid_mon, &MaxHSize, &MaxVSize);

            /*Find max resolution from Standard timing*/
            for (i = 0; i < 8; i++) {
                if (edid_mon->timings2[i].hsize > MaxHSize) {
                    MaxHSize = edid_mon->timings2[i].hsize;
                    MaxVSize = edid_mon->timings2[i].vsize;
                }            
            }

            /*Find max resolution from Detail timing*/
            for (i = 0; i < 4; i++) {
                if ((edid_mon->det_mon[i].type == DT) && 
                    (edid_mon->det_mon[i].section.d_timings.h_active > MaxHSize)) {

                    MaxHSize = edid_mon->det_mon[i].section.d_timings.h_active;
                    MaxVSize = edid_mon->det_mon[i].section.d_timings.v_active;

                    if (edid_mon->det_mon[i].section.d_timings.interlaced)                
                        MaxVSize = MaxVSize*2;                
                }       
            }

            /* To prevent to get a wrong EDID data. We should assign 
                   a default value.*/
            if ((MaxHSize == 0) || (MaxVSize == 0)) {
                MaxHSize = 1280;
                MaxVSize = 720;
            }
        
            *maxHeight = MaxVSize;
            *maxWidth = MaxHSize;
        }
    }else if (stdEdid[0x7E] > 0) {                
        /* Byte 04 is where the Data Block Collection begins */
        tmp = extEDID[4]; 

        /* Locate Video Block */                
        /* Bit[7:5]: Block Type Tag: value 2 is video specific*/
        while (((tmp & 0xE0) != 0x40) 
             && (VideoBlockIndex < (extEDID[2]))) {
            /* Bit[4:0]: 
            Total number of bytes in this block following this byte */
            tmp &= 0x1F;  

            /* The next byte is assumed to be the beginning of the 
               next data block. */
            VideoBlockIndex += (tmp+1); 

            /* Locate the next data block */
            tmp = extEDID[VideoBlockIndex]; 
        }
        
        if ((tmp & 0xE0) == 0x40) {
            DescriptorNum = tmp & 0x1F;

            for(i = 0; i< DescriptorNum; i++) {
            
                tmp = extEDID[VideoBlockIndex]+i;
                HDMIModeIndex = tmp & 0x7F;

                if(HDMIModeIndex <= 0) {                  
                    HDMIModeIndex = 1;
                }
                                    
                TmpHSize = CEATimingTable[HDMIModeIndex-1].HorSize;
                TmpVSize = CEATimingTable[HDMIModeIndex-1].VerSize;            

                if (TmpHSize*TmpVSize > MaxHSize*MaxVSize) {
                    MaxHSize = TmpHSize; 
                    MaxVSize = TmpVSize;
                }
            }        

            *maxHeight = MaxVSize;
            *maxWidth = MaxHSize;
            
            DEBUG(ErrorF("HDMI Max Support HSize = %d\n", MaxHSize)); 
            DEBUG(ErrorF("HDMI Max Support VSize = %d\n", MaxVSize));                
        }
    }
}

DisplayModePtr
viaAddExtDetailTiming(xf86OutputPtr output, DisplayModePtr modes,
                          CARD8* pEDIDData)
{
    int d=0;
    int i=0;
    unsigned int pixelClock;
    unsigned int hActive;
    unsigned int hBlanking;
    unsigned int horSyncOffset;
    unsigned int horSyncPluseWidth;
    unsigned int vActive;
    unsigned int vBlanking;
    unsigned int verSyncOffset;
    unsigned int verSyncPluseWidth;
    Bool         isInterlace;
    Bool         horSyncPolarity;
    Bool         verSyncPolarity;
    DisplayModePtr      newMode;
    
    unsigned char tmp, tmp1;
    d = pEDIDData[2];
    
    for (i = 0; (i+d) < pEDIDData[3]; i+=18) {            
        
        newMode = xnfalloc(sizeof(DisplayModeRec));
        memset(newMode, 0, sizeof(DisplayModeRec));    
        
        /* Pixel Clock */
        pixelClock = (((pEDIDData[1+i+d]) << 8) + pEDIDData[i+d]) * 10000;

        /* H Active */        
        tmp = pEDIDData[2+i+d];
        tmp1 = pEDIDData[4+i+d];
        hActive = ((tmp1 & 0xF0) << 4) + tmp;
        
        /* H Blanking */
        tmp= pEDIDData[3+i+d];
        tmp1 = pEDIDData[4+i+d];
        
        hBlanking = ((tmp1 & 0x0F) << 8) + tmp;

        /* H Sync Offset */
        tmp = pEDIDData[8+i+d];
        tmp1 = pEDIDData[11+i+d];
        horSyncOffset = ((tmp1 & 0xC0) << 2) + tmp;

        /* H Sync Pulse Width */
        tmp = pEDIDData[9+i+d];
        tmp1 = pEDIDData[11+i+d];
        horSyncPluseWidth = ((tmp1 & 0x30) << 4) + tmp;

        /* V Active */
        tmp = pEDIDData[5+i+d];
        tmp1 = pEDIDData[7+i+d];
        vActive = ((tmp1 & 0xF0) << 4) + tmp;

        /* V Blanking */
        tmp= pEDIDData[6+i+d];
        tmp1 = pEDIDData[7+i+d];
        vBlanking = ((tmp1 & 0x0F) << 8) + tmp;

        /* V Sync Offset */
        tmp = pEDIDData[10+i+d];
        tmp1 = pEDIDData[11+i+d];
        verSyncOffset = ((tmp1 & 0x0C) << 6) + ((tmp & 0xF0) >> 4);

        /* V Sync Pluse Width */
        tmp = pEDIDData[10+i+d];
        tmp1 = pEDIDData[11+i+d];
        verSyncPluseWidth = ((tmp1 & 0x03) << 8) + (tmp & 0x0F);

        /* Flag */
        tmp = pEDIDData[17+i+d];
        if (tmp & 0x80) 
            isInterlace = TRUE;
        else 
            isInterlace = FALSE;                

        /* horizontal polarity*/
        if (tmp & 0x1)
            horSyncPolarity = POSITIVE;
        else
            horSyncPolarity = NEGATIVE;

        /* verticle polarity*/
        if (tmp & 0x2)
            verSyncPolarity = POSITIVE;
        else
            verSyncPolarity = NEGATIVE;
            
        if (newMode) {
            newMode->type = M_T_DRIVER;
            newMode->Clock = pixelClock / 1000;
            newMode->HDisplay = hActive;
            newMode->HSyncStart = hActive + horSyncOffset;
            newMode->HSyncEnd = newMode->HSyncStart + horSyncPluseWidth;
            newMode->HTotal = hActive + hBlanking;
            newMode->VDisplay = vActive;
            newMode->VSyncStart = vActive + verSyncOffset;
            newMode->VSyncEnd = newMode->VSyncStart + verSyncPluseWidth;
            newMode->VTotal = vActive + vBlanking;

            if (isInterlace)
                newMode->Flags |= V_INTERLACE;

            if (horSyncPolarity == POSITIVE)
                newMode->Flags |= V_PHSYNC;
            else
                newMode->Flags |= V_NHSYNC;

            if (verSyncPolarity == POSITIVE)
                newMode->Flags |= V_PVSYNC;
            else
                newMode->Flags |= V_NVSYNC;

            if (newMode->HDisplay && newMode->VDisplay) {
                xf86SetModeDefaultName(newMode);
                modes = xf86ModesAdd(modes, newMode); 
            }
            
            DEBUG(ErrorF("Add new mode %s from extend EDID\n", newMode->name));
        }
    }    

    return modes;    
}

static void
loadHdmiDefaultDPASetting(CARD32 chipset, CARD32 port)
{
    if (port == DISP_DI_DVP1) {
        switch (chipset) {
            case VIA_CX700:    
                viaWriteVgaIoBits(REG_SR65, 0x0B, 0x0F);
                viaWriteVgaIoBits(REG_CR9B, 0x00, 0x0F);
                break;        
        
            case VIA_VX800:   
                viaWriteVgaIoBits(REG_SR65, 0x0B, 0x0F);
                viaWriteVgaIoBits(REG_CR9B, 0x0F, 0x0F);
                break;
        
            case VIA_VX855:
                viaWriteVgaIoBits(REG_SR65, 0x0B, 0x0F);
                viaWriteVgaIoBits(REG_CR9B, 0x0F, 0x0F);     
                break;
				
            case VIA_VX900:
                viaWriteVgaIoBits(REG_SR65, 0x0B, 0x0F);
                viaWriteVgaIoBits(REG_CR9B, 0x0F, 0x0F);     
                break;  
				
            default:
                break;
        }
    }
}

static void 
viaPatchDotClock(xf86OutputPtr output, DisplayModePtr mode, int igaPath)
{
    ScrnInfoPtr pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    int i;
    CARD32  patchPLL= 0;

    switch (pVia->Chipset) {        
        case VIA_VX855:
        case VIA_VX900:    
            for (i = 0; ViaDotClocks[i].DotClock; i++) {
                if (ViaDotClocks[i].DotClock == mode->Clock) {
                    patchPLL = ViaDotClocks[i].PLL3_409;
                    break;
                }
            }                
            break;
    
        default:
            for (i = 0; ViaDotClocks[i].DotClock; i++) {
                if (ViaDotClocks[i].DotClock == mode->Clock) {
                    patchPLL = ViaDotClocks[i].PLL3;
                    break;
                }
            }
            break;
    }
    
    viaSetVCLK(patchPLL, igaPath);
}

static void 
viaPrintModeList(DisplayModePtr modeList)
{
    DEBUG(ErrorF("viaPrintModeList\n"));   
    
    DisplayModePtr mode;

    for (mode = modeList; ; mode = mode->next) {        
        DEBUG(ErrorF("mode name = %s\n", mode->name));
        DEBUG(ErrorF("mode HDisplay = %d\n", mode->HDisplay));
        DEBUG(ErrorF("mode VDisplay = %d\n", mode->VDisplay));
        DEBUG(ErrorF("mode CrtcHDisplay = %d\n", mode->CrtcHDisplay));
        DEBUG(ErrorF("mode CrtcVDisplay = %d\n", mode->CrtcVDisplay));        
        if (mode->Flags & V_INTERLACE)
            DEBUG(ErrorF("interlace\n"));        
        else
            DEBUG(ErrorF("non-interlace\n"));

        if (mode->Flags & V_DBLSCAN)
            DEBUG(ErrorF("double scan\n"));        
        else
            DEBUG(ErrorF("non-double scan\n"));        
            
        if ((mode->next == modeList) || (mode->next ==  NULL))
            break;  /* The end of mode list. */
    }
}

static DisplayModePtr 
viaAddCEAModeLine(xf86OutputPtr output, CARD8* extEDID)                             
{
    int     i=0, k=0;
    CARD8   tmp = 0;
    CARD8   VideoBlockIndex = 0;      
    CARD32  DescriptorNum = 0;
    CARD8   HDMIModeIndex;
    DisplayModePtr  newMode = NULL;
    DisplayModePtr  modeList = NULL;
    ViaHdmiPrivateInfoPtr viaHdmiInfo = output->driver_private;

    if (viaHdmiInfo->attachAllModes) {
        while(ViaCEAModes[k].name) {
            if (CEATimingTable[k].IsBuilt) {
                newMode = xnfalloc(sizeof(DisplayModeRec));                
                if (newMode) { 
                    memcpy(newMode, &ViaCEAModes[k], sizeof(DisplayModeRec));                    
                    newMode->name = xnfstrdup(ViaCEAModes[k].name);
                    modeList = xf86ModesAdd(modeList, newMode);                        
                    DEBUG(ErrorF("Just add CEA mode ID:%d %dx%d@%d I:%d\n",
                          k,
                          CEATimingTable[k].HorSize,
                          CEATimingTable[k].VerSize,
                          CEATimingTable[k].RefreshRate,
                          CEATimingTable[k].ScanType));                        
                }
            }
            k++;
        }
        return modeList;
    }

    /* Byte 04 is where the Data Block Collection begins */
    VideoBlockIndex = 4;
    tmp = extEDID[4]; 
    /* Locate Video Block */                
    /* Bit[7:5]: Block Type Tag: value 2 is video specific*/
    while (((tmp & 0xE0) != 0x40) 
         && (VideoBlockIndex < (extEDID[2]))) {
        /* Bit[4:0]: 
        Total number of bytes in this block following this byte */
        tmp &= 0x1F;  
  
        /* The next byte is assumed to be the beginning of the 
           next data block. */
        VideoBlockIndex += (tmp+1); 
  
        /* Locate the next data block */
        tmp = extEDID[VideoBlockIndex]; 
    }
    if ((tmp & 0xE0) == 0x40) {
        DescriptorNum = tmp & 0x1F;

        for(i = 0; i< DescriptorNum; i++) {        
            tmp = extEDID[VideoBlockIndex+1+i];
            HDMIModeIndex = tmp & 0x7F;

            DEBUG(ErrorF("Monitor support CEA mode ID:%d %dx%d@%d I:%d\n",
                  HDMIModeIndex,
                  CEATimingTable[HDMIModeIndex-1].HorSize,
                  CEATimingTable[HDMIModeIndex-1].VerSize,
                  CEATimingTable[HDMIModeIndex-1].RefreshRate,
                  CEATimingTable[HDMIModeIndex-1].ScanType));
    
            if (CEATimingTable[HDMIModeIndex-1].IsBuilt &&
                ViaCEAModes[HDMIModeIndex-1].name != NULL) {

                newMode = xnfalloc(sizeof(DisplayModeRec));
                if (newMode) { 
                    memcpy(newMode, &ViaCEAModes[HDMIModeIndex-1], sizeof(DisplayModeRec));                    
                    newMode->name = xnfstrdup(ViaCEAModes[HDMIModeIndex-1].name);
                    modeList = xf86ModesAdd(modeList, newMode);
                    DEBUG(ErrorF("Driver add CEA mode ID:%d %dx%d@%d I:%d\n",
                          HDMIModeIndex,
                          CEATimingTable[HDMIModeIndex-1].HorSize,
                          CEATimingTable[HDMIModeIndex-1].VerSize,
                          CEATimingTable[HDMIModeIndex-1].RefreshRate,
                          CEATimingTable[HDMIModeIndex-1].ScanType));
                }
            }    
        }        
    }

    return modeList;
}

static CARD32
viaHotplugTimer(OsTimerPtr timer, CARD32 now, pointer arg)
{
    xf86OutputPtr output = (xf86OutputPtr) arg;
    ViaOutputInfo   commonInfo = ((ViaHdmiPrivateInfoPtr)(output->driver_private))->commonInfo;
    CARD32          serialPort = commonInfo.serialPort;                    
    CARD8   state;
    CARD8   state2;

    if (SUBCHIP_AD9389 == commonInfo.subChipName) {
        viaSerialReadByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x41, &state);
        viaSerialReadByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x42, &state2);

        /* if connected but disable */
        if ((state == 0x50) && (state2 & 0x40)) 
            viaInitializeAD9389(output);
    }

    /* 3 sec */
    return 3000; 
}

// calcute the BCH code and output 1 byte BCH code
CARD16 viaInsertBCHCode(CARD16* input, CARD32 length)
{
    CARD32   i, n;
    CARD32   reg[9]={0, 0, 0, 0, 0, 0, 0, 0, 0};
    CARD32   matrix[56];
    CARD16   output = 0;

    // initialize matrix
    for(i = 0; i < length; i++)
    {
        matrix[0 + i*8] = (input[i] >> 0) & BIT0;
        matrix[1 + i*8] = (input[i] >> 1) & BIT0;
        matrix[2 + i*8] = (input[i] >> 2) & BIT0;
        matrix[3 + i*8] = (input[i] >> 3) & BIT0;
        matrix[4 + i*8] = (input[i] >> 4) & BIT0;
        matrix[5 + i*8] = (input[i] >> 5) & BIT0;
        matrix[6 + i*8] = (input[i] >> 6) & BIT0;
        matrix[7 + i*8] = (input[i] >> 7) & BIT0;
    }

    for(i = 0; i < length * 8; i++)
    {
        reg[0] = reg[8] ^ matrix[i];
        reg[8] = reg[7] ^ reg[0];
        reg[7] = reg[6] ^ reg[0];

        reg[6]=reg[5];
        reg[5]=reg[4];
        reg[4]=reg[3];
        reg[3]=reg[2];
        reg[2]=reg[1];
        reg[1]=reg[0];
    }

    n = 0;
    for(i = 1; i <= 4; i++)
    {
        if(reg[i] == 1)
            n = 2*n + 1;
        else 
            n = 2*n;
    }

    output |= n << 4;

    n = 0;
    for(i = 5; i <= 8; i++)
    {
        if(reg[i] == 1)
            n = 2*n + 1;
        else 
            n = 2*n;
    }

    output |= n;

    return output;
}

void 
viaSetAVIInfoFrame(DisplayModePtr mode)
{
    viaHDMI_InfoFrameRec  InfoFramePacket;
    CARD32   i = 0;
    CARD16   checksum = 0;
    CARD32   mmC204, mmC208, mmC20C, mmC210, mmC214, mmC218, mmC21C, mmC220, mmC224;
    
    InfoFramePacket.PacketHeader[0] = 0x82;
    InfoFramePacket.PacketHeader[1] = 0x02;
    InfoFramePacket.PacketHeader[2] = 0x0D;
    InfoFramePacket.ECCCode = viaInsertBCHCode(InfoFramePacket.PacketHeader, 3);

    // set Packet Bytes
    for(i = 0; i < 28; i++)
        InfoFramePacket.PacketByte[i] = 0;
 
    // S1  S0   |   Scan Information
    // -----------------------------
    //  0   1   |   Overscan
    //  1   0   |   Underscan
    InfoFramePacket.PacketByte[1] |= 0x2;

    // R3  R2  R1  R0   |   Active Format Aspect Ratio
    // -----------------------------------------------
    //  1   0   0   0   |   Same as picture aspect ratio
    //  1   0   0   1   |   4:3
    //  1   0   1   0   |   16:9
    //  1   0   1   1   |   14:9
    InfoFramePacket.PacketByte[2] |= 0x08;

    // M1  M0   |   Picture Aspect Ratio
    // ---------------------------------
    //  0   1   |    4:3
    //  1   0   |   16:9
    switch(mode->CrtcHDisplay)
    {
        case 640:
        case 720:
            InfoFramePacket.PacketByte[2] |= 0x10;
            break;
        case 1280:
        case 1920:
            InfoFramePacket.PacketByte[2] |= 0x20;
            break;
        default:
            break;
    }

    // ITC bit
    InfoFramePacket.PacketByte[3] |= 0x80;


    // VIC code
    switch(mode->CrtcHDisplay)
    {
        case 640:
            InfoFramePacket.PacketByte[4] = 1;
            break;
        case 720:
            if (mode->CrtcVDisplay == 480)
                InfoFramePacket.PacketByte[4] = 2;
            else if (mode->CrtcVDisplay == 576)
                InfoFramePacket.PacketByte[4] = 17;
            break;
        case 1280:
            if (mode->VRefresh == 50)
                InfoFramePacket.PacketByte[4] = 19;
            else
                InfoFramePacket.PacketByte[4] = 4;
            break;
        case 1920:
            if (mode->CrtcVDisplay == 1080)
                InfoFramePacket.PacketByte[4] = 16;
            else if (mode->CrtcVDisplay == 540)
                InfoFramePacket.PacketByte[4] = 5;    
            break;
        default:
            break;
    }

    // count checksum
    for(i = 0; i < 28; i++)
        checksum += InfoFramePacket.PacketByte[i];
    checksum += ( 0x82 + 0x02 + 0x0D );
    
    checksum = 0x100 - checksum;
    InfoFramePacket.PacketByte[0] = checksum;

    // set BCHCode
    for(i = 0; i < 4; i++)
    {
        InfoFramePacket.BCHCode[i] = viaInsertBCHCode( &(InfoFramePacket.PacketByte[i*7]), 7);
    }

    mmC204 = (InfoFramePacket.ECCCode << 24 ) | (InfoFramePacket.PacketHeader[2] << 16) | 
             (InfoFramePacket.PacketHeader[1] << 8) | InfoFramePacket.PacketHeader[0];
    mmC208 = (InfoFramePacket.PacketByte[3] << 24) | (InfoFramePacket.PacketByte[2] << 16) |
             (InfoFramePacket.PacketByte[1] << 8) | InfoFramePacket.PacketByte[0];
    mmC20C = (InfoFramePacket.BCHCode[0] << 24) | (InfoFramePacket.PacketByte[6] << 16) |
             (InfoFramePacket.PacketByte[5] << 8) | InfoFramePacket.PacketByte[4];
    mmC210 = (InfoFramePacket.PacketByte[10] << 24) | (InfoFramePacket.PacketByte[9] << 16) |
             (InfoFramePacket.PacketByte[8] << 8) | InfoFramePacket.PacketByte[7];
    mmC214 = (InfoFramePacket.BCHCode[1] << 24) | (InfoFramePacket.PacketByte[13] << 16) |
             (InfoFramePacket.PacketByte[12] << 8) | InfoFramePacket.PacketByte[11];
    mmC218 = (InfoFramePacket.PacketByte[17] << 24) | (InfoFramePacket.PacketByte[16] << 16) |
             (InfoFramePacket.PacketByte[15] << 8) | InfoFramePacket.PacketByte[14];
    mmC21C = (InfoFramePacket.BCHCode[2] << 24) | (InfoFramePacket.PacketByte[20] << 16) |
             (InfoFramePacket.PacketByte[19] << 8) | InfoFramePacket.PacketByte[18];
    mmC220 = (InfoFramePacket.PacketByte[24] << 24) | (InfoFramePacket.PacketByte[23] << 16) |
             (InfoFramePacket.PacketByte[22] << 8) | InfoFramePacket.PacketByte[21];
    mmC224 = (InfoFramePacket.BCHCode[3] << 24) | (InfoFramePacket.PacketByte[27] << 16) |
             (InfoFramePacket.PacketByte[26] << 8) | InfoFramePacket.PacketByte[25];

    MMIO_WR(0xC204, mmC204);
    MMIO_WR(0xC208, mmC208);
    MMIO_WR(0xC20C, mmC20C);
    MMIO_WR(0xC210, mmC210);
    MMIO_WR(0xC214, mmC214);
    MMIO_WR(0xC218, mmC218);
    MMIO_WR(0xC21C, mmC21C);
    MMIO_WR(0xC220, mmC220);
    MMIO_WR(0xC224, mmC224);

    MMIO_WR(0xC200, 0x80);  
    
}

void 
viaSetAudioInfoFrame()
{
    viaHDMI_InfoFrameRec  InfoFramePacket;
    CARD32   i = 0;
    CARD16   checksum = 0;
    CARD32   mmC204, mmC208, mmC20C, mmC210, mmC214, mmC218, mmC21C, mmC220, mmC224;

    InfoFramePacket.PacketHeader[0] = 0x84;
    InfoFramePacket.PacketHeader[1] = 0x01;
    InfoFramePacket.PacketHeader[2] = 0x0A;
    InfoFramePacket.ECCCode = viaInsertBCHCode(InfoFramePacket.PacketHeader, 3);

    // set Packet Bytes
    for(i = 0; i < 28; i++)
        InfoFramePacket.PacketByte[i] = 0;
 
    // CC2 CC1 CC0  |   Audio Channel Count
    // ---------------------------------------
    //  0   0   0   |   Refer to Stream Header
    //  0   0   1   |   2ch
    //  0   1   0   |   3ch
    //  0   1   1   |   4ch
    InfoFramePacket.PacketByte[1] |= 0x0;

    // CT3 CT2 CT1 CT0  |   Audio Coding Type
    // -------------------------------------------
    //  0   0   0   0   |   Refer to Stream Header
    //  0   0   0   1   |   IEC 60958 PCM[30,31]
    //  0   0   1   0   |   AC-3
    //  0   0   1   1   |   MPEG1 (Layer 1 & 2)
    InfoFramePacket.PacketByte[1] |= 0x00;

    // SS1 SS0  |   Sample Size
    // --------------------------
    //  0   0   |   Refer to Stream Header
    //  0   1   |   16 bit
    //  1   0   |   20 bit
    //  1   1   |   24 bit
    InfoFramePacket.PacketByte[2] |= 0x00;

    // SF2 SF1 SF0  |   Sampling Frequency
    // ---------------------------------------
    //  0   0   0   |   Refer to Stream Header
    //  0   0   1   |   32 kHz
    //  0   1   0   |   44.1 kHz(CD)
    //  0   1   1   |   48 kHz
    //  1   0   0   |   88.2 kHz
    //  1   0   1   |   96 kHz
    //  1   1   0   |   176.4 kHz
    //  1   1   1   |   192 kHz
    InfoFramePacket.PacketByte[2] |= 0x00;

    // count checksum
    for(i = 0; i < 28; i++)
        checksum += InfoFramePacket.PacketByte[i];
    checksum += ( 0x84 + 0x01 + 0x0A );
    
    checksum = 0x100 - checksum;
    InfoFramePacket.PacketByte[0] = checksum;

    // set BCHCode
    for(i = 0; i < 4; i++)
    {
        InfoFramePacket.BCHCode[i] = viaInsertBCHCode( &(InfoFramePacket.PacketByte[i*7]), 7);
    }

    mmC204 = (InfoFramePacket.ECCCode << 24 ) | (InfoFramePacket.PacketHeader[2] << 16) | 
             (InfoFramePacket.PacketHeader[1] << 8) | InfoFramePacket.PacketHeader[0];
    mmC208 = (InfoFramePacket.PacketByte[3] << 24) | (InfoFramePacket.PacketByte[2] << 16) |
             (InfoFramePacket.PacketByte[1] << 8) | InfoFramePacket.PacketByte[0];
    mmC20C = (InfoFramePacket.BCHCode[0] << 24) | (InfoFramePacket.PacketByte[6] << 16) |
             (InfoFramePacket.PacketByte[5] << 8) | InfoFramePacket.PacketByte[4];
    mmC210 = (InfoFramePacket.PacketByte[10] << 24) | (InfoFramePacket.PacketByte[9] << 16) |
             (InfoFramePacket.PacketByte[8] << 8) | InfoFramePacket.PacketByte[7];
    mmC214 = (InfoFramePacket.BCHCode[1] << 24) | (InfoFramePacket.PacketByte[13] << 16) |
             (InfoFramePacket.PacketByte[12] << 8) | InfoFramePacket.PacketByte[11];
    mmC218 = (InfoFramePacket.PacketByte[17] << 24) | (InfoFramePacket.PacketByte[16] << 16) |
             (InfoFramePacket.PacketByte[15] << 8) | InfoFramePacket.PacketByte[14];
    mmC21C = (InfoFramePacket.BCHCode[2] << 24) | (InfoFramePacket.PacketByte[20] << 16) |
             (InfoFramePacket.PacketByte[19] << 8) | InfoFramePacket.PacketByte[18];
    mmC220 = (InfoFramePacket.PacketByte[24] << 24) | (InfoFramePacket.PacketByte[23] << 16) |
             (InfoFramePacket.PacketByte[22] << 8) | InfoFramePacket.PacketByte[21];
    mmC224 = (InfoFramePacket.BCHCode[3] << 24) | (InfoFramePacket.PacketByte[27] << 16) |
             (InfoFramePacket.PacketByte[26] << 8) | InfoFramePacket.PacketByte[25];

    MMIO_WR(0xC204, mmC204);
    MMIO_WR(0xC208, mmC208);
    MMIO_WR(0xC20C, mmC20C);
    MMIO_WR(0xC210, mmC210);
    MMIO_WR(0xC214, mmC214);
    MMIO_WR(0xC218, mmC218);
    MMIO_WR(0xC21C, mmC21C);
    MMIO_WR(0xC220, mmC220);
    MMIO_WR(0xC224, mmC224);

    MMIO_WR(0xC200, 0x81);  
    
}

Bool 
viaSetHDAudioSampleRate( DWORD dwSampleRate, DWORD dwPixelClock)
{
    CARD32 i = 0, mmC400, mmC404, mmC410, mmC414;

    mmC404 = MMIO_RD(0xC404) & 0xFFFFFF00;
    mmC414 = MMIO_RD(0xC414) & 0xFFFFFF00;

    MMIO_WR_MASK(0xC404, 0, 0xF); // Clear Wall Clock Ratio
    MMIO_WR_MASK(0xC400, 0, 0xFFFFFFFF);
    
    // Set Wall Clock Ratio: (24MHz / HDMI CLK frequency) * 2^40
    for(i = 0; i < WallClockReg_NUM; i++)
    {
        if(WallClockReg[i].PixelClock == dwPixelClock)
        {
            mmC400 = WallClockReg[i].WallClockRatio_L;
            mmC404 |= WallClockReg[i].WallClockRateo_H;
            break;
        }
    }
    MMIO_WR(0xC404, mmC404);
    MMIO_WR(0xC400, mmC400);

    MMIO_WR_MASK(0xC414, 0, 0xF); // Clear audio sample rate
    MMIO_WR_MASK(0xC410, 0, 0xFFFFFFFF);
    // Set audio sample rate: (Sample Rate / HDMI CLK frequency) * 2^40
    for(i = 0; i < PacketClockReg_NUM; i++)
    {
        if(PacketClockReg[i].SampleRate == dwSampleRate && PacketClockReg[i].PixelClock == dwPixelClock)
        {
            mmC410 = PacketClockReg[i].PacketClockRatio_L;
            mmC414 |= PacketClockReg[i].PacketClockRateo_H;
            break;
        }
    }

    MMIO_WR(0xC414, mmC414);
    MMIO_WR(0xC410, mmC410);

    return TRUE;
}

Bool 
viaSetHDAudioCTSandN(DWORD dwSampleRate, DWORD dwPixelClock)
{
    CARD32   i = 0, CTS = 0, N = 0, mmC294 = 0, mmC298 = 0, mmC29C = 0;
   
    for(i = 0; i < CTSandNReg_NUM; i++)
    {
        if(CTSandNReg[i].SampleRate == dwSampleRate && CTSandNReg[i].PixelClock == dwPixelClock)
        {
            CTS = CTSandNReg[i].CTS;
            N = CTSandNReg[i].N;
            break;
        }
    }

    mmC294 = 0x100000C0 | (CTS << 8);
    MMIO_WR(0xC294, mmC294);

    mmC298 = N | (( CTS & 0xFFF ) << 20);
    MMIO_WR(0xC298, mmC298);

    mmC29C = MMIO_RD(0xC29C) & 0xFFFFFF00;
    mmC29C |= ((CTS >> 12) & 0xFF);
    MMIO_WR(0xC29C, mmC29C);

    return TRUE;
}

void 
viaSetHDAudioParameter(DisplayModePtr mode)
{
    CARD32   mmC4D4, mmC4D8, mmC424, dwSampleRate, dwPixelClock;

    dwPixelClock = mode->Clock * 1000;

    // step 1. read out C4D8 
    mmC4D4 = MMIO_RD(0xC4D4) & 0x00FFFFFF;
    mmC4D4 |=  0x01000000;      // stream format
    MMIO_WR(0xC4D4, mmC4D4);

    mmC4D8 = MMIO_RD(0xC4D8); // [15:0]

    //step 2. decode C4D8 by definition of C45C
    // C45C [14] Converter_Format_Base
    // BASE. Converter Format Sample Base Rate. (PCM Format structure bit 14)
    //              0: 48kHz 
    //              1: 44.1kHz
    // C45C [13:11] Converter_Format_Mult
    // MULT. Converter Format Sample Base Rate Multiple. (PCM Format structure bits 13:11)
    //              000: 48kHz/ 44.1 kHz or lesss 
    //              001: x2 (96kHz, 88.2kHz, 32kHz) 
    //              010: x3 (144kHz) 
    //              011: x4 (192kHz, 176.4kHz) 
    //              100-111: Reserved
    // C45C [10:8] Converter_Format_Div
    // DIV. Converter Format Sample Base Rate Divisor. (PCM Format structure bits 10:8)
    //              000: Divide by 1 (48kHz, 44.1 kHz) 
    //              001: Divide by 2 (24kHz, 22.05kHz) 
    //              010: Divide by 3 (16kHz, 32kHz) 
    //              011: Divide by 4 (11.025kHz) 
    //              100: Divide by 5 (9.6kHz) 
    //              101: Divide by 6 (8kHz) 
    //              110: Divide by 7 
    //              111: Divide by 8 (6kHz)
    if( mmC4D8 & 0x4000 )
        dwSampleRate = 44100;
    else 
        dwSampleRate = 48000;
            
    switch((mmC4D8 & 0x3800) >> 11) 
    {
        case 1:
            dwSampleRate *= 2;
            break;
        case 2:
            dwSampleRate *= 3;
            break;
        case 3:
            dwSampleRate *= 4;
            break;
        default:
            break;
    }

    switch((mmC4D8 & 0x700) >> 8)
        {
        case 1:
            dwSampleRate /= 2;
            break;
        case 2:
            dwSampleRate /= 3;
            break;
        case 3:
            dwSampleRate /= 4;
            break;
        case 4:
            dwSampleRate /= 5;
            break;
        case 5:
            dwSampleRate /= 6;
            break;
        case 6:
            dwSampleRate /= 7;
            break;
        case 7:
            dwSampleRate /= 8;
            break;
        default:
            break;
    }

    mmC424 = MMIO_RD(0xC424) & 0xFFFF9FFF; // C424[14:13] codec type. related with mm c618, mm c698
    mmC424 |= (0 << 13);

    MMIO_WR(0xC424, mmC424);

    // step 3. calculate C400 and C410
    viaSetHDAudioSampleRate(dwSampleRate, dwPixelClock);

    // step 4. fill CTS & N according to Table7-1, 7-2, 7-3 of HDMI spec 1.3
    viaSetHDAudioCTSandN(dwSampleRate, dwPixelClock);

}

void 
viaSetInfoFrameData(DisplayModePtr mode)
{
    viaSetAVIInfoFrame(mode);
    viaSetAudioInfoFrame();
}

void
viaSetIntegratedHDMIMode(DisplayModePtr mode)
{
    // Infoframe setting
    viaSetInfoFrameData(mode);
    viaSetHDAudioParameter(mode);
				
    /* 135MHz ~ 270MHz */
    if (mode->Clock >= 135000)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x00000000, 0xC0000000);           
    /* 67.5MHz ~ <135MHz */
    else if (mode->Clock >= 67500)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x40000000, 0xC0000000);
    /* 33.75MHz ~ <67.5MHz */
    else if (mode->Clock >= 33750)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x80000000, 0xC0000000);
    /* 25MHz ~ <33.75MHz */
    else
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0xC0000000, 0xC0000000); 
		
    /* touch C282 when init HDMI by mode 720x576, 720x480, or other modes */
    if( ( mode->HDisplay == 720 ) && ( mode->VDisplay == 576 ) ) 
        MMIO_WR(0xC280, 0x18232402);
    else if ( ( mode->HDisplay == 720 ) && ( mode->VDisplay == 480 ) ) 
        MMIO_WR(0xC280, 0x181f2402);
    else
        MMIO_WR(0xC280, 0x18330002);
		
    // power down to reset
    MMIO_WR_MASK(0xC740, 0x00000000, 0x06000000);  
    // enable HDCP
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x00000001, 0x00000001);  
    // select HDMI mode
    MMIO_WR_MASK(0xC748, 0, BIT0);  
    // enable HDMI with HDMI mode
    MMIO_WR(0xC280, MMIO_RD(0xC280) | 0x18330002);  
    MMIO_WR_MASK(0xC280, 0x0, 0x40);
    // select AC mode
    MMIO_WR_MASK(0xC74C, 0x40, 0x40);  
    // enable InfoFrame
    MMIO_WR(0xC284, 0x20000102); 
    // set status of Lane0~3
    MMIO_WR_MASK(0xC744, 0x00FFFF82, 0x00FFFF82);  
    MMIO_WR(0xC0B4, 0x92000000);
    // enable audio packet
    MMIO_WR_MASK(0xC294, 0x10000000, 0x10000000); 
    // enable InfoFrame
    MMIO_WR(0xC284, 0x20000102);  
    MMIO_WR_MASK(0xC740, 0x1EC46E6F, 0x1EC46E6F);  
    /* Select PHY Function as HDMI */ 
    write_reg_mask(CRFF, VIACR, BIT0, BIT0);
    /* Select HDTV0 source */
    write_reg_mask(CRFF, VIACR, BIT1, BIT1);
}

static void 
via_hdmi_dpms(xf86OutputPtr output, int mode)
{
     DEBUG(ErrorF("via_hdmi_dpms, mode = %d\n", mode));
    
     ScrnInfoPtr    pScrn = output->scrn;
     VIAPtr         pVia = VIAPTR(pScrn);     
     ViaOutputInfo  commonInfo = ((ViaHdmiPrivateInfoPtr)
                                    (output->driver_private))->commonInfo;
     ViaHdmiPrivateInfoPtr hdmiInfo = output->driver_private;
     CARD32         subChipName = commonInfo.subChipName;
     CARD32         serialPort = hdmiInfo->commonInfo.serialPort;
     CARD32         diPort = commonInfo.diPort;
     CARD32         dwTemp;
     int            usedIga;
     
     switch (mode) {
        case DPMSModeOn: /* 0 */
            usedIga = ((VIACrtcPrivatePtr)(output->crtc->driver_private))->iga;               
            if (NONE_SUBCHIP == subChipName)
                return;
               
            /*1. Turn on DI port clock, set Iga source for DI port*/
            viaSetOutputPath(pScrn->scrnIndex, diPort, usedIga, pVia->Chipset);

            if (SUBCHIP_AD9389 == subChipName) { 
            /*2. Enable HDMI transmitter*/
                DEBUG(ErrorF("Initial AD9389\n"));
                viaInitializeAD9389(output);
            } else if (SUBCHIP_INTEGRATED_HDMI == subChipName) {
                // enable video
                MMIO_WR_MASK(0xC640, 0x8, 0x8);
                // enable HDMI
                MMIO_WR_MASK(0xC280, 0x2, 0x2);
            }

            /*3. Insert a hotplug timer */
            if ((commonInfo.hasHotplug) && (!hdmiInfo->hotplugTimer))
                    hdmiInfo->hotplugTimer = TimerSet(NULL, 0, 3000,
                                            viaHotplugTimer, output);          
            break;

        case DPMSModeStandby:   /* 1 */
        case DPMSModeSuspend:   /* 2 */
        case DPMSModeOff:       /* 3 */                           
                if (NONE_SUBCHIP == subChipName)
                    return;
                
                /* unload Timer */
                if (hdmiInfo->hotplugTimer) {
                    TimerCancel(hdmiInfo->hotplugTimer);
                    hdmiInfo->hotplugTimer=NULL; 
                }
                
                if (SUBCHIP_AD9389 == subChipName)                    
                    /* all circuit power off */
                    viaSerialWriteByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x41, 0x50);                       
                else if (SUBCHIP_INTEGRATED_HDMI == subChipName) {
                    // disable video
                    MMIO_WR_MASK(0xC640, 0x0, 0x8);
                    // disable HDMI
                    MMIO_WR_MASK(0xC280, 0x0, 0x2);
                }
                
                viaDIPortPadOff(pScrn->scrnIndex, diPort);
            break;
            
        default:
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Invalid DPMS mode %d\n", mode);
            break;
    }     
}

/*
* Saves the crtc's state for restoration on VT switch.

static void 
via_hdmi_save (xf86OutputPtr output)
{
}
*/
/*
* Restore's the crtc's state at VT switch.

static void 
via_hdmi_restore (xf86OutputPtr output)
{
}
*/
/*
 * Callback for testing a video mode for a given output.
 *
 * This function should only check for cases where a mode can't be supported
 * on the pipe specifically, and not represent generic CRTC limitations.
 *
 * return MODE_OK if the mode is valid, or another MODE_* otherwise.
 */
static int 
via_hdmi_mode_valid(xf86OutputPtr output, DisplayModePtr mode)
{
    ScrnInfoPtr    pScrn = output->scrn;
    VIAPtr         pVia = VIAPTR(pScrn);     

    /* Currently, only VX855 can support CEA interlace mode. */
    if((pVia->Chipset < VIA_VX855) && (mode->Flags & V_INTERLACE))
        return MODE_NOMODE;
    
    if (mode->type & M_T_DRIVER)
        return MODE_OK;
    else
        return MODE_NOMODE;
}

static Bool 
via_hdmi_mode_fixup(xf86OutputPtr output,
                         DisplayModePtr mode,
                         DisplayModePtr adjusted_mode)
{    
    return TRUE;
}

static void 
via_hdmi_mode_set(xf86OutputPtr output,
                       DisplayModePtr mode,
                       DisplayModePtr adjusted_mode)
{
    DEBUG(ErrorF("via_hdmi_mode_set\n"));  
    ScrnInfoPtr pScrn = output->scrn;
    VIAPtr      pVia = VIAPTR(pScrn);
    CARD8       data;
    ViaHdmiPrivateInfoPtr viaHdmiInfo = output->driver_private;
    CARD32      serialPort = viaHdmiInfo->commonInfo.serialPort;
    CARD32      diPort = viaHdmiInfo->commonInfo.diPort;
    CARD32      subChipName = viaHdmiInfo->commonInfo.subChipName;
    VIACrtcPrivatePtr   viaCrtc = VIACrtcPrivate(output->crtc);
    int usedIga = viaCrtc->iga;


    /* update sync polarity */
    setDiportSyncPolarity(pScrn->scrnIndex, viaHdmiInfo->commonInfo.diPort, adjusted_mode);      

    /* Important: Use HSkew to indicate CEA mode ID. Because the others 
       parameters will be clear by Xorg. I have no choice, there is only this
       parameter can be used. This should be fixed when we found a better 
       solution.
    */
    CEATimingInfoPtr ceaTimingInfoPtr = &CEATimingTable[mode->HSkew-1];

    /* Set aspect ratio */
    if (SUBCHIP_AD9389 == subChipName) {
        viaSerialReadByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x17, &data);  
        if (ceaTimingInfoPtr->Ratio == RATIO_16_9) 
            viaSerialWriteByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x17,
                                (data | 0x2));
        else if (ceaTimingInfoPtr->Ratio == RATIO_4_3)
            viaSerialWriteByte(serialPort, SUBCHIP_AD9389_SLAVE_ADDR, 0x17,
                                (data & 0xFD));        
    } else if (SUBCHIP_INTEGRATED_HDMI == subChipName) {
        viaSetIntegratedHDMIMode(adjusted_mode);
    }

    /* Set Pixel Timing register */
    if (usedIga == IGA1) {
        viaLoadCrtcPixelTiming(adjusted_mode);
    }
	
    /* Set Hsync Offset, delay one clock (To meet 861-D spec.)*/        
    viaWriteVgaIoBits(REG_CR8A, 0x01, 0x7);
    
    /* If CR8A +1, HSyc must -1 */
    viaWriteVgaIo(REG_CR56, viaReadVgaIo(REG_CR56)-1);
    viaWriteVgaIo(REG_CR57, viaReadVgaIo(REG_CR57)-1);        

    if (adjusted_mode->Flags & V_INTERLACE) {            
        if (viaCrtc->iga == IGA1)
            viaWriteVgaIoBits(REG_CR33, BIT6, BIT6);
        else if (viaCrtc->iga == IGA2) {
            switch (pVia->Chipset)
            {
                case VIA_VX855:
                    viaWriteVgaIoBits(REG_CR67, BIT5, BIT5);                
                    viaWriteVgaIoBits(REG_CRFB, ceaTimingInfoPtr->VSyncOffset&0xFF, 0xFF);
                    viaWriteVgaIoBits(REG_CRFC, (ceaTimingInfoPtr->VSyncOffset&0x700)>>8, 0x07);
                    break;
                case VIA_VX900:
                    viaWriteVgaIoBits(REG_CR67, BIT5, BIT5);                
                    viaWriteVgaIoBits(REG_CRAB, ceaTimingInfoPtr->VSyncOffset&0xFF, 0xFF);
                    viaWriteVgaIoBits(REG_CRAC, (ceaTimingInfoPtr->VSyncOffset&0x700)>>8, 0x07);
                    break;
                default:
                    viaWriteVgaIoBits(REG_CR67, BIT5, BIT5);                
                    viaWriteVgaIoBits(REG_CRFB, ceaTimingInfoPtr->VSyncOffset&0xFF, 0xFF);
                    viaWriteVgaIoBits(REG_CRFC, (ceaTimingInfoPtr->VSyncOffset&0x700)>>8, 0x07);
                    break;                    
            }
        }
    } else { /* non-interlace, clear interlace setting. */
        if (viaCrtc->iga == IGA1)
            viaWriteVgaIoBits(REG_CR33, 0, BIT6);
        else if (viaCrtc->iga == IGA2) {
            viaWriteVgaIoBits(REG_CR67, 0, BIT5);    
            viaWriteVgaIoBits(REG_CRFB, 0, 0xFF);
            viaWriteVgaIoBits(REG_CRFC, 0, 0x07);
        }
    }
    
    /* Patch dot clock for CEA mode. */
    viaPatchDotClock(output, adjusted_mode, viaCrtc->iga);
                
    /*Patch for clock skew*/
    loadHdmiDefaultDPASetting(pVia->Chipset, diPort);
    LoadUserGfxDPASetting(diPort, &viaHdmiInfo->userGfxDPA);
    
}


/* Need to fix to meet hot-plug's behavior */
static xf86OutputStatus 
via_hdmi_detect(xf86OutputPtr output)
{
    return XF86OutputStatusConnected;
}

static void 
via_hdmi_destroy (xf86OutputPtr output)
{
    ViaHdmiPrivateInfoPtr viaHdmiInfo = output->driver_private;

    if (viaHdmiInfo->commonInfo.rawEDID)
        xfree(viaHdmiInfo->commonInfo.rawEDID);

    if (viaHdmiInfo)
        xfree(viaHdmiInfo);
}


void 
via_hdmi_prepare(xf86OutputPtr output)
{
    output->funcs->dpms (output, DPMSModeOff);
}

void 
via_hdmi_commit (xf86OutputPtr output)
{
    DEBUG(ErrorF("via_hdmi_commit\n"));

    output->funcs->dpms (output, DPMSModeOn);
}

DisplayModePtr 
via_hdmi_get_modes (xf86OutputPtr output)
{    
    DEBUG(ErrorF("via_hdmi_get_modes\n"));
    ScrnInfoPtr pScrn = output->scrn;
    ViaHdmiPrivateInfoPtr viaHdmiInfo = output->driver_private;
    CARD32      subChipName = viaHdmiInfo->commonInfo.subChipName;
    unsigned char*  rawEDID = viaHdmiInfo->commonInfo.rawEDID;

    /* Xorg's EDID parser is incorrect to process interlace mode's modeline.
    *  So we don't get modeline from detail timing. we use the EDID's 
    *  video block to get the monitor's support capabilities. And we use this
    *  capabilities to map to our builit-in CEA mode table.
    */
       
    if (SUBCHIP_INTEGRATED_HDMI == subChipName) {
        /* Read Std and Ext EDID block */
        if (viaGetStdEDIDBlockByHDCP(rawEDID)) {
            DEBUG(ErrorF("Read Std EDID Success\n"));
            if(rawEDID[0x7E]) {
                if (viaGetExtEDIDBlockByHDCP(rawEDID+128, rawEDID[0x7E])) {
                    DEBUG(ErrorF("Read Ext EDID Success\n"));
                    viaHdmiInfo->hasExtEDID = TRUE;
                } else {
                    DEBUG(ErrorF("Read Ext EDID Fail\n"));                
                    viaHdmiInfo->hasExtEDID = FALSE;
                }
            } else {
                DEBUG(ErrorF("No extension block\n"));
                viaHdmiInfo->hasExtEDID = FALSE;
            }
        }else {
            DEBUG(ErrorF("Read Std EDID Fail\n"));
        }
        /* Detect the max support mode size */
        viaFindMaxSupportMode(pScrn, rawEDID, rawEDID+128,
                              &viaHdmiInfo->commonInfo.physicalWidth,
                              &viaHdmiInfo->commonInfo.physicalHeight);          
    }

    return viaAddCEAModeLine(output, rawEDID+128);

    
#if 0    /* Reservered. (see the above description).*/
    ViaHdmiPrivateInfoPtr hdmiInfo = output->driver_private;    
    ScrnInfoPtr         pScrn = output->scrn;
    DisplayModePtr      stdModes = NULL;
    DisplayModePtr      newModes = NULL;  
    DisplayModePtr      builtInMode = NULL;  
    xf86MonPtr          monitorInfo = NULL;

    /* Parse the EDID block */
    monitorInfo = xf86InterpretEDID(pScrn->scrnIndex,g_StdEdid);
    if (monitorInfo)
        xf86OutputSetEDID(output, monitorInfo);

    stdModes = xf86OutputGetEDIDModes(output);        
    viaPrintModeList(stdModes);  

    /* Add mode according to extension EDID. */
    if(hdmiInfo->hasExtEDID)
        newModes = viaAddExtDetailTiming(output, stdModes, g_ExtEdid);
    else 
        DEBUG(ErrorF("Extension EDID doesn't exists\n"));

    return newModes;
#endif
}


static const xf86OutputFuncsRec via_hdmi_output_funcs = {
    .dpms = via_hdmi_dpms,
    .save = NULL,
    .restore = NULL,
    .mode_valid = via_hdmi_mode_valid,
    .mode_fixup = via_hdmi_mode_fixup,
    .prepare = via_hdmi_prepare,
    .mode_set = via_hdmi_mode_set,
    .commit = via_hdmi_commit,
    .detect = via_hdmi_detect,    
    .get_modes = via_hdmi_get_modes,
    .destroy = via_hdmi_destroy,        
};

/*
Function Name:  checkHdmiSupport
Description:    Check if HDMI supported by platform
*/
static Bool
checkHdmiSupport(VIAPtr pVia, xf86OutputPtr output)
{
    Bool retVal = FALSE;
    ScrnInfoPtr pScrn = output->scrn;
    ViaHdmiPrivateInfoPtr hdmiInfo = output->driver_private;

    /* default HDMI type is internal */
    if (hdmiInfo->commonInfo.type == DISP_DEFAULT_SETTING) {
        
        if (pVia->GfxDispCaps & INTERNAL_HDMI)
            hdmiInfo->commonInfo.type = DISP_TYPE_INTERNAL;         
        else    
            hdmiInfo->commonInfo.type = DISP_TYPE_EXTERNAL;         
    }

    /* default HDMI DI port*/
    if (hdmiInfo->commonInfo.diPort == DISP_DEFAULT_SETTING) {
        if (hdmiInfo->commonInfo.type == DISP_TYPE_INTERNAL)
            /* 410 integrated HDMI doesn't have DI port */
            hdmiInfo->commonInfo.diPort = DISP_DI_NONE;
        else
            hdmiInfo->commonInfo.diPort = DISP_DI_DVP1;
    }
    
    /*if HDMI DI port is free*/
    if (checkDiPortUsage(pVia, hdmiInfo->commonInfo.diPort)) {

        /* external HDMI */
        if (hdmiInfo->commonInfo.type == DISP_TYPE_EXTERNAL) {            
            /* Sense AD9389 */
            viaDIPortPadOn(pScrn->scrnIndex, hdmiInfo->commonInfo.diPort);
            if (ad9389_module_loaded&&senseSubChip(&hdmiInfo->commonInfo.serialPort,
                                SUBCHIP_AD9389)) {
                hdmiInfo->commonInfo.subChipName = SUBCHIP_AD9389;
                hdmiInfo->commonInfo.slaveAddress = SUBCHIP_AD9389_SLAVE_ADDR;

                if (hdmiInfo->commonInfo.ddcPort == DISP_DEFAULT_SETTING)
                    hdmiInfo->commonInfo.ddcPort = hdmiInfo->commonInfo
                                                    .serialPort;

                pVia->MbDiPortUsedInfo |= hdmiInfo->commonInfo.diPort;
                retVal = TRUE;

                /* Init AD9389 at here to avoid enable/disable frequently. */
                viaInitializeAD9389(output);
            }else{
                if(!ad9389_module_loaded){
                    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
                    "Failed to load AD9389 module, "
                    "so we don't creat hdmi output "
                    "if you need hdmi output, please load the submodule.\n"));
                }
            }
            /*Turn off relevant di port clk set on previously to keep things consistent */
            viaDIPortPadOff(pScrn->scrnIndex, hdmiInfo->commonInfo.diPort);
            
        }
   } else { 
        /* internal HDMI */
        if (hdmiInfo->commonInfo.type == DISP_TYPE_INTERNAL) {
            if (pVia->GfxDispCaps & INTERNAL_HDMI) {
                hdmiInfo->commonInfo.subChipName = SUBCHIP_INTEGRATED_HDMI;
                hdmiInfo->commonInfo.slaveAddress = NONE_SUBCHIP_SLAVE_ADDR;
                pVia->GfxDispCaps &= ~INTERNAL_HDMI;
                // enable bangap
                MMIO_WR_MASK(0xC740,  0x1, 0x1);
                retVal = TRUE;
            }
        }
   }

   return retVal;
}

static void
parseHdmiOption(xf86OutputPtr output)
{
    char *s = NULL;
    int value = 0;    
    int sPort = DISP_SERIALP_NONE;
    int dPort = DISP_SERIALP_NONE;
    ViaHdmiPrivateInfoPtr hdmiInfo = output->driver_private;

    hdmiInfo->userGfxDPA.isClkPolarityUsed = FALSE;
    hdmiInfo->userGfxDPA.isClkAdjustUsed = FALSE;
    hdmiInfo->userGfxDPA.isClkDrivingSelUsed = FALSE;
    hdmiInfo->userGfxDPA.isDataDrivingSelUsed = FALSE;

    /* Default value */
    hdmiInfo->commonInfo.diPort = DISP_DEFAULT_SETTING;
    hdmiInfo->commonInfo.type = DISP_DEFAULT_SETTING;
    hdmiInfo->commonInfo.serialPort = DISP_DEFAULT_SETTING;
    hdmiInfo->commonInfo.ddcPort = DISP_DEFAULT_SETTING;
    hdmiInfo->commonInfo.hasHotplug = FALSE;    
    hdmiInfo->attachAllModes = FALSE;
    
    /* if no "HDMI" monitor section, all use default setting */ 
    if (output->conf_monitor)
        xf86ProcessOptions (output->scrn->scrnIndex,
                    output->conf_monitor->mon_option_lst,
                    ViaHdmiOptions);
 

    /* parse option "DIPort" */
    if ((s = xf86GetOptValString(ViaHdmiOptions, OPTION_HDMI_DIPORT)))
        hdmiInfo->commonInfo.diPort = transformDiPort(s);        

    /* parse option "Type" */
    if ((s = xf86GetOptValString(ViaHdmiOptions, OPTION_HDMI_TYPE)))
        hdmiInfo->commonInfo.type = transformOutputType(s);        

    /* parse option "SerialPort" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_SERIALPORT, &sPort))
        hdmiInfo->commonInfo.serialPort = transformPort(sPort);

    /* parse option "DDCPort" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_DDCPORT, &dPort))
        hdmiInfo->commonInfo.ddcPort = transformPort(dPort);

    /* parse option  "HotPlug" */
    if (xf86ReturnOptValBool(ViaHdmiOptions, OPTION_HDMI_Hotplug, FALSE))
        hdmiInfo->commonInfo.hasHotplug = TRUE;

    /* parse option  "DiscardEDID" */
    if (xf86ReturnOptValBool(ViaHdmiOptions, OPTION_HDMI_ATTACH_ALL_MODES, FALSE))
        hdmiInfo->attachAllModes = TRUE;
    
    /* parse user clock skew setting*/
    /* 1. parse option "ClockPolarity" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_CLOCK_POLARITY,
        &value)) {
        hdmiInfo->userGfxDPA.clkPolarity = value&BIT0;
        hdmiInfo->userGfxDPA.isClkPolarityUsed = TRUE;
    }

    /* 2. parse option "ClockAdjust" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_CLOCK_ADJUST,
        &value)) {
        hdmiInfo->userGfxDPA.clkAdjust = value;
        hdmiInfo->userGfxDPA.isClkAdjustUsed = TRUE;
    }

    /* 3. parse option "ClockDrivingSelection" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_CLOCK_DRIVING_SELECTION,
        &value)) {
        hdmiInfo->userGfxDPA.clkDrivingSel = value;
        hdmiInfo->userGfxDPA.isClkDrivingSelUsed = TRUE;
    }

    /* 4. parse option "DataDrivingSelection" */
    if (xf86GetOptValInteger(ViaHdmiOptions, OPTION_HDMI_DATA_DRIVING_SELECTION,
        &value)) {
        hdmiInfo->userGfxDPA.dataDrivingSel = value;
        hdmiInfo->userGfxDPA.isDataDrivingSelUsed = TRUE;
    }        
}

void
via_hdmi_init(ScrnInfoPtr pScrn, const char* name)
{    
    VIAPtr                  pVia = VIAPTR(pScrn);
    xf86OutputPtr           outputHdmi;
    ViaHdmiPrivateInfoPtr   hdmiInfo;  
    
    outputHdmi = xf86OutputCreate (pScrn, &via_hdmi_output_funcs, name);       
    
    if (!outputHdmi) {
        DEBUG(ErrorF("xf86OutputCreate %s Fail.\n", name));
        return;        
    }
    /* allocate HDMI private info structure */
    hdmiInfo = xnfcalloc (sizeof (ViaHdmiPrivateInfo), 1);
    if (!hdmiInfo) {
        xf86OutputDestroy(outputHdmi);
        DEBUG(ErrorF("Allocate %s private info Fail.\n", name));
        return; 
    }
    outputHdmi->driver_private = hdmiInfo;
    /* allocate memory to store EDID info */
    hdmiInfo->commonInfo.rawEDID = xcalloc(1,sizeof(unsigned char) * (256));
    if (!hdmiInfo->commonInfo.rawEDID) {
        xf86OutputDestroy(outputHdmi);
        DEBUG(ErrorF("Allocate %s EDID space Fail.\n", name));
        return; 
    }
    
    /* Read HDMI options in HDMI monitor section */
    parseHdmiOption(outputHdmi);

    if (checkHdmiSupport(pVia, outputHdmi)) {
        if (pVia->hdmi1Created)
            outputHdmi->possible_crtcs = 0x1;
        else
            outputHdmi->possible_crtcs = 0x2;
        outputHdmi->possible_clones = 0;
        outputHdmi->interlaceAllowed = TRUE;
        outputHdmi->doubleScanAllowed = FALSE;
        pScrn->adjustFlags |= INTERLACE_HALVE_V;

        if (!xf86NameCmp(name, OUTPUT_HDMI_NAME))
            pVia->hdmi1Created = TRUE;
		
    } else
        /* also free HDMI private date */
        xf86OutputDestroy(outputHdmi);  
}
