 /*
 * Copyright 2009 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2009 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "via_driver.h"
#include "via_dp.h"
#include "via_regs.h"
#include "via_mode.h"
#include "via_output.h"

void viaCalculate_TU(xf86OutputPtr output, unsigned long PixelClock, int bpp_byte);

/******************************************************************
*
*                                          DP1 Function                        
*
*******************************************************************/
void VIADPDisableEPHY()
{
    // Disable Bandgap to disable all EPHY functions
    MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x00000000, 0x00000001);
}

Bool VIADPInitEPHY(VIABIOSInfoPtr pBIOSInfo)
{
    CARD32 dwData;
     
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPInitEPHY\n"));
 
    // (1) enable Bandgap and set EPHY symbol rate
    dwData = MMIO_RD(DP_EPHY_PLL_REG) & 0xFFF800FF;
    // real EPHY symbol rate decided by HW link training
    switch (pBIOSInfo->DPSettingInfo.LinkSpeed)
    {
        case Speed_162MHz:
            dwData |= 0x0004BE00;
            break;
        case Speed_270MHz:
            dwData |= 0x00029E00;
            break;
        default:
            dwData |= 0x00029E00;
            break;	
    }
    dwData |= 0x00000001;
    MMIO_WR(DP_EPHY_PLL_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 1000); 
     
    // (2) start AUX channel, CMOP is on, Tx and Rx are off
    dwData = MMIO_RD(DP_EPHY_MISC_PWR_REG) & 0xFFFF3FFF;
    dwData |= 0x00004000;
    MMIO_WR(DP_EPHY_MISC_PWR_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 4000);
    dwData = MMIO_RD(DP_EPHY_MISC_PWR_REG) & 0xFFFF3FFF;
    dwData |= 0x00008000;
    MMIO_WR(DP_EPHY_MISC_PWR_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 4000);

    // (3) enable Mpll and Mpll reg power          
    dwData = MMIO_RD(DP_EPHY_PLL_REG) & 0xF9FFFFF9;
    dwData |= 0x00000006;
    MMIO_WR(DP_EPHY_PLL_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 1000);
    // enable DP
    MMIO_WR(DP_ENABLE_IF_REG, (MMIO_RD(DP_ENABLE_IF_REG) | 0x00000001));
    VIADelayIn_usec(pBIOSInfo, 1000);

    // (4) enable Tpll and Tpll reg power
    MMIO_WR(DP_EPHY_PLL_REG, (MMIO_RD(DP_EPHY_PLL_REG) | 0x06000000));
    VIADelayIn_usec(pBIOSInfo, 1000);

    // (5) Resister Tuning (RT), RT reset
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFD));
    // power up RT
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000001));
    // disable RT reset
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000002));
    VIADelayIn_usec(pBIOSInfo, 1000);
    // enable RT auto calibration mode
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000004));
    VIADelayIn_usec(pBIOSInfo, 1000);
    // disable RT mode
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFB));
    VIADelayIn_usec(pBIOSInfo, 1000);
    // power down RT
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFE));
     
    // (6)
    // [19:8] Tx power status default settings for all 4 lanes
    // [7:0] Tx power down mode for all 4 lanes, [15:8] HDMI TX output voltage swing 1200 mV for all 4 lanes
    dwData = MMIO_RD(DP_EPHY_TX_PWR_REG) & 0x0000F0FF;
    dwData |= 0xAAFF0F00;
    MMIO_WR(DP_EPHY_TX_PWR_REG, dwData); 
    VIADelayIn_usec(pBIOSInfo, 4000);
    // [7:0] Tx power control function CMOP on / PISO off for all 4 lanes
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0x00FFFFFF)); // Tx power control function on for all 4 lanes
    VIADelayIn_usec(pBIOSInfo, 1000);
    // Driver awaken for all 4 lanes
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x0000F000)); 
    MMIO_WR(DP_EPHY_MISC_PWR_REG, (MMIO_RD(DP_EPHY_MISC_PWR_REG) | 0x00000001));
    VIADelayIn_usec(pBIOSInfo, 2000);
}

void VIADPClearAuxWriteRegs()
{
    MMIO_WR(AUX_W_DATA0_REG, 0x0);
    MMIO_WR(AUX_W_DATA1_REG, 0x0);
    MMIO_WR(AUX_W_DATA2_REG, 0x0);
    MMIO_WR(AUX_W_DATA3_REG, 0x0);
}

void VIADPClearAuxReadRegs()
{
    MMIO_WR(AUX_R_DATA0_REG, 0x0);
    MMIO_WR(AUX_R_DATA1_REG, 0x0);
    MMIO_WR(AUX_R_DATA2_REG, 0x0);
    MMIO_WR(AUX_R_DATA3_REG, 0x0);
}

Bool VIADPReplyCmdStatus(VIABIOSInfoPtr pBIOSInfo)
{
    ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr      pVia = VIAPTR(pScrn);
    CARD32 dwAUXCMDStatus;
    CARD32 dwData;
    int i, j;
 
    for(i=0; i<NUM_RE_ISSUE_COMMAND; i++)
    {        
        for (j=0; j<NUM_RECHECK; j++)
        {
            VIADelayIn_usec(pBIOSInfo, 1000);
             
            dwAUXCMDStatus = MMIO_RD(AUX_TIMER_REG) & 0x3C000007;
             
            if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1 && dwAUXCMDStatus == ACK_OK)
            {
                return TRUE;
            }
            else if (dwAUXCMDStatus == 0x00000001) // hardware issue
            {
                return TRUE;
            }
            else if ((dwAUXCMDStatus & 0xFFFF) == RE_ISSUE_CMD)
            {
                // re-issue Cmd 
                dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
                dwData |= 0x03;
                MMIO_WR(AUX_TIMER_REG, dwData);
                break;
            }
        }
    }
 
    return FALSE;
}

Bool VIADPWriteDPCD(VIABIOSInfoPtr pBIOSInfo, CARD32 dwoffset, int ByteCount)
{
    CARD32 dwCmd;
    CARD32 dwData;
    
    DEBUG(ErrorF("VIADPWriteDPCD\n"));
     
    dwCmd = dwoffset << 12;   // AUX read offset
    dwCmd |= ByteCount << 4;  // AUX read length
    dwCmd |= AUX_CMD_WRITE;   // AUX write CMD
    MMIO_WR(AUX_CMD_REG, dwCmd);

    // issue request based on above
    dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwData); // trigger

    VIADelayIn_usec(pBIOSInfo, 10000);  
  
    return VIADPReplyCmdStatus(pBIOSInfo);
}

Bool VIADPReadDPCD(VIABIOSInfoPtr pBIOSInfo, CARD32 dwOffset)
{
    CARD32 dwCmd;
    CARD32 dwData;
     
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPReadDPCD\n")); 

    VIADPClearAuxReadRegs();

    // assign command
    dwCmd = dwOffset << 12;   // AUX read offset
    dwCmd |= 1 << 4;          // AUX read length
    dwCmd |= AUX_CMD_READ;    // AUX read CMD
    MMIO_WR(AUX_CMD_REG, dwCmd);

    // issue request based on above
    dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwData); 

    return VIADPReplyCmdStatus(pBIOSInfo);
}
 
void VIADPReadDPCDVersion(VIABIOSInfoPtr pBIOSInfo)
{   
    CARD32 dwDPCDVersion;
     
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPReadDPCDVersion\n"));
    
    // read DPCD Version
    if (VIADPReadDPCD(pBIOSInfo, 0x00))
    {  
        dwDPCDVersion = MMIO_RD(AUX_R_DATA0_REG) & 0xFF;
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DPCDVersion = %x\n",dwDPCDVersion));
         
        switch (dwDPCDVersion)
        {
            case 0x10:
                pBIOSInfo->DPSettingInfo.DPCDVersion |= DPCD_VER_1 | DPCD_REV_0; // Version 1.0
                break;
            case 0x11:
                pBIOSInfo->DPSettingInfo.DPCDVersion |= DPCD_VER_1 | DPCD_REV_1; // Version 1.1
                break;
            default:   
                pBIOSInfo->DPSettingInfo.DPCDVersion = NONEVERSION;
                break;
        }
    }
	
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
}

Bool VIASenseDP(VIABIOSInfoPtr pBIOSInfo)
{
    VIABIOSInfoPtr pVia = pBIOSInfo; 
    Bool IsEPHYEnabled = FALSE;
    int ret = FALSE;
     
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIASenseDP\n"));  
	
    VIADPReadDPCDVersion(pBIOSInfo);
     
    if (pBIOSInfo->DPSettingInfo.DPCDVersion)
    {
        ret = TRUE;
    }
     
    return ret;
}
  
Bool VIADPLinkTraining(VIABIOSInfoPtr pBIOSInfo)
{
    CARD32 dwData;
      
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPLinkTraining\n"));
      
    VIADPClearAuxReadRegs(); 
    VIADPClearAuxWriteRegs();
    pBIOSInfo->DPSettingInfo.IsLinkTrainingDone = FALSE;
	
    // reset link training
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP_NAUD_MUTE_REG, 0x80000000, 0x80000000); 
    VIADelayIn_usec(pBIOSInfo, 1000);
    MMIO_WR_MASK(DP_NAUD_MUTE_REG, 0x0, 0x80000000);
    VIADelayIn_usec(pBIOSInfo, 1000);
 
    // [9:7]=100b, 4 lanes, [10]=0, disable SW link training , 
    // [12:11]=01b, set clock recovery state
    dwData = MMIO_RD(DP_LINK_TRAINING_REG) & 0xFFFF0000; 
    dwData |= 0x00000200; 

      switch (pBIOSInfo->DPSettingInfo.LinkSpeed)
      {
          case Speed_162MHz:
              dwData &= 0xFFFFFFFD;
              break;
          case Speed_270MHz:
              dwData |= 0x00000002;
              break;
          default:
              dwData |= 0x00000002;
              break;          
      }    
    MMIO_WR(DP_LINK_TRAINING_REG, dwData);
  
    // Reset EPHY
    MMIO_WR_MASK(DP_EPHY_TX_PWR_REG, 0x55000000, 0xFF000000);
    VIADelayIn_usec(pBIOSInfo, 2000);
    MMIO_WR_MASK(DP_EPHY_TX_PWR_REG, 0x00000000, 0xFF000000);
 
    // enable scramble
    MMIO_WR_MASK(DP_VIDEO_CTRL_REG, 0x00000001, 0x00000001);
 
    // enable DP, set enhanced mode
    dwData = MMIO_RD(DP_ENABLE_IF_REG) & 0xC3000000; 
    dwData |= 0x00000005; // if enable standard mode , set 0x00000001
    MMIO_WR(DP_ENABLE_IF_REG, dwData);
 
    switch (pBIOSInfo->DPSettingInfo.LinkSpeed)
    {
        case Speed_162MHz:
            MMIO_WR(AUX_W_DATA0_REG, 0x000018406); // if enable standard mode , set 0x000010406
            break;
        case Speed_270MHz:
            MMIO_WR(AUX_W_DATA0_REG, 0x00001840A); // if enable standard mode , set 0x00001040A
            break;
        default:
            MMIO_WR(AUX_W_DATA0_REG, 0x00001840A);
            break;          
    }   
      
    MMIO_WR(AUX_W_DATA1_REG, 0x0);
      
    if (!VIADPWriteDPCD(pBIOSInfo, 0x100, 2))
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
        return FALSE;
    }
      
    //Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
    VIADelayIn_usec(pBIOSInfo, 20000);
  
    dwData = MMIO_RD(DP_LINK_TRAINING_REG) & 0x7FFFFFFF; 
    dwData |= 0x0000003C;  
   
    if (pBIOSInfo->DPSettingInfo.LinkSpeed == Speed_270MHz)
    {
         dwData |= 0x00000002;
    } 
    MMIO_WR(DP_LINK_TRAINING_REG, dwData);
          
    // enable HW link training
    MMIO_WR_MASK(DP_LINK_TRAINING_REG, 0x00000001, 0x00000001); 
  
    VIADelayIn_usec(pBIOSInfo, 20000);
      
    if ((MMIO_RD(DP_LINK_TRAINING_REG) & 0x80000000) != 0x80000000)
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
        return FALSE;
    }
 
    VIADelayIn_usec(pBIOSInfo, 4000);
     
    if ((MMIO_RD(DP_ENABLE_IF_REG) & 0x30000000) != 0x30000000)
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
          
        return FALSE;
    }

    pBIOSInfo->DPSettingInfo.IsLinkTrainingDone = TRUE;
	
    // clear [31], and [0]
    MMIO_WR_MASK(DP_LINK_TRAINING_REG, 0x0, 0x80000001);
     
    //Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
  
    return TRUE;
}

int VIADPGetLinkSpeed(VIABIOSInfoPtr pBIOSInfo)
{
    DPSETTINGINFOPTR pDPSettingInfo = &(pBIOSInfo->DPSettingInfo);
    Bool IsEPHYEnabled = FALSE;
    CARD32 dwLinkSpeed;
	
    if (pDPSettingInfo->IsLinkTrainingDone)
        return pDPSettingInfo->LinkSpeed;
	
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPGetLinkSpeed\n"));

    // Patch for driver flow
    IsEPHYEnabled = MMIO_RD(DP_EPHY_PLL_REG) & BIT0;
    if (!IsEPHYEnabled)
        VIADPInitEPHY(pBIOSInfo);

    VIADPReadDPCD(pBIOSInfo, 0x01);
    dwLinkSpeed = MMIO_RD(AUX_R_DATA0_REG) & 0xFF;
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);

    // Use default link speed of DP monitor
    switch (dwLinkSpeed) {
        case 0x06:
        pDPSettingInfo->LinkSpeed = Speed_162MHz;
        break;        
        case 0x0A:
        pDPSettingInfo->LinkSpeed = Speed_270MHz;
        break;          
        default:
        pDPSettingInfo->LinkSpeed = Speed_270MHz;
        break;
    }

    // Start link training to confirm link speed
    if ((!VIADPLinkTraining(pBIOSInfo)) && (pDPSettingInfo->LinkSpeed == Speed_270MHz))
    {
        // Reduce link speed to 162 MHz and retry link training if 270 MHz link training fails
        MMIO_WR(DP_LINK_TRAINING_REG, (MMIO_RD(DP_LINK_TRAINING_REG) & 0x7FFFFFFC));
        VIADPClearAuxWriteRegs();
        // Clear DPCD training pattern
        VIADPWriteDPCD(pBIOSInfo, 0x102, 1);
        pDPSettingInfo->LinkSpeed = Speed_162MHz;
        // Reset EPHY for changed link speed
        VIADPInitEPHY(pBIOSInfo);
        if (!VIADPLinkTraining(pBIOSInfo)) {
            pDPSettingInfo->LinkSpeed = Speed_NONE;
            pDPSettingInfo->IsLinkTrainingDone = TRUE;
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "There is no suitable link speed for DP monitor\n");
        }
    }

    return pDPSettingInfo->LinkSpeed;
}

void VIAEnableDP(VIABIOSInfoPtr pBIOSInfo)
{
    DPSETTINGINFOPTR pDPSettingInfo = &(pBIOSInfo->DPSettingInfo);
	
    DEBUG(ErrorF("VIAEnableDP\n"));
  
    // Must Enable HDCP first
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x0, 0xC0000000);
    
    // if DP mode, video input selection - sp1   
    MMIO_WR(DP_VIDEO_CTRL_REG, (MMIO_RD(DP_VIDEO_CTRL_REG) & 0x3FFC0001));

    // enable DP and video  
    MMIO_WR(DP_ENABLE_IF_REG, MMIO_RD(DP_ENABLE_IF_REG) | 0x00000009);
     
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0); 
}
 
void VIADisableDP(VIABIOSInfoPtr pBIOSInfo)
{
    DEBUG(ErrorF("VIADisableDP\n"));

    // disable DP and video
    MMIO_WR(DP_ENABLE_IF_REG, (MMIO_RD(DP_ENABLE_IF_REG) & 0xFFFFFFF6));
     
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0); 
}

void VIACalculate_TU(DPSETTINGINFOPTR pDPSettingInfo, CARD32 PixelClock, int bpp_byte)
{
    int LinkClock;
     
    DEBUG(ErrorF("VIACalculate_TU\n"));
 
    switch (pDPSettingInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }
    
    bpp_byte = 24;   

    // TU ratio = (Dclk * bpps) * 32768 / (Ls_clk * Lane # * 8)
    pDPSettingInfo->TURatio = ((PixelClock / 1000000) * bpp_byte * 4096)/(LinkClock * LANE);    
 }

void VIADpHorWidthAndTu(VIABIOSInfoPtr pBIOSInfo, struct display_timing crt_reg, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int TUSize = 48;
    int Width;
     
    VIACalculate_TU(&(pBIOSInfo->DPSettingInfo), PixelClock, pBIOSInfo->bitsPerPixel);
    Width = (crt_reg.hor_addr / 4) - 1;
    dwTemp = pBIOSInfo->DPSettingInfo.TURatio << 17 | (TUSize - 1) << 11 | Width;
    MMIO_WR(DP_HWIDTH_TUSIZE_REG, dwTemp);
    //MMIO_WR_MASK(DP_VER_EXT_PKT_HEAD_REG, Width/2048, 0x40000000); //  DP H width[11]
}

void VIADpHorLineDuration(VIABIOSInfoPtr pBIOSInfo, struct display_timing crt_reg, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int LinkClock;
    int HTotal;
    int HBlank;

    // units of Dclk and Ls_clk are MHz
    switch (pBIOSInfo->DPSettingInfo.LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }
 
    HBlank = ((((crt_reg.hor_total - crt_reg.hor_addr) * LinkClock) / (PixelClock / 10000)) - 15) & 0x00000FFF;
    HTotal = ((LinkClock * crt_reg.hor_total / (PixelClock / 1000000)) - 15) & 0x00007FFF;
 
    dwTemp = HBlank << 15 | HTotal;
    MMIO_WR(DP_HLINE_DUR_REG, dwTemp);
}

void VIADpMvid(VIABIOSInfoPtr pBIOSInfo, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int LinkClock;

    // units of Dclk and Ls_clk are MHz
    switch (pBIOSInfo->DPSettingInfo.LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }
 
    dwTemp = ((PixelClock / 1000000) * 32768 / LinkClock) & 0x19FFFFFF;
    dwTemp |=  0x20000000; // 8 bits, MMC6CC[31:29]=001b 
    dwTemp &=  0xFEFFFFFF;
    MMIO_WR(DP_MVID_MISC0_REG, dwTemp); // [24]=0: Asynchronous mode
    MMIO_WR_MASK(DP_ENABLE_IF_REG, 0x02000000, 0x02000000); // Generate MVID in Asynchronous mode
}

void VIADpHorVerTotal(struct display_timing crt_reg)
{
    CARD32 dwTemp;
    
    dwTemp = crt_reg.ver_total << 16 | crt_reg.hor_total;
    MMIO_WR(DP_H_ATTR_REG, dwTemp);
}
void VIADpHorVerOffset(struct display_timing crt_reg)
{
    CARD32 dwTemp;
 
    dwTemp = (crt_reg.ver_total - crt_reg.ver_sync_start) << 16 | 
             (crt_reg.hor_total - crt_reg.hor_sync_start);
    MMIO_WR(DP_HV_START_REG, dwTemp);
}

void VIADpHorVerSyncWidthAndPolarity(struct display_timing crt_reg, int HSync, int VSync)
{
    CARD32 dwTemp;
                                                  
    dwTemp = VSync << 31 | crt_reg.ver_sync_end << 16 |
             HSync << 15 | crt_reg.hor_sync_end;
    MMIO_WR(DP_POLARITY_WIDTH_REG, dwTemp);
}

void VIADpHorVerActive(struct display_timing crt_reg)
{
    CARD32 dwTemp;
 
    dwTemp = crt_reg.ver_addr << 16 | crt_reg.hor_addr;
    MMIO_WR(DP_ACITVE_WH_REG, dwTemp);
}

void VIADpNvid()
{
    // [23:0] Stream Clock Recovery :N  
    MMIO_WR_MASK(DP_ATTR_DATA_REG, 32768, 0x00FFFFFF);
}
 
void VIADpDelayReg()
{
    // Default: [29:24] = 20h for delay
    MMIO_WR_MASK(DP_VIDEO_CTRL_REG, 0x20000000, 0x3F000000);
}

Bool VIADPSetupMainLink(VIABIOSInfoPtr pBIOSInfo, struct crt_mode_table ModeInfo)
{
    VIABIOSInfoPtr pVia = pBIOSInfo;
    struct display_timing crt_reg = ModeInfo.crtc;
    CARD32 PixelClock = ModeInfo.clk;
    int hsync_polarity = ModeInfo.h_sync_polarity;
    int vsync_polarity = ModeInfo.v_sync_polarity;
	
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPSetupMainLink\n"));

    if (pBIOSInfo->DPSettingInfo.LinkSpeed != Speed_NONE) {
        // Field ID flag not invert
        MMIO_WR_MASK(DP_ENABLE_IF_REG, 0x0, BIT1);

        VIADpHorWidthAndTu(pBIOSInfo, crt_reg, PixelClock); 
        VIADpHorLineDuration(pBIOSInfo, crt_reg, PixelClock);
        VIADpMvid(pBIOSInfo, PixelClock);
        VIADpHorVerTotal(crt_reg);
        VIADpHorVerOffset(crt_reg);
        VIADpHorVerSyncWidthAndPolarity(crt_reg, hsync_polarity, vsync_polarity);
        VIADpHorVerActive(crt_reg);
        VIADpNvid();
        VIADpDelayReg();

        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "End Setup MainLink!!\n"));
    }
} 

Bool VIADPReadEDID(VIABIOSInfoPtr pBIOSInfo, CARD32 slave_addr, CARD32 dwOffset, CARD8 *pData)
{   
    VIABIOSInfoPtr pVia = pBIOSInfo;
 
    CARD32 dwCmd;

    VIADPClearAuxReadRegs();
    VIADPClearAuxWriteRegs();
 
    // 1. write slave addr offset for start
    dwCmd = dwOffset & 0x000000FF;
    MMIO_WR(AUX_W_DATA0_REG, dwCmd);
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_WRITE;
    MMIO_WR(AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwCmd);
 
    if(!VIADPReplyCmdStatus(pBIOSInfo))
    {
        return FALSE;
    }
 
    // 2. write command with read length
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_READ;
    MMIO_WR(AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwCmd);
 
    if(!VIADPReplyCmdStatus(pBIOSInfo))
    {
        return FALSE;
    }
 
    // 3. read 
    *pData = (MMIO_RD(AUX_R_DATA0_REG) & 0xFF);
 
    return TRUE;
}


 /******************************************************************
 *
 *                                          DP2 Function                        
 *
 *******************************************************************/
void VIADP2DisableEPHY()
{
    MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x00000000, 0x00000001);
}

 Bool VIADP2InitEPHY(VIABIOSInfoPtr pBIOSInfo)
 {
    CARD32 dwData;
     
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2InitEPHY\n"));
 
    // (1) enable Bandgap and set EPHY symbol rate
    dwData = MMIO_RD(DP2_EPHY_PLL_REG) & 0x0C7000FF;
    // real EPHY symbol rate decided by HW link training
    switch (pBIOSInfo->DPSettingInfo2.LinkSpeed)
    {
        case Speed_162MHz:
            dwData |= 0x0002BE00;
            break;
        case Speed_270MHz:
            dwData |= 0x00019E00;
            break;
        default:
            dwData |= 0x00019E00;
            break;	
    }
    dwData |= 0xF1080001;
    MMIO_WR(DP2_EPHY_PLL_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 1000); 
     
    // (2) start AUX channel, CMOP is on, Tx and Rx are off
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xCFFFFFFF;
    dwData |= 0x10000000;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 4000);
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xCFFFFFFF;
    dwData |= 0x20000000;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData);
    VIADelayIn_usec(pBIOSInfo, 4000);

    // (3) enable Mpll and Mpll reg power          
    // (4) enable Tpll and Tpll reg power    
    MMIO_WR(DP2_EPHY_PLL_REG, (MMIO_RD(DP2_EPHY_PLL_REG) & (~0x02000000)));
    VIADelayIn_usec(pBIOSInfo, 1000);

    // (5) Resister Tuning (RT), RT reset
    // (6)
    // [19:8] Tx power status default settings for all 4 lanes
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG2) & 0xFFF000FF;
    dwData |= 0x0000AA00;
    MMIO_WR(DP2_EPHY_TX_PWR_REG2, dwData); 
    // [7:0] Tx power down mode for all 4 lanes, [15:8] HDMI TX output voltage swing 1200 mV for all 4 lanes
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFF0000;
    dwData |= 0x0000FFFF;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData); 
    VIADelayIn_usec(pBIOSInfo, 4000);
    // [7:0] Tx power control function CMOP on / PISO off for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_PWR_REG, (MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFFFFAA)); // Tx power control function on for all 4 lanes
    VIADelayIn_usec(pBIOSInfo, 4000);
    // [7:0] Tx power control function on for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_PWR_REG, (MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFFFF00)); // Tx power control function on for all 4 lanes
    VIADelayIn_usec(pBIOSInfo, 1000);
    // [19:16] Driver awaken for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_IDLE_REG, (MMIO_RD(DP2_EPHY_TX_IDLE_REG) | 0x000F0000));

    // (7) DP2 enable
    MMIO_WR(DP2_ENABLE_IF_REG, (MMIO_RD(DP2_ENABLE_IF_REG) | 0x00000001)); 
    VIADelayIn_usec(pBIOSInfo, 2000);
 }

 void VIADP2ClearAuxWriteRegs()
{
    MMIO_WR(DP2_AUX_W_DATA0_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA1_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA2_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA3_REG, 0x0);
}

void VIADP2ClearAuxReadRegs()
{
    MMIO_WR(DP2_AUX_R_DATA0_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA1_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA2_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA3_REG, 0x0);
}

 Bool VIADP2ReplyCmdStatus(VIABIOSInfoPtr pBIOSInfo)
 {
     ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];
     VIAPtr      pVia = VIAPTR(pScrn);
     CARD32 dwAUXCMDStatus;
     CARD32 dwData;
     int i, j;
 
     //DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2ReplyCmdStatus\n"));
     
     for(i=0; i<NUM_RE_ISSUE_COMMAND; i++)
     {        
         for (j=0; j<NUM_RECHECK; j++)
         {
             VIADelayIn_usec(pBIOSInfo, 1000);
             
             dwAUXCMDStatus = MMIO_RD(DP2_AUX_TIMER_REG) & 0x3C000007;
             if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1 && dwAUXCMDStatus == ACK_OK)
             {
                 return TRUE;
             }
             else if (dwAUXCMDStatus == 0x00000001) // hardware issue
             {
                 return TRUE;
             }
             else if ((dwAUXCMDStatus & 0xFFFF) == RE_ISSUE_CMD)
             {
                 // re-issue Cmd 
                 dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
                 dwData |= 0x03;
                 MMIO_WR(DP2_AUX_TIMER_REG, dwData);
                 break;
             }
         }
     }
     return FALSE;
 }

 void VIADp2TurnOnAuxPowerSequence(VIABIOSInfoPtr pBIOSInfo)
 {
     VIADelayIn_usec(pBIOSInfo, 10);  // Delay TD0 time to wait HPD signal from sink
     VIADelayIn_usec(pBIOSInfo, 10);  // Delay TD5 time
 }
 
 void VIADp2TurnOffAuxPowerSequence(VIABIOSInfoPtr pBIOSInfo)
 {
     VIADelayIn_usec(pBIOSInfo, 10);  // Delay TD5 time
 }

 Bool VIADP2WriteDPCD(VIABIOSInfoPtr pBIOSInfo, CARD32 dwoffset, int ByteCount)
 {
     CARD32 dwCmd;
     CARD32 dwData;
     
     DEBUG(ErrorF("VIADP2WriteDPCD\n"));

     // if EPHY disable, turn on power sequence 
     if(!(MMIO_RD(DP2_ENABLE_IF_REG) & BIT0))
     {
         VIADp2TurnOnAuxPowerSequence(pBIOSInfo);
     }
     
     dwCmd = dwoffset << 12;   // AUX read offset
     dwCmd |= ByteCount << 4;  // AUX read length
     dwCmd |= AUX_CMD_WRITE;   // AUX write CMD
     MMIO_WR(DP2_AUX_CMD_REG, dwCmd);

     // issue request based on above
     dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
     dwData |= 0x03;
     MMIO_WR(DP2_AUX_TIMER_REG, dwData); // trigger

     VIADelayIn_usec(pBIOSInfo, 10000);  

     // when leave, turn off power sequence 
     if(!(MMIO_RD(DP2_ENABLE_IF_REG) & BIT0))
     {
         VIADp2TurnOffAuxPowerSequence(pBIOSInfo);
     }
  
     return VIADP2ReplyCmdStatus(pBIOSInfo);
 }
     
 Bool VIADP2ReadDPCD(VIABIOSInfoPtr pBIOSInfo, CARD32 dwOffset)
 {
     CARD32 dwCmd;
     CARD32 dwData;
     
     DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2ReadDPCD\n")); 

     VIADP2ClearAuxReadRegs();

     // assign command
     dwCmd = dwOffset << 12;   // AUX read offset
     dwCmd |= 1 << 4;          // AUX read length
     dwCmd |= AUX_CMD_READ;    // AUX read CMD
     MMIO_WR(DP2_AUX_CMD_REG, dwCmd);

     // issue request based on above
     dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
     dwData |= 0x03;
     MMIO_WR(DP2_AUX_TIMER_REG, dwData); 

     return VIADP2ReplyCmdStatus(pBIOSInfo);
 }

void VIADP2ReadDPCDVersion(VIABIOSInfoPtr pBIOSInfo)
{   
    CARD32 dwDPCDVersion;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2ReadDPCDVersion\n"));
    
    // read DPCD Version
    if (VIADP2ReadDPCD(pBIOSInfo, 0x00))
    {  
        dwDPCDVersion = MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF;
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DPCDVersion = %x\n",dwDPCDVersion));
        
        switch (dwDPCDVersion)
        {
            case 0x10:
                pBIOSInfo->DPSettingInfo2.DPCDVersion |= DPCD_VER_1 | DPCD_REV_0; // Version 1.0
                break;
            case 0x11:
                pBIOSInfo->DPSettingInfo2.DPCDVersion |= DPCD_VER_1 | DPCD_REV_1; // Version 1.1
                break;
            default:   
                pBIOSInfo->DPSettingInfo2.DPCDVersion = NONEVERSION;
                break;
        }
    }
	
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
}

 Bool VIASenseDP2(VIABIOSInfoPtr pBIOSInfo)
 {
     VIABIOSInfoPtr pVia = pBIOSInfo; 
     Bool IsEPHYEnabled = FALSE;
     int ret = FALSE;
     
     DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIASenseDP2\n"));  
	 
     VIADP2ReadDPCDVersion(pBIOSInfo);
     
     if (pBIOSInfo->DPSettingInfo2.DPCDVersion)
     {
         ret = TRUE;
     }
     
     return ret;
 }
 
 Bool VIADP2LinkTraining(VIABIOSInfoPtr pBIOSInfo)
 {
     CARD32 dwData;
     
     DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2LinkTraining\n"));
     
    VIADP2ClearAuxReadRegs();
    VIADP2ClearAuxWriteRegs();
    pBIOSInfo->DPSettingInfo2.IsLinkTrainingDone = FALSE;
	
    // reset link training
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP2_NAUD_MUTE_REG, 0x80000000, 0x80000000); 
    VIADelayIn_usec(pBIOSInfo, 1000);
    MMIO_WR_MASK(DP2_NAUD_MUTE_REG, 0x0, 0x80000000);
    VIADelayIn_usec(pBIOSInfo, 1000);

    // [9:7]=100b, 4 lanes, [10]=0, disable SW link training , 
    // [12:11]=01b, set clock recovery state
    dwData = MMIO_RD(DP2_LINK_TRAINING_REG) & 0xFFFF0000; 
    dwData |= 0x0000023C; 
   
    switch (pBIOSInfo->DPSettingInfo2.LinkSpeed)
    {
        case Speed_162MHz:
            dwData &= 0xFFFFFFFD;
            break;
        case Speed_270MHz:
            dwData |= 0x00000002;             break;
        default:
            dwData |= 0x00000002;
            break;          
    }    
    MMIO_WR(DP2_LINK_TRAINING_REG, dwData);
 
    // Reset EPHY
    MMIO_WR_MASK(DP2_EPHY_TX_PWR_REG, 0x00000055, 0x000000FF);
    VIADelayIn_usec(pBIOSInfo, 2000);
    MMIO_WR_MASK(DP2_EPHY_TX_PWR_REG, 0x00000000, 0x000000FF);

    // enable scramble
    MMIO_WR_MASK(DP2_VIDEO_CTRL_REG, 0x00000001, 0x00000001);

    // enable DP, set enhanced mode
    dwData = MMIO_RD(DP2_ENABLE_IF_REG) & 0xC3000000; 
    dwData |= 0x00000005; // if enable standard mode , set 0x00000001
    MMIO_WR(DP2_ENABLE_IF_REG, dwData);

    switch (pBIOSInfo->DPSettingInfo2.LinkSpeed)
    {
        case Speed_162MHz:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x000018406); // if enable standard mode , set 0x000010406
            break;
        case Speed_270MHz:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x00001840A); // if enable standard mode , set 0x00001040A
            break;
        default:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x00001840A);
            break;          
    }   
     
    MMIO_WR(DP2_AUX_W_DATA1_REG, 0x0);
     
    if (!VIADP2WriteDPCD(pBIOSInfo, 0x100, 2))
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
        return FALSE;
    }
     
    //Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
    VIADelayIn_usec(pBIOSInfo, 20000);
 
    dwData = MMIO_RD(DP2_LINK_TRAINING_REG) & 0x7FFFFFFF; 
    dwData |= 0x0000003C;  
   
    if (pBIOSInfo->DPSettingInfo2.LinkSpeed == Speed_270MHz)
    {
         dwData |= 0x00000002;
    } 
    MMIO_WR(DP2_LINK_TRAINING_REG, dwData);
         
    // enable HW link training
    MMIO_WR_MASK(DP2_LINK_TRAINING_REG, 0x00000001, 0x00000001); 
 
    VIADelayIn_usec(pBIOSInfo, 20000);
     
    if ((MMIO_RD(DP2_LINK_TRAINING_REG) & 0x80000000) != 0x80000000)
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
        return FALSE;
    }

    VIADelayIn_usec(pBIOSInfo, 4000);
     
    if ((MMIO_RD(DP2_ENABLE_IF_REG) & 0x30000000) != 0x30000000)
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
         
        return FALSE;
    }

    pBIOSInfo->DPSettingInfo2.IsLinkTrainingDone = TRUE;

    // clear [31], and [0]
    MMIO_WR_MASK(DP2_LINK_TRAINING_REG, 0x0, 0x80000001);
    
    //Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
 
    return TRUE;
}

int VIADP2GetLinkSpeed(VIABIOSInfoPtr pBIOSInfo)
{
    DPSETTINGINFOPTR pDPSettingInfo = &(pBIOSInfo->DPSettingInfo2);
    Bool IsEPHYEnabled = FALSE;
    CARD32 dwLinkSpeed;

    if (pDPSettingInfo->IsLinkTrainingDone)
        return pDPSettingInfo->LinkSpeed;
	
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2GetLinkSpeed\n"));

    // Patch for driver flow
    IsEPHYEnabled = MMIO_RD(DP2_ENABLE_IF_REG) & BIT0;
    if (!IsEPHYEnabled)
        VIADP2InitEPHY(pBIOSInfo);

    VIADP2ReadDPCD(pBIOSInfo, 0x01);
    dwLinkSpeed = MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF;
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);

    // Use default link speed of DP monitor
    switch (dwLinkSpeed) {
        case 0x06:
        pDPSettingInfo->LinkSpeed = Speed_162MHz;
        break;        
        case 0x0A:
        pDPSettingInfo->LinkSpeed = Speed_270MHz;
        break;          
        default:
        pDPSettingInfo->LinkSpeed = Speed_270MHz;
        break;
    }
    
    // Start link training to confirm link speed
    if ((!VIADP2LinkTraining(pBIOSInfo)) && (pDPSettingInfo->LinkSpeed == Speed_270MHz)) {
        // Reduce link speed to 162 MHz and retry link training if 270 MHz link training fail
        MMIO_WR(DP2_LINK_TRAINING_REG, (MMIO_RD(DP2_LINK_TRAINING_REG) & 0x7FFFFFFC));
        VIADP2ClearAuxWriteRegs();
        // Clear DPCD training pattern
        VIADP2WriteDPCD(pBIOSInfo, 0x102, 1);
        pDPSettingInfo->LinkSpeed = Speed_162MHz;
        // Reset EPHY for changed link speed
        VIADP2InitEPHY(pBIOSInfo);
        if (!VIADP2LinkTraining(pBIOSInfo)) {
            pDPSettingInfo->LinkSpeed = Speed_NONE;
            pDPSettingInfo->IsLinkTrainingDone = TRUE;
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "There is no suitable link speed for DP monitor\n");
        }
    }
    
    return pDPSettingInfo->LinkSpeed;
}
 
void VIAEnableDP2(VIABIOSInfoPtr pBIOSInfo)
{
    DPSETTINGINFOPTR pDPSettingInfo = &(pBIOSInfo->DPSettingInfo2);
	
    DEBUG(ErrorF("VIAEnableDP2\n"));
 
    // Must Enable HDCP first
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);

    // if DP mode, video input selection - sp1
    MMIO_WR(DP2_VIDEO_CTRL_REG, (MMIO_RD(DP2_VIDEO_CTRL_REG) & 0x3FFC0001));

    // enable DP and video
    MMIO_WR(DP2_ENABLE_IF_REG, MMIO_RD(DP2_ENABLE_IF_REG) | 0x00000009);

    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0); 
}
 
void VIADisableDP2(VIABIOSInfoPtr pBIOSInfo)
{
    DEBUG(ErrorF("VIADisableDP2\n"));
         
    // disable DP and video
    MMIO_WR(DP2_ENABLE_IF_REG, MMIO_RD(DP2_ENABLE_IF_REG) & 0xFFFFFFF6);
    
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0); 
}

void VIADp2HorWidthAndTu(VIABIOSInfoPtr pBIOSInfo, struct display_timing crt_reg, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int TUSize = 48;
    int Width;
    
    VIACalculate_TU(&(pBIOSInfo->DPSettingInfo2), PixelClock, pBIOSInfo->bitsPerPixel);
    Width = (crt_reg.hor_addr / 4) - 1;
    dwTemp = pBIOSInfo->DPSettingInfo2.TURatio << 17 | (TUSize - 1) << 11 | Width;
    MMIO_WR(DP2_HWIDTH_TUSIZE_REG, dwTemp);
    //MMIO_WR_MASK(DP2_EXT_REG, Width/2048, 0x40000000);
}

void VIADp2HorLineDuration(VIABIOSInfoPtr pBIOSInfo, struct display_timing crt_reg, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int LinkClock;
    int HTotal;
    int HBlank;

    // units of Dclk and Ls_clk are MHz
    switch (pBIOSInfo->DPSettingInfo2.LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }

    HBlank = ((((crt_reg.hor_total - crt_reg.hor_addr) * LinkClock) / (PixelClock / 10000)) - 15) & 0x00000FFF;
    HTotal = ((LinkClock * crt_reg.hor_total / (PixelClock / 1000000)) - 15) & 0x00007FFF;

    dwTemp = HBlank << 15 | HTotal;
    MMIO_WR(DP2_HLINE_DUR_REG, dwTemp);
}

void VIADp2Mvid(VIABIOSInfoPtr pBIOSInfo, CARD32 PixelClock)
{
    CARD32 dwTemp;
    int LinkClock;
	
    // units of Dclk and Ls_clk are MHz
    switch (pBIOSInfo->DPSettingInfo2.LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }

    dwTemp = ((PixelClock / 1000000) * 32768 / LinkClock) & 0x19FFFFFF;
    dwTemp |=  0x20000000; // 8 bits, MMC6CC[31:29]=001b 
    dwTemp &=  0xFEFFFFFF;
    MMIO_WR(DP2_MVID_MISC0_REG, dwTemp); // [24]=0: Asynchronous mode
    MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x02000000, 0x02000000); // Generate MVID in Asynchronous mode
}

void VIADp2HorVerTotal(struct display_timing crt_reg)
{
    CARD32 dwTemp;
   
    dwTemp = crt_reg.ver_total << 16 | crt_reg.hor_total;
    MMIO_WR(DP2_H_ATTR_REG, dwTemp);
}

void VIADp2HorVerOffset(struct display_timing crt_reg)
{
    CARD32 dwTemp;

    dwTemp = (crt_reg.ver_total - crt_reg.ver_sync_start) << 16 | 
             (crt_reg.hor_total - crt_reg.hor_sync_start);
    MMIO_WR(DP2_HV_START_REG, dwTemp);
}

void VIADp2HorVerSyncWidthAndPolarity(struct display_timing crt_reg, int HSync, int VSync)
{
    CARD32 dwTemp;
                                                 
    dwTemp = VSync << 31 | crt_reg.ver_sync_end << 16 |
             HSync << 15 | crt_reg.hor_sync_end;
    MMIO_WR(DP2_POLARITY_WIDTH_REG, dwTemp);  
}

void VIADp2HorVerActive(struct display_timing crt_reg)
{
    CARD32 dwTemp;

    dwTemp = crt_reg.ver_addr << 16 | crt_reg.hor_addr;
    MMIO_WR(DP2_ACITVE_WH_REG, dwTemp);
}

void VIADp2Nvid()
{
    // [23:0] Stream Clock Recovery :N  
    MMIO_WR_MASK(DP2_NVID_MISC0_REG, 32768, 0x00FFFFFF);
}

void VIADp2DelayReg()
{
    // Default: [29:24] = 20h for delay
    MMIO_WR_MASK(DP2_VIDEO_CTRL_REG, 0x20000000, 0x3F000000);
}

Bool VIADP2SetupMainLink(VIABIOSInfoPtr pBIOSInfo, struct crt_mode_table  ModeInfo)
{
    VIABIOSInfoPtr pVia = pBIOSInfo;
    struct display_timing crt_reg = ModeInfo.crtc;
    CARD32 PixelClock = ModeInfo.clk;
    int hsync_polarity = ModeInfo.h_sync_polarity;
    int vsync_polarity = ModeInfo.v_sync_polarity;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADP2SetupMainLink\n"));

    if (pBIOSInfo->DPSettingInfo2.LinkSpeed != Speed_NONE) {

        // Field ID flag not invert
        MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x0, BIT1);

        VIADp2HorWidthAndTu(pBIOSInfo, crt_reg, PixelClock); 
        VIADp2HorLineDuration(pBIOSInfo, crt_reg, PixelClock);
        VIADp2Mvid(pBIOSInfo, PixelClock);
        VIADp2HorVerTotal(crt_reg);
        VIADp2HorVerOffset(crt_reg);
        VIADp2HorVerSyncWidthAndPolarity(crt_reg, hsync_polarity, vsync_polarity);
        VIADp2HorVerActive(crt_reg);
        VIADp2Nvid();
        VIADp2DelayReg();

        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "End Setup MainLink!!\n"));
    }
}

Bool VIADP2ReadEDID(VIABIOSInfoPtr pBIOSInfo, CARD32 slave_addr, CARD32 dwOffset, CARD8 *pData)
{   
    VIABIOSInfoPtr pVia = pBIOSInfo;

    CARD32 dwCmd;

    VIADP2ClearAuxReadRegs();
    VIADP2ClearAuxWriteRegs();
	 
    // 1. write slave addr offset for start
    dwCmd = dwOffset & 0x000000FF;
    MMIO_WR(DP2_AUX_W_DATA0_REG, dwCmd);
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_WRITE;
    MMIO_WR(DP2_AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(DP2_AUX_TIMER_REG, dwCmd);

    if(!VIADP2ReplyCmdStatus(pBIOSInfo))
    {
        return FALSE;
    }

    // 2. write command with read length
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_READ;
    MMIO_WR(DP2_AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(DP2_AUX_TIMER_REG, dwCmd);

    if(!VIADP2ReplyCmdStatus(pBIOSInfo))
    {
        return FALSE;
    }

    // 3. read 
    *pData = (MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF);

    return TRUE;
}


/******************************************************************
*
*                                          Common Function                        
*
*******************************************************************/
Bool VIAIdentifyDP(VIABIOSInfoPtr pBIOSInfo)
{
    int ret= NOT_FOUND_DEV;
    Bool IsEPHYEnabled = FALSE;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAIdentifyDP\n"));
    
    if (pBIOSInfo->Chipset == VIA_VX900) {
        if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP) {
            pBIOSInfo->DPSettingInfo.ChipID = VIA_INTEGRATED_DP;
            pBIOSInfo->DPSettingInfo.I2CPort = PORT_INDEX_GPIO_3D;
            // initialize EPHY once
            IsEPHYEnabled = MMIO_RD(DP_EPHY_PLL_REG) & BIT0;
            if (!IsEPHYEnabled)
                VIADPInitEPHY(pBIOSInfo);
        } else if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP2) {
            pBIOSInfo->DPSettingInfo2.ChipID = VIA_INTEGRATED_DP2;
            pBIOSInfo->DPSettingInfo2.I2CPort = PORT_INDEX_GPIO_3D;
            // initialize EPHY once
            IsEPHYEnabled = MMIO_RD(DP2_ENABLE_IF_REG) & BIT0;
            if (!IsEPHYEnabled)
                VIADP2InitEPHY(pBIOSInfo);
        }
        ret = FOUND_INTERNAL_DEV;
    }

    return ret;
}

CARD32 VIAGetDeviceDetectDP(VIABIOSInfoPtr pBIOSInfo)
{
    CARD32    ret = 0x0;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAGetDeviceDetectDP\n"));

    if(pBIOSInfo->DPSettingInfo.ChipID)
    {
        if(VIASenseDP(pBIOSInfo))
        {
            ret |= VIA_DEVICE_DP;
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "DP has Attachment.\n"));
        }
        else
        {
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "DP hasn't Attachment.\n"));
        }
    } 
    else if (pBIOSInfo->DPSettingInfo2.ChipID)
    {
        if(VIASenseDP2(pBIOSInfo))
        {
            ret |= VIA_DEVICE_DP2;
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "DP2 has Attachment.\n"));
        }
        else
        {
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "DP2 hasn't Attachment.\n"));
        }
    }

    return ret;
}


static void VIASetDPTiming(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo, int DPType)
{
	ScrnInfoPtr             pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr                  pVia = VIAPTR(pScrn);  
    struct VideoModeTable *DPMode = NULL;
    struct display_timing   mode_crt_reg;
    struct crt_mode_table   *mode_crt_table = NULL;
    BOOL   useModeline = FALSE;
    int i, set_iga = pDPSettingInfo->IGAPath;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIASetDPTiming\n"));

    /* Use the mode line in two cases:
        1. SAMM and Mode line is user-setting
        2. DuoView/single, EDIDType=true, and Mode line is user-setting*/
    if (!(pBIOSInfo->Is3DScalingEnable == (DISP_3D_SCALING_ENABLE | DPType)) &&
		(pVia->pVIAEnt->HasSecondary ||
         (!pVia->pVIAEnt->HasSecondary && (pBIOSInfo->EDIDType == DPType))) && 
         (pBIOSInfo->ModeLineStatus & (MODE_LINE_TYPE_USER | MODE_LINE_TYPE_EDID))) {
        useModeline = TRUE;
    }

    DPMode = &VIARoutineModes[SearchModeSetting(pDPSettingInfo->ModeIndex)];
    mode_crt_table = DPMode->crtc;
    mode_crt_reg = mode_crt_table->crtc;

    /* Enable Interlace mode. */
    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        if (set_iga == IGA1)
        {
            write_reg_mask(CR33, VIACR, BIT6, BIT6);
        }
        else if (set_iga == IGA2)
        {
            write_reg_mask(CR67, VIACR, BIT5, BIT5);
        }
    }

    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "-----DP TIMING-----\n");

    /* Write CRTC Timing */
    fill_crtc_timing(pBIOSInfo, mode_crt_table, DPMode->mode_array,
                     pBIOSInfo->bitsPerPixel >> 3, DPType,
                     set_iga, useModeline);   

    if (useModeline) {
        if (DPType == VIA_DEVICE_DP)
            VIADPSetupMainLink(pBIOSInfo, pBIOSInfo->UserSpecifiedMode);
        else if (DPType == VIA_DEVICE_DP2)
            VIADP2SetupMainLink(pBIOSInfo, pBIOSInfo->UserSpecifiedMode);
    } else {
        for (i = 0; i < DPMode->mode_array; i++) {
            if (mode_crt_table[i].refresh_rate == pBIOSInfo->FoundRefresh)
                break;
        }

        if (DPType == VIA_DEVICE_DP)
            VIADPSetupMainLink(pBIOSInfo, mode_crt_table[i]);
        else if (DPType == VIA_DEVICE_DP2)
            VIADP2SetupMainLink(pBIOSInfo, mode_crt_table[i]);             
    }
	
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIASetDPTiming --Done--\n"));
}

void VIAPreSetDP(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo)
{
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAPreSetDP\n"));

    switch (pDPSettingInfo->ChipID)
    {
        case VIA_INTEGRATED_DP:
            // Embedded DP channel selection.
            if (pDPSettingInfo->IGAPath == IGA1)
	     {
                write_reg_mask(CRFF, VIACR, 0x00, BIT1);
            }
            else
            {
                write_reg_mask(CRFF, VIACR, BIT1, BIT1);
            }
            // disable video
            MMIO_WR(DP_ENABLE_IF_REG, MMIO_RD(DP_ENABLE_IF_REG) & 0xFFFFFFF7);  
            // select DP mode
            MMIO_WR_MASK(DP_EPHY_MISC_PWR_REG, BIT0, BIT0); 
            //source select: DP
            write_reg_mask(CRFF, VIACR, 0x0, BIT0);
            break;
        case VIA_INTEGRATED_DP2:
            // Embedded DP channel selection.
            if (pDPSettingInfo->IGAPath == IGA1)
            {
                write_reg_mask(CRFF, VIACR, 0x00, BIT2);
            }
            else
            {
                write_reg_mask(CRFF, VIACR, BIT2, BIT2);
            }
            // disable video
            MMIO_WR(DP2_ENABLE_IF_REG, MMIO_RD(DP2_ENABLE_IF_REG) & 0xFFFFFFF7);  
            // select DP mode
            MMIO_WR_MASK(DP2_ENABLE_IF_REG, BIT0, BIT0);  
            break;
        default:
            break;
    }
}


Bool VIASetDPMode(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo)
{
    ScrnInfoPtr             pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr                  pVia = VIAPTR(pScrn);
    Bool                    TrainingSuccess;
    int DPType;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIASetDPMode\n"));

    VIAPreSetDP(pBIOSInfo, pDPSettingInfo);

    if (pDPSettingInfo->ChipID == VIA_INTEGRATED_DP)
    {
        DPType = VIA_DEVICE_DP;   
    }
    else
    {
        DPType = VIA_DEVICE_DP2;     
    }
    
    VIASetDPTiming(pBIOSInfo, pDPSettingInfo, DPType);

    // Let HW use AUX channel
    if (pDPSettingInfo->ChipID == VIA_INTEGRATED_DP)
    {
        MMIO_WR(AUX_TIMER_REG, 0x0); 
    }
    else
    { 
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0); 
    }  
    
    if ((pVia->IsSecondary)||(pBIOSInfo->DuoView)) 
    {
        /* Second Display Channel Enable */
        EnableSecondDisplayChannel();
            
        /* Fixed Set DVI+TV/DVT+CRT/CRT+TV SAMM mode,
            the secondary monitor without display */
        write_reg_mask(CR91, VIACR, 0x0, BIT5);
    }
    
    /* Disable LCD Scaling */
    if(pDPSettingInfo->IGAPath == IGA2)
    {
        write_reg(CR79, VIACR, 0x0);
    }
         
    /* Set display channel. */
    if ((pBIOSInfo->Chipset == VIA_CX700) || (pBIOSInfo->Chipset == VIA_VX800))
        VIASetDisplayChannel(pBIOSInfo);
       
    return TRUE;

}

void VIAGetDPEDIDBlockData(VIABIOSInfoPtr pBIOSInfo, CARD8 ChipID, CARD8* pEDIDData)
{
    CARD8   EDIDHeader[3];
    CARD8   i = 0;
    Bool IsEPHYEnabled = FALSE;

    switch (ChipID) {
        case VIA_INTEGRATED_DP:
            /* First Read EDID Block 0 */
            for (i = 0; i < 3; i++) {    
                VIADPReadEDID(pBIOSInfo, 0xA0, i, EDIDHeader+i);
            }

            /* EDID Header is 00 FF FF FF FF FF FF 00 */
            if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
                (EDIDHeader[2] == 0xFF)) {
                /* Read Block 0 Raw Data. */
                for (i = 0; i < 128; i++) {
                    VIADPReadEDID(pBIOSInfo, 0xA0, i, pEDIDData+i);
                }    
            }
            break;
        case VIA_INTEGRATED_DP2:
            /* First Read EDID Block 0 */
            for (i = 0; i < 3; i++) {    
                VIADP2ReadEDID(pBIOSInfo, 0xA0, i, EDIDHeader+i);
            }

            /* EDID Header is 00 FF FF FF FF FF FF 00 */
            if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
                (EDIDHeader[2] == 0xFF)) {
                /* Read Block 0 Raw Data. */
                for (i = 0; i < 128; i++) {
                    VIADP2ReadEDID(pBIOSInfo, 0xA0, i, pEDIDData+i);
                }    
            }
            break;
        default:
           break;
    }
}

void VIAGetDPPanelSize(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo)
{        
    xf86MonPtr          edid_mon;      
    int                 i, max_h, max_v;
    unsigned char       *edidBuffer = NULL;  

    if (pBIOSInfo->NoDDCValue) {
        max_h = 2048;
        max_v = 1536;
        pDPSettingInfo->DPSize = VIAGetModeIndex(max_h, max_v);
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "Default DP Panel Size ID = 0x%x\n", pBIOSInfo->DPSettingInfo.DPSize));
        return;
    }
    
    //  hardware issue, we can't read EDID on VT3410 A0 A1
    //  so, we set max size 2048x1536
    //  if-else conditions may be removed future 
    ScrnInfoPtr pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);

    if (VIAGetRevisionOfVX900(pVia) < REVISION_VX900_A2) {
        max_h = 2048;
        max_v = 1536;
        pDPSettingInfo->DPSize = VIAGetModeIndex(max_h, max_v);
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DP Panel Size ID = 0x%x\n", pBIOSInfo->DPSettingInfo.DPSize));
        return;
    }
   
    edidBuffer = (CARD8*) xcalloc(1, sizeof(CARD8)*128);
      
    // Read EDID v1.1 Block 
    VIAGetDPEDIDBlockData(pBIOSInfo, pDPSettingInfo->ChipID, edidBuffer);  

    // Parse the EDID block 
    edid_mon = xf86InterpretEDID(pBIOSInfo->scrnIndex,edidBuffer);
    
    if (edid_mon)
    {             
        //Find max resolution from Established timing
        VIAFindMaxResFromEstablishedTiming(edid_mon, &max_h, &max_v);

        //Find max resolution from Standard timing
        for (i = 0; i < 8; i++)
        {
            if (edid_mon->timings2[i].hsize > max_h)
            {
                max_h = edid_mon->timings2[i].hsize;
                max_v = edid_mon->timings2[i].vsize;
            }            
        }

        //Find max resolution from Detail timing
        for (i = 0; i < 4; i++)
        {
            if ((edid_mon->det_mon[i].type == DT) && 
                (edid_mon->det_mon[i].section.d_timings.h_active > max_h))
            {
                max_h = edid_mon->det_mon[i].section.d_timings.h_active;
                max_v = edid_mon->det_mon[i].section.d_timings.v_active;

                if (edid_mon->det_mon[i].section.d_timings.interlaced)                
                    max_v = max_v*2;                
            }       
        }
        
        // Correct monitor size to avoid invalid mode. 
        VIACorrectMonitorSize(&max_h, &max_v);            

        // To prevent to get a wrong EDID data. We should assign a default value.
        if ((max_h == 0) || (max_v == 0))
        {
            max_h = 1024;
            max_v = 768;

        }
		
        pDPSettingInfo->DPSize = VIAGetModeIndex(max_h, max_v);
    }                                	
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DP Panel Size ID = 0x%x\n", pBIOSInfo->DPSettingInfo.DPSize));
}

unsigned char VIAGetDPInfo(VIABIOSInfoPtr pBIOSInfo, DPSETTINGINFOPTR pDPSettingInfo)
{
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAGetDPInfo\n"));

    VIAGetDPPanelSize(pBIOSInfo, pDPSettingInfo);
    VIAGetModeSizeByIndex(pDPSettingInfo->DPSize, &pDPSettingInfo->PanelSizeH, &pDPSettingInfo->PanelSizeV);

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "DPSize = %d\n", pDPSettingInfo->DPSize));
    return (unsigned char)(pDPSettingInfo->DPSize);
}

void VIADPInit(VIABIOSInfoPtr pBIOSInfo)
{  
    DPSETTINGINFOPTR pDPSettingInfo;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIADPInit\n"));

    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP)
    {
        pDPSettingInfo = &(pBIOSInfo->DPSettingInfo);
    }
    else
    {
        pDPSettingInfo = &(pBIOSInfo->DPSettingInfo2);
    }
    
    if (pDPSettingInfo->ChipID)
    {
        VIAGetDPInfo(pBIOSInfo, pDPSettingInfo);
    }
}



/******************************************************************
*
*                                          DP1 Function                        
*
*******************************************************************/
void viaDPDisableEPHY()
{
    //Disable Bandgap to disable all EPHY functions
    MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x00000000, 0x00000001);
}

Bool viaDPInitEPHY(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;


    DEBUG(ErrorF("viaDPInitEPHY\n"));
 
    // (1) enable Bandgap and set EPHY symbol rate
    dwData = MMIO_RD(DP_EPHY_PLL_REG) & 0xFFF800FF;
    // real EPHY symbol rate decided by HW link training
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            dwData |= 0x0004BE00;
            break;
        case Speed_270MHz:
            dwData |= 0x00029E00;
            break;
        default:
            dwData |= 0x00029E00;
            break;	
    }
    dwData |= 0x00000001;
    MMIO_WR(DP_EPHY_PLL_REG, dwData);
    viaDelayIn_usec(pVia, 1000);
     
    // (2) start AUX channel, CMOP is on, Tx and Rx are off
    dwData = MMIO_RD(DP_EPHY_MISC_PWR_REG) & 0xFFFF3FFF;
    dwData |= 0x00004000;
    MMIO_WR(DP_EPHY_MISC_PWR_REG, dwData);
    viaDelayIn_usec(pVia, 4000);
    dwData = MMIO_RD(DP_EPHY_MISC_PWR_REG) & 0xFFFF3FFF;
    dwData |= 0x00008000;
    MMIO_WR(DP_EPHY_MISC_PWR_REG, dwData);
    viaDelayIn_usec(pVia, 4000);

    // (3) enable Mpll and Mpll reg power          
    dwData = MMIO_RD(DP_EPHY_PLL_REG) & 0xF9FFFFF9;
    dwData |= 0x00000006;
    MMIO_WR(DP_EPHY_PLL_REG, dwData);
    viaDelayIn_usec(pVia, 1000);
    //enable DP
    MMIO_WR(DP_ENABLE_IF_REG, (MMIO_RD(DP_ENABLE_IF_REG) | 0x00000001));
    viaDelayIn_usec(pVia, 1000);

    // (4) enable Tpll and Tpll reg power
    MMIO_WR(DP_EPHY_PLL_REG, (MMIO_RD(DP_EPHY_PLL_REG) | 0x06000000));
    viaDelayIn_usec(pVia, 1000);

    // (5) Resister Tuning (RT), RT reset
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFD));
    // power up RT
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000001));
    // disable RT reset
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000002));
    viaDelayIn_usec(pVia, 1000);
    // enable RT auto calibration mode
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x00000004));
    viaDelayIn_usec(pVia, 1000);
    // disable RT mode
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFB));
    viaDelayIn_usec(pVia, 1000);
    // power down RT
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0xFFFFFFFE));
     
    // (6)
    // [19:8] Tx power status default settings for all 4 lanes
    // [7:0] Tx power down mode for all 4 lanes, [15:8] HDMI TX output voltage swing 1200 mV for all 4 lanes
    dwData = MMIO_RD(DP_EPHY_TX_PWR_REG) & 0x0000F0FF;
    dwData |= 0xAAFF0F00;
    MMIO_WR(DP_EPHY_TX_PWR_REG, dwData); 
    viaDelayIn_usec(pVia, 4000);
    // [7:0] Tx power control function CMOP on / PISO off for all 4 lanes
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) & 0x00FFFFFF)); // Tx power control function on for all 4 lanes
    viaDelayIn_usec(pVia, 1000);
    // Driver awaken for all 4 lanes
    MMIO_WR(DP_EPHY_TX_PWR_REG, (MMIO_RD(DP_EPHY_TX_PWR_REG) | 0x0000F000)); 
    MMIO_WR(DP_EPHY_MISC_PWR_REG, (MMIO_RD(DP_EPHY_MISC_PWR_REG) | 0x00000001));
    viaDelayIn_usec(pVia, 2000);
}

void viaDPClearAuxWriteRegs()
{
    MMIO_WR(AUX_W_DATA0_REG, 0x0);
    MMIO_WR(AUX_W_DATA1_REG, 0x0);
    MMIO_WR(AUX_W_DATA2_REG, 0x0);
    MMIO_WR(AUX_W_DATA3_REG, 0x0);
}

void viaDPClearAuxReadRegs()
{
    MMIO_WR(AUX_R_DATA0_REG, 0x0);
    MMIO_WR(AUX_R_DATA1_REG, 0x0);
    MMIO_WR(AUX_R_DATA2_REG, 0x0);
    MMIO_WR(AUX_R_DATA3_REG, 0x0);
}

Bool viaDPReplyCmdStatus(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwAUXCMDStatus;
    CARD32 dwData;
    int i, j;
 
    //DEBUG(ErrorF("viaDPReplyCmdStatus\n"));
     
    for(i=0; i<NUM_RE_ISSUE_COMMAND; i++)
    {        
        for (j=0; j<NUM_RECHECK; j++)
        {
            viaDelayIn_usec(pVia, 1000);
             
            dwAUXCMDStatus = MMIO_RD(AUX_TIMER_REG) & 0x3C000007;
             
            if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1 && dwAUXCMDStatus == ACK_OK)
            {
                return TRUE;
            }
            else if (dwAUXCMDStatus == 0x00000001) // hardware issue
            {
                return TRUE;
            }            
            else if ((dwAUXCMDStatus & 0xFFFF) == RE_ISSUE_CMD)
            {
                // re-issue Cmd 
                dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
                dwData |= 0x03;
                MMIO_WR(AUX_TIMER_REG, dwData);
                break;
            }
        }
    }
 
    return FALSE;
}

Bool viaDPWriteDPCD(xf86OutputPtr output, CARD32 dwoffset, int ByteCount)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwCmd;
    CARD32 dwData;
    
    DEBUG(ErrorF("viaDPWriteDPCD\n"));
     
    dwCmd = dwoffset << 12;   // AUX read offset
    dwCmd |= ByteCount << 4;  // AUX read length
    dwCmd |= AUX_CMD_WRITE;   // AUX write CMD
    MMIO_WR(AUX_CMD_REG, dwCmd);

    // issue request based on above
    dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwData); // trigger

    viaDelayIn_usec(pVia, 10000);  
  
    return viaDPReplyCmdStatus(output);
}

Bool viaDPReadDPCD(xf86OutputPtr output, CARD32 dwOffset)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwCmd;
    CARD32 dwData;
     
    DEBUG(ErrorF("viaDPReadDPCD\n")); 

    viaDPClearAuxReadRegs();

    // assign command
    dwCmd = dwOffset << 12;   // AUX read offset
    dwCmd |= 1 << 4;          // AUX read length
    dwCmd |= AUX_CMD_READ;    // AUX read CMD
    MMIO_WR(AUX_CMD_REG, dwCmd);

    // issue request based on above
    dwData = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwData); 

    return viaDPReplyCmdStatus(output);
}

void viaDPReadDPCDVersion(xf86OutputPtr output)
{   
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;
    CARD32 dwDPCDVersion;

    DEBUG(ErrorF("viaDPReadDPCDVersion\n"));
    
    // read DPCD Version
    if (viaDPReadDPCD(output, 0x00))
    {  
        dwDPCDVersion = MMIO_RD(AUX_R_DATA0_REG) & 0xFF;
         
        switch (dwDPCDVersion)
        {
            case 0x10:
                viaDPInfo->DPCDVersion |= DPCD_VER_1 | DPCD_REV_0; // Version 1.0
                break;
            case 0x11:
                viaDPInfo->DPCDVersion |= DPCD_VER_1 | DPCD_REV_1; // Version 1.1
                break;
            default:   
                viaDPInfo->DPCDVersion = NONEVERSION;
                break;
        }
    }

    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
}

Bool viaDPLinkTraining(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;
    
    DEBUG(ErrorF("viaDPLinkTraining\n"));

    viaDPClearAuxReadRegs();
    viaDPClearAuxWriteRegs();
    viaDPInfo->IsLinkTrainingDone = FALSE;
	
    // reset link training
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP_NAUD_MUTE_REG, 0x80000000, 0x80000000); 
    viaDelayIn_usec(pVia, 1000);
    MMIO_WR_MASK(DP_NAUD_MUTE_REG, 0x0, 0x80000000);
    viaDelayIn_usec(pVia, 1000);
 
    // [9:7]=100b, 4 lanes, [10]=0, disable SW link training , 
    // [12:11]=01b, set clock recovery state
    dwData = MMIO_RD(DP_LINK_TRAINING_REG) & 0xFFFF0000; 
    dwData |= 0x00000200; 
   
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            dwData &= 0xFFFFFFFD;
            break;
        case Speed_270MHz:
            dwData |= 0x00000002;
            break;
        default:
            dwData |= 0x00000002;
            break;          
    }    
    MMIO_WR(DP_LINK_TRAINING_REG, dwData);
  
    // Reset EPHY
    MMIO_WR_MASK(DP_EPHY_TX_PWR_REG, 0x55000000, 0xFF000000);
    viaDelayIn_usec(pVia, 2000);
    MMIO_WR_MASK(DP_EPHY_TX_PWR_REG, 0x00000000, 0xFF000000);
 
    // enable scramble
    MMIO_WR_MASK(DP_VIDEO_CTRL_REG, 0x00000001, 0x00000001);
 
    // enable DP, set enhanced mode
    dwData = MMIO_RD(DP_ENABLE_IF_REG) & 0xC3000000; 
    dwData |= 0x00000005; // if enable standard mode , set 0x00000001
    MMIO_WR(DP_ENABLE_IF_REG, dwData);
 
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            MMIO_WR(AUX_W_DATA0_REG, 0x000018406); // if enable standard mode , set 0x000010406
            break;
        case Speed_270MHz:
            MMIO_WR(AUX_W_DATA0_REG, 0x00001840A); // if enable standard mode , set 0x00001040A
            break;
        default:
            MMIO_WR(AUX_W_DATA0_REG, 0x00001840A);
            break;          
    }   
      
    MMIO_WR(AUX_W_DATA1_REG, 0x0);
      
    if (!viaDPWriteDPCD(output, 0x100, 2))
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
        return FALSE;
    }
      
    //Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
    viaDelayIn_usec(pVia, 20000);
  
    dwData = MMIO_RD(DP_LINK_TRAINING_REG) & 0x7FFFFFFF; 
    dwData |= 0x0000003C;  
   
    if (viaDPInfo->LinkSpeed == Speed_270MHz)
    {
         dwData |= 0x00000002;
    } 
    MMIO_WR(DP_LINK_TRAINING_REG, dwData);
          
    // enable HW link training
    MMIO_WR_MASK(DP_LINK_TRAINING_REG, 0x00000001, 0x00000001); 
  
    viaDelayIn_usec(pVia, 20000);
      
    if ((MMIO_RD(DP_LINK_TRAINING_REG) & 0x80000000) != 0x80000000)
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
        return FALSE;
    }
 
    viaDelayIn_usec(pVia, 4000);
     
    if ((MMIO_RD(DP_ENABLE_IF_REG) & 0x30000000) != 0x30000000)
    {
        //Let HW use AUX channel
        MMIO_WR(AUX_TIMER_REG, 0x0);
          
        return FALSE;
    }

    viaDPInfo->IsLinkTrainingDone = TRUE;
	
    // clear [31], and [0]
    MMIO_WR_MASK(DP_LINK_TRAINING_REG, 0x0, 0x80000001);
     
    //Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);
  
    return TRUE;
}

int viaDPGetLinkSpeed(xf86OutputPtr output)
{   
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwLinkSpeed;

    if (viaDPInfo->IsLinkTrainingDone)
        return viaDPInfo->LinkSpeed;

    DEBUG(ErrorF("viaDPGetLinkSpeed\n"));
	
    viaDPReadDPCD(output, 0x01);
    dwLinkSpeed = MMIO_RD(AUX_R_DATA0_REG) & 0xFF;
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0);

    // Use default link speed of DP monitor
    switch (dwLinkSpeed)
    {
        case 0x06:
            viaDPInfo->LinkSpeed = Speed_162MHz;
            break;        
        case 0x0A:
            viaDPInfo->LinkSpeed = Speed_270MHz;
            break;          
        default:
            viaDPInfo->LinkSpeed = Speed_270MHz;
            break;    
    }

    // Start link training to confirm link speed
    if ((!viaDPLinkTraining(output)) && (viaDPInfo->LinkSpeed == Speed_270MHz)) {
        // Reduce link speed to 162 MHz and retry link training if 270 MHz link training fails
        MMIO_WR(DP_LINK_TRAINING_REG, (MMIO_RD(DP_LINK_TRAINING_REG) & 0x7FFFFFFC));
        viaDPClearAuxWriteRegs();
        // Clear DPCD training pattern
        viaDPWriteDPCD(output, 0x102, 1);
        viaDPInfo->LinkSpeed = Speed_162MHz;
        //Reset EPHY for changed link speed
        viaDPInitEPHY(output);
        if (!viaDPLinkTraining(output)) {
            viaDPInfo->LinkSpeed = Speed_NONE;
            viaDPInfo->IsLinkTrainingDone = FALSE;
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Output %s link training fails\n", output->name);
        }
    }

    return viaDPInfo->LinkSpeed;
}

void viaEnableDP(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    xf86CrtcPtr crtc = output->crtc;

    DEBUG(ErrorF("viaEnableDP\n"));
  
    // Must Enable HDCP first
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x0, 0xC0000000);
    
    // if DP mode, video input selection - sp1   
    MMIO_WR(DP_VIDEO_CTRL_REG, (MMIO_RD(DP_VIDEO_CTRL_REG) & 0x3FFC0001));

    // enable DP and video 
    MMIO_WR(DP_ENABLE_IF_REG, MMIO_RD(DP_ENABLE_IF_REG) | 0x00000009);
     
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0); 
}
 
void viaDisableDP(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
	
    DEBUG(ErrorF("viaDisableDP\n"));

    // disable DP and video
    MMIO_WR(DP_ENABLE_IF_REG, (MMIO_RD(DP_ENABLE_IF_REG) & 0xFFFFFFF6));
     
    // Let HW use AUX channel
    MMIO_WR(AUX_TIMER_REG, 0x0); 
}

void viaDpHorWidthAndTu(xf86OutputPtr output, DisplayModePtr mode)
{
    ScrnInfoPtr     pScrn = output->scrn;
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;
    int TUSize = 48;
    int Width;  

    viaCalculate_TU(output, mode->Clock, pScrn->bitsPerPixel);
    Width = (mode->CrtcHDisplay / 4) - 1;
    dwTemp = viaDPInfo->TURatio << 17 | (TUSize - 1) << 11 | Width;
    MMIO_WR(DP_HWIDTH_TUSIZE_REG, dwTemp);
}

void viaDpHorLineDuration(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;
    int LinkClock;
    int HTotal;
    int HBlank;

    // units of Dclk and Ls_clk are MHz
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }    

    HBlank = ((((mode->CrtcHTotal - mode->CrtcHDisplay) * LinkClock) / (mode->Clock / 10000)) - 15) & 0x00000FFF;
    HTotal = ((LinkClock * mode->CrtcHTotal / (mode->Clock / 1000)) - 15) & 0x00007FFF;

    dwTemp = HBlank << 15 | HTotal;
    MMIO_WR(DP_HLINE_DUR_REG, dwTemp);
}

void viaDpMvid(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;
    int LinkClock;
    int StreamClockRecovery_M;
	
    // units of Dclk and Ls_clk are MHz
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }   

    dwTemp = ((mode->Clock / 1000) * 32768 / LinkClock) & 0x19FFFFFF;
    dwTemp |=  0x20000000; // 8 bits, MMC6CC[31:29]=001b 
    dwTemp &=  0xFEFFFFFF;
    MMIO_WR(DP_MVID_MISC0_REG, dwTemp); // [24]=0: Asynchronous mode
    MMIO_WR_MASK(DP_ENABLE_IF_REG, 0x02000000, 0x02000000); // Generate MVID in Asynchronous mode
}

void viaDpHorVerTotal(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;

    dwTemp = mode->CrtcVTotal << 16 | mode->CrtcHTotal;
    MMIO_WR(DP_H_ATTR_REG, dwTemp);
}

void viaDpHorVerOffset(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;

    dwTemp = (mode->CrtcVTotal - mode->CrtcVSyncStart) << 16 | 
             (mode->CrtcHTotal - mode->CrtcHSyncStart);
    MMIO_WR(DP_HV_START_REG, dwTemp);
}

void viaDpHorVerSyncWidthAndPolarity(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;
    int HSyncPolarity, VSyncPolarity, VSyncInterval, HSyncInterval;

    if (mode->Flags & V_PHSYNC)
        HSyncPolarity = POSITIVE;

    if (mode->Flags & V_NHSYNC)
        HSyncPolarity = NEGATIVE;

    if (mode->Flags & V_PVSYNC)
        VSyncPolarity = POSITIVE;

    if (mode->Flags & V_NVSYNC)
        VSyncPolarity = NEGATIVE;                                                   

    VSyncInterval = mode->CrtcVSyncEnd - mode->CrtcVSyncStart;
    HSyncInterval = mode->CrtcHSyncEnd - mode->CrtcHSyncStart;
	
    dwTemp = VSyncPolarity << 31 | VSyncInterval << 16 |
             HSyncPolarity << 15 | HSyncInterval;
    MMIO_WR(DP_POLARITY_WIDTH_REG, dwTemp);
}

void viaDpHorVerActive(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;

    dwTemp = mode->CrtcVDisplay << 16 | mode->CrtcHDisplay;
    MMIO_WR(DP_ACITVE_WH_REG, dwTemp);
}

void viaDpNvid()
{
    // [23:0] Stream Clock Recovery :N  
    MMIO_WR_MASK(DP_ATTR_DATA_REG, 32768, 0x00FFFFFF);
}
 
void viaDpDelayReg()
{
    // Default: [29:24] = 20h for delay
    MMIO_WR_MASK(DP_VIDEO_CTRL_REG, 0x20000000, 0x3F000000);
}

Bool viaDPSetupMainLink(xf86OutputPtr output, DisplayModePtr mode)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
	
    DEBUG(ErrorF("viaDPSetupMainLink\n"));

    if (viaDPInfo->LinkSpeed) {
        // Field ID flag not invert
        MMIO_WR_MASK(DP_ENABLE_IF_REG, 0x0, BIT1);

        viaDpHorWidthAndTu(output, mode); 
        viaDpHorLineDuration(output, mode);
        viaDpMvid(output, mode);
        viaDpHorVerTotal(output, mode);
        viaDpHorVerOffset(output, mode);
        viaDpHorVerSyncWidthAndPolarity(output, mode);
        viaDpHorVerActive(output,mode);
        viaDpNvid();
        viaDpDelayReg();    
    
        DEBUG(ErrorF("End Setup MainLink!!\n"));
    }
}


Bool viaDPReadEDID(xf86OutputPtr output, CARD32 slave_addr, CARD32 dwOffset, CARD8 *pData)
{    
    CARD32 dwCmd;

    viaDPClearAuxReadRegs();
    viaDPClearAuxWriteRegs();

    // 1. write slave addr offset for start
    dwCmd = dwOffset & 0x000000FF;
    MMIO_WR(AUX_W_DATA0_REG, dwCmd);
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_WRITE;
    MMIO_WR(AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwCmd);
 
    if(!viaDPReplyCmdStatus(output))
    {
        return FALSE;
    }
 
    // 2. write command with read length
    dwCmd = slave_addr << 11 | 1 << 4  | AUX_CMD_I2C_READ;
    MMIO_WR(AUX_CMD_REG, dwCmd);
    // issue request based on above
    dwCmd = MMIO_RD(AUX_TIMER_REG) & 0xFFFFFFF0;
    dwCmd |= 0x03;
    MMIO_WR(AUX_TIMER_REG, dwCmd);
 
    if(!viaDPReplyCmdStatus(output))
    {
        return FALSE;
    }
 
    // 3. read 
    *pData = (MMIO_RD(AUX_R_DATA0_REG) & 0xFF);
 
    return TRUE;
}

/******************************************************************
*
*                                          DP2 Function                        
*
*******************************************************************/
void viaDP2DisableEPHY()
{
    MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x00000000, 0x00000001);
}

Bool viaDP2InitEPHY(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;

    DEBUG(ErrorF("viaDP2InitEPHY\n"));

    // (1) enable Bandgap and set EPHY symbol rate
    dwData = MMIO_RD(DP2_EPHY_PLL_REG) & 0x0C7000FF;
    // real EPHY symbol rate decided by HW link training
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            dwData |= 0x0002BE00;
            break;
        case Speed_270MHz:
            dwData |= 0x00019E00;
            break;
        default:
            dwData |= 0x00019E00;
            break;	
    }
    dwData |= 0xF1080001;
    MMIO_WR(DP2_EPHY_PLL_REG, dwData);
    viaDelayIn_usec(pVia, 1000); 
    
    // (2) start AUX channel, CMOP is on, Tx and Rx are off
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xCFFFFFFF;
    dwData |= 0x10000000;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData);
    viaDelayIn_usec(pVia, 4000);
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xCFFFFFFF;
    dwData |= 0x20000000;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData);
    viaDelayIn_usec(pVia, 4000);
    
    // (3) enable Mpll and Mpll reg power          
    // (4) enable Tpll and Tpll reg power
    MMIO_WR(DP2_EPHY_PLL_REG, (MMIO_RD(DP2_EPHY_PLL_REG) & (~0x02000000)));
    viaDelayIn_usec(pVia, 1000);
    
    // (5) Resister Tuning (RT), RT reset
    // (6)
    // [19:8] Tx power status default settings for all 4 lanes
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG2) & 0xFFF000FF;
    dwData |= 0x0000AA00;
    MMIO_WR(DP2_EPHY_TX_PWR_REG2, dwData); 
    // [7:0] Tx power down mode for all 4 lanes, [15:8] HDMI TX output voltage swing 1200 mV for all 4 lanes
    dwData = MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFF0000;
    dwData |= 0x0000FFFF;
    MMIO_WR(DP2_EPHY_TX_PWR_REG, dwData); 
    viaDelayIn_usec(pVia, 4000);
    // [7:0] Tx power control function CMOP on / PISO off for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_PWR_REG, (MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFFFFAA)); // Tx power control function on for all 4 lanes
    viaDelayIn_usec(pVia, 4000);
    // [7:0] Tx power control function on for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_PWR_REG, (MMIO_RD(DP2_EPHY_TX_PWR_REG) & 0xFFFFFF00)); // Tx power control function on for all 4 lanes
    viaDelayIn_usec(pVia, 1000);
    // [19:16] Driver awaken for all 4 lanes
    MMIO_WR(DP2_EPHY_TX_IDLE_REG, (MMIO_RD(DP2_EPHY_TX_IDLE_REG) | 0x000F0000));
    
    //(7) DP2 enable
    MMIO_WR(DP2_ENABLE_IF_REG, (MMIO_RD(DP2_ENABLE_IF_REG) | 0x00000001)); 
    viaDelayIn_usec(pVia, 2000);
}

void viaDP2ClearAuxWriteRegs()
{
    MMIO_WR(DP2_AUX_W_DATA0_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA1_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA2_REG, 0x0);
    MMIO_WR(DP2_AUX_W_DATA3_REG, 0x0);
}

void viaDP2ClearAuxReadRegs()
{
    MMIO_WR(DP2_AUX_R_DATA0_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA1_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA2_REG, 0x0);
    MMIO_WR(DP2_AUX_R_DATA3_REG, 0x0);
}

Bool viaDP2ReplyCmdStatus(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwAUXCMDStatus;
    CARD32 dwData;
    int i, j;

    //DEBUG(ErrorF("viaDP2ReplyCmdStatus\n"));
    
    for(i=0; i<NUM_RE_ISSUE_COMMAND; i++)
    {        
        for (j=0; j<NUM_RECHECK; j++)
        {
            viaDelayIn_usec(pVia, 1000);
            
            dwAUXCMDStatus = MMIO_RD(DP2_AUX_TIMER_REG) & 0x3C000007;
            
            if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1 && dwAUXCMDStatus == ACK_OK)
            {
                return TRUE;
            }
            else if (dwAUXCMDStatus == 0x00000001) // hardware issue
            {
                return TRUE;
            } 
            else if ((dwAUXCMDStatus & 0xFFFF) == RE_ISSUE_CMD)
            {
                // re-issue Cmd 
                dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
                dwData |= 0x03;
                MMIO_WR(DP2_AUX_TIMER_REG, dwData);
                break;
            }
        }
    }

    return FALSE;
}

void viaDp2TurnOnAuxPowerSequence(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);

    viaDelayIn_usec(pVia, 10);  // Delay TD0 time to wait HPD signal from sink
    viaDelayIn_usec(pVia, 10);  // Delay TD5 time
}

void viaDp2TurnOffAuxPowerSequence(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);

    viaDelayIn_usec(pVia, 10);  // Delay TD5 time
}

Bool viaDP2WriteDPCD(xf86OutputPtr output, CARD32 dwoffset, int ByteCount)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwCmd;
    CARD32 dwData;
    
    DEBUG(ErrorF("viaDP2WriteDPCD\n"));

    // if EPHY disable, turn on power sequence 
    if(!(MMIO_RD(DP2_ENABLE_IF_REG) & BIT0))
    {
        viaDp2TurnOnAuxPowerSequence(output);
    }
     
    dwCmd = dwoffset << 12;   // AUX read offset
    dwCmd |= ByteCount << 4;  // AUX read length
    dwCmd |= AUX_CMD_WRITE;   // AUX write CMD
    MMIO_WR(DP2_AUX_CMD_REG, dwCmd);
    
    // issue request based on above
    dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(DP2_AUX_TIMER_REG, dwData); // trigger

    viaDelayIn_usec(pVia, 10000); 

    // when leave, turn off power sequence 
    if(!(MMIO_RD(DP2_ENABLE_IF_REG) & BIT0))
    {
        viaDp2TurnOffAuxPowerSequence(output);
    }
     
    return viaDP2ReplyCmdStatus(output);
}
    
CARD32 viaDP2ReadDPCD(xf86OutputPtr output, CARD32 dwOffset)
{   
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    CARD32 dwCmd;
    CARD32 dwData;
    
    DEBUG(ErrorF("viaDP2ReadDPCD\n"));
    
    viaDP2ClearAuxReadRegs();
    
    // assign command
    dwCmd = dwOffset << 12;   // AUX read offset
    dwCmd |= 1 << 4;          // AUX read length
    dwCmd |= AUX_CMD_READ;    // AUX read CMD
    MMIO_WR(DP2_AUX_CMD_REG, dwCmd);
    
    // issue request based on above
    dwData = MMIO_RD(DP2_AUX_TIMER_REG) & 0xFFFFFFF0;
    dwData |= 0x03;
    MMIO_WR(DP2_AUX_TIMER_REG, dwData); 

    return viaDP2ReplyCmdStatus(output);
}

void viaDP2ReadDPCDVersion(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;
    CARD32 dwDPCDVersion;
    
    DEBUG(ErrorF("viaDP2ReadDPCDVersion\n"));
    
    // read DPCD Version
    if (viaDP2ReadDPCD(output, 0x00))
    {  
        dwDPCDVersion = MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF;
             
        switch (dwDPCDVersion)
        {
            case 0x10:
                viaDPInfo->DPCDVersion |= DPCD_VER_1 | DPCD_REV_0; // Version 1.0
                break;
            case 0x11:
                viaDPInfo->DPCDVersion |= DPCD_VER_1 | DPCD_REV_1; // Version 1.1
                break;
            default:   
                viaDPInfo->DPCDVersion = NONEVERSION;
                break;
        }
    }
	
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
}

Bool viaDP2LinkTraining(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwData;
    
    DEBUG(ErrorF("viaDP2LinkTraining\n"));
    
    viaDP2ClearAuxReadRegs();
    viaDP2ClearAuxWriteRegs();
    viaDPInfo->IsLinkTrainingDone = FALSE;
	
    // reset link training
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);
    MMIO_WR_MASK(DP2_NAUD_MUTE_REG, 0x80000000, 0x80000000); 
    viaDelayIn_usec(pVia, 1000);
    MMIO_WR_MASK(DP2_NAUD_MUTE_REG, 0x0, 0x80000000);
    viaDelayIn_usec(pVia, 1000);

    // [9:7]=100b, 4 lanes, [10]=0, disable SW link training , 
    // [12:11]=01b, set clock recovery state
    dwData = MMIO_RD(DP2_LINK_TRAINING_REG) & 0xFFFF0000; 
    dwData |= 0x0000023C; 

    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            dwData &= 0xFFFFFFFD;
            break;
        case Speed_270MHz:
            dwData |= 0x00000002;
            break;
        default:
            dwData |= 0x00000002;
            break;          
    }   
    MMIO_WR(DP2_LINK_TRAINING_REG, dwData);

    // Reset EPHY
    MMIO_WR_MASK(DP2_EPHY_TX_PWR_REG, 0x00000055, 0x000000FF);
    viaDelayIn_usec(pVia, 2000);
    MMIO_WR_MASK(DP2_EPHY_TX_PWR_REG, 0x00000000, 0x000000FF);

    // enable scramble
    MMIO_WR_MASK(DP2_VIDEO_CTRL_REG, 0x00000001, 0x00000001);

    // enable DP, set enhanced mode
    dwData = MMIO_RD(DP2_ENABLE_IF_REG) & 0xC3000000; 
    dwData |= 0x00000005; // if enable standard mode , set 0x00000001
    MMIO_WR(DP2_ENABLE_IF_REG, dwData);

    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x000018406); // if enable standard mode , set 0x000010406
            break;
        case Speed_270MHz:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x00001840A); // if enable standard mode , set 0x00001040A
            break;
        default:
            MMIO_WR(DP2_AUX_W_DATA0_REG, 0x00001840A);
            break;          
    }   
    
    MMIO_WR(DP2_AUX_W_DATA1_REG, 0x0);
    
    if (!viaDP2WriteDPCD(output, 0x100, 2))
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
        return FALSE;
    }
    
    //Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
    viaDelayIn_usec(pVia, 20000);

    dwData = MMIO_RD(DP2_LINK_TRAINING_REG) & 0x7FFFFFFF; 
    dwData |= 0x0000003C;  
   
    if (viaDPInfo->LinkSpeed == Speed_270MHz)
    {
         dwData |= 0x00000002;
    } 
    MMIO_WR(DP2_LINK_TRAINING_REG, dwData);
        
    // enable HW link training
    MMIO_WR_MASK(DP2_LINK_TRAINING_REG, 0x00000001, 0x00000001); 

    viaDelayIn_usec(pVia, 20000);;
    
    if ((MMIO_RD(DP2_LINK_TRAINING_REG) & 0x80000000) != 0x80000000)
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
        return FALSE;
    }

    viaDelayIn_usec(pVia, 4000);
    
    if ((MMIO_RD(DP2_ENABLE_IF_REG) & 0x30000000) != 0x30000000)
    {
        //Let HW use AUX channel
        MMIO_WR(DP2_AUX_TIMER_REG, 0x0);
        
        return FALSE;
    }

    viaDPInfo->IsLinkTrainingDone = TRUE;

    // clear [31], and [0]
    MMIO_WR_MASK(DP2_LINK_TRAINING_REG, 0x0, 0x80000001);
   
    //Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);

    return TRUE;

}

int viaDP2GetLinkSpeed(xf86OutputPtr output)
{   
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwLinkSpeed;

    if (viaDPInfo->IsLinkTrainingDone)
        return viaDPInfo->LinkSpeed;

    DEBUG(ErrorF("viaDP2GetLinkSpeed\n"));
	
    viaDP2ReadDPCD(output, 0x01);
    dwLinkSpeed = MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF;
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0);

    // Use default link speed of DP monitor
    switch (dwLinkSpeed)
    {
        case 0x06:
            viaDPInfo->LinkSpeed = Speed_162MHz;
            break;        
        case 0x0A:
            viaDPInfo->LinkSpeed = Speed_270MHz;
            break;          
        default:
            viaDPInfo->LinkSpeed = Speed_270MHz;
            break;    
    }

    // Start link training to confirm link speed
    if ((!viaDP2LinkTraining(output)) && (viaDPInfo->LinkSpeed == Speed_270MHz)) {
        // Reduce link speed to 162 MHz and retry link training if 270 MHz link training fails
        MMIO_WR(DP2_LINK_TRAINING_REG, (MMIO_RD(DP2_LINK_TRAINING_REG) & 0x7FFFFFFC));
        viaDP2ClearAuxWriteRegs();
        // Clear DPCD training pattern
        viaDP2WriteDPCD(output, 0x102, 1);
        viaDPInfo->LinkSpeed = Speed_162MHz;
        // Reset EPHY for changed link speed
        viaDP2InitEPHY(output);
        if (!viaDP2LinkTraining(output)) {
            viaDPInfo->LinkSpeed = Speed_NONE;
            viaDPInfo->IsLinkTrainingDone = FALSE;
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Output %s link training fails\n", output->name);
        }
    }

    return viaDPInfo->LinkSpeed;
}

void viaEnableDP2(xf86OutputPtr output)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    xf86CrtcPtr crtc = output->crtc;
	
    DEBUG(ErrorF("viaEnableDP2\n"));
	 
    // Must Enable HDCP first
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x01, 0x01);

    // if DP mode, video input selection - sp1
    MMIO_WR(DP2_VIDEO_CTRL_REG, (MMIO_RD(DP2_VIDEO_CTRL_REG) & 0x3FFC0001));

    // enable DP and video
    MMIO_WR(DP2_ENABLE_IF_REG, MMIO_RD(DP2_ENABLE_IF_REG) | 0x00000009);
    
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0); 
}
 
void viaDisableDP2(xf86OutputPtr output)
{
    DEBUG(ErrorF("viaDisableDP2\n"));

    // disable DP and video
    MMIO_WR(DP2_ENABLE_IF_REG, MMIO_RD(DP2_ENABLE_IF_REG) & 0xFFFFFFF6);
    
    // Let HW use AUX channel
    MMIO_WR(DP2_AUX_TIMER_REG, 0x0); 
}

void viaDp2HorWidthAndTu(xf86OutputPtr output, DisplayModePtr mode)
{
    ScrnInfoPtr     pScrn = output->scrn;
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;
    int TUSize = 48;
    int Width;  

    viaCalculate_TU(output, mode->Clock, pScrn->bitsPerPixel);
    Width = (mode->CrtcHDisplay / 4) - 1;
    dwTemp = viaDPInfo->TURatio << 17 | (TUSize - 1) << 11 | Width;
    MMIO_WR(DP2_HWIDTH_TUSIZE_REG, dwTemp);
}

void viaDp2HorLineDuration(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;
    int LinkClock;
    int HTotal;
    int HBlank;

    // units of Dclk and Ls_clk are MHz
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }    

    HBlank = ((((mode->CrtcHTotal - mode->CrtcHDisplay) * LinkClock) / (mode->Clock / 1000)) - 15) & 0x00000FFF;
    HTotal = ((LinkClock * mode->CrtcHTotal / (mode->Clock / 10000)) - 15) & 0x00007FFF;

    dwTemp = HBlank << 15 | HTotal;
    MMIO_WR(DP2_HLINE_DUR_REG, dwTemp);
}

void viaDp2Mvid(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;
    int LinkClock;
    int StreamClockRecovery_M;

    // units of Dclk and Ls_clk are MHz
    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 2700;
            break;          
        default:
            LinkClock = 270;
            break;
    }   

    dwTemp = ((mode->Clock / 1000) * 32768 / LinkClock) & 0x19FFFFFF;
    dwTemp |=  0x20000000; // 8 bits, MMC6CC[31:29]=001b 
    dwTemp &=  0xFEFFFFFF;
    MMIO_WR(DP2_MVID_MISC0_REG, dwTemp); // [24]=0: Asynchronous mode
    MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x02000000, 0x02000000); // Generate MVID in Asynchronous mode
}

void viaDp2HorVerTotal(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private; 
    CARD32 dwTemp;

    dwTemp = mode->CrtcVTotal << 16 | mode->CrtcHTotal;
    MMIO_WR(DP2_H_ATTR_REG, dwTemp);
}

void viaDp2HorVerOffset(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;

    dwTemp = (mode->CrtcVTotal - mode->CrtcVSyncStart) << 16 | 
             (mode->CrtcHTotal - mode->CrtcHSyncStart);
    MMIO_WR(DP2_HV_START_REG, dwTemp);
}

void viaDp2HorVerSyncWidthAndPolarity(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;
    int HSyncPolarity, VSyncPolarity, VSyncInterval, HSyncInterval;

    if (mode->Flags & V_PHSYNC)
        HSyncPolarity = POSITIVE;

    if (mode->Flags & V_NHSYNC)
        HSyncPolarity = NEGATIVE;

    if (mode->Flags & V_PVSYNC)
        VSyncPolarity = POSITIVE;

    if (mode->Flags & V_NVSYNC)
        VSyncPolarity = NEGATIVE;                                                   

    VSyncInterval = mode->CrtcVSyncEnd - mode->CrtcVSyncStart;
    HSyncInterval = mode->CrtcHSyncEnd - mode->CrtcHSyncStart;
	
    dwTemp = VSyncPolarity << 31 | VSyncInterval << 16 |
             HSyncPolarity << 15 | HSyncInterval;
    MMIO_WR(DP2_POLARITY_WIDTH_REG, dwTemp);
}

void viaDp2HorVerActive(xf86OutputPtr output, DisplayModePtr mode)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD32 dwTemp;

    dwTemp = mode->CrtcVDisplay << 16 | mode->CrtcHDisplay;
    MMIO_WR(DP2_ACITVE_WH_REG, dwTemp);
}

void viaDp2Nvid()
{
    // [23:0] Stream Clock Recovery :N  
    MMIO_WR_MASK(DP2_NVID_MISC0_REG, 32768, 0x00FFFFFF);
}

void viaDp2DelayReg()
{
    // Default: [29:24] = 20h for delay
    MMIO_WR_MASK(DP2_VIDEO_CTRL_REG, 0x20000000, 0x3F000000);
}

Bool viaDP2SetupMainLink(xf86OutputPtr output, DisplayModePtr mode)
{
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
	
    DEBUG(ErrorF("viaDP2SetupMainLink\n"));

    if (viaDPInfo->LinkSpeed) {
        // Field ID flag not invert
        MMIO_WR_MASK(DP2_ENABLE_IF_REG, 0x0, BIT1);

        viaDp2HorWidthAndTu(output, mode); 
        viaDp2HorLineDuration(output, mode);
        viaDp2Mvid(output, mode);
        viaDp2HorVerTotal(output, mode);
        viaDp2HorVerOffset(output, mode);
        viaDp2HorVerSyncWidthAndPolarity(output, mode);
        viaDp2HorVerActive(output,mode);
        viaDp2Nvid();
        viaDp2DelayReg();    
    
        DEBUG(ErrorF("End Setup MainLink!!\n"));
    }
}


Bool viaDP2ReadEDID(xf86OutputPtr output, CARD32 slave_addr, CARD32 dwOffset, CARD8 *pData)
{   
    CARD32 dwData;
	
    viaDP2ClearAuxReadRegs();
    viaDP2ClearAuxWriteRegs();

    // 1. write slave addr offset for start
    dwData = dwOffset & 0x000000FF;
    MMIO_WR(DP2_AUX_W_DATA0_REG, dwData);
    dwData = slave_addr << 11 | 1 << 4 | AUX_CMD_I2C_WRITE;
    MMIO_WR(DP2_AUX_CMD_REG, dwData);

    MMIO_WR(DP2_AUX_TIMER_REG, (MMIO_RD(DP2_AUX_TIMER_REG)|0x3)); // AUX trigger
    if(!viaDP2ReplyCmdStatus(output))
    {
        return FALSE;
    }

    // 2. write command with read length
    dwData = slave_addr << 11 | 1 << 4 | AUX_CMD_I2C_READ;
    MMIO_WR(DP2_AUX_CMD_REG, dwData);

    MMIO_WR(DP2_AUX_TIMER_REG, (MMIO_RD(DP2_AUX_TIMER_REG)|0x3)); // AUX trigger
    if(!viaDP2ReplyCmdStatus(output))
    {
        return FALSE;
    }

    // 3. read 
    *pData = (MMIO_RD(DP2_AUX_R_DATA0_REG) & 0xFF);

    return TRUE;
}

/******************************************************************
*
*                                          Common Function                        
*
*******************************************************************/
void viaCalculate_TU(xf86OutputPtr output, unsigned long PixelClock, int bpp_byte)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;

    int LinkClock;
    
    DEBUG(ErrorF("viaCalculate_TU\n"));

    switch (viaDPInfo->LinkSpeed)
    {
        case Speed_162MHz:
            LinkClock = 162;
            break;        
        case Speed_270MHz:
            LinkClock = 270;
            break;          
        default:
            LinkClock = 270;
            break;
    }
    
    bpp_byte = 24; 

    viaDPInfo->TURatio = ((PixelClock / 1000) * bpp_byte * 4096)/(LinkClock * LANE);   
}



void viaGetDPEDIDBlockData(xf86OutputPtr output, CARD8* pEDIDData)
{
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    CARD8   EDIDHeader[3];
    CARD8   i = 0;


    switch (viaDPInfo->commonInfo.subChipName) {
        case SUBCHIP_INTEGRATED_DP:
            /* First Read EDID Block 0 */
            for (i = 0; i < 3; i++) {    
                viaDPReadEDID(output, 0xA0, i, EDIDHeader+i);
            }

            /* EDID Header is 00 FF FF FF FF FF FF 00 */
            if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
                (EDIDHeader[2] == 0xFF)) {
                /* Read Block 0 Raw Data. */
                for (i = 0; i < 128; i++) {
                    viaDPReadEDID(output, 0xA0, i, pEDIDData+i);
                }    
            }
            break;
        case SUBCHIP_INTEGRATED_DP2:
            /* First Read EDID Block 0 */
            for (i = 0; i < 3; i++) {    
                viaDP2ReadEDID(output, 0xA0, i, EDIDHeader+i);
            }

            /* EDID Header is 00 FF FF FF FF FF FF 00 */
            if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) &&
                (EDIDHeader[2] == 0xFF)) {
                /* Read Block 0 Raw Data. */
                for (i = 0; i < 128; i++) {
                    viaDP2ReadEDID(output, 0xA0, i, pEDIDData+i);
                }    
            }
            break;
        default:
           break;
    }
}

CARD32 viaDPHotplugTimer(OsTimerPtr timer, CARD32 now, pointer arg)
{
    xf86OutputPtr output = (xf86OutputPtr) arg;
    ViaDPPrivateInfoPtr viaDPInfo = output->driver_private;
    ScrnInfoPtr     pScrn = output->scrn;
    VIAPtr  pVia = VIAPTR(pScrn);
    
    DEBUG(ErrorF("viaDPHotplugTimer\n"));

    switch (viaDPInfo->commonInfo.subChipName)
    {
        case SUBCHIP_INTEGRATED_DP:
            if (viaDPInfo->IsDPConnected) {
                // Detect unhotplug
                if(MMIO_RD(0xC730) & 0x80000000) {
                    viaDPInfo->IsDPConnected = FALSE;
                    viaDPInfo->DPCDVersion = NONEVERSION;
                    viaDPInfo->LinkSpeed = Speed_NONE;
                    viaDPInfo->IsLinkTrainingDone = FALSE;
                }
            } else {
                // Detect hotplug
                if (MMIO_RD(0xC730) & 0x40000000) {
                    viaDPReadDPCDVersion(output);
                    if (viaDPInfo->DPCDVersion) {
                        viaDPInfo->IsDPConnected = TRUE;
                        if (xf86InitialConfiguration(pScrn, FALSE))
                            xf86SetDesiredModes(pScrn);
                        else
                            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                                    "No valid initial configuration found\n");  
                    }
                }    
            }
            break;
        case SUBCHIP_INTEGRATED_DP2:
            if (viaDPInfo->IsDPConnected) {
                // Detect unhotplug
                if(MMIO_RD(0xC7B0) & 0x80000000) {
                    viaDPInfo->IsDPConnected = FALSE;
                    viaDPInfo->DPCDVersion = NONEVERSION;
                    viaDPInfo->LinkSpeed = Speed_NONE;
                    viaDPInfo->IsLinkTrainingDone = FALSE;
                }
            } else {
                // Detect hotplug
                if (MMIO_RD(0xC7B0) & 0x40000000) {
                    viaDP2ReadDPCDVersion(output);
                    if (viaDPInfo->DPCDVersion) {
                        viaDPInfo->IsDPConnected = TRUE;
                        if (xf86InitialConfiguration(pScrn, FALSE))
                            xf86SetDesiredModes(pScrn);
                        else
                            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                                    "No valid initial configuration found\n"); 
                    }
                }    
            }
            break;
    }
    
    /* 5 sec */
    return 5000; 
}
