/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 

/*************************************************************************
 *
 *  File:       via_exa.c
 *  Content:    2D acceleration function for VIA/S3G UniChrome
 *
 ************************************************************************/
#include "via_exa.h"
#include "via_exa_h5.h"
#include "via_exa_h6.h"
#include "via_dri.h"
#include "via_dmabuffer.h"
#include "via_rop.h"
#define RESERVED_FIX_SIZE (1024*1024)

ExaDriverPtr viaInitExa(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    int dimension;
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;

    /*3*mode size + 1*Onscreen size for RandR rotate and 3D Desktop*/
    int reserve4drm = ((pScrn->virtualY * pVia->Bpl)+3*pBIOSInfo->HDisplay*pBIOSInfo->VDisplay*4);
    double exaOffscreenSize; 

    ExaDriverPtr pExa = exaDriverAlloc();
    if (!pExa)
        return NULL;
	
    memset(pExa, 0, sizeof(*pExa));
    pExa->exa_major = EXA_VERSION_MAJOR;
    pExa->exa_minor = EXA_VERSION_MINOR;
    pExa->memoryBase = pVia->FBBase;

    if (pVia->FBFreeEnd -pVia->FBFreeStart < RESERVED_FIX_SIZE )
    { /*If Onscreen is too large ,almost cover the full framebuffer , return FALSE*/
      /* For there are still some fixed framebuffer needed(sync marker ...etc), so no less than 1M*/
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		"Video MEMORY is too small,Please Expand Video MM or resize your Virtual size!!\n");	

    	return FALSE;
    }
	
    if (!pVia->IsSecondary) {

	    /*default offscreen size is 3*Onscreen*/
	    exaOffscreenSize = 3;
	
        if (pVia->directRenderingEnabled) {
            pExa->memorySize = exaOffscreenSize*pScrn->virtualY * pVia->Bpl + pVia->FBFreeStart;
	        /*If not enough reserve for DRM , Split VRAM to offscreem and DRM as proportion 1:2,
		    if leaving VRAM size >4*modes(Least for 3D desktop and RandR) */

            if (pVia->FBFreeStart + reserve4drm >= pVia->FBFreeEnd)
            {
         	    if ( pVia->FBFreeEnd - pVia->FBFreeStart > 4*pBIOSInfo->HDisplay*pBIOSInfo->VDisplay*4) {
                    pExa->memorySize = pVia->FBFreeStart + (pVia->FBFreeEnd - pVia->FBFreeStart)/3;
                    exaOffscreenSize =  (double)((pVia->FBFreeEnd -pVia->FBFreeStart)/3)/(pScrn->virtualY * pVia->Bpl);
  		        } else {/*can't start 3D desktop, no reserve for drm*/
                    pExa->memorySize = pVia->FBFreeEnd;
                    exaOffscreenSize =  (double)(pVia->FBFreeEnd -pVia->FBFreeStart)/(pScrn->virtualY * pVia->Bpl);;
		        }
            }

            /*Reserve (3*mode size+ Onscreen) for  DRM,	which RandR rotate and 3D Desktop needs */
            while ((pExa->memorySize + reserve4drm) >= pVia->FBFreeEnd &&
				(pVia->FBFreeStart + reserve4drm) < pVia->FBFreeEnd )
            { /*needs to adjust offscreen size*/

                if (exaOffscreenSize <=2)
                    exaOffscreenSize -=0.1;
                else
                    exaOffscreenSize -=1;

                if (exaOffscreenSize <=0) {
                   /*can't reserve enough for DRM,The OnScreen is too large that fit almost full Framebuffer */
                   /*So there's no offscreen*/
                    pExa->memorySize = pVia->FBFreeEnd - reserve4drm;
                    break; 
                }
                pExa->memorySize = pVia->FBFreeStart+exaOffscreenSize*pScrn->virtualY * pVia->Bpl;
            }

            xf86DrvMsg(pScrn->scrnIndex, X_INFO,"[Memory_layout] EXA Offscreen size=%f\n",exaOffscreenSize);
            pExa->offScreenBase = pVia->FBFreeStart; 
            pVia->FBFreeStart = pExa->memorySize;
        }
    	else {
            pExa->memorySize = pVia->FBFreeEnd-RESERVED_FIX_SIZE;
            pExa->offScreenBase = pVia->FBFreeStart; 
            pVia->FBFreeStart = pExa->memorySize;
        }

    }
    else {
        pExa->memorySize = pVia->FBFreeEnd; /*Change for all 2nd offscreen for exa now*/
        pExa->offScreenBase =  pScrn->virtualY * pVia->Bpl + 1;
    }
	
    pVia->nPOT[0] = 0;
    pVia->nPOT[1] = 0;
    pExa->flags = EXA_OFFSCREEN_PIXMAPS;
    pExa->pixmapOffsetAlign = 256;
    pExa->pixmapPitchAlign = 32;

    /* Deliver the pitch limitation to exa up layer, and relax the max dimension
     * limitation for pixmap which will be accelerated by our HW.
     * From doing this, our driver will receive pixmaps who's dimension may be
     * larger than our HW's limitation. But that is what we want, since we want
     * to accelerate pixmaps who's dimension is larger than our hw limitation,
     * while its pitch still falls into our hw range.
     */
     /* Exa version < 2.3.0 hasn't gotten "maxPitchBytes" yet.
      * So it may throw pixmap whose pitch is bigger than 
      * our hw limitation to driver, and our driver need to care about this
      */
#if EXA_VERSION_MAJOR > 2 || (EXA_VERSION_MAJOR == 2 && EXA_VERSION_MINOR >= 3)
    if(H2_UMA_CHIPID) {
        pExa->maxPitchBytes = TEX_PITCH_LIMIT_INBYTE;
    } else {
        pExa->maxPitchBytes = (TEX_PITCH_LIMIT_INBYTE << 1) - 1;
    }
    ErrorF("DEBUG: maxPitchBytes %d\n", pExa->maxPitchBytes);
#endif
    if(H2_UMA_CHIPID) {
        dimension = TEX_PITCH_LIMIT_INBYTE / (pScrn->bitsPerPixel >> 3);
    } else {
        dimension = (TEX_PITCH_LIMIT_INBYTE << 1) / (pScrn->bitsPerPixel >> 3) - 1;
    }
    /* Do not let max dimension larger than twice TEX_DIMENSION_LIMIT_INPIXEL
     * Just for helping simplify the case.
     */
    pExa->maxX = pExa->maxY = (dimension > (TEX_DIMENSION_LIMIT_INPIXEL << 1)) ?
        (TEX_DIMENSION_LIMIT_INPIXEL << 1) : dimension;
	

    viaExaSetupFuncTable(pExa, pScreen);
	
    if (!exaDriverInit(pScreen, pExa)) {
        xfree(pExa);
        return NULL;
    }

    viaExa3DInit(pScreen);
    viaExa2DInit(pScreen);

    xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Enable EXA Now\n");

    return pExa;
}

void viaExaSetupFuncTable(ExaDriverPtr pExa,ScreenPtr pScreen)
{
	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
	
	switch (pVia->ChipId) {
	case PCI_CHIP_VT3336:
	case PCI_CHIP_VT3364:
		pExa->WaitMarker = viaAccelWaitMarker_H5;
    	pExa->MarkSync = viaAccelMarkSync_H5;

		pExa->PrepareCopy = viaExaPrepareCopy_H5;
    	pExa->Copy = viaExaCopy_H5;
    	pExa->DoneCopy = viaExaDoneSolidCopy_H5;
    	pExa->PrepareSolid = viaExaPrepareSolid_H5;
    	pExa->Solid = viaExaSolid_H5;
    	pExa->DoneSolid = viaExaDoneSolidCopy_H5;
#if XORG_VERSION_CURRENT < XF86_VERSION_NUMERIC(1,6,99,0,0) || \
                XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(7,0,0,0,0)
      pExa->PixmapIsOffscreen = viaExaIsOffscreen;
#endif
        if (!pVia->NoComposite) {
    	    pExa->CheckComposite = viaExaCheckComposite_H5;
    	    pExa->PrepareComposite = viaExaPrepareComposite_H5;
    	    pExa->Composite = viaExaComposite_H5;
            pExa->DoneComposite = viaExaDoneComposite_H5;
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "[EXA] Disabling EXA accelerated composite.\n");
        }

		break;
	case PCI_CHIP_VT3353:
	case PCI_CHIP_VT3409:
    case PCI_CHIP_VT3410:
		pExa->WaitMarker = viaAccelWaitMarker_H6;
    	pExa->MarkSync = viaAccelMarkSync_H6;

		pExa->PrepareCopy = viaExaPrepareCopy_H6;
    	pExa->Copy = viaExaCopy_H6;
    	pExa->DoneCopy = viaExaDoneSolidCopy_H6;
    	pExa->PrepareSolid = viaExaPrepareSolid_H6;
    	pExa->Solid = viaExaSolid_H6;
    	pExa->DoneSolid = viaExaDoneSolidCopy_H6;
#if XORG_VERSION_CURRENT < XF86_VERSION_NUMERIC(1,6,99,0,0) || \
                XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(7,0,0,0,0)
	pExa->PixmapIsOffscreen = viaExaIsOffscreen;
#endif
        if (!pVia->NoComposite) {
		    pExa->CheckComposite = viaExaCheckComposite_H5;
            pExa->PrepareComposite = viaExaPrepareComposite_H5;
            pExa->Composite = viaExaComposite_H5;
            pExa->DoneComposite = viaExaDoneComposite_H5;
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "[EXA] Disabling EXA accelerated composite.\n");
        }
            
		break;
	default:
		pExa->WaitMarker = viaAccelWaitMarker;
    	pExa->MarkSync = viaAccelMarkSync;

		pExa->PrepareCopy = viaExaPrepareCopy;
    	pExa->Copy = viaExaCopy;
    	pExa->DoneCopy = viaExaDoneSolidCopy;
    	pExa->PrepareSolid = viaExaPrepareSolid;
    	pExa->Solid = viaExaSolid;
    	pExa->DoneSolid = viaExaDoneSolidCopy;
#if XORG_VERSION_CURRENT < XF86_VERSION_NUMERIC(1,6,99,0,0) || \
                XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(7,0,0,0,0)
	pExa->PixmapIsOffscreen = viaExaIsOffscreen;
#endif		
        if (!pVia->NoComposite) {
    	    pExa->CheckComposite = viaExaCheckComposite;
    	    pExa->PrepareComposite = viaExaPrepareComposite;
    	    pExa->Composite = viaExaComposite;
    	    pExa->DoneComposite = viaExaDoneSolidCopy;
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "[EXA] Disabling EXA accelerated composite.\n");
        }

    	break;
	}

}

void viaExa2DInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);

    switch (pVia->ChipId) {
        case PCI_CHIP_VT3336:
        case PCI_CHIP_VT3364:
            viaInit2DState_H5(&pVia->td);
            break;
        case PCI_CHIP_VT3353:
        case PCI_CHIP_VT3409:
        case PCI_CHIP_VT3410:
            viaInit2DState_H6(&pVia->td, pVia->ChipId);
            break;
        default:
            viaInit2DState(&pVia->td);
            break;
        }
}

void viaExa3DInit(ScreenPtr pScreen)
{
	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);

	switch (pVia->ChipId) {
	case PCI_CHIP_VT3336:
	case PCI_CHIP_VT3364:
		viaInit3DState_H5(&pVia->v3d);
		break;
	case PCI_CHIP_VT3353:
	case PCI_CHIP_VT3409:
        pVia->v3d.chipset410=FALSE;
		viaInit3DState_H6(&pVia->v3d);
        break;
    case PCI_CHIP_VT3410:
        pVia->v3d.chipset410=TRUE;
		viaInit3DState_H6(&pVia->v3d);
		break;
	default:
		EXAInitialize3DEngine(pScrn);
		viaInit3DState(&pVia->v3d);
		break;
	}
}
/*
 * Mark Sync using the 2D blitter for AGP. NoOp for PCI.
 * In the future one could even launch a NULL PCI DMA command
 * to have an interrupt generated, provided it is possible to
 * write to the PCI DMA engines from the AGP command stream.
 */
 
int
viaAccelMarkSync(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);

    RING_VARS;

    ++pVia->curMarker;

    /* Wrap around without affecting the sign bit. */
    pVia->curMarker &= 0x7FFFFFFF;

    if (!pVia->IsPCI) {
        BEGIN_RING(2);
        BEGIN_HEAD1_2D(VIA_REG_KEYCONTROL, 0x00);
        viaAccelSolidHelper(cb, 0, 0, 1, 1, pVia->markerOffset,
                            VIA_GEM_32bpp, 4, pVia->curMarker,
                            (0xF0 << 24) | VIA_GEC_BLT | VIA_GEC_FIXCOLOR_PAT);
        ADVANCE_RING;
    }
    return pVia->curMarker;
}

void viaInit2DState(ViaTwodContext *td)
{
    td->setModeHelper = viaAccelSetMode;
    td->planeMaskHelper = viaAccelPlaneMaskHelper;
    td->transparentHelper = viaAccelTransparentHelper;
    td->copyHelper = viaAccelCopyHelper;
    td->solidHelper = viaAccelSolidHelper;
    td->ScalingHelper_H6_410 = NULL;
    td->CFCHelper_H6_410 = NULL;
    td->Set410Helper_H6 = NULL;
}

/*
 * Wait for the value to get blitted, or in the PCI case for engine idle.
 */
void
viaAccelWaitMarker(ScreenPtr pScreen, int marker)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    CARD32 uMarker = marker;

    if (!pVia->IsPCI) {
        while ((pVia->lastMarkerRead - uMarker) > (1 << 24))
            pVia->lastMarkerRead = *(volatile CARD32 *)pVia->markerBuf;
    } else {
        viaAccelSync(pScrn);
    }
}

/*
 * Wait for acceleration engines idle. An expensive way to sync.
 */
void
viaAccelSync(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    int loop = 0;

    mem_barrier();

	switch (pVia->ChipId) {
		case PCI_CHIP_VT3336:
		case PCI_CHIP_VT3364:
			while ((VIAGETREG(VIA_REG_STATUS) & \
				(VIA_CMD_RGTR_BUSY_H5| VIA_2D_ENG_BUSY_H5| VIA_3D_ENG_BUSY_H5)) \
				&& (loop++ < MAXLOOP)) ;
        	break;							
							
		case PCI_CHIP_VT3353:
		case PCI_CHIP_VT3409:
		case PCI_CHIP_VT3410:
			while ((VIAGETREG(VIA_REG_STATUS) & \
				(VIA_CMD_RGTR_BUSY_M1 | VIA_2D_ENG_BUSY_M1 | VIA_3D_ENG_BUSY_M1)) \
				&& (loop++ < MAXLOOP)) ;
        	break;

		default:
			while (!(VIAGETREG(VIA_REG_STATUS) & VIA_VR_QUEUE_EMPTY)
                   && (loop++ < MAXLOOP)) ;
			
            while ((VIAGETREG(VIA_REG_STATUS) &
                    (VIA_CMD_RGTR_BUSY | VIA_2D_ENG_BUSY | VIA_3D_ENG_BUSY))
                   && (loop++ < MAXLOOP)) ;
			
			break;
	}
}
static Bool
viaExaPrepareCopy(PixmapPtr pSrcPixmap, PixmapPtr pDstPixmap, int xdir,
                  int ydir, int alu, Pixel planeMask)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;

    RING_VARS;

    if (pSrcPixmap->drawable.bitsPerPixel != pDstPixmap->drawable.bitsPerPixel)
        return FALSE;

    if ((tdc->srcPitch = exaGetPixmapPitch(pSrcPixmap)) & 3)
        return FALSE;

    if (exaGetPixmapPitch(pDstPixmap) & 7)
        return FALSE;

    tdc->srcOffset = exaGetPixmapOffset(pSrcPixmap) + pScrn->fbOffset;
    tdc->cmd = VIA_GEC_BLT | VIAACCELCOPYROP(alu);
    if (xdir < 0)
        tdc->cmd |= VIA_GEC_DECX;
    if (ydir < 0)
        tdc->cmd |= VIA_GEC_DECY;

    if (!viaAccelSetMode(pDstPixmap->drawable.bitsPerPixel,0,0,0,0, tdc))
        return FALSE;

    if (!viaAccelPlaneMaskHelper(tdc, planeMask))
        return FALSE;
    viaAccelTransparentHelper(tdc, cb, 0x0, 0x0, TRUE);

    return TRUE;
}


static void
viaExaCopy(PixmapPtr pDstPixmap, int srcX, int srcY, int dstX, int dstY,
           int width, int height)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    CARD32 srcOffset = tdc->srcOffset;
    CARD32 dstOffset = exaGetPixmapOffset(pDstPixmap) + pScrn->fbOffset;

    RING_VARS;

    if (!width || !height)
        return;

    viaAccelCopyHelper(cb, srcX, srcY, dstX, dstY, width, height,
                   srcOffset, dstOffset, tdc->mode, tdc->srcPitch,
                   exaGetPixmapPitch(pDstPixmap), tdc->cmd);
    
    ADVANCE_RING;
}

static void
viaExaDoneSolidCopy(PixmapPtr pPixmap)
{
}


/*
 * Update our 2D state (TwoDContext) with a new mode.
 */
 #if 0
Bool
viaAccelSetMode(int bpp, ViaTwodContext * tdc)
{
    switch (bpp) {
        case 16:
            tdc->mode = VIA_GEM_16bpp;
            tdc->bytesPPShift = 1;
            return TRUE;
        case 32:
            tdc->mode = VIA_GEM_32bpp;
            tdc->bytesPPShift = 2;
            return TRUE;
        case 8:
            tdc->mode = VIA_GEM_8bpp;
            tdc->bytesPPShift = 0;
            return TRUE;
        default:
            tdc->bytesPPShift = 0;
            return FALSE;
    }
}
 #endif
static Bool
viaAccelSetMode(int bpp, unsigned format,unsigned rotate, unsigned alphaCnst,
                unsigned alphaVaule8, ViaTwodContext *tdc)
{
    switch (bpp) {
        case 16:
            tdc->mode = VIA_GEM_16bpp;
            tdc->bytesPPShift = 1;
            break;
        case 32:
            tdc->mode = VIA_GEM_32bpp;
            tdc->bytesPPShift = 2;
            break;
        case 8:
            tdc->mode = VIA_GEM_8bpp;
            tdc->bytesPPShift = 0;
            break;
/*old tdc->mode*/            
        default:
            /*410 set according to PicFormat*/
            if(format==PICT_x1r5g5b5){
                tdc->mode = VIA_GEM_16bpp;
                tdc->bytesPPShift = 1;
            }else if(format==PICT_r5g6b5){
                tdc->mode = VIA_GEM_16bpp;
                tdc->bytesPPShift = 1;
            }else if((format==PICT_x8r8g8b8)||(format==PICT_a8r8g8b8)){
                tdc->mode = VIA_GEM_32bpp;
                tdc->bytesPPShift = 2;
            }else{
                tdc->bytesPPShift = 0;
                return FALSE;
            }
            break;
/*add fm setting mode*/
    }
}


/*
 * Check if we can use a planeMask and update the 2D context accordingly.
 */
static Bool
viaAccelPlaneMaskHelper(ViaTwodContext * tdc, CARD32 planeMask)
{
    CARD32 modeMask = (1 << ((1 << tdc->bytesPPShift) << 3)) - 1;
    CARD32 curMask = 0x00000000;
    CARD32 curByteMask;
    int i;

    if ((planeMask & modeMask) != modeMask) {

        /* Masking doesn't work in 8bpp. */
        if (modeMask == 0xFF) {
            tdc->keyControl &= 0x0FFFFFFF;
            return FALSE;
        }

        /* Translate the bit planemask to a byte planemask. */
        for (i = 0; i < (1 << tdc->bytesPPShift); ++i) {
            curByteMask = (0xFF << (i << 3));

            if ((planeMask & curByteMask) == 0) {
                curMask |= (1 << i);
            } else if ((planeMask & curByteMask) != curByteMask) {
                tdc->keyControl &= 0x0FFFFFFF;
                return FALSE;
            }
        }
        ErrorF("DEBUG: planeMask 0x%08x, curMask 0%02x\n",
               (unsigned)planeMask, (unsigned)curMask);

        tdc->keyControl = (tdc->keyControl & 0x0FFFFFFF) | (curMask << 28);
    }

    return TRUE;
}


/*
 * Emit transparency state and color to the command buffer.
 */
static void
viaAccelTransparentHelper(ViaTwodContext * tdc, ViaCommandBuffer * cb,
                          CARD32 keyControl, CARD32 transColor,
                          Bool usePlaneMask)
{
    tdc->keyControl &= ((usePlaneMask) ? 0xF0000000 : 0x00000000);
    tdc->keyControl |= (keyControl & 0x0FFFFFFF);
    BEGIN_RING(4);
    BEGIN_HEAD1_2D(VIA_REG_KEYCONTROL, tdc->keyControl);
    if (keyControl) {
        BEGIN_HEAD1_2D(VIA_REG_SRCCOLORKEY, transColor);
    }
}


/*
 * Emit a copy blit operation to the command buffer.
 */
static void
viaAccelCopyHelper(ViaCommandBuffer * cb, int xs, int ys, int xd, int yd,
                   int w, int h, unsigned srcFbBase, unsigned dstFbBase,
                   CARD32 mode, unsigned srcPitch, unsigned dstPitch,
                   CARD32 cmd)
{
    if (cmd & VIA_GEC_DECY) {
        ys += h - 1;
        yd += h - 1;
    }

    if (cmd & VIA_GEC_DECX) {
        xs += w - 1;
        xd += w - 1;
    }

    BEGIN_RING(16);
    BEGIN_HEAD1_2D(VIA_REG_GEMODE, mode);
    BEGIN_HEAD1_2D(VIA_REG_SRCBASE, srcFbBase >> 3);
    BEGIN_HEAD1_2D(VIA_REG_DSTBASE, dstFbBase >> 3);
    BEGIN_HEAD1_2D(VIA_REG_PITCH, VIA_PITCH_ENABLE |
                ((dstPitch >> 3) << 16) | (srcPitch >> 3));
    BEGIN_HEAD1_2D(VIA_REG_SRCPOS, (ys << 16) | (xs & 0xFFFF));
    BEGIN_HEAD1_2D(VIA_REG_DSTPOS, (yd << 16) | (xd & 0xFFFF));
    BEGIN_HEAD1_2D(VIA_REG_DIMENSION, ((h - 1) << 16) | (w - 1));
    BEGIN_HEAD1_2D(VIA_REG_GECMD, cmd);
}


/*
 * Emit a solid blit operation to the command buffer. 
 */
static void
viaAccelSolidHelper(ViaCommandBuffer * cb, int x, int y, int w, int h,
                    unsigned fbBase, CARD32 mode, unsigned pitch,
                    CARD32 fg, CARD32 cmd)
{
    BEGIN_RING(14);
    BEGIN_HEAD1_2D(VIA_REG_GEMODE, mode);
    BEGIN_HEAD1_2D(VIA_REG_DSTBASE, fbBase >> 3);
    BEGIN_HEAD1_2D(VIA_REG_PITCH, VIA_PITCH_ENABLE | (pitch >> 3) << 16);
    BEGIN_HEAD1_2D(VIA_REG_DSTPOS, (y << 16) | (x & 0xFFFF));
    BEGIN_HEAD1_2D(VIA_REG_DIMENSION, ((h - 1) << 16) | (w - 1));
    BEGIN_HEAD1_2D(VIA_REG_FGCOLOR, fg);
    BEGIN_HEAD1_2D(VIA_REG_GECMD, cmd);
}

static Bool
viaExaPrepareSolid(PixmapPtr pPixmap, int alu, Pixel planeMask, Pixel fg)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;

    RING_VARS;

    if (exaGetPixmapPitch(pPixmap) & 7)
        return FALSE;

    if (!viaAccelSetMode(pPixmap->drawable.bitsPerPixel,0,0,0,0, tdc))
        return FALSE;

    if (!viaAccelPlaneMaskHelper(tdc, planeMask))
        return FALSE;
    viaAccelTransparentHelper(tdc, cb, 0x0, 0x0, TRUE);

    tdc->cmd = VIA_GEC_BLT | VIA_GEC_FIXCOLOR_PAT | VIAACCELPATTERNROP(alu);

    tdc->fgColor = fg;

    return TRUE;
}


static void
viaExaSolid(PixmapPtr pPixmap, int x1, int y1, int x2, int y2)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    CARD32 dstPitch, dstOffset;

    RING_VARS;

    int w = x2 - x1, h = y2 - y1;

    dstPitch = exaGetPixmapPitch(pPixmap);
    dstOffset = exaGetPixmapOffset(pPixmap) + pScrn->fbOffset;
    viaAccelSolidHelper(cb, x1, y1, w, h, dstOffset,
                        tdc->mode, dstPitch, tdc->fgColor, tdc->cmd);
    ADVANCE_RING;
}

static Bool
viaExaCheckComposite(int op, PicturePtr pSrcPicture,
                     PicturePtr pMaskPicture, PicturePtr pDstPicture)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPicture->pDrawable->pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    Via3DState *v3d = &pVia->v3d;

    /* Reject small composites early. They are done much faster in software. */
    if (pSrcPicture && pSrcPicture->pDrawable) {
        if (!pSrcPicture->repeat &&
            pSrcPicture->pDrawable->width *
            pSrcPicture->pDrawable->height < VIA_MIN_COMPOSITE)
            return FALSE;
    } else {
        return FALSE;
    }

    if (pMaskPicture && pMaskPicture->pDrawable) {
        if (!pMaskPicture->repeat &&
            pMaskPicture->pDrawable->width *
            pMaskPicture->pDrawable->height < VIA_MIN_COMPOSITE)
            return FALSE;
    }

    if (pMaskPicture && pMaskPicture->componentAlpha) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "use componentAlpha \n"));
        /*return FALSE;*/
        if((op==PictOpAdd)||(op==PictOpOver)||(op==PictOpAtop)||(op==PictOpAtopReverse)||(op==PictOpXor))
            return FALSE;  	
    }

    if (!v3d->opSupported(op)) {
#ifdef VIA_DEBUG_COMPOSITE
#warning Composite verbose debug turned on.
        ErrorF("Operator not supported\n");
        viaExaPrintComposite(op, pSrcPicture, pMaskPicture, pDstPicture);
#endif
        return FALSE;
    }

    if (pSrcPicture && pSrcPicture->pDrawable && pMaskPicture && pMaskPicture->pDrawable
                && pDstPicture && pDstPicture->pDrawable) {
        if (((pSrcPicture->pDrawable->width > TEX_DIMENSION_LIMIT_INPIXEL ||
            pSrcPicture->pDrawable->height > TEX_DIMENSION_LIMIT_INPIXEL) || 
            (pMaskPicture && (pMaskPicture->pDrawable->width > TEX_DIMENSION_LIMIT_INPIXEL ||
            pMaskPicture->pDrawable->height > TEX_DIMENSION_LIMIT_INPIXEL))) &&
            (pDstPicture->pDrawable->width > TEX_DIMENSION_LIMIT_INPIXEL || 
            pDstPicture->pDrawable->height > TEX_DIMENSION_LIMIT_INPIXEL)) {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "The Composite is not support the region bigger than 2048 \n");
            return FALSE;
        }
    }

    /*
     * FIXME: A8 destination formats are currently not supported and do not
     * seem supported by the hardware, although there are some leftover
     * register settings apparent in the via_3d_reg.h file. We need to fix this
     * (if important), by using component ARGB8888 operations with bitmask.
     */

    if (!v3d->dstSupported(pDstPicture->format)) {
#ifdef VIA_DEBUG_COMPOSITE
        ErrorF("Destination format not supported:\n");
        viaExaPrintComposite(op, pSrcPicture, pMaskPicture, pDstPicture);
#endif
        return FALSE;
    }

    if (v3d->texSupported(pSrcPicture->format)) {
        if (pMaskPicture && (PICT_FORMAT_A(pMaskPicture->format) == 0 ||
                             !v3d->texSupported(pMaskPicture->format))) {
#ifdef VIA_DEBUG_COMPOSITE
            ErrorF("Mask format not supported:\n");
            viaExaPrintComposite(op, pSrcPicture, pMaskPicture, pDstPicture);
#endif
            return FALSE;
        }
        return TRUE;
    }
#ifdef VIA_DEBUG_COMPOSITE
    ErrorF("Src format not supported:\n");
    viaExaPrintComposite(op, pSrcPicture, pMaskPicture, pDstPicture);
#endif
    return FALSE;
}


static Bool
viaExaPrepareComposite(int op, PicturePtr pSrcPicture,
                       PicturePtr pMaskPicture, PicturePtr pDstPicture,
                       PixmapPtr pSrc, PixmapPtr pMask, PixmapPtr pDst)
{
    CARD32 dstOffset, pot_width, pot_height;
    ScrnInfoPtr pScrn = xf86Screens[pDst->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    Via3DState *v3d = &pVia->v3d;
    ViaTextureUnit *vTex = v3d->tex;
    int curTex = 0;
    ViaTexBlendingModes srcMode,maskMode;
    unsigned long offset;
    v3d->componentAlpha = v3d->srcRepeat = v3d->maskRepeat = FALSE;

    dstOffset = exaGetPixmapOffset(pDst) + pScrn->fbOffset;
    v3d->setDestination(v3d, dstOffset/*exaGetPixmapOffset(pDst)*/,
                        exaGetPixmapPitch(pDst), pDstPicture->format);
    if(pMaskPicture && pMaskPicture->componentAlpha)
        v3d->componentAlpha = TRUE;
    else
        v3d->componentAlpha = FALSE;
    v3d->setCompositeOperator(v3d, op, v3d->componentAlpha);
    v3d->setDrawing(v3d, 0x0c, 0xFFFFFFFF, 0x000000FF, 0xFF);

    /*
     * For one-pixel repeat mask pictures we avoid using multitexturing by
     * modifying the src's texture blending equation and feed the pixel
     * value as a constant alpha for the src's texture. Multitexturing on the
     * Unichromes seems somewhat slow, so this speeds up translucent windows.
     */

    srcMode = via_src;
    pVia->maskP = NULL;
    if (pMaskPicture &&
        (pMaskPicture->pDrawable->height == 1) &&
        (pMaskPicture->pDrawable->width == 1) &&
        pMaskPicture->repeat && viaExpandablePixel(pMaskPicture->format)) {
        pVia->maskP = exaGetPixmapOffset(pMask) + pVia->FBBase;
        pVia->componentAlpha = pMaskPicture->componentAlpha;
        v3d->maskRepeat = TRUE;
        srcMode = ((pMaskPicture->componentAlpha)
                   ? via_src_onepix_comp_mask : via_src_onepix_mask);
    }

    if(pMaskPicture) {
        pVia->maskFormat = pMaskPicture->format;
    }

    /*
     * One-Pixel repeat src pictures go as solid color instead of textures.
     * Speeds up window shadows.
     */

    pVia->srcP = NULL;
    if (pSrcPicture && pSrcPicture->repeat
        && (pSrcPicture->pDrawable->height == 1)
        && (pSrcPicture->pDrawable->width == 1)
        && viaExpandablePixel(pSrcPicture->format)) {
        pVia->srcP = exaGetPixmapOffset(pSrc) + pVia->FBBase;
        v3d->srcRepeat = TRUE;
    }

    if(pSrcPicture) {
        pVia->srcFormat = pSrcPicture->format;
    }
    
    /* Exa should be smart enough to eliminate this IN operation. */
    if (pVia->srcP && pVia->maskP) {
        ErrorF("Bad one-pixel IN composite operation. "
               "EXA needs to be smarter.\n");
        return FALSE;
    }
    maskMode=via_mask;
    if(pMaskPicture && pMaskPicture->componentAlpha){
	     switch(op){
            case PictOpOutReverse:
			case PictOpInReverse:
			    srcMode =src_Aa;
			    maskMode =mask_Ca;
			    break;
			case PictOpOverReverse:
			case PictOpIn:
			case PictOpOut:
			case PictOpAdd:
			case PictOpSrc:
			case PictOpDisjointSrc:
			case PictOpConjointSrc:
			    srcMode=src_Ca;
				maskMode =mask_Ca;
				break;
			case PictOpClear:
			case PictOpDisjointClear:
			case PictOpConjointClear:
			case PictOpDst:
			case PictOpDisjointDst:
			case PictOpConjointDst:
			    pVia->srcP=NULL;
			    pVia->maskP = NULL;
			    curTex=0;
			    goto AlphaBlendOnly;
			    break;
			default:
			    break;
		 }
		 pVia->srcP=NULL;
		 pVia->maskP = NULL;
	}
    if (!pVia->srcP) {
        offset = exaGetPixmapOffset(pSrc) + pScrn->fbOffset;
        vTex = v3d->tex + curTex;
        vTex->PictureOffset= offset;
        vTex->npot = pVia->nPOT[curTex];
        vTex->textureLevel0Pitch = exaGetPixmapPitch(pSrc);
        vTex->pictureFormat= pSrcPicture->format;
        vTex->textureBlendMode = srcMode;
        vTex->transform = pSrcPicture->transform;
        vTex->bytePerPixel = pSrcPicture->pDrawable->bitsPerPixel >> 3;
        vTex->textureDrawable_width = pSrc->drawable.width;
        vTex->textureDrawable_height = pSrc->drawable.height;
        vTex->pictureFilter = pSrcPicture->filter;
        vTex->textureRepeat = v3d->srcRepeat;
        if(pSrcPicture->repeat) {
                viaOrder(vTex->textureDrawable_width, &pot_width);
                viaOrder(vTex->textureDrawable_height, &pot_height);
                pot_width = 1 << pot_width;
                pot_height = 1 << pot_height;
                if((vTex->textureDrawable_width != pot_width) || 
                    (vTex->textureDrawable_height != pot_height))
                    return FALSE;
            }
        curTex++;
    }

    if (pMaskPicture && !pVia->maskP) {
        offset = exaGetPixmapOffset(pMask) + pScrn->fbOffset;
        vTex = v3d->tex + curTex;                             
        vTex->PictureOffset = offset;
        vTex->npot = pVia->nPOT[curTex];
        vTex->textureLevel0Pitch = exaGetPixmapPitch(pMask);
        vTex->pictureFormat = pMaskPicture->format;
        vTex->textureBlendMode = maskMode;
        vTex->transform = pMaskPicture->transform;
        vTex->bytePerPixel = pMaskPicture->pDrawable->bitsPerPixel >> 3;
        vTex->textureDrawable_width = pMask->drawable.width;
        vTex->textureDrawable_height= pMask->drawable.height;
        vTex->pictureFilter = pMaskPicture->filter;
        vTex->textureRepeat = v3d->maskRepeat;
        if(pMaskPicture->repeat) {
                viaOrder(vTex->textureDrawable_width, &pot_width);
                viaOrder(vTex->textureDrawable_height, &pot_height);
                pot_width = 1 << pot_width;
                pot_height = 1 << pot_height;
                if((vTex->textureDrawable_width != pot_width) || 
                    (vTex->textureDrawable_height != pot_height))
                    return FALSE;
            }
        curTex++;
    }
AlphaBlendOnly:
    v3d->setFlags(v3d, curTex, TRUE, TRUE, TRUE);

    return TRUE;
}

static void
viaExaComposite(PixmapPtr pDst, int srcX, int srcY, int maskX, int maskY,
                int dstX, int dstY, int width, int height)
{
    ScrnInfoPtr pScrn = xf86Screens[pDst->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    Via3DState *v3d = &pVia->v3d;
    CARD32 dstOffset;
    ViaTextureUnit *vTex = v3d->tex;
    CARD32 col, i, j,textureDrawable_width,textureDrawable_height;
    int dst_X_wasted =0,dst_Y_wasted=0;
    v3d->destOffset = exaGetPixmapOffset(pDst) + pScrn->fbOffset;
    v3d->forceUpload = viaCheckUpload(pScrn, v3d);
    RING_VARS;
    for(i = 0; i < v3d->numTextures; i++) {
        vTex = v3d->tex + i;
        textureDrawable_width =
            vTex->textureDrawable_width > TEX_DIMENSION_LIMIT_INPIXEL ?
            TEX_DIMENSION_LIMIT_INPIXEL : vTex->textureDrawable_width;
        textureDrawable_height =
            vTex->textureDrawable_height > TEX_DIMENSION_LIMIT_INPIXEL ?
            TEX_DIMENSION_LIMIT_INPIXEL : vTex->textureDrawable_height;
              
        v3d->setTexture(v3d, i, vTex->PictureOffset,
                 vTex->textureLevel0Pitch, vTex->npot,
                 textureDrawable_width,textureDrawable_height,
                 vTex->pictureFormat,
                 via_repeat, via_repeat,
                  vTex->textureBlendMode, FALSE, 
                  vTex->transform, vTex->pictureFilter); 
    }


    if (pVia->maskP) {
        viaPixelARGB8888(pVia->maskFormat, pVia->maskP, &col);
        v3d->setTexBlendCol(v3d, 0, pVia->componentAlpha, col);
        }
    if (pVia->srcP) {
        viaPixelARGB8888(pVia->srcFormat, pVia->srcP, &col);
        v3d->setDrawing(v3d, 0x0c, 0xFFFFFFFF, col & 0x00FFFFFF, col >> 24);
        srcX = maskX;
        srcY = maskY;
        }
    if (((dstX + width) >= TEX_DIMENSION_LIMIT_INPIXEL) ||
        ((dstY + height) >= TEX_DIMENSION_LIMIT_INPIXEL)) {
        dst_X_wasted = dstX % 256;
        dst_Y_wasted = dstY % 256;
        dstX = dstX - dst_X_wasted;
        dstY = dstY - dst_Y_wasted;
        v3d->destOffset +=
            dstY * v3d->destPitch + dstX * (pDst->drawable.bitsPerPixel >> 3);
        dstX = dst_X_wasted;
        dstY = dst_Y_wasted;
        }
    
    v3d->emitState(v3d, cb, 1); /*force upload all Setting*/
    v3d->emitClipRect(v3d, &pVia->cb, dstX, dstY, width,height);
    v3d->emitQuad(v3d, &pVia->cb, dstX, dstY, srcX, srcY, maskX, maskY,width, height);
    ADVANCE_RING;
}

void
EXAInitialize3DEngine(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    int i;

    VIASETREG(VIA_REG_TRANSET, 0x00010000);
    for (i = 0; i <= 0x7D; i++)
        VIASETREG(VIA_REG_TRANSPACE, (CARD32) i << 24);

    VIASETREG(VIA_REG_TRANSET, 0x00020000);
    for (i = 0; i <= 0x94; i++)
        VIASETREG(VIA_REG_TRANSPACE, (CARD32) i << 24);
    VIASETREG(VIA_REG_TRANSPACE, 0x82400000);

    VIASETREG(VIA_REG_TRANSET, 0x01020000);
    for (i = 0; i <= 0x94; i++)
        VIASETREG(VIA_REG_TRANSPACE, (CARD32) i << 24);
    VIASETREG(VIA_REG_TRANSPACE, 0x82400000);

    VIASETREG(VIA_REG_TRANSET, 0xfe020000);
    for (i = 0; i <= 0x03; i++)
        VIASETREG(VIA_REG_TRANSPACE, (CARD32) i << 24);

    VIASETREG(VIA_REG_TRANSET, 0x00030000);
    for (i = 0; i <= 0xff; i++)
        VIASETREG(VIA_REG_TRANSPACE, 0);

    VIASETREG(VIA_REG_TRANSET, 0x00100000);
    VIASETREG(VIA_REG_TRANSPACE, 0x00333004);
    VIASETREG(VIA_REG_TRANSPACE, 0x10000002);
    VIASETREG(VIA_REG_TRANSPACE, 0x60000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x61000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x62000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x63000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x64000000);

    VIASETREG(VIA_REG_TRANSET, 0x00fe0000);
    VIASETREG(VIA_REG_TRANSPACE, 0x4000800f);
    VIASETREG(VIA_REG_TRANSPACE, 0x44000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x45080C04);
    VIASETREG(VIA_REG_TRANSPACE, 0x46800408);
    VIASETREG(VIA_REG_TRANSPACE, 0x50000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x51000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x52000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x53000000);

    VIASETREG(VIA_REG_TRANSET, 0x00fe0000);
    VIASETREG(VIA_REG_TRANSPACE, 0x08000001);
    VIASETREG(VIA_REG_TRANSPACE, 0x0A000183);
    VIASETREG(VIA_REG_TRANSPACE, 0x0B00019F);
    VIASETREG(VIA_REG_TRANSPACE, 0x0C00018B);
    VIASETREG(VIA_REG_TRANSPACE, 0x0D00019B);
    VIASETREG(VIA_REG_TRANSPACE, 0x0E000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x0F000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x10000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x11000000);
    VIASETREG(VIA_REG_TRANSPACE, 0x20000000);
}

Bool
viaIsOffscreen(VIAPtr pVia, PixmapPtr pPix)
{
    return ((unsigned long)pPix->devPrivate.ptr -
            (unsigned long)pVia->FBBase) < pVia->videoRambytes;
}

Bool
viaExaIsOffscreen(PixmapPtr pPix)
{	
	ScreenPtr	pScreen = pPix->drawable.pScreen;
	ScrnInfoPtr	pScrn = xf86Screens[pScreen->myNum];
    VIAPtr		pVia = VIAPTR(pScrn);
	
    return ((unsigned long)pPix->devPrivate.ptr -
            (unsigned long)pVia->FBBase) < pVia->videoRambytes;
}

/*
 * Check if the above function will work.
 */
Bool
viaExpandablePixel(int format)
{
    int formatType = PICT_FORMAT_TYPE(format);

    return (formatType == PICT_TYPE_A ||
            formatType == PICT_TYPE_ABGR || formatType == PICT_TYPE_ARGB);
}

/*
 * Check if we need to force upload of the whole 3D state (when other
 * clients or subsystems have touched the 3D engine). Also tell DRI
 * clients and subsystems that we have touched the 3D engine.
 */
Bool
viaCheckUpload(ScrnInfoPtr pScrn, Via3DState * v3d)
{
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    Bool forceUpload;

    forceUpload = (pVia->lastToUpload != v3d);
    pVia->lastToUpload = v3d;

#ifdef XF86DRI
    if (pVia->directRenderingEnabled) {
        volatile drm_via_sarea_t *saPriv = (drm_via_sarea_t *)
                DRIGetSAREAPrivate(pScrn->pScreen);
        int myContext = DRIGetContext(pScrn->pScreen);

        forceUpload = forceUpload || (saPriv->ctxOwner != myContext);
        saPriv->ctxOwner = myContext;
    }
#endif
    return (forceUpload || (pVia->DISP3DScalingCaps & pBIOSInfo->ActiveDevice));
}

/*
 * Extract the components from a pixel of the given format to an argb8888 pixel.
 * This is used to extract data from one-pixel repeat pixmaps.
 * Assumes little endian.
 */
void
viaPixelARGB8888(unsigned format, void *pixelP, CARD32 * argb8888)
{
    CARD32 bits, shift, pixel, bpp;

    bpp = PICT_FORMAT_BPP(format);

    if (bpp <= 8) {
        pixel = *((CARD8 *) pixelP);
    } else if (bpp <= 16) {
        pixel = *((CARD16 *) pixelP);
    } else {
        pixel = *((CARD32 *) pixelP);
    }

    switch (PICT_FORMAT_TYPE(format)) {
        case PICT_TYPE_A:
            bits = PICT_FORMAT_A(format);
            *argb8888 = viaBitExpandHelper(pixel, bits) << 24;
            return;
        case PICT_TYPE_ARGB:
            shift = 0;
            bits = PICT_FORMAT_B(format);
            *argb8888 = viaBitExpandHelper(pixel, bits);
            shift += bits;
            bits = PICT_FORMAT_G(format);
            *argb8888 |= viaBitExpandHelper(pixel >> shift, bits) << 8;
            shift += bits;
            bits = PICT_FORMAT_R(format);
            *argb8888 |= viaBitExpandHelper(pixel >> shift, bits) << 16;
            shift += bits;
            bits = PICT_FORMAT_A(format);
            *argb8888 |= ((bits) ? viaBitExpandHelper(pixel >> shift,
                                                      bits) : 0xFF) << 24;
            return;
        case PICT_TYPE_ABGR:
            shift = 0;
            bits = PICT_FORMAT_B(format);
            *argb8888 = viaBitExpandHelper(pixel, bits) << 16;
            shift += bits;
            bits = PICT_FORMAT_G(format);
            *argb8888 |= viaBitExpandHelper(pixel >> shift, bits) << 8;
            shift += bits;
            bits = PICT_FORMAT_R(format);
            *argb8888 |= viaBitExpandHelper(pixel >> shift, bits);
            shift += bits;
            bits = PICT_FORMAT_A(format);
            *argb8888 |= ((bits) ? viaBitExpandHelper(pixel >> shift,
                                                      bits) : 0xFF) << 24;
            return;
        default:
            break;
    }
    return;
}

/*
 * Helper for bitdepth expansion.
 */
static CARD32
viaBitExpandHelper(CARD32 pixel, CARD32 bits)
{
    CARD32 component, mask, tmp;

    component = pixel & ((1 << bits) - 1);
    mask = (1 << (8 - bits)) - 1;
    tmp = component << (8 - bits);
    return ((component & 1) ? (tmp | mask) : tmp);
}

void via_ExaMoveDwords(CARD32* dest,CARD32* src,int dwords )
{
     while(dwords & ~0x03) {
	*dest = *src;
	*(dest + 1) = *(src + 1);
	*(dest + 2) = *(src + 2);
	*(dest + 3) = *(src + 3);
	src += 4;
	dest += 4;
	dwords -= 4;
     }	
     if(!dwords) return;
     *dest = *src;
     if(dwords == 1) return;
     *(dest + 1) = *(src + 1);
     if(dwords == 2) return;
     *(dest + 2) = *(src + 2);
}


