/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 

/*************************************************************************
 *
 *  File:       via_exa_h6.c
 *  Content:    2D acceleration function for VIA/S3G UniChrome
 *
 ************************************************************************/
#include "via_exa_h6.h"
#include "via_rop.h"

#define H6_EXA_PITCH_CHECK(pitch) (((pitch) >= 0x4000) || ((pitch) & 0x000F))
#define H6_EXA_OFFSET_CHECK(offset) ((offset) & 0xFF)

void viaInit2DState_H6(ViaTwodContext *td ,int Chipid)
{
    td->setModeHelper = viaAccelSetMode_H6;
    td->planeMaskHelper = viaAccelPlaneMaskHelper_H6;
    td->transparentHelper = viaAccelTransparentHelper_H6;
    td->copyHelper = viaAccelCopyHelper_H6;
    td->solidHelper = viaAccelSolidHelper_H6;
    td->ScalingHelper_H6_410= NULL;
    td->CFCHelper_H6_410= NULL;
    td->Set410Helper_H6= NULL;
    td->Chipid=0;
    if(Chipid == PCI_CHIP_VT3410){
        td->Chipid= PCI_CHIP_VT3410;
        td->ScalingHelper_H6_410= viaAccelScalingHelper_H6_410;
        td->CFCHelper_H6_410= viaAccelCFCHelper_H6_410;
        td->Set410Helper_H6= viaAccelSet410Helper_H6;
    }
}

/*
 * Wait for the value to get blitted, or in the PCI case for engine idle.
 */
void
viaAccelWaitMarker_H6(ScreenPtr pScreen, int marker)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    CARD32 uMarker = marker;

    if (!pVia->IsPCI) {
        while ((pVia->lastMarkerRead - uMarker) > (1 << 24))
            pVia->lastMarkerRead = *(volatile CARD32 *)pVia->markerBuf;
    } else {
        viaAccelSync_H6(pScrn);
    }
}
/*
 * Wait for acceleration engines idle. An expensive way to sync.
 */
void
viaAccelSync_H6(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    int loop = 0;

    mem_barrier();

    switch (pVia->ChipId) {
        case PCI_CHIP_VT3353:
        case PCI_CHIP_VT3409:
        case PCI_CHIP_VT3410:
            while ((VIAGETREG(VIA_REG_STATUS) &
                    (VIA_CMD_RGTR_BUSY_M1 | VIA_2D_ENG_BUSY_M1 | VIA_3D_ENG_BUSY_M1))
                   && (loop++ < MAXLOOP)) ;
            if (loop >= MAXLOOP)
                ErrorF("DEBUG: Stalled LOOP: %d\n", loop);
            break;

		default:
			break;
    }
}

/*
 * Mark Sync using the 2D blitter for AGP. NoOp for PCI.
 * In the future one could even launch a NULL PCI DMA command
 * to have an interrupt generated, provided it is possible to
 * write to the PCI DMA engines from the AGP command stream.
 */
 
int
viaAccelMarkSync_H6(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    RING_VARS;

    ++pVia->curMarker;

    /* Wrap around without affecting the sign bit. */
    pVia->curMarker &= 0x7FFFFFFF;
    if (!pVia->IsPCI) {
        Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
        BEGIN_HEADER0_2D_H5(1);
        OUT_RING_QW(VIA_REG_KEYCONTROL_M1, 0x00);
        if(pVia->ChipId==PCI_CHIP_VT3410){
            viaAccelScalingHelper_H6_410 (0,0,0,0,0,0,0,tdc);
            viaAccelCFCHelper_H6_410(0,0,0,0,0,0,tdc);
            viaAccelSet410Helper_H6(cb,tdc);
        }
        viaAccelSolidHelper_H6(cb, 0, 0, 1, 1, pVia->markerOffset,
                            VIA_GEM_32bpp, 4, pVia->curMarker,
                            (0xF0 << 24) | VIA_GEC_BLT | VIA_GEC_FIXCOLOR_PAT);
        Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
        ADVANCE_RING;
    }   
    return pVia->curMarker;    
}

/*
 * Emit a solid blit operation to the command buffer. 
 */
static void
viaAccelSolidHelper_H6(ViaCommandBuffer * cb, int x, int y, int w, int h,
                    unsigned fbBase, CARD32 mode, unsigned pitch,
                    CARD32 fg, CARD32 cmd)
{
    BEGIN_HEADER0_2D_H5(7);
    OUT_RING_QW(VIA_REG_GEMODE_M1, mode);
    OUT_RING_QW(VIA_REG_DSTBASE_M1, fbBase >> 3);
    OUT_RING_QW(VIA_REG_PITCH_M1, (pitch >> 3) << 16);
    OUT_RING_QW(VIA_REG_DSTPOS_M1, (y << 16) | (x & 0xFFFF));
    OUT_RING_QW(VIA_REG_DIMENSION_M1, ((h - 1) << 16) | (w - 1));
    OUT_RING_QW(VIA_REG_MONOPATFGC_M1, fg);
    OUT_RING_QW(VIA_REG_GECMD_M1, cmd);
}

Bool viaExaPrepareCopy_H6 (PixmapPtr pSrcPixmap, PixmapPtr pDstPixmap, int xdir,
                  int ydir, int alu, Pixel planeMask)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    CARD32 srcpitch, srcoffset, dstpitch, dstoffset;

    RING_VARS;

    if (pSrcPixmap->drawable.bitsPerPixel != pDstPixmap->drawable.bitsPerPixel)
        return FALSE;

    srcpitch = exaGetPixmapPitch(pSrcPixmap);
    if (H6_EXA_PITCH_CHECK(srcpitch)) {
        ErrorF("DEBUG: %s Invalid source pitch: %d\n",
                __func__, (unsigned int) srcpitch);
        return FALSE;
    }

    tdc->srcPitch = srcpitch;

    srcoffset = exaGetPixmapOffset(pSrcPixmap) + pScrn->fbOffset;
    if (H6_EXA_OFFSET_CHECK(srcoffset)) {
        ErrorF("DEBUG: %s Invalid source offset: %d\n",
                __func__, (unsigned int) srcoffset);
        return FALSE;
    }
    tdc->srcOffset = srcoffset;

    dstpitch = exaGetPixmapPitch(pDstPixmap);
    if (H6_EXA_PITCH_CHECK(dstpitch)) {
        ErrorF("DEBUG: %s Invalid destination pitch: %d\n",
                __func__, (unsigned int) dstpitch);
        return FALSE;
    }

    dstoffset = exaGetPixmapOffset(pDstPixmap) + pScrn->fbOffset;
    if (H6_EXA_OFFSET_CHECK(dstoffset)) {
        ErrorF("DEBUG: %s Invalid destination offset: %d\n",
                __func__, (unsigned int) dstoffset);
        return FALSE;
    }

    tdc->cmd = VIA_GEC_BLT | VIAACCELCOPYROP(alu);
    if (xdir < 0)
        if ((pDstPixmap->drawable.x + pDstPixmap->drawable.width) < pSrcPixmap->drawable.x) 
            tdc->cmd |= VIA_GEC_DECX;
        else
            return FALSE;
    if (ydir < 0)
        tdc->cmd |= VIA_GEC_DECY;

    if (!viaAccelSetMode_H6(pDstPixmap->drawable.bitsPerPixel,0,0,0,0,tdc))
        return FALSE;

    if (!viaAccelPlaneMaskHelper_H6(tdc, planeMask))
        return FALSE;

    return TRUE;
}

/*
 * Emit transparency state and color to the command buffer.
 */
static void
viaAccelTransparentHelper_H6 (ViaTwodContext * tdc, ViaCommandBuffer * cb,
                          CARD32 keyControl, CARD32 transColor,
                          Bool usePlaneMask)
{
    tdc->keyControl &= ((usePlaneMask) ? 0xF0000000 : 0x00000000);
    tdc->keyControl |= (keyControl & 0x0FFFFFFF);
    int i = 1;
    if (keyControl)
        i = 2; 

    BEGIN_HEADER0_2D_H5(i);
    OUT_RING_QW(VIA_REG_KEYCONTROL_M1, tdc->keyControl);

    if (keyControl)
        OUT_RING_QW(VIA_REG_SRCCOLORKEY_M1, transColor);
}

void viaExaCopy_H6 (PixmapPtr pDstPixmap, int srcX, int srcY, int dstX, int dstY,
           int width, int height)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    CARD32 srcOffset = tdc->srcOffset;
    CARD32 dstOffset = exaGetPixmapOffset(pDstPixmap) + pScrn->fbOffset;

    RING_VARS;

    if (!width || !height)
        return;

//     ErrorF("DEBUG: Copy srcOffset 0x%7x, dstOffset 0x%7x srcPitch %d, dstPitch %d\n\t\tsrc: %d,%d dst: %d,%d width: %d, height: %d\n", srcOffset, dstOffset, tdc->srcPitch, exaGetPixmapPitch(pDstPixmap), srcX, srcY, dstX, dstY, width, height);

    Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
    viaAccelTransparentHelper_H6(tdc, cb, 0x0, 0x0, TRUE);
    if(pVia->ChipId==PCI_CHIP_VT3410){
        viaAccelScalingHelper_H6_410 (0,0,0,0,0,0,0,tdc);
        viaAccelCFCHelper_H6_410(0,0,0,0,0,0,tdc);
        viaAccelSet410Helper_H6(cb,tdc);
    }
    viaAccelCopyHelper_H6(cb, srcX, srcY, dstX, dstY, width, height,
                       srcOffset, dstOffset, tdc->mode, tdc->srcPitch,
                       exaGetPixmapPitch(pDstPixmap), tdc->cmd);
    Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
    ADVANCE_RING;
}



/*
 * Emit a copy blit operation to the command buffer.
 */
static void
viaAccelCopyHelper_H6 (ViaCommandBuffer * cb, int xs, int ys, int xd, int yd,
                   int w, int h, unsigned srcFbBase, unsigned dstFbBase,
                   CARD32 mode, unsigned srcPitch, unsigned dstPitch,
                   CARD32 cmd)
{
    if (cmd & VIA_GEC_DECY) {
        ys += h - 1;
        yd += h - 1;
    }

    if (cmd & VIA_GEC_DECX) {
        xs += w - 1;
        xd += w - 1;
    }

    BEGIN_HEADER0_2D_H5(8);
    OUT_RING_QW(VIA_REG_GEMODE_M1, mode);
    OUT_RING_QW(VIA_REG_SRCBASE_M1, srcFbBase >> 3);
    OUT_RING_QW(VIA_REG_DSTBASE_M1, dstFbBase >> 3);
    OUT_RING_QW(VIA_REG_PITCH_M1, ((dstPitch >> 3) << 16) | (srcPitch >> 3));
    OUT_RING_QW(VIA_REG_SRCPOS_M1, (ys << 16) | (xs & 0xFFFF));
    OUT_RING_QW(VIA_REG_DSTPOS_M1, (yd << 16) | (xd & 0xFFFF));
    OUT_RING_QW(VIA_REG_DIMENSION_M1, ((h - 1) << 16) | (w - 1));
    OUT_RING_QW(VIA_REG_GECMD_M1, cmd);
	
}

void viaExaDoneSolidCopy_H6(PixmapPtr pPixmap)
{
    // ErrorF("DEBUG: viaExaDoneSolidCopy_H6\n");
}


Bool viaExaPrepareSolid_H6(PixmapPtr pPixmap, int alu, Pixel planeMask, Pixel fg)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;

    RING_VARS;

    if (H6_EXA_PITCH_CHECK(exaGetPixmapPitch(pPixmap))) {
        ErrorF("DEBUG: %s Invalid destination pitch: %d\n",
                __func__, (unsigned int) exaGetPixmapPitch(pPixmap));
        return FALSE;
    }

    if (!viaAccelSetMode_H6(pPixmap->drawable.bitsPerPixel,0,0,0,0, tdc))
        return FALSE;

    if (!viaAccelPlaneMaskHelper_H6(tdc, planeMask))
        return FALSE;

    tdc->cmd = VIA_GEC_BLT | VIA_GEC_FIXCOLOR_PAT | VIAACCELPATTERNROP(alu);

    tdc->fgColor = fg;

    return TRUE;
}

void viaExaSolid_H6(PixmapPtr pPixmap, int x1, int y1, int x2, int y2)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ViaTwodContext *tdc = &pVia->td;
    CARD32 dstPitch, dstOffset;

    RING_VARS;

    int w = x2 - x1, h = y2 - y1;

    dstPitch = exaGetPixmapPitch(pPixmap);
    dstOffset = exaGetPixmapOffset(pPixmap) + pScrn->fbOffset;

    // ErrorF("DEBUG: Solit dstPitch 0x%lx, dstOffset 0x%lx\n, %dx%d",
    //        dstPitch, dstOffset, w, h);

    Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
    viaAccelTransparentHelper_H6(tdc, cb, 0x0, 0x0, TRUE);

    if(pVia->ChipId==PCI_CHIP_VT3410){
        viaAccelScalingHelper_H6_410 (0,0,0,0,0,0,0,tdc);
        viaAccelCFCHelper_H6_410(0,0,0,0,0,0,tdc);
        viaAccelSet410Helper_H6(cb,tdc);
    }
    viaAccelSolidHelper_H6(cb, x1, y1, w, h, dstOffset,
                        tdc->mode, dstPitch, tdc->fgColor, tdc->cmd);
    Command_CRSync_2D3D(&pVia->cb,(unsigned int)pVia->ChipId, FLAG_WAIT_2D_IDLE);
    ADVANCE_RING;
}

/*
 * Update our 2D state (TwoDContext) with a new mode.
 */
static Bool
viaAccelSetMode_H6(int bpp, unsigned format,unsigned rotate, unsigned alphaCnst,
                unsigned alphaVaule, ViaTwodContext *tdc)
{
    switch (bpp) {
        case 16:
            tdc->mode = VIA_GEM_16bpp;
            tdc->bytesPPShift = 1;
            break;
        case 32:
            tdc->mode = VIA_GEM_32bpp;
            tdc->bytesPPShift = 2;
            break;
        case 8:
            tdc->mode = VIA_GEM_8bpp;
            tdc->bytesPPShift = 0;
            break;
/*old tdc->mode*/            
        default:
            /*410 set according to PicFormat*/
            if(format==PICT_x1r5g5b5){
                tdc->mode = VIA_GEM_16bpp;
                tdc->bytesPPShift = 1;
            }else if(format==PICT_r5g6b5){
                tdc->mode = VIA_GEM_16bpp;
                tdc->bytesPPShift = 1;
            }else if((format==PICT_x8r8g8b8)||(format==PICT_a8r8g8b8)){
                tdc->mode = VIA_GEM_32bpp;
                tdc->bytesPPShift = 2;
            }else{
                tdc->bytesPPShift = 0;
                return FALSE;
            }
            break;
/*add fm setting mode*/
    }
    if(tdc->Chipid== PCI_CHIP_VT3410){
        if(alphaCnst){
            tdc->mode |= VIA_GEM_ALPHA_CNT_EN;
            tdc->mode |= VIA_GEM_SETALPHA8(alphaVaule);
            
        }
        switch(rotate){
            case RR_Rotate_90:
                tdc->mode |= VIA_GEM_degree90;
            case RR_Rotate_180:
                tdc->mode |= VIA_GEM_degree180;
            case RR_Rotate_270:
                tdc->mode |= VIA_GEM_degree270;
        }
    }

    return TRUE;
}

Bool
viaAccelScalingHelper_H6_410 (Bool ENscale,unsigned Vfactor, unsigned Hfactor,
        Bool Shrink,unsigned scaleMode, unsigned h, unsigned w,ViaTwodContext *tdc)
{
    if(ENscale){
        tdc->Scale = VIA_SCALEM_En;
        if(Shrink)
            tdc->Scale |= VIA_SCALEM_Shrink;
        tdc->Scale |= scaleMode;
/*scale mode should be :VIA_SCALEM_BW,VIA_SCALEM_WB,*/
/*                      VIA_SCALEM_CC,VIA_SCALEM_CC*/
        tdc->Scale |= (((h - 1) << 16) | (w - 1)| 0x0fffffff);
        tdc->ScaleFactor =0;
        if((Vfactor)||(Hfactor))
            tdc->ScaleFactor = (VIA_SCALEF_VFactor8(Vfactor)|
                                    VIA_SCALEF_HFactor8(Hfactor)|0x0ff00ff0);
    }else{
        tdc->Scale = 0;
        tdc->ScaleFactor =0;
    }
    return TRUE;
}

Bool
viaAccelCFCHelper_H6_410 (Bool enCFC,Bool Extend0, Bool DirthModeRouding,
        unsigned dstCFMode,unsigned srcCFMode, unsigned srcDep,ViaTwodContext *tdc)
{
    if(enCFC){
        tdc->Cfc= VIA_CFC_en;
        if(Extend0)
            tdc->Cfc |= VIA_CFC_EXT_en;  /*extent with 0, else with High Color*/
        if(DirthModeRouding)
            tdc->Cfc |= VIA_CFC_DITH_en; /*Mode with rounding, else with Dit table*/
        tdc->Cfc |= dstCFMode;           /*VIA_CFC_DesCF16/32_555,565,888,000(10,10,10),*/
        tdc->Cfc |= srcCFMode;           /*VIA_CFC_DesCF16/32_555,565,888,000(10,10,10),*/
        tdc->Cfc |= srcDep;              /*VIA_CFC_SrcDep8/16/32,*/
    }else{
        tdc->Cfc = 0;
    }
    return TRUE;
}

/*Call before Solid /CopyHelper ,to set CFC Scaling etc.410 issue*/
Bool
viaAccelSet410Helper_H6(ViaCommandBuffer * cb,ViaTwodContext *tdc )
{
    BEGIN_HEADER0_2D_H5(3);
    OUT_RING_QW(VIA_REG_CFC, tdc->Cfc);
    OUT_RING_QW(VIA_REG_SCALEFACTOR, tdc->ScaleFactor);
    OUT_RING_QW(VIA_REG_SCALINGMODE, tdc->Scale);
}

/*
 * Check if we can use a planeMask and update the 2D context accordingly.
 */
static Bool
viaAccelPlaneMaskHelper_H6(ViaTwodContext * tdc, CARD32 planeMask)
{
    CARD32 modeMask = (1 << ((1 << tdc->bytesPPShift) << 3)) - 1;
    CARD32 curMask = 0x00000000;
    CARD32 curByteMask;
    int i;

    if ((planeMask & modeMask) != modeMask) {

        /* Masking doesn't work in 8bpp. */
        if (modeMask == 0xFF) {
            tdc->keyControl &= 0x0FFFFFFF;
            return FALSE;
        }

        /* Translate the bit planemask to a byte planemask. */
        for (i = 0; i < (1 << tdc->bytesPPShift); ++i) {
            curByteMask = (0xFF << (i << 3));

            if ((planeMask & curByteMask) == 0) {
                curMask |= (1 << i);
            } else if ((planeMask & curByteMask) != curByteMask) {
                tdc->keyControl &= 0x0FFFFFFF;
                return FALSE;
            }
        }
        // ErrorF("DEBUG: planeMask 0x%08x, curMask 0%02x\n",
        //       (unsigned)planeMask, (unsigned)curMask);

        tdc->keyControl = (tdc->keyControl & 0x0FFFFFFF) | (curMask << 28);
    }

    return TRUE;
}
