/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include "via_driver.h"
#include "via_dvi.h"
#include "via_hdmi.h"

#include "via_common.h"
#include "via_serial.h"
#include "via_output.h"
#include "via_dp.h"

extern Bool ad9389_module_loaded;

static void 
viaInitHDMIDIPort(VIABIOSInfoPtr pBIOSInfo)
{
    /* 410 Integrated HDMI doesn't have DIPort */
    if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_INTEGRATED_HDMI)
        return;
	
    if (pBIOSInfo->HDMISettingInfo.HDMIDIPort != VIA_DI_NONE) {
        /* Do nothing, because DFP DIPort can be assigned by option. */
        return;
    }
        
    switch (pBIOSInfo->Chipset) {
        case VIA_CX700:
        case VIA_VX800:
        case VIA_VX855:
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DVP1;            
            break;
            
        case VIA_K8M890:
        case VIA_P4M900:
        case VIA_P4M890:
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DFPLOW;            
            break;

        default:
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DVP1;                       
            break;
    }
}

CARD32
viaGetIntegratedHDMIEEDIDBlock(VIABIOSInfoPtr pBIOSInfo, CARD8* pEDIDData)
{
    CARD8   EDIDHeader[3];
    CARD8   ExtensionFlag;  /* Extension Number */
    CARD8   ExtensionTag;   /* Extension Type */
    CARD8   RevisionNum;
    CARD8   i = 0;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "viaGetIntegratedHDMIEEDIDBlock\n"));
    
    /* First Read EDID Block 0 */
    for (i = 0; i < 3; i++) {    
        viaSerialReadByteByHDCP(0xA0, i, EDIDHeader+i);
    }

    /* EDID Header is 00 FF FF FF FF FF FF 00 */
    if ((EDIDHeader[0] == 0) && (EDIDHeader[1] == 0xFF) && 
        (EDIDHeader[2] == 0xFF)) {

        viaSerialReadByteByHDCP(0xA0, 0x7E, &ExtensionFlag);

        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
              "EDID ExtensionFlag = %d\n", ExtensionFlag));

        if (ExtensionFlag == 0) {
        
            /* There is no extension block. */
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
                  "No extension block\n"));
            
            return EDID_EXIST;
        } else if (ExtensionFlag == 1) {
            DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
                  "Only one extension block\n"));
        
            viaSerialReadByteByHDCP(0xA0, 128, &ExtensionTag); 
        
            /* 0x02 is timing extension block */
            if (ExtensionTag == 0x02) {
                viaSerialReadByteByHDCP(0xA0, 129, &RevisionNum);
                /* Currently, we only support rev.3 */
                if (RevisionNum == 0x03) {                
                    /* This have determined as HDMI EDID. */
                    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
                          "Found CEA EDID\n"));
        
                    /* Read Block 2 Raw Data. */
                    viaSerialReadBytesByHDCP(0xA0, 128, pEDIDData, 128);
                }
            }
        }
        return EDID_EXTENSION_EXIST;
    }
    return EDID_NOT_EXIST;;
}

/* This function will determine the conntected device and panel size. */
Bool 
viaGetHDMIInfo(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    CARD32 DescriptorNum = 0;
    CARD8  i = 0;
    CARD8  HDMIModeIndex;
    int    TmpHSize = 0;
    int    TmpVSize = 0;
    int    MaxHSize = 0;
    int    MaxVSize = 0;
    CARD8  CEAEdidData[128];
    CARD8  EDIDData[128];
    CARD8  TmpValue = 0;
    CARD8  VendorBlockIndex = 0;      
    CARD8  VideoBlockIndex = 0;      
    CARD32 ret = 0;
    xf86MonPtr          edid_mon;      

    /* Initial */
    for (i = 0; i < 128; i++) {
        CEAEdidData[i] = 0;
        EDIDData[i] = 0;
    }

    pHDMISettingInfo->SinkDevType = 0x0;
	
    /* Check if extend EDID exist. */
    if (pHDMISettingInfo->ChipID == VIA_INTEGRATED_HDMI)
        ret = viaGetIntegratedHDMIEEDIDBlock(pBIOSInfo, CEAEdidData);
    else
        ret = viaGetAD9389EEDIDBlock(pBIOSInfo, CEAEdidData);
    
    if (ret == EDID_EXIST) {
        pHDMISettingInfo->SinkDevType = CONNECT_DVI;

        /* Read EDID 1.3, the following code shoud be refined. */
        if (pHDMISettingInfo->ChipID == VIA_INTEGRATED_HDMI) {
            viaSerialReadBytesByHDCP(0xA0, 0, EDIDData, 128);  
        } else {
            for (i = 0; i < 128; i++) {
                readEDID_AD9389(pBIOSInfo, i, EDIDData+i);
            }
        }
        /* Parse the EDID block */
        edid_mon = xf86InterpretEDID(pBIOSInfo->scrnIndex, EDIDData);
        if (edid_mon) {             
            /*Find max resolution from Established timing*/
            VIAFindMaxResFromEstablishedTiming(edid_mon, &MaxHSize, &MaxVSize);

            /*Find max resolution from Standard timing*/
            for (i = 0; i < 8; i++) {
                if (edid_mon->timings2[i].hsize > MaxHSize) {
                    MaxHSize = edid_mon->timings2[i].hsize;
                    MaxVSize = edid_mon->timings2[i].vsize;
                }            
            }

            /*Find max resolution from Detail timing*/
            for (i = 0; i < 4; i++) {
                if ((edid_mon->det_mon[i].type == DT) && 
                    (edid_mon->det_mon[i].section.d_timings.h_active > MaxHSize)) {
                    
                    MaxHSize = edid_mon->det_mon[i].section.d_timings.h_active;
                    MaxVSize = edid_mon->det_mon[i].section.d_timings.v_active;

                    if (edid_mon->det_mon[i].section.d_timings.interlaced)                
                        MaxVSize = MaxVSize*2;                
                }       
            }
            
            /* Correct monitor size to avoid invalid mode. */
            VIACorrectMonitorSize(&MaxHSize, &MaxVSize);            

            /* To prevent to get a wrong EDID data. We should assign 
               a default value.*/
            if ((MaxHSize == 0) || (MaxVSize == 0)) {
                MaxHSize = 1280;
                MaxVSize = 720;
            }
            
            pHDMISettingInfo->NativeSizeID =
                VIAGetModeIndex(MaxHSize, MaxVSize); 
        }
        
        return TRUE;
    } else if (ret == EDID_EXTENSION_EXIST) {        
        /* Check Data Block amount. If no data block exists, return DVI connect. 
           Byte 2: byte number offset d where detail timing data begin. */
        /* if 0, No Data Block, No Detailed Timing Descriptor.*/
        /* if 4, No Data Block.*/
        if(!(CEAEdidData[2] == 0 ||    
             CEAEdidData[2] == 4)) {
             
            /* Byte 04 is where the Data Block Collection begins */
            VendorBlockIndex = 4; 
            TmpValue = CEAEdidData[VendorBlockIndex]; 
            
            /* Locate Vendor Description Block to identify if connect to HDMI.*/
            /* Bit[7:5]: Block Type Tag: value 3 is vendor specific*/
            while (((TmpValue & 0xE0) != 0x60) 
                 && (VendorBlockIndex < (CEAEdidData[2]))) {
                 
                /* Bit[4:0]: Total number of bytes in this block following 
                   this byte */
                TmpValue &= 0x1F;  

                /* The next byte is assumed to be the beginning of the next 
                   data block. */
                VendorBlockIndex += (TmpValue+1); 

                /* Locate the next data block */
                TmpValue = CEAEdidData[VendorBlockIndex]; 
            }
            
            if((TmpValue & 0xE0) == 0x60) {
                /* Read 24-bit IEEE registration.*/            
                /* HDMI identifier = 0x000C03.*/
                if( (CEAEdidData[VendorBlockIndex+1] == 0x03)
                 && (CEAEdidData[VendorBlockIndex+2] == 0x0C)
                 && (CEAEdidData[VendorBlockIndex+3] == 0x00)) {                        
                    pHDMISettingInfo->SinkDevType =  CONNECT_HDMI;
                } else {
                    pHDMISettingInfo->SinkDevType = CONNECT_DVI;  
                }
            }

            if (pHDMISettingInfo->SinkDevType ==  CONNECT_HDMI) {                
                /* Byte 04 is where the Data Block Collection begins */
                TmpValue = CEAEdidData[4]; 

                /* Locate Video Block */                
                /* Bit[7:5]: Block Type Tag: value 2 is video specific*/
                while (((TmpValue & 0xE0) != 0x40) 
                     && (VideoBlockIndex < (CEAEdidData[2]))) {
                    /* Bit[4:0]: 
                    Total number of bytes in this block following this byte */
                    TmpValue &= 0x1F;  

                    /* The next byte is assumed to be the beginning of the 
                       next data block. */
                    VideoBlockIndex += (TmpValue+1); 

                    /* Locate the next data block */
                    TmpValue = CEAEdidData[VideoBlockIndex]; 
                }
                
                if ((TmpValue & 0xE0) == 0x40) {
                    DescriptorNum = TmpValue & 0x1F;
                    
                    for(i = 0; i< DescriptorNum; i++) {
                        TmpValue = CEAEdidData[VideoBlockIndex]+i;
                        HDMIModeIndex = TmpValue & 0x7F;

                        if(HDMIModeIndex <= 0) {                  
                            HDMIModeIndex = 1;
                        }
                                            
                        TmpHSize = CEATimingTable[HDMIModeIndex-1].HorSize;
                        TmpVSize = CEATimingTable[HDMIModeIndex-1].VerSize;            

                        if (TmpHSize*TmpVSize > MaxHSize*MaxVSize) {
                            MaxHSize = TmpHSize; 
                            MaxVSize = TmpVSize;
                        }
                    }        

                    /* Map the flat panel size to mode index. */
                    pHDMISettingInfo->NativeSizeID =
                        VIAGetModeIndex(MaxHSize, MaxVSize);    
                    
                    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, 
                        "HDMI Max Support HSize = %d\n", MaxHSize)); 
                    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED,
                        "HDMI Max Support VSize = %d\n", MaxVSize));                
                }
            }

            return TRUE;
        }
    } 
    return FALSE;
}

int
viaIdentify_IntegratedHDMI(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    int ret= NOT_FOUND_DEV;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "viaIdentify_IntegratedHDMI\n"));

    if (pBIOSInfo->Chipset == VIA_VX900) {	
        pHDMISettingInfo->ChipID = VIA_INTEGRATED_HDMI;
        /* Integrated HDMI have no I2C slave address */
        pHDMISettingInfo->I2CAddress = 0x0;
        pHDMISettingInfo->HDMIDIPort = VIA_DI_NONE;
        // Enable bangap
        MMIO_WR_MASK(0xC740, 0x1, 0x1);
        viaGetHDMIInfo(pBIOSInfo, pHDMISettingInfo);
        ret = FOUND_INTERNAL_DEV;
    }    

    return ret;
}

Bool 
viaIdentifyHDMI(VIABIOSInfoPtr pBIOSInfo)
{
    int             IdentifyState = NOT_FOUND_DEV;
    unsigned char   sr2a = 0, sr1e = 0, sr3e = 0;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "viaIdentifyHDMI\n"));

    switch (pBIOSInfo->Chipset) {
        case VIA_K8M890:
            /* DFP Low Pad on */
            sr2a = read_reg(VIASR, SR2A);
            write_reg_mask(SR2A, VIASR, 0x03, BIT0+BIT1);
            break;

        case VIA_P4M900:
        case VIA_P4M890:
            /* DFP Low Pad on */
            sr2a = read_reg(VIASR, SR2A);
            write_reg_mask(SR2A, VIASR, 0x03, BIT0+BIT1);

            /* VT3364 also supports DVI on AMR, so we need to sense DVP0, too.*/
            /* DVP0 Pad on */
            sr1e = read_reg(VIASR, SR1E);
            write_reg_mask(SR1E, VIASR, 0xC0, BIT6+BIT7);
            break;

        default:
            /* DVP0/DVP1 Pad on */
            sr1e = read_reg(VIASR, SR1E);
            write_reg_mask(SR1E, VIASR, 0xF0, BIT4+BIT5+BIT6+BIT7);
            /* SR3E[1]Multi-function selection:
               0 = Emulate I2C and DDC bus by GPIO2/3/4. */
            sr3e = read_reg(VIASR, SR3E);
            write_reg_mask(SR3E, VIASR, 0x0, BIT5);
            break;
    }
        
    if (ad9389_module_loaded)    
        IdentifyState = viaIdentify_AD9389(pBIOSInfo);

    if (!pBIOSInfo->HDMISettingInfo.ChipID)
        IdentifyState |= viaIdentify_IntegratedHDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo));
    else
        IdentifyState |= viaIdentify_IntegratedHDMI(pBIOSInfo, &(pBIOSInfo->HDMISettingInfo2));
    
    /* Restore status */
    switch (pBIOSInfo->Chipset) {
        case VIA_K8M890:
            write_reg(SR2A, VIASR, sr2a);
            break;

        case VIA_P4M900:
        case VIA_P4M890:
            write_reg(SR2A, VIASR, sr2a);
            write_reg(SR1E, VIASR, sr1e);
            break;

        default: 
            write_reg(SR1E, VIASR, sr1e);
            write_reg(SR3E, VIASR, sr3e);
            break;
    }

    if (IdentifyState) {
        viaInitHDMIDIPort(pBIOSInfo);
        return TRUE;  
    } else {        
        return FALSE;
    }
    
}


void 
viaInitHDMI(VIABIOSInfoPtr pBIOSInfo)
{
    unsigned char   sr2a = 0, sr1e = 0, sr3e = 0;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "viaInitHDMI\n"));        

    sr1e = read_reg(VIASR, SR1E);
    sr2a = read_reg(VIASR, SR2A);
    
    VIADIPortPadOn(pBIOSInfo->HDMISettingInfo.HDMIDIPort);
    
    /* SR3E[1]Multi-function selection:
       0 = Emulate I2C and DDC bus by GPIO2/3/4. */
    sr3e = read_reg(VIASR, SR3E);
    write_reg_mask(SR3E, VIASR, 0x0, BIT5);
    
    if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
        viaInitReg_AD9389(pBIOSInfo);        
    }

    /* Restore status */
    write_reg(SR1E, VIASR, sr1e);
    write_reg(SR2A, VIASR, sr2a);
    write_reg(SR3E, VIASR, sr3e);
}


static void 
viaPatchHDMISkew_DVP1(VIABIOSInfoPtr pBIOSInfo)
{
    switch (pBIOSInfo->Chipset) {
        case VIA_CX700:        
            if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
                write_reg_mask(SR65, VIASR, 0x0B, 0x0F);

                if (pBIOSInfo->HDMISettingInfo.ModeIndex == M1280x720)
                    write_reg_mask(CR9B, VIACR, 0x04, 0x0F);
                else
                    write_reg_mask(CR9B, VIACR, 0x02, 0x0F);
            }
            break;        

        case VIA_VX800:        
            if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
                write_reg_mask(SR65, VIASR, 0x01, 0x0F);
                write_reg_mask(CR9B, VIACR, 0x0C, 0x0F);
            }   
            break;
			
        case VIA_VX855:
            if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
                write_reg_mask(SR65, VIASR, 0x0D, 0x0F);
                write_reg_mask(CR9B, VIACR, 0x0F, 0x0F);
            }   
            break;
			
        case VIA_VX900:
            if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
                write_reg_mask(SR65, VIASR, 0x0D, 0x0F);
                write_reg_mask(CR9B, VIACR, 0x0F, 0x0F);
            }   
            break;   
    }
}


static void 
viaPatchCEADotClock(VIABIOSInfoPtr pBIOSInfo, int dotClock, int igaPath)
{
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "viaPatchCEADotClock\n"));        

    int i;
    int  patchPLL= 0;
    switch (pBIOSInfo->Chipset) {        
        case VIA_VX855:
        case VIA_VX900:    
            for (i = 0; ViaDotClocks[i].DotClock; i++) {
                if (ViaDotClocks[i].DotClock == dotClock) {
                    patchPLL = ViaDotClocks[i].PLL3_409;
                    break;
                }
            }                
            break;
    
        default:
            for (i = 0; ViaDotClocks[i].DotClock; i++) {
                if (ViaDotClocks[i].DotClock == dotClock) {
                    patchPLL = ViaDotClocks[i].PLL3;
                    break;
                }
            }
            break;
    }
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
          "patchPLL = 0x%x\n", patchPLL));        
    SetVCLK(pBIOSInfo, patchPLL, igaPath);

    /* 135MHz ~ 270MHz */
    if (dotClock*1000 >= CLK_135_000M)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x00000000, 0xC0000000);           
    /* 67.5MHz ~ <135MHz */
    else if (dotClock*1000 >= CLK_67_500M)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x40000000, 0xC0000000);
    /* 33.75MHz ~ <67.5MHz */
    else if (dotClock*1000 >= CLK_33_750M)
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0x80000000, 0xC0000000);
    /* 25MHz ~ <33.75MHz */
    else
        MMIO_WR_MASK(DP_EPHY_PLL_REG, 0xC0000000, 0xC0000000);     
}


void 
viaSetHDMIOutputPath(VIABIOSInfoPtr pBIOSInfo, int set_iga, 
                                int output_interface)
{
    switch (output_interface) {
        case VIA_DI_DVP0:
            write_reg_mask(CR6B, VIACR, 0x01, BIT0);

            if (set_iga == IGA1)
                write_reg_mask(CR96, VIACR, 0x00, BIT4);
            else
                write_reg_mask(CR96, VIACR, 0x10, BIT4);
                 
            write_reg_mask(SR1E, VIASR, 0xC0, BIT7+BIT6);
            break;

        case VIA_DI_DVP1:
            if (set_iga== IGA1)
                write_reg_mask(CR9B, VIACR, 0x00, BIT4);
            else
                write_reg_mask(CR9B, VIACR, 0x10, BIT4);

            write_reg_mask(SR1E, VIASR, 0x30, BIT4+BIT5);
            break;
    }
}

void
viaAdjustTimingForHDMI(VIABIOSInfoPtr pBIOSInfo)
{
    if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389) {
        viaAdjustTimingForAD9389(pBIOSInfo);
    }
    
    /* Set Hsync Offset, delay one clock (To meet 861-D spec.) */
    write_reg_mask(CR8A, VIACR, 0x1, 0x7);

    /* If CR8A +1, HSyc must -1 */
    write_reg(CR56, VIACR, read_reg(VIACR, CR56)-1);
    write_reg(CR57, VIACR, read_reg(VIACR, CR57)-1);

}

// calcute the BCH code and output 1 byte BCH code
CARD16 InsertBCHCode(CARD16* input, CARD32 length)
{
    CARD32   i, n;
    CARD32   reg[9]={0, 0, 0, 0, 0, 0, 0, 0, 0};
    CARD32   matrix[56];
    CARD16   output = 0;

    // initialize matrix
    for(i = 0; i < length; i++)
    {
        matrix[0 + i*8] = (input[i] >> 0) & BIT0;
        matrix[1 + i*8] = (input[i] >> 1) & BIT0;
        matrix[2 + i*8] = (input[i] >> 2) & BIT0;
        matrix[3 + i*8] = (input[i] >> 3) & BIT0;
        matrix[4 + i*8] = (input[i] >> 4) & BIT0;
        matrix[5 + i*8] = (input[i] >> 5) & BIT0;
        matrix[6 + i*8] = (input[i] >> 6) & BIT0;
        matrix[7 + i*8] = (input[i] >> 7) & BIT0;
    }

    for(i = 0; i < length * 8; i++)
    {
        reg[0] = reg[8] ^ matrix[i];
        reg[8] = reg[7] ^ reg[0];
        reg[7] = reg[6] ^ reg[0];

        reg[6]=reg[5];
        reg[5]=reg[4];
        reg[4]=reg[3];
        reg[3]=reg[2];
        reg[2]=reg[1];
        reg[1]=reg[0];
    }

    n = 0;
    for(i = 1; i <= 4; i++)
    {
        if(reg[i] == 1)
            n = 2*n + 1;
        else 
            n = 2*n;
    }

    output |= n << 4;

    n = 0;
    for(i = 5; i <= 8; i++)
    {
        if(reg[i] == 1)
            n = 2*n + 1;
        else 
            n = 2*n;
    }

    output |= n;

    return output;
}

void 
SetAVIInfoFrame(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    HDMI_InfoFrameRec  InfoFramePacket;
    CARD32   i = 0;
    CARD16   checksum = 0;
    CARD32   mmC204, mmC208, mmC20C, mmC210, mmC214, mmC218, mmC21C, mmC220, mmC224;
    
    InfoFramePacket.PacketHeader[0] = 0x82;
    InfoFramePacket.PacketHeader[1] = 0x02;
    InfoFramePacket.PacketHeader[2] = 0x0D;
    InfoFramePacket.ECCCode = InsertBCHCode(InfoFramePacket.PacketHeader, 3);

    // set Packet Bytes
    for(i = 0; i < 28; i++)
        InfoFramePacket.PacketByte[i] = 0;
 
    // S1  S0   |   Scan Information
    // -----------------------------
    //  0   1   |   Overscan
    //  1   0   |   Underscan
    InfoFramePacket.PacketByte[1] |= 0x2;

    // R3  R2  R1  R0   |   Active Format Aspect Ratio
    // -----------------------------------------------
    //  1   0   0   0   |   Same as picture aspect ratio
    //  1   0   0   1   |   4:3
    //  1   0   1   0   |   16:9
    //  1   0   1   1   |   14:9
    InfoFramePacket.PacketByte[2] |= 0x08;

    // M1  M0   |   Picture Aspect Ratio
    // ---------------------------------
    //  0   1   |    4:3
    //  1   0   |   16:9
    switch(pHDMISettingInfo->ModeIndex)
    {
        case M640x480:
        case M720x480:
        case M720x576:
            InfoFramePacket.PacketByte[2] |= 0x10;
            break;
        case M1280x720:
        case M1920x1080:
        case M1920x1080_I:
            InfoFramePacket.PacketByte[2] |= 0x20;
            break;
        default:
            break;
    }

    // ITC bit
    InfoFramePacket.PacketByte[3] |= 0x80;

    // VIC code
    switch(pHDMISettingInfo->ModeIndex)
    {
        case M640x480:
            InfoFramePacket.PacketByte[4] = 1;
            break;

        case M720x480:
            InfoFramePacket.PacketByte[4] = 2;
            break;

        case M720x576:
            InfoFramePacket.PacketByte[4] = 17;
            break;

        case M1280x720:
            if (pBIOSInfo->FoundRefresh == 50)
                InfoFramePacket.PacketByte[4] = 19;
            else
                InfoFramePacket.PacketByte[4] = 4;
            break;

        case M1920x1080:
            InfoFramePacket.PacketByte[4] = 16;
            break;

        case M1920x1080_I:
            InfoFramePacket.PacketByte[4] = 5;
            break;

        default:
            break;
    }

    // count checksum
    for(i = 0; i < 28; i++)
        checksum += InfoFramePacket.PacketByte[i];
    checksum += ( 0x82 + 0x02 + 0x0D );
    
    checksum = 0x100 - checksum;
    InfoFramePacket.PacketByte[0] = checksum;

    // set BCHCode
    for(i = 0; i < 4; i++)
    {
        InfoFramePacket.BCHCode[i] = InsertBCHCode( &(InfoFramePacket.PacketByte[i*7]), 7);
    }

    mmC204 = (InfoFramePacket.ECCCode << 24 ) | (InfoFramePacket.PacketHeader[2] << 16) | 
             (InfoFramePacket.PacketHeader[1] << 8) | InfoFramePacket.PacketHeader[0];
    mmC208 = (InfoFramePacket.PacketByte[3] << 24) | (InfoFramePacket.PacketByte[2] << 16) |
             (InfoFramePacket.PacketByte[1] << 8) | InfoFramePacket.PacketByte[0];
    mmC20C = (InfoFramePacket.BCHCode[0] << 24) | (InfoFramePacket.PacketByte[6] << 16) |
             (InfoFramePacket.PacketByte[5] << 8) | InfoFramePacket.PacketByte[4];
    mmC210 = (InfoFramePacket.PacketByte[10] << 24) | (InfoFramePacket.PacketByte[9] << 16) |
             (InfoFramePacket.PacketByte[8] << 8) | InfoFramePacket.PacketByte[7];
    mmC214 = (InfoFramePacket.BCHCode[1] << 24) | (InfoFramePacket.PacketByte[13] << 16) |
             (InfoFramePacket.PacketByte[12] << 8) | InfoFramePacket.PacketByte[11];
    mmC218 = (InfoFramePacket.PacketByte[17] << 24) | (InfoFramePacket.PacketByte[16] << 16) |
             (InfoFramePacket.PacketByte[15] << 8) | InfoFramePacket.PacketByte[14];
    mmC21C = (InfoFramePacket.BCHCode[2] << 24) | (InfoFramePacket.PacketByte[20] << 16) |
             (InfoFramePacket.PacketByte[19] << 8) | InfoFramePacket.PacketByte[18];
    mmC220 = (InfoFramePacket.PacketByte[24] << 24) | (InfoFramePacket.PacketByte[23] << 16) |
             (InfoFramePacket.PacketByte[22] << 8) | InfoFramePacket.PacketByte[21];
    mmC224 = (InfoFramePacket.BCHCode[3] << 24) | (InfoFramePacket.PacketByte[27] << 16) |
             (InfoFramePacket.PacketByte[26] << 8) | InfoFramePacket.PacketByte[25];

    MMIO_WR(0xC204, mmC204);
    MMIO_WR(0xC208, mmC208);
    MMIO_WR(0xC20C, mmC20C);
    MMIO_WR(0xC210, mmC210);
    MMIO_WR(0xC214, mmC214);
    MMIO_WR(0xC218, mmC218);
    MMIO_WR(0xC21C, mmC21C);
    MMIO_WR(0xC220, mmC220);
    MMIO_WR(0xC224, mmC224);

    MMIO_WR(0xC200, 0x80);  
    
}

void 
SetAudioInfoFrame()
{
    HDMI_InfoFrameRec  InfoFramePacket;
    CARD32   i = 0;
    CARD16   checksum = 0;
    CARD32   mmC204, mmC208, mmC20C, mmC210, mmC214, mmC218, mmC21C, mmC220, mmC224;

    InfoFramePacket.PacketHeader[0] = 0x84;
    InfoFramePacket.PacketHeader[1] = 0x01;
    InfoFramePacket.PacketHeader[2] = 0x0A;
    InfoFramePacket.ECCCode = InsertBCHCode(InfoFramePacket.PacketHeader, 3);

    // set Packet Bytes
    for(i = 0; i < 28; i++)
        InfoFramePacket.PacketByte[i] = 0;
 
    // CC2 CC1 CC0  |   Audio Channel Count
    // ---------------------------------------
    //  0   0   0   |   Refer to Stream Header
    //  0   0   1   |   2ch
    //  0   1   0   |   3ch
    //  0   1   1   |   4ch
    InfoFramePacket.PacketByte[1] |= 0x0;

    // CT3 CT2 CT1 CT0  |   Audio Coding Type
    // -------------------------------------------
    //  0   0   0   0   |   Refer to Stream Header
    //  0   0   0   1   |   IEC 60958 PCM[30,31]
    //  0   0   1   0   |   AC-3
    //  0   0   1   1   |   MPEG1 (Layer 1 & 2)
    InfoFramePacket.PacketByte[1] |= 0x00;

    // SS1 SS0  |   Sample Size
    // --------------------------
    //  0   0   |   Refer to Stream Header
    //  0   1   |   16 bit
    //  1   0   |   20 bit
    //  1   1   |   24 bit
    InfoFramePacket.PacketByte[2] |= 0x00;

    // SF2 SF1 SF0  |   Sampling Frequency
    // ---------------------------------------
    //  0   0   0   |   Refer to Stream Header
    //  0   0   1   |   32 kHz
    //  0   1   0   |   44.1 kHz(CD)
    //  0   1   1   |   48 kHz
    //  1   0   0   |   88.2 kHz
    //  1   0   1   |   96 kHz
    //  1   1   0   |   176.4 kHz
    //  1   1   1   |   192 kHz
    InfoFramePacket.PacketByte[2] |= 0x00;

    // count checksum
    for(i = 0; i < 28; i++)
        checksum += InfoFramePacket.PacketByte[i];
    checksum += ( 0x84 + 0x01 + 0x0A );
    
    checksum = 0x100 - checksum;
    InfoFramePacket.PacketByte[0] = checksum;

    // set BCHCode
    for(i = 0; i < 4; i++)
    {
        InfoFramePacket.BCHCode[i] = InsertBCHCode( &(InfoFramePacket.PacketByte[i*7]), 7);
    }

    mmC204 = (InfoFramePacket.ECCCode << 24 ) | (InfoFramePacket.PacketHeader[2] << 16) | 
             (InfoFramePacket.PacketHeader[1] << 8) | InfoFramePacket.PacketHeader[0];
    mmC208 = (InfoFramePacket.PacketByte[3] << 24) | (InfoFramePacket.PacketByte[2] << 16) |
             (InfoFramePacket.PacketByte[1] << 8) | InfoFramePacket.PacketByte[0];
    mmC20C = (InfoFramePacket.BCHCode[0] << 24) | (InfoFramePacket.PacketByte[6] << 16) |
             (InfoFramePacket.PacketByte[5] << 8) | InfoFramePacket.PacketByte[4];
    mmC210 = (InfoFramePacket.PacketByte[10] << 24) | (InfoFramePacket.PacketByte[9] << 16) |
             (InfoFramePacket.PacketByte[8] << 8) | InfoFramePacket.PacketByte[7];
    mmC214 = (InfoFramePacket.BCHCode[1] << 24) | (InfoFramePacket.PacketByte[13] << 16) |
             (InfoFramePacket.PacketByte[12] << 8) | InfoFramePacket.PacketByte[11];
    mmC218 = (InfoFramePacket.PacketByte[17] << 24) | (InfoFramePacket.PacketByte[16] << 16) |
             (InfoFramePacket.PacketByte[15] << 8) | InfoFramePacket.PacketByte[14];
    mmC21C = (InfoFramePacket.BCHCode[2] << 24) | (InfoFramePacket.PacketByte[20] << 16) |
             (InfoFramePacket.PacketByte[19] << 8) | InfoFramePacket.PacketByte[18];
    mmC220 = (InfoFramePacket.PacketByte[24] << 24) | (InfoFramePacket.PacketByte[23] << 16) |
             (InfoFramePacket.PacketByte[22] << 8) | InfoFramePacket.PacketByte[21];
    mmC224 = (InfoFramePacket.BCHCode[3] << 24) | (InfoFramePacket.PacketByte[27] << 16) |
             (InfoFramePacket.PacketByte[26] << 8) | InfoFramePacket.PacketByte[25];

    MMIO_WR(0xC204, mmC204);
    MMIO_WR(0xC208, mmC208);
    MMIO_WR(0xC20C, mmC20C);
    MMIO_WR(0xC210, mmC210);
    MMIO_WR(0xC214, mmC214);
    MMIO_WR(0xC218, mmC218);
    MMIO_WR(0xC21C, mmC21C);
    MMIO_WR(0xC220, mmC220);
    MMIO_WR(0xC224, mmC224);

    MMIO_WR(0xC200, 0x81);  
    
}

Bool 
SetHDAudioSampleRate( DWORD dwSampleRate, DWORD dwPixelClock)
{
    CARD32 i = 0, mmC400, mmC404, mmC410, mmC414;

    mmC404 = MMIO_RD(0xC404) & 0xFFFFFF00;
    mmC414 = MMIO_RD(0xC414) & 0xFFFFFF00;

    MMIO_WR_MASK(0xC404, 0, 0xF); // Clear Wall Clock Ratio
    MMIO_WR_MASK(0xC400, 0, 0xFFFFFFFF);
    
    // Set Wall Clock Ratio: (24MHz / HDMI CLK frequency) * 2^40
    for(i = 0; i < WallClockReg_NUM; i++)
    {
        if(WallClockReg[i].PixelClock == dwPixelClock)
        {
            mmC400 = WallClockReg[i].WallClockRatio_L;
            mmC404 |= WallClockReg[i].WallClockRateo_H;
            break;
        }
    }
    MMIO_WR(0xC404, mmC404);
    MMIO_WR(0xC400, mmC400);

    MMIO_WR_MASK(0xC414, 0, 0xF); // Clear audio sample rate
    MMIO_WR_MASK(0xC410, 0, 0xFFFFFFFF);
    // Set audio sample rate: (Sample Rate / HDMI CLK frequency) * 2^40
    for(i = 0; i < PacketClockReg_NUM; i++)
    {
        if(PacketClockReg[i].SampleRate == dwSampleRate && PacketClockReg[i].PixelClock == dwPixelClock)
        {
            mmC410 = PacketClockReg[i].PacketClockRatio_L;
            mmC414 |= PacketClockReg[i].PacketClockRateo_H;
            break;
        }
    }

    MMIO_WR(0xC414, mmC414);
    MMIO_WR(0xC410, mmC410);

    return TRUE;
}

Bool 
SetHDAudioCTSandN(DWORD dwSampleRate, DWORD dwPixelClock)
{
    CARD32   i = 0, CTS = 0, N = 0, mmC294 = 0, mmC298 = 0, mmC29C = 0;

    for(i = 0; i < CTSandNReg_NUM; i++)
    {
        if(CTSandNReg[i].SampleRate == dwSampleRate && CTSandNReg[i].PixelClock == dwPixelClock)
        {
            CTS = CTSandNReg[i].CTS;
            N = CTSandNReg[i].N;
            break;
        }
    }

    mmC294 = 0x100000C0 | (CTS << 8);
    MMIO_WR(0xC294, mmC294);

    mmC298 = N | (( CTS & 0xFFF ) << 20);
    MMIO_WR(0xC298, mmC298);

    mmC29C = MMIO_RD(0xC29C) & 0xFFFFFF00;
    mmC29C |= ((CTS >> 12) & 0xFF);
    MMIO_WR(0xC29C, mmC29C);

    return TRUE;
}

void 
SetHDAudioParameter(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    CARD32   mmC4D4, mmC4D8, mmC424, dwSampleRate, dwPixelClock;

    dwPixelClock = pHDMISettingInfo->PixelClock * 1000;

    // step 1. read out C4D8 
    mmC4D4 = MMIO_RD(0xC4D4) & 0x00FFFFFF;
    mmC4D4 |=  0x01000000;      // stream format
    MMIO_WR(0xC4D4, mmC4D4);

    mmC4D8 = MMIO_RD(0xC4D8); // [15:0]

    //step 2. decode C4D8 by definition of C45C
    // C45C [14] Converter_Format_Base
    // BASE. Converter Format Sample Base Rate. (PCM Format structure bit 14)
    //              0: 48kHz 
    //              1: 44.1kHz
    // C45C [13:11] Converter_Format_Mult
    // MULT. Converter Format Sample Base Rate Multiple. (PCM Format structure bits 13:11)
    //              000: 48kHz/ 44.1 kHz or lesss 
    //              001: x2 (96kHz, 88.2kHz, 32kHz) 
    //              010: x3 (144kHz) 
    //              011: x4 (192kHz, 176.4kHz) 
    //              100-111: Reserved
    // C45C [10:8] Converter_Format_Div
    // DIV. Converter Format Sample Base Rate Divisor. (PCM Format structure bits 10:8)
    //              000: Divide by 1 (48kHz, 44.1 kHz) 
    //              001: Divide by 2 (24kHz, 22.05kHz) 
    //              010: Divide by 3 (16kHz, 32kHz) 
    //              011: Divide by 4 (11.025kHz) 
    //              100: Divide by 5 (9.6kHz) 
    //              101: Divide by 6 (8kHz) 
    //              110: Divide by 7 
    //              111: Divide by 8 (6kHz)
    if( mmC4D8 & 0x4000 )
        dwSampleRate = 44100;
    else 
        dwSampleRate = 48000;
            
    switch((mmC4D8 & 0x3800) >> 11) 
    {
        case 1:
            dwSampleRate *= 2;
            break;
        case 2:
            dwSampleRate *= 3;
            break;
        case 3:
            dwSampleRate *= 4;
            break;
        default:
            break;
    }

    switch((mmC4D8 & 0x700) >> 8)
        {
        case 1:
            dwSampleRate /= 2;
            break;
        case 2:
            dwSampleRate /= 3;
            break;
        case 3:
            dwSampleRate /= 4;
            break;
        case 4:
            dwSampleRate /= 5;
            break;
        case 5:
            dwSampleRate /= 6;
            break;
        case 6:
            dwSampleRate /= 7;
            break;
        case 7:
            dwSampleRate /= 8;
            break;
        default:
            break;
    }

    mmC424 = MMIO_RD(0xC424) & 0xFFFF9FFF; // C424[14:13] codec type. related with mm c618, mm c698
    mmC424 |= (0 << 13);

    MMIO_WR(0xC424, mmC424);

    // step 3. calculate C400 and C410
    SetHDAudioSampleRate(dwSampleRate, dwPixelClock);

    // step 4. fill CTS & N according to Table7-1, 7-2, 7-3 of HDMI spec 1.3
    SetHDAudioCTSandN(dwSampleRate, dwPixelClock);

}

void 
SetInfoFrameData(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    SetAVIInfoFrame(pBIOSInfo, pHDMISettingInfo);
    SetAudioInfoFrame();
}

/* Search HDMI Mode Index */
int SearchHDMIModeIndex(int ModeInfoIndex )
{
    int i=0;

    while ((i < NUM_TOTAL_HDMI_MODETABLE) && (ModeInfoIndex != CEAModes[i].ModeIndex))
    {
        i++;
    }

    if (i >= NUM_TOTAL_HDMI_MODETABLE)        
        i=0;
    
    return i;
}

static void 
viaSetHDMITiming(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo, int activeDevice)
{
    struct VideoModeTable   *HDMIMode = NULL;
    struct display_timing   mode_crt_reg;
    struct crt_mode_table   *mode_crt_table = NULL;
    BOOL                    useModeline = FALSE;
    int i = 0, index = 0, set_iga = pHDMISettingInfo->IGAPath;
    ScrnInfoPtr             pScrn = xf86Screens[pBIOSInfo->scrnIndex];
    VIAPtr                  pVia = VIAPTR(pScrn);    
	
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
          "viaSetHDMITiming: Set IGA%d Mode!\n", set_iga));

    /* Use the mode line in two cases:
    1. SAMM and Mode line is user-setting
    2. DuoView/single, EDIDType=true, and Mode line is user-setting*/
    if (!(pBIOSInfo->Is3DScalingEnable == (DISP_3D_SCALING_ENABLE | activeDevice)) &&
		(pVia->pVIAEnt->HasSecondary ||
         (!pVia->pVIAEnt->HasSecondary && (pBIOSInfo->EDIDType == activeDevice))) && 
         (pBIOSInfo->ModeLineStatus & (MODE_LINE_TYPE_USER | MODE_LINE_TYPE_EDID))) {
        useModeline = TRUE;
    }
    HDMIMode = &CEAModes[SearchHDMIModeIndex(pHDMISettingInfo->ModeIndex)];
    mode_crt_table = HDMIMode->crtc;
    mode_crt_reg = mode_crt_table->crtc;

    if ((pBIOSInfo->IsEnableInterlaceMode) &&
        (pHDMISettingInfo->ModeIndex == M1920x1080_I)) {
        if (pHDMISettingInfo->IGAPath == IGA1)
            write_reg_mask(CR33, VIACR, BIT6, BIT6);
        else if (pHDMISettingInfo->IGAPath == IGA2)
            write_reg_mask(CR67, VIACR, BIT5, BIT5);

        switch (pBIOSInfo->Chipset)
        {
            case VIA_VX855:
                write_reg_mask(CRFB, VIACR, (mode_crt_reg.hor_total/2)&0xFF, 0xFF);
                write_reg_mask(CRFC, VIACR, ((mode_crt_reg.hor_total/2)&0x700)>>8, 0x07);
                break;
            case VIA_VX900:
                write_reg_mask(CRAB, VIACR, (mode_crt_reg.hor_total/2)&0xFF, 0xFF);
                write_reg_mask(CRAC, VIACR, ((mode_crt_reg.hor_total/2)&0x700)>>8, 0x07);
                break;
            default:
                write_reg_mask(CRFB, VIACR, (mode_crt_reg.hor_total/2)&0xFF, 0xFF);
                write_reg_mask(CRFC, VIACR, ((mode_crt_reg.hor_total/2)&0x700)>>8, 0x07);
                break;
        }

    }

    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "-----HDMI TIMING-----\n");

    /* Write CRTC */
    fill_crtc_timing(pBIOSInfo, mode_crt_table, HDMIMode->mode_array,
                     pBIOSInfo->bitsPerPixel >> 3, activeDevice,
                     set_iga, useModeline);

    for (i=0; i<HDMIMode->mode_array; i++) {
            index = i;
            if (mode_crt_table[i].refresh_rate == pBIOSInfo->FoundRefresh)
                break;
    }
                                                                             
    viaPatchCEADotClock(pBIOSInfo, mode_crt_table[index].clk, set_iga);

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, 
          "viaSetHDMITiming --Done--\n"));
}

void
VIASetIntegratedHDMIMode(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo)
{
    // power down to reset
    MMIO_WR_MASK(0xC740, 0x00000000, 0x06000000);  
    // enable HDCP
    MMIO_WR_MASK(HDCP_ENABLE_REG, 0x00000001, 0x00000001);  
    // Infoframe setting
    SetInfoFrameData(pBIOSInfo, pHDMISettingInfo);
    SetHDAudioParameter(pBIOSInfo, pHDMISettingInfo);
    // select HDMI mode
    MMIO_WR_MASK(0xC748, 0, BIT0); 
    // enable HDMI with HDMI mode
    MMIO_WR(0xC280, MMIO_RD(0xC280) | 0x18330002);  
    MMIO_WR_MASK(0xC280, 0x0, 0x40);
    // select AC mode
    MMIO_WR_MASK(0xC74C, 0x40, 0x40);  
    // enable InfoFrame
    MMIO_WR(0xC284, 0x20000102); 
    // set status of Lane0~3
    MMIO_WR_MASK(0xC744, 0x00FFFF82, 0x00FFFF82);  
    // InfoFrame setting
    MMIO_WR(0xC0B4, 0x92000000);
    // enable audio packet
    MMIO_WR_MASK(0xC294, 0x10000000, 0x10000000); 
    // enable InfoFrame
    MMIO_WR(0xC284, 0x20000102);  
    MMIO_WR_MASK(0xC740, 0x1EC46E6F, 0x1EC46E6F);  
    //Select PHY Function as HDMI  
    write_reg_mask(CRFF, VIACR, BIT0, BIT0);
    //Select HDTV0 source 
    write_reg_mask(CRFF, VIACR, BIT1, BIT1);
}

Bool 
viaSetHDMIMode(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo, int activeDevice)
{
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "viaSetHDMIMode\n"));

    viaSetHDMITiming(pBIOSInfo, pHDMISettingInfo, activeDevice);
    viaAdjustTimingForHDMI(pBIOSInfo);

    if (pHDMISettingInfo->ChipID == VIA_INTEGRATED_HDMI)
        VIASetIntegratedHDMIMode(pBIOSInfo, pHDMISettingInfo);

    if (pHDMISettingInfo->IGAPath == IGA2) {
        /* Second Display Channel Enable */
        EnableSecondDisplayChannel();
        
        /* Fixed Set DVI+TV/DVT+CRT/CRT+TV SAMM mode,
           the secondary monitor without display */
        write_reg_mask(CR91, VIACR, 0x0, BIT5);

        /* Disable LCD Scaling */
        write_reg(CR79, VIACR, 0x0);
    }

    VIASetOutputPath(pBIOSInfo, activeDevice,
                     pHDMISettingInfo->IGAPath, pHDMISettingInfo->HDMIDIPort);

    if (pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx)
        VIASetDPA_Gfx(pHDMISettingInfo->HDMIDIPort,
                      &(pBIOSInfo->UserDPASettingInfo.GfxDPA));
    else
        viaPatchHDMISkew_DVP1(pBIOSInfo);  /* Patch DVP1 Clock Skew for External HDMI */
    
    /* Set display channel. */
    if ((pBIOSInfo->Chipset == VIA_CX700) || (pBIOSInfo->Chipset == VIA_VX800))
        VIASetDisplayChannel(pBIOSInfo);
    
    return TRUE;
}

CARD32 
viaGetDeviceDetectHDMI(VIABIOSInfoPtr pBIOSInfo)
{
    CARD32  ret = 0x0;
    
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "viaGetDeviceDetectHDMI\n"));

    if (pBIOSInfo->HDMISettingInfo.SinkDevType)
        ret |= VIA_DEVICE_HDMI;
    if (pBIOSInfo->HDMISettingInfo2.SinkDevType)
        ret |= VIA_DEVICE_HDMI2;

    return ret;    
}

void VIAInitSettingInfo_HDMI(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo, CARD16 activeDevice)
{
    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAInitSettingInfo_HDMI!!\n");
    
    int     HActive, VActive;
    Bool    IsPanning = pHDMISettingInfo->IsPanning;
    Bool    IsDownScaling = pHDMISettingInfo->IsDownScaling;
    
    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;

        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO,
                   "HDMI Max Resolution %lux%lu, set virtual desktop!!\n", 
                   pHDMISettingInfo->PanelSizeH,
                   pHDMISettingInfo->PanelSizeV));
    }
    else if (IsDownScaling)
    {
        HActive = pBIOSInfo->HorDownScaleSize;
        VActive = pBIOSInfo->VerDownScaleSize;
    }
    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }
    
    VIAInitSettingInfo_IGA (pBIOSInfo, pHDMISettingInfo->IGAPath,
                            IsPanning, IsDownScaling, HActive, VActive,
                            activeDevice);

    pHDMISettingInfo->ModeIndex = VIAGetModeIndex(HActive, VActive);
    pHDMISettingInfo->HActive = HActive;
    pHDMISettingInfo->VActive = VActive;        
    pBIOSInfo->IsDownScaleEnable = IsDownScaling;        

    if (pBIOSInfo->IsEnableInterlaceMode)
    {
        switch (pHDMISettingInfo->ModeIndex)
        {
            case M1920x1080:                
                pHDMISettingInfo->ModeIndex |= 0x200;
                break;
        }
    }
}

void VIAAdjustActiveSize_HDMI(VIABIOSInfoPtr pBIOSInfo, 
                                   HDMISETTINGINFOPTR pHDMISettingInfo,
                                   int *pHActive,
                                   int *pVActive, 
                                   int whichHDMI)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;

    if (pBIOSInfo->Is3DScalingEnable & whichHDMI)
    {
         pHDMISettingInfo->IsPanning = FALSE; 
    }
    else
    {
        if ((pHDMISettingInfo->PanelSizeH == 0) &&
            (pHDMISettingInfo->PanelSizeV == 0) &&
            (pHDMISettingInfo->PanelSizeH != 0xFFF))
        {
            VIAGetModeSizeByIndex(pHDMISettingInfo->NativeSizeID,
                                 &pHDMISettingInfo->PanelSizeH,
                                 &pHDMISettingInfo->PanelSizeV);
        }
        
        if ((HResToSet > pHDMISettingInfo->PanelSizeH) ||
            (VResToSet > pHDMISettingInfo->PanelSizeV))
        {        
            if ((*pHActive > pHDMISettingInfo->PanelSizeH) ||
                (*pVActive > pHDMISettingInfo->PanelSizeV))
            {            
                /* Panning */
                *pHActive = pHDMISettingInfo->PanelSizeH;
                *pVActive = pHDMISettingInfo->PanelSizeV;
                pHDMISettingInfo->IsPanning = TRUE;
                pHDMISettingInfo->IsDownScaling = FALSE;
            }            
        }
        else if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
        {
            /* If use Virtual Desktop, and the size is larger than the resolution, */
            /* we need to do panning, too. */
            pHDMISettingInfo->IsPanning = TRUE;
        }
        else
        {
            pHDMISettingInfo->IsPanning = FALSE;
        }
    }
}

/* For HDMI 3DScaling */
Bool VIASetHDMI3DScalScreenParas(VIABIOSInfoPtr  pBIOSInfo)
{
       VIA3DSclSCNParasPtr pHDMI3DScalInfo = &(pBIOSInfo->HDMISettingInfo.HDMI3DScalInfo);
       /*Get the  physical size*/
       pHDMI3DScalInfo->XSIZEMax = pBIOSInfo->HDMISettingInfo.HActive;
       pHDMI3DScalInfo->YSIZEMax = pBIOSInfo->HDMISettingInfo.VActive;
       
       /*make adjust SIZE range 50% of physical mode, 25% up and down*/
       pHDMI3DScalInfo->XSIZEMaxRange= (pBIOSInfo->HDMISettingInfo.HActive >> 1);
       pHDMI3DScalInfo->YSIZEMaxRange= (pBIOSInfo->HDMISettingInfo.VActive >> 1);    


       pHDMI3DScalInfo->XScaleStep = 3;
       pHDMI3DScalInfo->YScaleStep = 3;

       pHDMI3DScalInfo->XScaleMaximum = (pHDMI3DScalInfo->XSIZEMaxRange / pHDMI3DScalInfo->XScaleStep);
       pHDMI3DScalInfo->YScaleMaximum = (pHDMI3DScalInfo->YSIZEMaxRange / pHDMI3DScalInfo->YScaleStep);
       
       pHDMI3DScalInfo->XScaleLevel = pHDMI3DScalInfo->XScaleDefault = pHDMI3DScalInfo->XScaleMaximum >>1;
       pHDMI3DScalInfo->YScaleLevel = pHDMI3DScalInfo->YScaleDefault = pHDMI3DScalInfo->YScaleMaximum >>1;
       
       pHDMI3DScalInfo->XSIZEOffset = (pHDMI3DScalInfo->XScaleDefault - pHDMI3DScalInfo->XScaleLevel)*pHDMI3DScalInfo->XScaleStep;
       pHDMI3DScalInfo->YSIZEOffset = (pHDMI3DScalInfo->YScaleDefault - pHDMI3DScalInfo->YScaleLevel)*pHDMI3DScalInfo->YScaleStep;

       pHDMI3DScalInfo->XSIZENow = pHDMI3DScalInfo->XSIZEMax - (pHDMI3DScalInfo->XScaleLevel*pHDMI3DScalInfo->XScaleStep);
       pHDMI3DScalInfo->YSIZENow = pHDMI3DScalInfo->YSIZEMax - (pHDMI3DScalInfo->YScaleLevel*pHDMI3DScalInfo->YScaleStep);

       /*make adjust POSITION range 50% of physical mode, 25% up and down       */
       pHDMI3DScalInfo->XPOSITIONMaxRange = (pBIOSInfo->HDMISettingInfo.HActive >> 1);
       pHDMI3DScalInfo->YPOSITIONMaxRange = (pBIOSInfo->HDMISettingInfo.VActive >> 1);
       
       pHDMI3DScalInfo->XPOSITIONStep = 3;
       pHDMI3DScalInfo->YPOSITIONStep = 3;

       pHDMI3DScalInfo->XPOSITIONMaximum = (pHDMI3DScalInfo->XPOSITIONMaxRange / pHDMI3DScalInfo->XPOSITIONStep);
       pHDMI3DScalInfo->YPOSITIONMaximum = (pHDMI3DScalInfo->YPOSITIONMaxRange / pHDMI3DScalInfo->YPOSITIONStep);
       
       pHDMI3DScalInfo->XPOSITIONLevel = pHDMI3DScalInfo->XPOSITIONDefault = pHDMI3DScalInfo->XPOSITIONMaximum >>1;
       pHDMI3DScalInfo->YPOSITIONLevel = pHDMI3DScalInfo->YPOSITIONDefault = pHDMI3DScalInfo->YPOSITIONMaximum >>1;

       pHDMI3DScalInfo->XPOSITIONOffset= (pHDMI3DScalInfo->XPOSITIONLevel - pHDMI3DScalInfo->XPOSITIONDefault)*pHDMI3DScalInfo->XPOSITIONStep;
       pHDMI3DScalInfo->YPOSITIONOffset= (pHDMI3DScalInfo->YPOSITIONLevel - pHDMI3DScalInfo->YPOSITIONDefault)*pHDMI3DScalInfo->YPOSITIONStep;

       pHDMI3DScalInfo->XPOSITIONNow =( pHDMI3DScalInfo->XPOSITIONLevel - pHDMI3DScalInfo->XPOSITIONDefault)*pHDMI3DScalInfo->XPOSITIONStep;
       pHDMI3DScalInfo->YPOSITIONNow =( pHDMI3DScalInfo->YPOSITIONLevel - pHDMI3DScalInfo->YPOSITIONDefault)*pHDMI3DScalInfo->YPOSITIONStep;

       pHDMI3DScalInfo->bNeedBlack3dScalingBorders = FALSE;

       if((pBIOSInfo->HDMISettingInfo.IsPreSetting))
       {
            pHDMI3DScalInfo->XScaleLevel = pBIOSInfo->HDMISettingInfo.XScalingLevel;
            pHDMI3DScalInfo->YScaleLevel = pBIOSInfo->HDMISettingInfo.YScalingLevel; 
            pHDMI3DScalInfo->XSIZEOffset = (pHDMI3DScalInfo->XScaleDefault - pHDMI3DScalInfo->XScaleLevel)*pHDMI3DScalInfo->XScaleStep;
            pHDMI3DScalInfo->YSIZEOffset = (pHDMI3DScalInfo->YScaleDefault - pHDMI3DScalInfo->YScaleLevel)*pHDMI3DScalInfo->YScaleStep;
            
            pHDMI3DScalInfo->XPOSITIONLevel = pBIOSInfo->HDMISettingInfo.XScalPosLevel;
            pHDMI3DScalInfo->YPOSITIONLevel = pBIOSInfo->HDMISettingInfo.YScalPosLevel;
            pHDMI3DScalInfo->XPOSITIONOffset= (pHDMI3DScalInfo->XPOSITIONLevel - pHDMI3DScalInfo->XPOSITIONDefault)*pHDMI3DScalInfo->XPOSITIONStep;
            pHDMI3DScalInfo->YPOSITIONOffset= (pHDMI3DScalInfo->YPOSITIONLevel - pHDMI3DScalInfo->YPOSITIONDefault)*pHDMI3DScalInfo->YPOSITIONStep;
       }
       return TRUE;
       
}
