/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#ifndef __VIA_HDMI_H__
#define __VIA_HDMI_H__

#define CONNECT_HDMI    0x1
#define CONNECT_DVI     0x2

#define RATIO_4_3       1
#define RATIO_5_4       2
#define RATIO_16_10     3
#define RATIO_16_9      4


#define INTERLACE       1
#define PROGRESSIVE     0

/* For Detect EDID */
#define EDID_NOT_EXIST          0
#define EDID_EXIST              1
#define EDID_EXTENSION_EXIST    2

#define DCLK_25_200M        25200
#define DCLK_27_000M        27000
#define DCLK_54_000M        54000
#define DCLK_72_300M        72300
#define DCLK_74_270M        74270
#define DCLK_108_000M      108000
#define DCLK_148_500M      148500

static struct ViaDotClock {
    int DotClock;
    CARD32 PLL3;
    CARD32 PLL3_409;
} ViaDotClocks[] = {
    /*               PLL3      PLL3_409       */    
    {  DCLK_25_200M, 0x8B9083, 0x8D9085}, 
    {  DCLK_27_000M, 0x959083, 0x979085}, 
    {  DCLK_54_000M, 0x958C83, 0x978C85}, 
    {  DCLK_72_300M, 0xC88C83, 0xCA8C85},     
    {  DCLK_74_270M, 0xA48C82, 0xA68C84}, 
    { DCLK_108_000M, 0x958883, 0x978885}, 
    { DCLK_148_500M, 0xA48882, 0xA68884},     
    {      0,      0,        0 }
};

typedef struct {
    CARD16 ID;
    CARD16 HorSize;
    CARD16 VerSize;
    CARD16 RefreshRate;
    int    ScanType;
    CARD16 VSyncOffset;
    int    Ratio;
    Bool   IsBuilt; /* If the timing table exists. */
}CEATimingInfoRec, *CEATimingInfoPtr;

/* Sync CBIOS's CEA timing table at 20090327 */
/* The table include whole CEA861.D's mode, but we only open some familiar
   modes. If we want to open the others mode, we just need to set the "Use"
   field as TRUE. */
static CEATimingInfoRec CEATimingTable[] = {
    /* ID     H_Res    V_Res   Refresh  Inter/Prog  VSynOff   Ratio      Use */
    {  1,     640,     480,    60,    PROGRESSIVE,     0,    RATIO_4_3,  TRUE },
    {  2,     720,     480,    60,    PROGRESSIVE,     0,    RATIO_4_3,  TRUE},
    {  3,     720,     480,    60,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    {  4,    1280,     720,    60,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    {  5,    1920,    1080,    60,      INTERLACE,   1100,  RATIO_16_9,  TRUE },
    {  6,     720,     480,    60,      INTERLACE,     22,   RATIO_4_3,  FALSE},
    {  7,     720,     480,    60,      INTERLACE,     22,  RATIO_16_9,  FALSE},
    {  8,     720,     240,    60,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    {  9,     720,     240,    60,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 10,    2880,     480,    60,      INTERLACE,     22,   RATIO_4_3,  FALSE},
    { 11,    2880,     480,    60,      INTERLACE,     22,  RATIO_16_9,  FALSE},
    { 12,    2880,     240,    60,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 13,    2880,     240,    60,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 14,    1440,     480,    60,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},    
    { 15,    1440,     480,    60,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},    
    { 16,    1920,    1080,    60,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    { 17,     720,     576,    50,    PROGRESSIVE,     0,    RATIO_4_3,  TRUE },
    { 18,     720,     576,    50,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    { 19,    1280,     720,    50,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    { 20,    1920,    1080,    50,      INTERLACE,   1320,  RATIO_16_9,  TRUE },
    { 21,     720,     576,    50,      INTERLACE,    864,   RATIO_4_3,  FALSE},
    { 22,     720,     576,    50,      INTERLACE,    864,  RATIO_16_9,  FALSE},
    { 23,     720,     288,    50,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 24,     720,     288,    50,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 25,    2880,     576,    50,      INTERLACE,   1728,   RATIO_4_3,  FALSE},
    { 26,    2880,     576,    50,      INTERLACE,   1728,  RATIO_16_9,  FALSE},
    { 27,    2880,     288,    50,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 28,    2880,     288,    50,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 29,    1440,     576,    50,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},    
    { 30,    1440,     576,    50,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},    
    { 31,    1920,    1080,    50,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    { 32,    1920,    1080,    24,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },    
    { 33,    1920,    1080,    25,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },    
    { 34,    1920,    1080,    30,    PROGRESSIVE,     0,   RATIO_16_9,  TRUE },
    { 35,    2880,     480,    60,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 36,    2880,     480,    60,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 37,    2880,     576,    50,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 38,    2880,     576,    50,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 39,    1920,    1080,    50,      INTERLACE,   1152,  RATIO_16_9,  TRUE },
    { 40,    1920,    1080,   100,      INTERLACE,   1320,  RATIO_16_9,  FALSE},
    { 41,    1280,     720,   100,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 42,     720,     576,   100,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 43,     720,     576,   100,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 44,     720,     576,   100,      INTERLACE,    864,  RATIO_16_9,  FALSE},
    { 45,     720,     576,   100,      INTERLACE,    864,   RATIO_4_3,  FALSE},
    { 46,    1920,    1080,   120,      INTERLACE,   1100,  RATIO_16_9,  FALSE},
    { 47,    1280,     720,   120,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 48,     720,     480,   120,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 49,     720,     480,   120,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 50,     720,     480,   120,      INTERLACE,   1134,   RATIO_4_3,  FALSE},
    { 51,     720,     480,   120,      INTERLACE,   1134,  RATIO_16_9,  FALSE},
    { 52,     720,     576,   200,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 53,     720,     576,   200,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 54,     720,     576,   200,      INTERLACE,    864,   RATIO_4_3,  FALSE},
    { 55,     720,     576,   200,      INTERLACE,    864,  RATIO_16_9,  FALSE},
    { 56,     720,     480,   240,    PROGRESSIVE,     0,    RATIO_4_3,  FALSE},
    { 57,     720,     480,   240,    PROGRESSIVE,     0,   RATIO_16_9,  FALSE},
    { 58,     720,     480,   240,      INTERLACE,   1134,   RATIO_4_3,  FALSE},
    { 59,     720,     480,   240,      INTERLACE,   1134,  RATIO_16_9,  FALSE},
};

/*    Name                      Dot Clock                           */
/*    HA   HSS  HSE  HT   HS VA    VSS   VSE   VT   VS              */
/*    Flags                                                         */
/* Important! HS=HSkew, we use skew to indicate the video ID codec  */
static DisplayModeRec ViaCEAModes[] = {    
    { MODEPREFIX("640x480"),    DCLK_25_200M,
      640, 656, 752, 800, 1, 480,  490,  492,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },
        
    { MODEPREFIX("720x480P_4_3"),    DCLK_27_000M,
      720, 736, 798, 858, 2, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },
        
    { MODEPREFIX("720x480P_16_9"),    DCLK_27_000M,
      720, 736, 798, 858, 3, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },        
        
    { MODEPREFIX("1280x720P"),    DCLK_74_270M,
      1280, 1390, 1430, 1650, 4, 720,  725,  730,  750, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },
        
    { MODEPREFIX("1920x1080I"),   DCLK_74_270M,
      1920, 2008, 2052, 2200, 5, 1080, 1084, 1094, 1124, 0,
      V_PHSYNC | V_PVSYNC | V_INTERLACE, MODESUFFIX },                
        
    { MODEPREFIX("720x480I_4_3"),   DCLK_27_000M,
      1440, 1478, 1602, 1716, 6, 480, 488, 494, 524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },
        
    { MODEPREFIX("720x480I_16_9"),   DCLK_27_000M,
      1440, 1478, 1602, 1716, 7, 480, 488, 494, 524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },        
        
    { MODEPREFIX("720x240P_4_3"),   DCLK_27_000M,
      1440, 1478, 1602, 1716, 8, 240, 244, 247, 262, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },        
        
    { MODEPREFIX("720x240P_16_9"),   DCLK_27_000M,
      1440, 1478, 1602, 1716, 9, 240, 244, 247, 262, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                
        
    { MODEPREFIX("2880x480I_4_3"),   DCLK_54_000M,
      2880, 2956, 3204, 3432, 10, 480, 488, 494, 524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },
        
    { MODEPREFIX("2880x480I_16_9"),   DCLK_54_000M,
      2880, 2956, 3204, 3432, 11, 480, 488, 494, 524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },        
        
    { MODEPREFIX("2880x240P_4_3"),   DCLK_54_000M,
      2880, 2956, 3204, 3432, 12, 240, 244, 247, 262, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },        
       
    { MODEPREFIX("2880x240P_16_9"),   DCLK_54_000M,
      2880, 2956, 3204, 3432, 13, 240, 244, 247, 262, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },                
        
    { MODEPREFIX("1440x480P_4_3"),   DCLK_54_000M,
      1440, 1472, 1596, 1716, 14, 480, 489, 495, 525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                
        
    { MODEPREFIX("1440x480P_16_9"),   DCLK_54_000M,
      1440, 1472, 1596, 1716, 15, 480, 489, 495, 525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        
        
    { MODEPREFIX("1920x1080P"),  DCLK_148_500M,
      1920, 2008, 2052, 2200, 16, 1080, 1084, 1089, 1125, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },
        
    { MODEPREFIX("720x576P_4_3"),    DCLK_27_000M,
      720, 732, 796, 864, 17, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                
        
    { MODEPREFIX("720x576P_16_9"),    DCLK_27_000M,
      720, 732, 796, 864, 18, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1280x720P"),    DCLK_74_270M,
      1280, 1720, 1760, 1980, 19, 720,  725,  730,  750, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1920x1080I"),    DCLK_74_270M,
      1920, 2448, 2492, 2640, 20, 1080,  1084,  1094,  1124, 0,
      V_PHSYNC | V_PVSYNC | V_INTERLACE, MODESUFFIX },                        

    { MODEPREFIX("720x576I_4_3"),    DCLK_27_000M,
      1440, 1464, 1590, 1728, 21, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },                        

    { MODEPREFIX("720x576I_16_9"),    DCLK_27_000M,
      1440, 1464, 1590, 1728, 22, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },                        

    { MODEPREFIX("720x288P_4_3"),    DCLK_27_000M,
      1440, 1464, 1590, 1728, 23, 288,  290,  293,  312, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("720x288P_16_9"),    DCLK_27_000M,
      1440, 1464, 1590, 1728, 24, 288,  290,  293,  312, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("2880x576I_4_3"),    DCLK_54_000M,
      2880, 2928, 3180, 3456, 25, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },                        

    { MODEPREFIX("2880x576I_16_9"),    DCLK_54_000M,
      2880, 2928, 3180, 3456, 26, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },                        

    { MODEPREFIX("2880x288P_4_3"),    DCLK_54_000M,
      2880, 2928, 3180, 3456, 27, 288,  290,  293,  312, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("2880x288P_16_9"),    DCLK_54_000M,
      2880, 2928, 3180, 3456, 28, 288,  290,  293,  312, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1440x576P_4_3"),    DCLK_54_000M,
      1440, 1464, 1592, 1728, 29, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1440x576P_16_9"),    DCLK_54_000M,
      1440, 1464, 1592, 1728, 30, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1920x1080P"),    DCLK_148_500M,
      1920, 2448, 2492, 2640, 31, 1080,  1084,  1089,  1125, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1920x1080P"),    DCLK_74_270M,
      1920, 2558, 2602, 2750, 32, 1080,  1084,  1089,  1125, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1920x1080P"),    DCLK_74_270M,
      1920, 2448, 2492, 2640, 33, 1080,  1084,  1089,  1125, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },                        

    { MODEPREFIX("1920x1080P"),    DCLK_74_270M,
      1920, 2008, 2052, 2200, 34, 1080,  1084,  1089,  1125, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },                        

    { MODEPREFIX("2880x480P_4_3"),    DCLK_108_000M,
      2880, 2944, 3192, 3432, 35, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },
      
    { MODEPREFIX("2880x480P_16_9"),    DCLK_108_000M,
      2880, 2944, 3192, 3432, 36, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("2880x576P_4_3"),    DCLK_108_000M,
      2880, 2928, 3184, 3456, 37, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("2880x576P_16_9"),    DCLK_108_000M,
      2880, 2928, 3184, 3456, 38, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("1920x1080I"),    DCLK_72_300M,
      1920, 1952, 2120, 2304, 39, 1080,  1126,  1136,  1250, 0,
      V_PHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("1920x1080I"),    DCLK_148_500M,
      1920, 2448, 2492, 2640, 40, 1080,  1084,  1094,  1124, 0,
      V_PHSYNC | V_PVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("1280x720P"),    DCLK_148_500M,
      1280, 1720, 1760, 1980, 41, 720,  725,  730,  750, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },

    { MODEPREFIX("720x576P_4_3"),    DCLK_54_000M,
      720, 732, 796, 864, 42, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x576P_16_9"),    DCLK_54_000M,
      720, 732, 796, 864, 43, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x576I_4_3"),    DCLK_54_000M,
      1440, 1464, 1590, 1728, 44, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },
    
    { MODEPREFIX("720x576I_16_9"),    DCLK_54_000M,
      1440, 1464, 1590, 1728, 45, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },
      
    { MODEPREFIX("1920x1080I"),    DCLK_148_500M,
      1920, 2008, 2052, 2200, 46, 1080,  1084,  1094,  1124, 0,
      V_PHSYNC | V_PVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("1280x720P"),    DCLK_148_500M,
      1280, 1390, 1430, 1650, 47, 720,  725,  730,  750, 0,
      V_PHSYNC | V_PVSYNC, MODESUFFIX },
      
    { MODEPREFIX("720x480P_4_3"),    DCLK_54_000M,
      720, 736, 798, 858, 48, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x480P_16_9"),    DCLK_54_000M,
      720, 736, 798, 858, 49, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x480I_4_3"),    DCLK_54_000M,
      1440, 1478, 1602, 1716, 50, 480,  488,  494,  524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("720x480I_16_9"),    DCLK_54_000M,
      1440, 1478, 1602, 1716, 51, 480,  488,  494,  524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("720x576P_4_3"),    DCLK_108_000M,
      720, 732, 796, 864, 52, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x576P_16_9"),    DCLK_108_000M,
      720, 732, 796, 864, 53, 576,  581,  586,  625, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x576I_4_3"),    DCLK_108_000M,
      1440, 1464, 1590, 1728, 54, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("720x576I_16_9"),    DCLK_108_000M,
      1440, 1464, 1590, 1728, 55, 576,  580,  586,  624, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("720x480P_4_3"),    DCLK_108_000M,
      720, 736, 798, 858, 56, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x480P_16_9"),    DCLK_108_000M,
      720, 736, 798, 858, 57, 480,  489,  495,  525, 0,
      V_NHSYNC | V_NVSYNC, MODESUFFIX },

    { MODEPREFIX("720x480I_4_3"),    DCLK_108_000M,
      1440, 1478, 1602, 1716, 58, 480,  488,  494,  524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },

    { MODEPREFIX("720x480I_16_9"),    DCLK_108_000M,
      1440, 1478, 1602, 1716, 59, 480,  488,  494,  524, 0,
      V_NHSYNC | V_NVSYNC | V_INTERLACE, MODESUFFIX },
        
/* End */        
    { MODEPREFIX(NULL), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, MODESUFFIX }
};

typedef struct _HDAuioCTSandNRegSetting{
    DWORD SampleRate;
    DWORD PixelClock;
    DWORD CTS;
    DWORD N;
}HDAuioCTSandNRegSetting, *PHDAuioCTSandNRegSetting;

typedef struct _HDAuioPacketClockRegSetting{
    CARD32 SampleRate;
    CARD32 PixelClock;
    CARD32 PacketClockRatio_L;
    CARD32 PacketClockRateo_H;
}HDAuioPacketClockRegSetting, *PHDAuioPacketClockRegSetting;

typedef struct _HDMI_InfoFrameRec
{
    CARD16 PacketHeader[3];
    CARD16 ECCCode;
    CARD16 PacketByte[28];
    CARD16 BCHCode[4];
} HDMI_InfoFrameRec;

typedef struct _HDAuioWallClockRegSetting{
    CARD32 PixelClock;
    CARD32 WallClockRatio_L;
    CARD32 WallClockRateo_H;
}HDAuioWallClockRegSetting, *PHDAuioWallClockRegSetting;


static HDAuioWallClockRegSetting  WallClockReg[] =
{
    { DCLK_25_200M,  0xCF3CF3CF, 0xF3 },
    { DCLK_27_000M,  0x8E38E38E, 0xE3 },
    { DCLK_54_000M,  0xC71C71C7, 0x71 },
    { DCLK_72_300M,  0xFAB054FA, 0x54 },
    { DCLK_74_270M,  0xB9A62C9A, 0x52 },
    { DCLK_108_000M, 0x8E38E38E3,0x03 },
    { DCLK_148_500M, 0x95FAD40A5,0x02 }
};

static HDAuioPacketClockRegSetting  PacketClockReg[] =
{
    { 32000,  DCLK_25_200M,  0x533866B9, 0x00 },
    { 32000,  DCLK_27_000M,  0x4DAC1B9C, 0x00 },
    { 32000,  DCLK_54_000M,  0x26D60DCE, 0x00 },
    { 32000,  DCLK_72_300M,  0x1D0199B6, 0x00 },
    { 32000,  DCLK_74_270M,  0x1C3CA337, 0x00 },
    { 32000,  DCLK_108_000M, 0x136B06E7, 0x00 },    
    { 32000,  DCLK_148_500M, 0x0E1F4AD6, 0x00 }, 
    { 44100,  DCLK_25_200M,  0x72B020C4, 0x00 },
    { 44100,  DCLK_27_000M,  0x6B0AC940, 0x00 },
    { 44100,  DCLK_54_000M,  0x358564A0, 0x00 },
    { 44100,  DCLK_72_300M,  0x27F967D6, 0x00 },
    { 44100,  DCLK_74_270M,  0x26E9F755, 0x00 },
    { 44100,  DCLK_108_000M, 0x1AC2B250, 0x00 },  
    { 44100,  DCLK_148_500M, 0x13765322, 0x00 }, 
    { 88200,  DCLK_25_200M,  0xE5604189, 0x00 },
    { 88200,  DCLK_27_000M,  0xD6159280, 0x00 },
    { 88200,  DCLK_54_000M,  0x6B0AC940, 0x00 },
    { 88200,  DCLK_72_300M,  0x4FF2CFAC, 0x00 },
    { 88200,  DCLK_74_270M,  0x4DD3EEAA, 0x00 },
    { 88200,  DCLK_108_000M, 0x358564A0, 0x00 },  
    { 88200,  DCLK_148_500M, 0x26ECA645, 0x00 }, 
    { 176400, DCLK_25_200M,  0xCAC08312, 0x01 },
    { 176400, DCLK_27_000M,  0xAC2B2500, 0x01 },
    { 176400, DCLK_54_000M,  0xD6159280, 0x00 },
    { 176400, DCLK_72_300M,  0x9FE59F58, 0x00 },
    { 176400, DCLK_74_270M,  0x9BA7DD55, 0x00 },
    { 176400, DCLK_108_000M, 0x6B0AC940, 0x00 },    
    { 176400, DCLK_148_500M, 0x4DD94C8B, 0x00 }, 
    { 48000,  DCLK_25_200M,  0x7CD49A16, 0x00 },
    { 48000,  DCLK_27_000M,  0x7482296A, 0x00 },
    { 48000,  DCLK_54_000M,  0x3A4114B5, 0x00 },
    { 48000,  DCLK_72_300M,  0x2B826691, 0x00 },
    { 48000,  DCLK_74_270M,  0x2A5AF4D3, 0x00 },
    { 48000,  DCLK_108_000M, 0x1D208A5A, 0x00 },    
    { 48000,  DCLK_148_500M, 0x152EF041, 0x00 }, 
    { 96000,  DCLK_25_200M,  0xF9A9342C, 0x00 },
    { 96000,  DCLK_27_000M,  0xE90452D4, 0x00 },
    { 96000,  DCLK_54_000M,  0x7482296A, 0x00 },
    { 96000,  DCLK_72_300M,  0x5704CD23, 0x00 },
    { 96000,  DCLK_74_270M,  0x54B5E9A6, 0x00 },
    { 96000,  DCLK_108_000M, 0x3A4114B5, 0x00 },   
    { 96000,  DCLK_148_500M, 0x2A5DE083, 0x00 },
    { 192000, DCLK_25_200M,  0xF3526859, 0x01 },
    { 192000, DCLK_27_000M,  0xD208A5A9, 0x01 },
    { 192000, DCLK_54_000M,  0xE90452D4, 0x00 },
    { 192000, DCLK_72_300M,  0xAE099A47, 0x00 },
    { 192000, DCLK_74_270M,  0xA96BD34D, 0x00 },
    { 192000, DCLK_108_000M, 0x7482296A, 0x00 },
    { 192000, DCLK_148_500M, 0x54BBC107, 0x00 }
};

static HDAuioCTSandNRegSetting  CTSandNReg[] =
{
    // CTS = (clk*4096)/(128*32000)=clk/1000
    { 32000,  DCLK_25_200M, 25200,  4096},
    { 32000,  DCLK_27_000M, 27000,  4096},
    { 32000,  DCLK_54_000M, 54000,  4096},
    { 32000,  DCLK_72_300M, 72300,  4096},
    { 32000,  DCLK_74_270M, 74270,  4096},
    { 32000,  DCLK_108_000M,108000, 4096},
    { 32000,  DCLK_148_500M,148500, 4096},
    // CTS = (clk*6272)/(128*44100)=clk/900
    { 44100,  DCLK_25_200M, 28000,  6272},
    { 44100,  DCLK_27_000M, 30000,  6272},
    { 44100,  DCLK_54_000M, 60000,  6272},
    { 44100,  DCLK_72_300M, 80333,  6272},
    { 44100,  DCLK_74_270M, 82522,  6272},
    { 44100,  DCLK_108_000M,120000, 6272},
    { 44100,  DCLK_148_500M,165000, 6272},
    // CTS = (clk*12544)/(128*88200)=clk/900
    { 88200,  DCLK_25_200M, 28000,  12544},
    { 88200,  DCLK_27_000M, 30000,  12544},
    { 88200,  DCLK_54_000M, 60000,  12544},
    { 88200,  DCLK_72_300M, 80333,  12544},
    { 88200,  DCLK_74_270M, 82522,  12544},
    { 88200,  DCLK_108_000M,120000, 12544},
    { 88200,  DCLK_148_500M,165000, 12544},
    // CTS = (clk*25088)/(128*176400)=clk/900
    { 176400, DCLK_25_200M, 28000,  25088},
    { 176400, DCLK_27_000M, 30000,  25088},
    { 176400, DCLK_54_000M, 60000,  25088},
    { 176400, DCLK_72_300M, 80333,  25088},
    { 176400, DCLK_74_270M, 82522,  25088},
    { 176400, DCLK_108_000M,120000, 25088},
    { 176400, DCLK_148_500M,165000, 25088},
    // CTS = (clk*6144)/(128*48000)=clk/1000
    { 48000, DCLK_25_200M, 25200,  6144},
    { 48000, DCLK_27_000M, 27000,  6144},
    { 48000, DCLK_54_000M, 54000,  6144},
    { 48000, DCLK_72_300M, 72300,  6144},
    { 48000, DCLK_74_270M, 74270,  6144},
    { 48000, DCLK_108_000M,108000, 6144},
    { 48000, DCLK_148_500M,148500, 6144},
    // CTS = (clk*12288)/(128*96000)=clk/1000
    { 96000, DCLK_25_200M, 25200,  12288},
    { 96000, DCLK_27_000M, 27000,  12288},
    { 96000, DCLK_54_000M, 54000,  12288},
    { 96000, DCLK_72_300M, 72300,  12288},
    { 96000, DCLK_74_270M, 74270,  12288},
    { 96000, DCLK_108_000M,108000, 12288},
    { 96000, DCLK_148_500M,148500, 12288},
    // CTS = (clk*24576)/(128*192000)=clk/1000
    { 192000, DCLK_25_200M, 25200,  24576},
    { 192000, DCLK_27_000M, 27000,  24576},
    { 192000, DCLK_54_000M, 54000,  24576},
    { 192000, DCLK_72_300M, 72300,  24576},
    { 192000, DCLK_74_270M, 74270,  24576},
    { 192000, DCLK_108_000M,108000, 24576},
    { 192000, DCLK_148_500M,148500, 24576}
};

#define CTSandNReg_NUM sizeof(CTSandNReg)/sizeof(HDAuioCTSandNRegSetting)
#define WallClockReg_NUM sizeof(WallClockReg)/sizeof(HDAuioWallClockRegSetting)
#define PacketClockReg_NUM  sizeof(PacketClockReg)/sizeof(HDAuioPacketClockRegSetting)

/*** Function ***/
extern _X_EXPORT Bool viaGetHDMIInfo(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo);
Bool viaIdentifyHDMI(VIABIOSInfoPtr pBIOSInfo);
void viaInitHDMI(VIABIOSInfoPtr pBIOSInfo);
void viaSetHDMIOutputPath(VIABIOSInfoPtr pBIOSInfo, int set_iga, 
                       int output_interface);
void viaEnableHDMI(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo);
void viaDisableHDMI(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo);
Bool viaSetHDMIMode(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo, int activeDevice);
CARD32 viaGetDeviceDetectHDMI(VIABIOSInfoPtr pBIOSInfo);
void viaAdjustTimingForHDMI(VIABIOSInfoPtr pBIOSInfo);
void VIAAdjustActiveSize_HDMI(VIABIOSInfoPtr pBIOSInfo, HDMISETTINGINFOPTR pHDMISettingInfo,
                                   int *pHActive, int *pVActive, int whichHDMI);
Bool VIASetHDMI3DScalScreenParas(VIABIOSInfoPtr  pBIOSInfo);
#endif

