/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include "via_driver.h"
#include "via_hw.h"
#include "via_tv.h"
#include "via_tblDPASetting.h"
#include "via_gpioi2c.h"

extern Bool vt1625_module_loaded;
extern Bool integratedtv_module_loaded;

Bool IsTVMode(VIABIOSInfoPtr pBIOSInfo, int ModeIndex, TVSETTINGINFOPTR pTVSettingInfo)
{
    BOOL IsTVMode = FALSE;
    
    switch (pTVSettingInfo->TVEncoder)
    {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                IsTVMode = IsVT1625TVMode(ModeIndex); 
            break;
            
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                IsTVMode = IsIntegratedTVMode(pBIOSInfo, ModeIndex, pTVSettingInfo);
            break;
    }
    return IsTVMode;
}

void VIATVSetMode(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    BOOL    setTV = FALSE;
    int     resMode;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIATVSetMode!!\n"));

    resMode = pTVSettingInfo->ModeIndex;
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIATVSetMode: modeindex= %d !!\n",resMode));

    unlock_crt();

    /* Found TV resmod index */
    switch (resMode)
    {
        case M640x480:
            pTVSettingInfo->resTVMode = VIA_640X480;
            break;
        case M800x600:
            pTVSettingInfo->resTVMode = VIA_800X600;
            break;
        case M1024x768:
            pTVSettingInfo->resTVMode = VIA_1024X768;
            break;
        case M848x480:
            pTVSettingInfo->resTVMode = VIA_848X480;
            break;
        case M720x480:
            pTVSettingInfo->resTVMode = VIA_720X480;
            if (pTVSettingInfo->TVType == TVTYPE_480P ||
                pTVSettingInfo->TVType == TVTYPE_720P ||
                pTVSettingInfo->TVType == TVTYPE_1080I)
            {
                /* SDTV only support RGB & YPBPR output signal type */
                if((pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR) ||
                   (pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO))
                {
                    pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR;
                }
                else
                {
                    pTVSettingInfo->TVOutput = TV_OUTPUT_RGB;
                }
            }
            else
            {
                pTVSettingInfo->TVType = TVTYPE_NTSC;
            }
            break;
        case M720x576:
            pTVSettingInfo->resTVMode = VIA_720X576;
            /* Bug fix: SV-VT3259-Linux-071(CR63643): CRT+HDTV Simultaneous: Set TV output format as 576P */
            /* then change resolution to 720x576, caused CRT, TV, HDTV icons all be enabled. (CN400 Mandrake9.2) */
            /* Reason: Use wrong TV type to check the active state of TV and HDTV when utility query UT_XV_FUNC_DEVICE_GetActiveState */
            /* to get active state. */
            /* Solution: Don't change TV type from 576P to PAL in mode 720X576 for HDTV case. */
            if (pTVSettingInfo->TVType == TVTYPE_576P ||
                pTVSettingInfo->TVType == TVTYPE_720P || 
                pTVSettingInfo->TVType == TVTYPE_1080I)
            {
                /* SDTV only support RGB & YPBPR output signal type */
                if((pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR) ||
                   (pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO))
                {
                    pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR;
                }
                else
                {
                    pTVSettingInfo->TVOutput = TV_OUTPUT_RGB;
                }
            }
            else
            {
                pTVSettingInfo->TVType = TVTYPE_PAL;
            }
            break;
        case M1280x720:
            pTVSettingInfo->resTVMode = VIA_1280X720;
            pTVSettingInfo->TVType = TVTYPE_720P;
            /* HDTV only support RGB & YPBPR output signal type */
            if((pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR) ||
               (pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO))
            {
                pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR;
            }
            else
            {
                pTVSettingInfo->TVOutput = TV_OUTPUT_RGB;
            }
            break;
        case M1920x1080:
            pTVSettingInfo->resTVMode = VIA_1920X1080;
            pTVSettingInfo->TVType = TVTYPE_1080I;
            /* HDTV only support RGB & YPBPR output signal type */
            if((pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR) ||
               (pTVSettingInfo->TVOutput == TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO))
            {
                pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR;
            }
            else
            {
                pTVSettingInfo->TVOutput = TV_OUTPUT_RGB;
            }
            break;
        default :
            pTVSettingInfo->resTVMode = resMode;
            ErrorF("No TV suitable mode found!! mode:%d\n", resMode);
            break;
    }

    /*if now TVType still none, we give a default value*/
    if(!pTVSettingInfo->TVType)
        pTVSettingInfo->TVType = TVTYPE_NTSC;
    
    switch (pTVSettingInfo->TVEncoder)
    {
        case VIA_VT1625:
            if (IsTVMode(pBIOSInfo, resMode, pTVSettingInfo))
            {
                setTV = TRUE;

                if (vt1625_module_loaded)
                    load_vt1625_regs(pBIOSInfo, pTVSettingInfo);
            }
            break;

        case VIA_INTEGRATED_TV:
            if (IsTVMode(pBIOSInfo, resMode, pTVSettingInfo))
            {
                setTV = TRUE;

                if (integratedtv_module_loaded)
                    SetIntegratedTVMode(pBIOSInfo, pTVSettingInfo);
            }
            
            break;
            
    }

    if (setTV)
    {
        /* Adjust clock skew. */
        VIATVPatch(pBIOSInfo, pTVSettingInfo);
        
        VIAEnableTV(pBIOSInfo, pTVSettingInfo);

        /* get driver default TV setting */
        UTGetTVInfo(pBIOSInfo, pTVSettingInfo);
		
        /*For Zinwell: support overscan on Internal TV with NTSC/PAL.
        	At last, save tv setting.*/
        if(pTVSettingInfo->TVEncoder == VIA_INTEGRATED_TV)
        {
            Bool IsDISP3DScaling = FALSE;
            IsDISP3DScaling = (pTVSettingInfo->IGAPath == IGA1)? 
                                  pBIOSInfo->IGA1SettingInfo.IsDISP3DScaling:
                                  pBIOSInfo->IGA2SettingInfo.IsDISP3DScaling; 

            if(IsDISP3DScaling)
            {
                if(pTVSettingInfo->TVType==TVTYPE_NTSC)
                {      
                    pTVSettingInfo->TVInfo.CurrentScalH = 4;
                    pTVSettingInfo->TVInfo.CurrentScalV = 7;
                    pTVSettingInfo->TVInfo.CurrentPositionH = 17;
                    pTVSettingInfo->TVInfo.CurrentPositionV = 7;
                    AdjustTVSize_IntegratedTV(pBIOSInfo, pTVSettingInfo, 12, 15);
                    AdjustTVPos_IntegratedTV(pBIOSInfo, pTVSettingInfo, 4, 12); 
                } else if(pTVSettingInfo->TVType==TVTYPE_PAL){
                    pTVSettingInfo->TVInfo.CurrentScalH = 10;
                    pTVSettingInfo->TVInfo.CurrentScalV = 7;
                    pTVSettingInfo->TVInfo.CurrentPositionH = 9;
                    pTVSettingInfo->TVInfo.CurrentPositionV = 7;
                    AdjustTVSize_IntegratedTV(pBIOSInfo, pTVSettingInfo, 17, 22);
                    AdjustTVPos_IntegratedTV(pBIOSInfo, pTVSettingInfo, 5, 16);
                } else {
                    pTVSettingInfo->TVInfo.CurrentScalH = 10;
                    pTVSettingInfo->TVInfo.CurrentScalV = 7;
                    pTVSettingInfo->TVInfo.CurrentPositionH = 9;
                    pTVSettingInfo->TVInfo.CurrentPositionV = 7;                
                    AdjustTVSize_IntegratedTV(pBIOSInfo, pTVSettingInfo, 21, 15);
                    AdjustTVPos_IntegratedTV(pBIOSInfo, pTVSettingInfo, 11, 12);
                }
            }
        }

        VIASaveUserTVSetting(pBIOSInfo, pTVSettingInfo, "/.VIATVRC");
	
        /* restore utility user TV setting */
        VIARestoreUserTVSetting(pBIOSInfo, pTVSettingInfo);        

        if (pTVSettingInfo->IGAPath == IGA2)
        {
            load_fetch_count_reg(pTVSettingInfo->HActive, (pBIOSInfo->bitsPerPixel>>3),IGA2);
            load_offset_reg(pTVSettingInfo->HActive,(pBIOSInfo->bitsPerPixel>>3),IGA2);
            
            load_FIFO_reg(pBIOSInfo, pBIOSInfo->CrtcHDisplay, pBIOSInfo->CrtcVDisplay, IGA2);
        }

	/*set iga color depth*/
		set_color_depth(pBIOSInfo, pBIOSInfo->bitsPerPixel>>3, pTVSettingInfo->IGAPath);
	
        if (pBIOSInfo->TVSettingInfo.TVTypeIsNTSCJ == TRUE)
            GPIOI2CWrite_TV(pBIOSInfo, 0x0B, 0);
    }
}

void VIASetTVOutputPath(VIABIOSInfoPtr pBIOSInfo, int set_iga, int output_interface)
{
    switch(set_iga)
    {
        case IGA1:
            /* Set IGA1 display channel to normal case beside CLE266.AX chip */
            /* and disable simultaneous display. */
            write_reg_mask(CR6B, VIACR, 0x00, BIT3+BIT6+BIT7);
            break;
        case IGA2:
            /* Set IGA2 display channel to normal case beside CLE266.AX chip */
            /* and disable simultaneous display. */
            /* Modify for sequence issue. */
            EnableSecondDisplayChannel();

            write_reg_mask(CR6B, VIACR, 0x00, BIT3+BIT6+BIT7);

            /* Disable LCD Scaling */
            write_reg_mask(CR79, VIACR, 0x00, BIT0);
            break;
    }

    switch(output_interface)
    {
        case VIA_DI_DVP0:
            if (set_iga== IGA1)
            {
                write_reg_mask(CR6C, VIACR, 0x00, BIT7);
            }
            else
            {
                write_reg_mask(CR6C, VIACR, 0x80, BIT7);
            }

            write_reg_mask(CR6B, VIACR, 0x00, BIT0);
            write_reg_mask(CR6C, VIACR, 0x21, BIT0+BIT5);
            write_reg_mask(SR1E, VIASR, 0xC0, BIT6+BIT7);
            break;
        case VIA_DI_DVP1:
            if (set_iga== IGA1)
            {
                write_reg_mask(CR6C, VIACR, 0x00, BIT7);
                write_reg_mask(CR9B, VIACR, 0x00, BIT4);
            }
            else
            {
                write_reg_mask(CR6C, VIACR, 0x80, BIT7);
                write_reg_mask(CR9B, VIACR, 0x10, BIT4);
            }

            write_reg_mask(CR6C, VIACR, 0x21, BIT0+BIT5);
            write_reg_mask(SR1E, VIASR, 0x30, BIT4+BIT5);
            break;
        case VIA_DI_DFPHIGH:
            if (set_iga== IGA1)
            {
                write_reg_mask(CR6C, VIACR, 0x00, BIT7);
                write_reg_mask(CR96, VIACR, 0x00, BIT4);
                write_reg_mask(CR97, VIACR, 0x00, BIT4);
            }
            else
            {
                write_reg_mask(CR6C, VIACR, 0x80, BIT7);
                write_reg_mask(CR96, VIACR, 0x10, BIT4);
                write_reg_mask(CR97, VIACR, 0x10, BIT4);
            }

            write_reg_mask(CR6C, VIACR, 0x21, BIT0+BIT5);
            write_reg_mask(SR2A, VIASR, 0x0C, BIT2+BIT3);
            break;
        case VIA_DI_DFPLOW:
            if (set_iga== IGA1)
            {
                write_reg_mask(CR6C, VIACR, 0x00, BIT7);
                write_reg_mask(CR9B, VIACR, 0x00, BIT4);
                write_reg_mask(CR99, VIACR, 0x00, BIT4);
            }
            else
            {
                write_reg_mask(CR6C, VIACR, 0x80, BIT7);
                write_reg_mask(CR9B, VIACR, 0x10, BIT4);
                write_reg_mask(CR99, VIACR, 0x10, BIT4);
            }

            write_reg_mask(CR6C, VIACR, 0x21, BIT0+BIT5);
            write_reg_mask(SR2A, VIASR, 0x03, BIT0+BIT1);
            break;
    }
}


void VIATVSetGfxDPA(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, TV_DPA_TABLE_PTR pTVDPATable)
{
    GFX_DPA_VALUE_PTR pGfxDPASettingValue = NULL;
    GFX_DPA_TVTYPE_MAP_TABLE_PTR pGfxDPATVTypeMapTable = NULL;
    GFX_DPA_TVSIZE_MAP_TABLE_PTR pGfxDPATVSizeMapTable = NULL;
    
    if (pTVDPATable)
    {
        pGfxDPATVTypeMapTable = pTVDPATable->pGfxDPATVTypeMapTable;
    }

    if (pGfxDPATVTypeMapTable)
    {
        switch (pTVSettingInfo->TVType)
        {
            case TVTYPE_NTSC:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->StdNTSC);
                break;

            case TVTYPE_PAL:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->StdPAL);
                break;

            case TVTYPE_480P:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->Std480p);
                break;            
                
            case TVTYPE_576P:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->Std576P);
                break;

            case TVTYPE_720P:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->Std720p);
                break;
                
            case TVTYPE_1080I:
                pGfxDPATVSizeMapTable = &(pGfxDPATVTypeMapTable->Std1080i);
                break;

            default:
                break;
        }
    }

    if (pGfxDPATVSizeMapTable)
    {
        switch (pTVSettingInfo->TVVScan)
        {
            case VIA_TVNORMAL:
                pGfxDPASettingValue = pGfxDPATVSizeMapTable->pNormalscan;
                break;

            case VIA_TVFIT:
                pGfxDPASettingValue = pGfxDPATVSizeMapTable->pFitscan;
                break;

            case VIA_TVOVER:
                pGfxDPASettingValue = pGfxDPATVSizeMapTable->pOverscan;
                break;

            default:
                break;
        }
    }

    if (pGfxDPASettingValue)
    {
        VIASetDPA_Gfx(pBIOSInfo->TVDIPort, pGfxDPASettingValue);
    }
}


void VIATVSetTVEncoderDPA(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, TV_DPA_TABLE_PTR pTVDPATable)
{
    TV_DPA_TVTYPE_MAP_TABLE_PTR pTVDPATVTypeMapTable = NULL;
    TV_DPA_TVSIZE_MAP_TABLE_PTR pTVDPATVSizeMapTable = NULL;
    CARD32 *pTVDPASettingValue = NULL;
    int index;
    CARD8 data, mask, tmp_data;
    
    if (pTVDPATable)
    {
        pTVDPATVTypeMapTable = pTVDPATable->pTVDPATVTypeMapTable;
    }

    if (pTVDPATVTypeMapTable)
    {
        switch (pTVSettingInfo->TVType)
        {
            case TVTYPE_NTSC:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->StdNTSC);
                break;

            case TVTYPE_PAL:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->StdPAL);
                break;

            case TVTYPE_480P:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->Std480p);
                break;            
                
            case TVTYPE_576P:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->Std576P);
                break;

            case TVTYPE_720P:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->Std720p);
                break;
                
            case TVTYPE_1080I:
                pTVDPATVSizeMapTable = &(pTVDPATVTypeMapTable->Std1080i);
                break;

            default:
                break;
        }
    }

    if (pTVDPATVSizeMapTable)
    {
        switch (pTVSettingInfo->TVVScan)
        {
            case VIA_TVNORMAL:
                pTVDPASettingValue = pTVDPATVSizeMapTable->pNormalscan;
                break;

            case VIA_TVFIT:
                pTVDPASettingValue = pTVDPATVSizeMapTable->pFitscan;
                break;

            case VIA_TVOVER:
                pTVDPASettingValue = pTVDPATVSizeMapTable->pOverscan;
                break;

            default:
                break;
        }
    }

    if (pTVDPASettingValue)
    {
        while (*pTVDPASettingValue != 0xFFFFFFFF)
        {
            index = *pTVDPASettingValue & 0xFF;
            mask = (*pTVDPASettingValue >> 8) & 0xFF;
            data = (*pTVDPASettingValue >> 16) & 0xFF;
            
            tmp_data = GPIOI2CRead_TV(pBIOSInfo, index);
            tmp_data = (tmp_data & (~mask));
            GPIOI2CWrite_TV(pBIOSInfo, index, tmp_data | data);

            pTVDPASettingValue++;
        }
    }
}


void VIATVPatch(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    switch(pTVSettingInfo->TVEncoder)
    {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                VIATVPatchSkew_VT1625(pBIOSInfo, pTVSettingInfo);
            break;
            
        default:
            break;
    }

    /* Replace the default patch if get user DPA settings from xorg: */
    if (pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx)
    {
        VIASetDPA_Gfx(pBIOSInfo->TVDIPort, &(pBIOSInfo->UserDPASettingInfo.GfxDPA));
    }

    if (pBIOSInfo->UserDPASettingInfo.HasUserSetting_VT1625 && (pTVSettingInfo->TVEncoder == VIA_VT1625))
    {
        if (vt1625_module_loaded)
            VIASetDPA_VT1625(pBIOSInfo, pBIOSInfo->UserDPASettingInfo.VT1625DPA);
    }
    
}

/*=*
 *
 * int VIACheckTVExist(VIABIOSInfoPtr) - Check TV Endcoder
 *
 * Return Type:    int
 *
 * The Definition of Input Value:
 *
 *                 VIABIOSInfoPtr
 *
 * The Definition of Return Value:
 *
 *                 0 - None
 *                 4 - Chrontel 7019
 *=*/
int VIACheckTVExist(VIABIOSInfoPtr pBIOSInfo)
{
    unsigned char   sr1e = 0, sr3e = 0, cr6c = 0;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIACheckTVExist\n"));
    

/*User can specify tv encoder from XServer configure file*/
    if (vt1625_module_loaded) {
        TVEncoderIdentify_VT1625(pBIOSInfo, &(pBIOSInfo->TVSettingInfo));
        if (pBIOSInfo->TVSettingInfo.TVEncoder == VIA_VT1625)
        {           
            InitTV_VT1625(pBIOSInfo);

            /* If TV2 is exist, we should keep finding second TV encoder. */
            if (!(pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2))
                return pBIOSInfo->TVSettingInfo.TVEncoder;
        }
    }    

    /*=* DVP0/DVP1 Pad on *=*/
    sr1e = read_reg(VIASR, SR1E);
    write_reg_mask(SR1E, VIASR, 0xF0, BIT4+BIT5+BIT6+BIT7);

    /*=* SR3E[1]Multi-function selection:
                0 = Emulate I2C and DDC bus by GPIO2/3/4. *=*/
    sr3e = read_reg(VIASR, SR3E);
    write_reg_mask(SR3E, VIASR, 0x0, BIT5);
    /*=* CR6C[0]TV Clock Input Enable *=*/
    cr6c = read_reg(VIACR, CR6C);
    write_reg_mask(CR6C, VIACR, 0x21, BIT0+BIT5);

    /*=* Due to TV3 has 3 level H-Scal, others only has
     *   2 level. If not TV3, we let overscan equal Fit for coding convenience.
     *=*/
    if (pBIOSInfo->TVSettingInfo.TVVScan == VIA_TVOVER &&
        pBIOSInfo->TVSettingInfo.TVEncoder != VIA_VT1625)
    {
    xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED,"VIACheckTVExist:TVVScan=%d!\n", pBIOSInfo->TVSettingInfo.TVVScan);    
        pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVFIT;
    }

    write_reg(SR1E, VIASR, sr1e);
    write_reg(SR3E, VIASR, sr3e);
    write_reg(CR6C, VIACR, cr6c);

    if (pBIOSInfo->TVSettingInfo.TVEncoder == VIA_NONETV)
    {
        if (integratedtv_module_loaded) {
            if (TVEncoderIdentify_IntegratedTV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo)))
            {
                return pBIOSInfo->TVSettingInfo.TVEncoder;
            }
        }    
    }

    if ((pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2) && integratedtv_module_loaded)
    {        
        TVEncoderIdentify_IntegratedTV(pBIOSInfo, &(pBIOSInfo->TVSettingInfo2));
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED,
                "TV2 Encoder:%d!\n", pBIOSInfo->TVSettingInfo2.TVEncoder));
    }    
    
    return pBIOSInfo->TVSettingInfo.TVEncoder;
}

void load_tv_output_signal(VIABIOSInfoPtr pBIOSInfo, CARD16 *tv_func_reg, TVSETTINGINFOPTR pTVSettingInfo)
{
    int             i = 0;

    while(tv_func_reg[i]!= 0xFFFF)
    {
        if(pTVSettingInfo->TVEncoder == VIA_VT1625)
        {
            GPIOI2CWrite_TV(pBIOSInfo, GET_LOW_BYTE(tv_func_reg[i]), GET_HIGH_BYTE(tv_func_reg[i]));
        }
        else
        {
            GPIOI2C_Write(pBIOSInfo, GET_LOW_BYTE(tv_func_reg[i]), GET_HIGH_BYTE(tv_func_reg[i]));
        }
        DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "Write TV Reg Offset %#x value %#x!\n" , GET_LOW_BYTE(tv_func_reg[i]), GET_HIGH_BYTE(tv_func_reg[i])));
        i++;
    }
}


void VIAEnableTV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    VIABIOSInfoPtr  pVia = pBIOSInfo;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "VIAEnableTV\n"));

    if ((pTVSettingInfo->TVEncoder == VIA_INTEGRATED_TV) && integratedtv_module_loaded)
    {       
        EnableTV_IntegratedTV(pBIOSInfo, ITV_DAC_PRIMARY);
    }
    else
    {
        if ((pTVSettingInfo->TVEncoder == VIA_VT1625) && vt1625_module_loaded)
        {
            EnableTV_VT1625(pBIOSInfo, pTVSettingInfo);
            pTVSettingInfo->TVAutoSense = TRUE;                    
        }

        /* Pad on/off according PMS */
        if (pBIOSInfo->TVDIPort & VIA_DI_DVP0)
        {
            VGAOUT8(0x3c4, 0x1E);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) | 0xC0);
        }
        
        if (pBIOSInfo->TVDIPort & VIA_DI_DVP1)
        {
            VGAOUT8(0x3c4, 0x1E);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) | 0x30);
        }
        
        if (pBIOSInfo->TVDIPort & VIA_DI_DFPHIGH)
        {
            VGAOUT8(0x3c4, 0x2A);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) | 0x0C);
        }
        
        if (pBIOSInfo->TVDIPort & VIA_DI_DFPLOW)
        {
            VGAOUT8(0x3c4, 0x2A);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) | 0x0C);
        }
    }
}

void VIADisableTV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    VIABIOSInfoPtr  pVia = pBIOSInfo;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "VIADisableTV\n"));

    if ((pTVSettingInfo->TVEncoder == VIA_INTEGRATED_TV) && integratedtv_module_loaded)
    {
        if (pBIOSInfo->Chipset == VIA_CX700)
        {
            DisableTV_IntegratedTV(pBIOSInfo, ITV_DAC_PRIMARY);
        }
    }
    else
    {
        /* Digital Interface Pad off */
        if (pBIOSInfo->TVDIPort & VIA_DI_DVP0)
        {
            VGAOUT8(0x3c4, 0x1E);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0x3F);
        }
        if (pBIOSInfo->TVDIPort & VIA_DI_DVP1)
        {
            VGAOUT8(0x3c4, 0x1E);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xCF);
        }
        if (pBIOSInfo->TVDIPort & VIA_DI_DFPHIGH)
        {
            VGAOUT8(0x3c4, 0x2A);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xF3);
        }
        if (pBIOSInfo->TVDIPort & VIA_DI_DFPLOW)
        {
            VGAOUT8(0x3c4, 0x2A);
            VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xFC);
        }

        if ((pTVSettingInfo->TVEncoder == VIA_VT1625) && vt1625_module_loaded)
        {
            DisableTV_VT1625(pBIOSInfo);
            pTVSettingInfo->TVAutoSense = FALSE;
        }

    }
}


/* Sense TV to determine whether TV is connect and choice the output signal. */
unsigned char VIAGetDeviceDetectTV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    unsigned char   ret = 0x0, sense = 0;
    int DACSelect  = 0; /* for integrated tv */
        
    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIAGetDeviceDetectTV\n"));

    switch(pTVSettingInfo->TVEncoder)
    {
        case VIA_NONETV:
            break;

        case VIA_VT1625:
            if (vt1625_module_loaded)
                sense = VIASensorVT1625(pBIOSInfo, pTVSettingInfo->TVAutoSense);

            if (!pTVSettingInfo->TVOutput && sense == 0x3F)
            {
            }
            else
            {
                ret = VIA_DEVICE_TV;
                if (!pTVSettingInfo->TVOutput)
                {
                    switch (sense)
                    {
                        case 0x07:  /* DAC A, B, C */
                            pTVSettingInfo->TVOutput = TV_OUTPUT_COMPOSITE_SVIDEO;
                            break;
                        case 0x0F:  /* DAC A, B */
                            pTVSettingInfo->TVOutput = TV_OUTPUT_SVIDEO;
                            break;
                        case 0x38:  /* DAC D, E, F */
                            pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR;
                            break;
                        case 0x37:  /* DAC C */
                            pTVSettingInfo->TVOutput = TV_OUTPUT_COMPOSITE;
                            break;
                        case 0x0:  /* DAC A, B, C, D, E, F */
                        case 0x10:  /* S-Video C may not be detected. */
                            pTVSettingInfo->TVOutput = TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO;
                            break;
                        default:
                            ErrorF("Unknown signal type!\n");
                            pTVSettingInfo->TVOutput = TV_OUTPUT_COMPOSITE;
                            break;
                    }
                }
                DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_PROBED, "VT1625 sense = %x ,TVOutput = %d\n", sense, pTVSettingInfo->TVOutput));
            }
            ret |= VIA_DEVICE_CRT2; /*ian*/            
            break;

        case VIA_INTEGRATED_TV:
            /* Determine which DAC we should use according to chipset. */
            if (pBIOSInfo->Chipset == VIA_CX700)
            {
                DACSelect = ITV_DAC_PRIMARY;
            }

            /* start to sense */ 
            if (integratedtv_module_loaded) {
                if (SenseIntegratedTV(pBIOSInfo, DACSelect, pTVSettingInfo) == ITV_VALUE_TV_CONNECTED)
                {
                    ret = VIA_DEVICE_TV;
                }
                else
                {
                }
            }    
            
            break;
    }
    
    return ret;
}

void VIASaveTV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, VIARegPtr save)
{

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_NONETV:
            break;

        case VIA_INTEGRATED_TV:
            /* It's better to save the original values before initialize, so */
            /* we move the call SaveTVRegs_IntegratedTV() before InitIntegratedTVRegs(). */
            break;

        case VIA_VT1625:
            if (vt1625_module_loaded)
                SaveTVRegs_VT1625(pBIOSInfo, save);
            break;
           
    }

}

void VIARestoreTV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, VIARegPtr restore)
{
    switch (pTVSettingInfo->TVEncoder)
    {          
        case VIA_VT1625:
            if (vt1625_module_loaded)
                RestoreTVRegs_VT1625(pBIOSInfo, restore);
            break;
            
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                RestoreTVRegs_IntegratedTV(pBIOSInfo, restore);
            break;
    }
}

/* Add this function to control TV clock source by case. */
void SetTVClockSource(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{   
    if (pTVSettingInfo->TVEncoder == VIA_INTEGRATED_TV)
    {
        /* From TVPLL. */
        if (pTVSettingInfo->IGAPath == IGA1)
        {
            write_reg_mask(CR6C, VIACR, 0x50, 0xF0);
        }
        else
        {
            write_reg_mask(CR6C, VIACR, 0x05, 0x0F);
        }
    }
    else
    {
        /* External TV: */
        /* Correct CX700 External TV output port from DVP0 to DVP1 */
        switch(pBIOSInfo->Chipset)
        {
            case VIA_CX700:
            case VIA_VX800:
			case VIA_VX855:
            case VIA_VX900:
                if (pTVSettingInfo->IGAPath == IGA1)
                {
                    if(pBIOSInfo->TVDIPort & VIA_DI_DVP1)
                    {
                        write_reg_mask(CR6C, VIACR, 0xB0, 0xF0);
                    }
                    else if(pBIOSInfo->TVDIPort & VIA_DI_DVP0)
                    {
                        write_reg_mask(CR6C, VIACR, 0x90, 0xF0);
                    }
                }
                else
                {
                    if(pBIOSInfo->TVDIPort & VIA_DI_DVP1)
                    {
                        write_reg_mask(CR6C, VIACR, 0x0b, 0x0f);
                    }
                    else if(pBIOSInfo->TVDIPort & VIA_DI_DVP0)
                    {
                        write_reg_mask(CR6C, VIACR, 0x09, 0x0f);
                    }
                }               
                break;              
            default:
                if (pTVSettingInfo->IGAPath == IGA1)
                {
                    write_reg_mask(CR6C, VIACR, BIT5+BIT0, BIT5+BIT0);
                }
                else
                {
                    write_reg_mask(CR6C, VIACR, BIT7+BIT5+BIT0, BIT7+BIT5+BIT0);
                }    
        }    
    }
}

void DoTVSWReset(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    /* Do TV Software Reset can help TV image sync well with IGA1. */
    if ((pTVSettingInfo->TVEncoder == VIA_INTEGRATED_TV) && integratedtv_module_loaded)
    {
        DoTVSWReset_IntegratedTV(pBIOSInfo);
    }
    else if ((pTVSettingInfo->TVEncoder == VIA_VT1625) && vt1625_module_loaded)
    {
        DoTVSWReset_VT1625(pBIOSInfo);
    }
}


void 
VIAAdjustTVPos(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, 
             CARD32 desired_hpos, CARD32 desired_vpos)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVPos_VT1625(pBIOSInfo, pTVSettingInfo, 
                               desired_hpos, desired_vpos);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVPos_IntegratedTV(pBIOSInfo, pTVSettingInfo,
                                     desired_hpos, desired_vpos);                                        
            break;
        default:
            pTVInfo->CurrentPositionH = 0;
            pTVInfo->CurrentPositionV = 0;
            break;
    }    
}


void 
VIAAdjustTVBrightness(VIABIOSInfoPtr pBIOSInfo,  TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* Set default Brightness */
    if (desired_value == 0xFFFF)
        desired_value = pTVInfo->DefaultBrightness;

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVBrightness_VT1625(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVBrightness_IntegratedTV(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        default:
            pTVInfo->CurrentBrightness = 0;
            break;
    }
}

void 
VIAAdjustTVContrast(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* Set default Contrast */
    if (desired_value == 0xFFFF)
        desired_value = pTVInfo->DefaultContrast;

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVContrast_VT1625(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVContrast_IntegratedTV(pBIOSInfo,  pTVSettingInfo, desired_value);
            break;
        default:
            pTVInfo->CurrentContrast = 0;
            break;
    }
}


void 
VIAAdjustTVSaturation(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* Set default Saturation */
    if (desired_value == 0xFFFF)
        desired_value = pTVInfo->DefaultSaturation;

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVSaturation_VT1625(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVSaturation_IntegratedTV(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        default:
            pTVInfo->CurrentSaturation = 0;
            break;
    }
}


void 
VIAAdjustTVHue(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* Set default TINT */
    if (desired_value == 0xFFFF)
        desired_value = pTVInfo->DefaultTINT;

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVHue_VT1625(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVHue_IntegratedTV(pBIOSInfo, pTVSettingInfo, desired_value);
            break;
        default:
            pTVInfo->CurrentTINT = 0;
            break;
    }
}

void 
VIAAdjustTVFFilter(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVFFilter_VT1625(pBIOSInfo, pTVSettingInfo, VT1625_FILTER_TYPE_NORMAL, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVFFilter_IntegratedTV(pBIOSInfo, pTVSettingInfo, ITV_FILTER_TYPE_NORMAL, desired_value);
            break;
        default:
            pTVInfo->CurrentFFilter = 0;
            break;
    }
}

void 
VIAEnableTVFFilter(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* FFliter on, AFFliter off */
    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                EnableTVFFilter_VT1625(pBIOSInfo, pTVSettingInfo, VT1625_FILTER_TYPE_NORMAL);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                EnableTVFFilter_IntegratedTV(pBIOSInfo, ITV_FILTER_TYPE_NORMAL);
            break;
        default:
            pTVInfo->CurrentAFFilter = 0;
            break;
    }
    pTVInfo->AdaptiveFFilterOn = FALSE;
}


void 
VIAAdjustTVAFFilter(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, CARD32 desired_value)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    switch (pTVSettingInfo->TVEncoder) {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                AdjustTVFFilter_VT1625(pBIOSInfo, pTVSettingInfo, VT1625_FILTER_TYPE_ADAPTIVE, desired_value);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                AdjustTVFFilter_IntegratedTV(pBIOSInfo, pTVSettingInfo, ITV_FILTER_TYPE_ADAPTIVE, desired_value);
            break;
        default:
            pTVInfo->CurrentAFFilter = 0;
            break;
    }
}


void 
VIAEnableTVAFFilter(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* FFliter off, AFFliter on */
    switch (pTVSettingInfo->TVEncoder){
        case VIA_VT1625:
            if (vt1625_module_loaded) {
                EnableTVFFilter_VT1625(pBIOSInfo, pTVSettingInfo, VT1625_FILTER_TYPE_ADAPTIVE);
                pTVInfo->AdaptiveFFilterOn = TRUE;
            }
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded) {
                EnableTVFFilter_IntegratedTV(pBIOSInfo, ITV_FILTER_TYPE_ADAPTIVE);
                pTVInfo->AdaptiveFFilterOn = TRUE;
            }    
            break;
        default:
            pTVInfo->AdaptiveFFilterOn = FALSE;
            pTVInfo->CurrentAFFilter = 0;
            break;
    }
}

/* This function is used by utility ,when user switch the dot crawl status */
void 
VIASetTVDotCrawl(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    if ((pTVSettingInfo->TVEncoder == VIA_VT1625) && vt1625_module_loaded) 
        load_vt1625_regs(pBIOSInfo, pTVSettingInfo);

    VIATVPatch(pBIOSInfo, pTVSettingInfo);
    VIAEnableTV(pBIOSInfo, pTVSettingInfo);  
}


void 
UTGetTVInfo(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "UTGetTVInfo\n"));

    switch (pTVSettingInfo->TVEncoder){
        case VIA_VT1625:
            if (vt1625_module_loaded)
                GetTVInfo_VT1625(pBIOSInfo, pTVSettingInfo);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                GetTVInfo_IntegratedTV(pBIOSInfo, pTVSettingInfo);
            break;
        default:
            pTVInfo->DefaultScalH = 0;
            pTVInfo->DefaultScalV = VIA_TVNORMAL;
            pTVInfo->ScalHLevel = 0;
            pTVInfo->ScalVLevel = 2;
            pTVInfo->DefaultPositionV = 0;
            pTVInfo->PositionVLevel = 0;
            pTVInfo->DefaultPositionH = 0;
            pTVInfo->PositionHLevel = 0;
            pTVInfo->AdaptiveFFilterOn = FALSE;
            pTVInfo->DefaultFFilter = 0;
            pTVInfo->FFilterLevel = 0;
            pTVInfo->DefaultAFFilter = 0;
            pTVInfo->AFFilterLevel = 0;
            pTVInfo->DefaultBrightness = 0;
            pTVInfo->BrightnessLevel = 0;
            pTVInfo->DefaultContrast = 0;
            pTVInfo->ContrastLevel = 0;
            pTVInfo->DefaultSaturation = 0;
            pTVInfo->SaturationLevel = 0;
            pTVInfo->DefaultTINT = 0;
            pTVInfo->TINTLevel = 0;
            break;
    }

    if (!pTVInfo->HasUserSetting)
    {
        pTVInfo->CurrentScalH = pTVInfo->DefaultScalH;
        pTVInfo->CurrentScalV = pTVInfo->DefaultScalV;
        pTVInfo->CurrentPositionV = pTVInfo->DefaultPositionV;
        pTVInfo->CurrentPositionH = pTVInfo->DefaultPositionH;
        pTVInfo->CurrentFFilter = pTVInfo->DefaultFFilter;
        pTVInfo->CurrentAFFilter = pTVInfo->DefaultAFFilter;
        pTVInfo->CurrentBrightness = pTVInfo->DefaultBrightness;
        pTVInfo->CurrentContrast = pTVInfo->DefaultContrast;
        pTVInfo->CurrentSaturation = pTVInfo->DefaultSaturation;
        pTVInfo->CurrentTINT = pTVInfo->DefaultTINT;
        DEBUG(xf86Msg(X_INFO, "No user TV setting exist! Current setting equal default setting!\n"));
    }
}


void 
VIARestoreUserTVSetting(VIABIOSInfoPtr  pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);

    /* if no user setting, we do not need to restore. */
    if (!pTVInfo->HasUserSetting)
        return;

    switch (pTVSettingInfo->TVEncoder)
    {
        case VIA_VT1625:
            if (vt1625_module_loaded)
                RestoreUserTVSettings_VT1625(pBIOSInfo, pTVSettingInfo);
            break;
        case VIA_INTEGRATED_TV:
            if (integratedtv_module_loaded)
                RestoreUserTVSettings_IntegratedTV(pBIOSInfo, pTVSettingInfo);
            break;
    }
}


Bool 
VIASaveUserTVSetting(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo,
                   const char* pchFileName)
{
    char filename[40], tmpname[40], buffer[200], *home = NULL;
    Bool found = FALSE, ret = FALSE;
    FILE *VIATVRC=NULL, *VIAtmp=NULL;
    int  HDisplay, VDisplay, bitsPerPixel, tmp, TVStandard, TVOutput;

    DEBUG(xf86Msg(X_ERROR, "VIASaveUserTVSetting!\n"));    
    
    TVInfoPtr pTVInfo = &(pTVSettingInfo->TVInfo);
    
    filename[0] = '\0';
    found = FALSE;
    
    home="/etc/X11";
    strcpy(filename, home);
    strcat(filename, pchFileName);
    if ((VIATVRC = fopen(filename, "r")) == NULL)
    {
        xf86Msg(X_DEFAULT, "Open user TV setting file failed.\n");
        if ((VIATVRC = fopen(filename, "a")) == NULL)
        {
            DEBUG(xf86Msg(X_ERROR, "Create .VIATVRC failure!\n"));
            return FALSE;
        }
    }

    pTVInfo->HasUserSetting = TRUE;
    
    /* save user setting to temp file */
    strcpy(tmpname, home);
    strcat(tmpname, "/.VIAtmp");
        
    VIAtmp = fopen(tmpname, "w+");
    while (fgets(buffer, 200, VIATVRC) != NULL)
    {
        /* TV output should be a condition of saving TV advanced settings because */
        /* not all of TV outputs support the same TV advanced setting items. */
        sscanf(buffer,"%d %d %d %d %d %d %d %d", &HDisplay, &VDisplay, &bitsPerPixel, &tmp, &tmp, &tmp, &TVStandard, &TVOutput);
        
        if (HDisplay == pBIOSInfo->SaveHDisplay &&
            VDisplay == pBIOSInfo->SaveVDisplay &&
            bitsPerPixel == pBIOSInfo->bitsPerPixel &&
            TVStandard == pTVSettingInfo->TVType &&
            TVOutput == pTVSettingInfo->TVOutput)
        {
            fprintf(VIAtmp, "%d %d %d %d %lu %lu %x %x %lu %lu %lu %lu %lu %lu %lu %lu %d %lu %lu %lu %lu\n",
            pBIOSInfo->SaveHDisplay,
            pBIOSInfo->SaveVDisplay,
            pBIOSInfo->bitsPerPixel,
            pTVInfo->AdaptiveFFilterOn,
            pTVInfo->CurrentScalH,
            pTVInfo->CurrentScalV,
            pTVSettingInfo->TVType,
            pTVSettingInfo->TVOutput,
            pTVInfo->CurrentPositionH,
            pTVInfo->CurrentPositionV,
            pTVInfo->CurrentFFilter,
            pTVInfo->CurrentAFFilter,
            pTVInfo->CurrentBrightness,
            pTVInfo->CurrentContrast,
            pTVInfo->CurrentSaturation,
            pTVInfo->CurrentTINT,
            pTVSettingInfo->TVDotCrawl,
            pTVSettingInfo->TV3DScalInfo.XScaleLevel,
            pTVSettingInfo->TV3DScalInfo.YScaleLevel,
            pTVSettingInfo->TV3DScalInfo.XPOSITIONLevel,
            pTVSettingInfo->TV3DScalInfo.YPOSITIONLevel);            
            found = TRUE;
            xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "After save TV info: XScaleLevel=%ld,YScaleLevel=%ld,XPOSITIONLevel=%ld,YPOSITIONLevel=%ld\n",pTVSettingInfo->TV3DScalInfo.XScaleLevel,pTVSettingInfo->TV3DScalInfo.YScaleLevel,pTVSettingInfo->TV3DScalInfo.XPOSITIONLevel,pTVSettingInfo->TV3DScalInfo.YPOSITIONLevel);
        }
        else
        {            
            fputs(buffer, VIAtmp);
        }
    }
    if (!found)
    {        
        fprintf(VIAtmp, "%d %d %d %d %lu %lu %x %x %lu %lu %lu %lu %lu %lu %lu %lu %d %lu %lu %lu %lu\n",
                pBIOSInfo->SaveHDisplay,
                pBIOSInfo->SaveVDisplay,
                pBIOSInfo->bitsPerPixel,
                pTVInfo->AdaptiveFFilterOn,
                pTVInfo->CurrentScalH,
                pTVInfo->CurrentScalV,
                pTVSettingInfo->TVType,
                pTVSettingInfo->TVOutput,
                pTVInfo->CurrentPositionH,
                pTVInfo->CurrentPositionV,
                pTVInfo->CurrentFFilter,
                pTVInfo->CurrentAFFilter,
                pTVInfo->CurrentBrightness,
                pTVInfo->CurrentContrast,
                pTVInfo->CurrentSaturation,
                pTVInfo->CurrentTINT,
                pTVSettingInfo->TVDotCrawl,
                pTVSettingInfo->TV3DScalInfo.XScaleLevel,
                pTVSettingInfo->TV3DScalInfo.YScaleLevel,
                pTVSettingInfo->TV3DScalInfo.XPOSITIONLevel,
                pTVSettingInfo->TV3DScalInfo.YPOSITIONLevel);
    }

    fclose(VIATVRC);
    fclose(VIAtmp);
    if(remove(filename) != -1)
    {
        if(rename(tmpname, filename) == 0)
        {
            ret = TRUE;
        }
    }

    return ret;

}


Bool 
VIALoadUserTVSetting(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo,
                  const char* pchFileName)
{
    TVInfoPtr       pTVInfo = &(pTVSettingInfo->TVInfo);
    FILE *VIATVRC=NULL;
    char filename[40], buffer[200];
    int  HDisplay, VDisplay, bitsPerPixel, tmp, TVStandard, TVOutput;
    Bool            UT_TV_DOTCRAWL, found=FALSE;
    CARD32          UT_TV_STD=0, UT_TV_SGNL=0;

    DEBUG(xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "VIALoadUserTVSetting\n"));
    if (pTVSettingInfo->TVEncoder)
    {
        filename[0] = '\0';
        strcpy(filename, "/etc/X11");
        strcat(filename, pchFileName);
        if ((VIATVRC = fopen(filename, "r")) == NULL)
        {
            DEBUG(xf86Msg(X_WARNING, "No user TV setting file exist.\n"));
            return FALSE;
        }
        else
        {
            while (fgets(buffer, 200, VIATVRC) != NULL)
            {
                sscanf(buffer,"%d %d %d %d %d %d %d %d", &HDisplay, &VDisplay, &bitsPerPixel, &tmp, &tmp, &tmp, &TVStandard, &TVOutput);                    
                
                if (HDisplay == pBIOSInfo->SaveHDisplay &&
                    VDisplay == pBIOSInfo->SaveVDisplay &&
                    bitsPerPixel == pBIOSInfo->bitsPerPixel &&
                    TVStandard == pTVSettingInfo->TVType &&
                    TVOutput == pTVSettingInfo->TVOutput)
                {
                    sscanf(buffer,"%d %d %d %d %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %d %lu %lu %lu %lu",
                            &tmp, &tmp, &tmp, &pTVInfo->AdaptiveFFilterOn,
                            &pTVInfo->CurrentScalH, &pTVInfo->CurrentScalV,
                            &UT_TV_STD,
                            &UT_TV_SGNL,
                            &pTVInfo->CurrentPositionH,
                            &pTVInfo->CurrentPositionV,
                            &pTVInfo->CurrentFFilter,
                            &pTVInfo->CurrentAFFilter,
                            &pTVInfo->CurrentBrightness,
                            &pTVInfo->CurrentContrast,
                            &pTVInfo->CurrentSaturation,
                            &pTVInfo->CurrentTINT,
                            &UT_TV_DOTCRAWL,
                            &pTVInfo->XScalingLevel,
                            &pTVInfo->YScalingLevel,
                            &pTVInfo->XScalPosLevel,
                            &pTVInfo->YScalPosLevel);
                    xf86DrvMsg(pBIOSInfo->scrnIndex, X_INFO, "After load TV info: XScaleLevel=%lu,YScaleLevel=%lu,XPOSITIONLevel=%lu,YPOSITIONLevel=%lu\n",pTVInfo->XScalingLevel,pTVInfo->YScalingLevel,pTVInfo->XScalPosLevel,pTVInfo->YScalPosLevel);

                    pTVSettingInfo->TVVScan = pTVInfo->CurrentScalV;

                    pTVSettingInfo->TVType = UT_TV_STD;
                    pTVSettingInfo->TVOutput = UT_TV_SGNL;
                    pTVSettingInfo->TVDotCrawl = UT_TV_DOTCRAWL;
                    pTVInfo->HasUserSetting = TRUE;
                    found = TRUE;
                    DEBUG(xf86Msg(X_INFO, "Found user TV setting!!\n"));
                    break;
                }
            }
        }

        if (!found)
        {
            pTVInfo->HasUserSetting = FALSE;
            pTVInfo->AdaptiveFFilterOn = FALSE;
            pTVInfo->CurrentScalH = pTVInfo->DefaultScalH;
            pTVInfo->CurrentScalV = pTVInfo->DefaultScalV;
            pTVInfo->CurrentPositionV = pTVInfo->DefaultPositionV;
            pTVInfo->CurrentPositionH = pTVInfo->DefaultPositionH;
            pTVInfo->CurrentFFilter = pTVInfo->DefaultFFilter;
            pTVInfo->CurrentAFFilter = pTVInfo->DefaultAFFilter;
            pTVInfo->CurrentBrightness = pTVInfo->DefaultBrightness;
            pTVInfo->CurrentContrast = pTVInfo->DefaultContrast;
            pTVInfo->CurrentSaturation = pTVInfo->DefaultSaturation;
            pTVInfo->CurrentTINT = pTVInfo->DefaultTINT;
            DEBUG(xf86Msg(X_INFO, "No user TV setting exist!!\n"));
            fclose(VIATVRC);
            return FALSE;
        }

        fclose(VIATVRC);
        return TRUE;
    }
    return FALSE;
}


/* CheckMactovisionEnabled: to check if MV is enabled 
   according to TV encoder's registers 0x0F and 0x30
   Currently support VT1625 */
Bool 
CheckMacrovisionEnabled(VIABIOSInfoPtr  pBIOSInfo)
{
    unsigned char   R_Buffer[3];

    if (pBIOSInfo->TVSettingInfo.TVEncoder == VIA_VT1625)
    {
        R_Buffer[1] = GPIOI2CRead_TV(pBIOSInfo, 0x0F);
        R_Buffer[2] = GPIOI2CRead_TV(pBIOSInfo, 0x30);

        if((R_Buffer[1]&0x80) && (R_Buffer[2]&0x30))
            return TRUE;
    }

    return FALSE;  
}


void 
VIAAdjustActiveSize_TV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo, 
                    int* pHActive, int* pVActive, unsigned long whichTV)
{
    int HResToSet = pBIOSInfo->CrtcHDisplay;
    int VResToSet = pBIOSInfo->CrtcVDisplay;
    int MaxTVHRes, MaxTVVRes;
    
    /* When TV scaling function is enable, we neednt to take care of Panning. */
    if (pBIOSInfo->Is3DScalingEnable & whichTV)
    {
         pTVSettingInfo->IsPanning = FALSE; 
    }
    else
    {
    
        /* Try to find max size of TV: */
        pTVSettingInfo->ModeIndex = VIAGetModeIndex(HResToSet, VResToSet);

        if (!IsTVMode(pBIOSInfo, pTVSettingInfo->ModeIndex, pTVSettingInfo))
        {
            /* The mode is invalid, set a TV mode as default: */
            if ((*pHActive < 800) || (*pVActive < 600))
            {
                pTVSettingInfo->ModeIndex = M640x480;
                MaxTVHRes = 640;
                MaxTVVRes = 480;
            }
            else if ((*pHActive < 1024) || (*pVActive < 768))
            {
                pTVSettingInfo->ModeIndex = M800x600;
                MaxTVHRes = 800;
                MaxTVVRes = 600;
            }
            else
            {
                pTVSettingInfo->ModeIndex = M1024x768;
                MaxTVHRes = 1024;
                MaxTVVRes = 768;
            }

            /* Change view size if needed: */
            if ((HResToSet > MaxTVHRes) || (VResToSet > MaxTVVRes))
            {
                if ((*pHActive > MaxTVHRes) || (*pVActive > MaxTVVRes))
                {
                    *pHActive = MaxTVHRes;
                    *pVActive = MaxTVVRes;
                }

                pTVSettingInfo->IsPanning = TRUE;
            }
            else
            {
                pTVSettingInfo->IsPanning = FALSE;
            }
        }        
        else
        {
            pTVSettingInfo->IsPanning = FALSE;
        }

        if ((pBIOSInfo->VirtualX > HResToSet) || (pBIOSInfo->VirtualY > VResToSet))
        {
            /* If use Virtual Desktop, and the size is larger than the resolution, */
            /* we need to do panning, too. */
            pTVSettingInfo->IsPanning = TRUE;
        }
    }        

}


void 
VIAInitSettingInfo_TV(VIABIOSInfoPtr pBIOSInfo, TVSETTINGINFOPTR pTVSettingInfo,  CARD16 activeDevice)
{
    int     HActive, VActive;
    Bool    IsPanning = pTVSettingInfo->IsPanning;
	Bool    IsDownScaling = FALSE;


    if (IsPanning)
    {
        HActive = pBIOSInfo->HDisplay;
        VActive = pBIOSInfo->VDisplay;
    }
    else
    {
        HActive = pBIOSInfo->CrtcHDisplay;
        VActive = pBIOSInfo->CrtcVDisplay;
    }
    
    VIAInitSettingInfo_IGA (pBIOSInfo, pTVSettingInfo->IGAPath, IsPanning, IsDownScaling, HActive, VActive, activeDevice);

    pTVSettingInfo->HActive = HActive;
    pTVSettingInfo->VActive = VActive;
}

