#ifndef CS_FLAG_NO_XVMC
 /*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "via_driver.h" 
#include "compiler.h"
#include "xf86PciInfo.h"
#include "xf86Pci.h"
#include "xf86fbman.h"
#include "regionstr.h"

#include "xf86xv.h"
#include "xf86xvmc.h"
#include "Xv.h"
#include "XvMC.h"
#include "dixstruct.h"
#include "fourcc.h"
#include "via_video.h"
#include "via_xvmc.h"

#include "xf86.h"
#include "xf86_OSproc.h"
#if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 6 
#include "xf86Resources.h"
#endif


/*#define XVMC_DEBUG*/

void viaInitMC(ScreenPtr pScreen);

int viaXvMCCreateContext (ScrnInfoPtr pScrn, XvMCContextPtr pContext,
                           int *num_priv, long **priv );
void viaXvMCDestroyContext (ScrnInfoPtr pScrn, XvMCContextPtr pContext);

int viaXvMCCreateSurface (ScrnInfoPtr pScrn, XvMCSurfacePtr pSurf,
                           int *num_priv, long **priv );

void viaXvMCDestroySurface (ScrnInfoPtr pScrn, XvMCSurfacePtr pSurf);
int viaXvMCCreateSubpicture (ScrnInfoPtr pScrn, XvMCSubpicturePtr pSurf,
                               int *num_priv, long **priv );

void viaXvMCDestroySubpicture (ScrnInfoPtr pScrn, XvMCSubpicturePtr pSurf);

/*
 * Proposed XvMC via driver extension:
 */

#define XVMC_VLD 0x0020000
#define XVMC_WMV_9 0x00000009
#define XVMC_WVC_1 0x0000000a

typedef struct {
    unsigned int bviaXVMC_ON;
    unsigned int fbBase;
    unsigned int OverlayOffset;
    unsigned int OverlaySize;
    unsigned int SurfacesOffset;
    unsigned int SurfacesSize;
    char busIdString[10];
    char pad[2];
} viaXvMCCreateContextRec;


static int yv12_subpicture_index_list[1] = 
{
    FOURCC_AI44
};

static XF86MCImageIDList yv12_subpicture_list =
{
    1,
    yv12_subpicture_index_list
};
 
    /* test for xvmc support mpeg4 */
static XF86MCSurfaceInfoRec via_mpeg1_surface =
{
    FOURCC_VIA,
    XVMC_CHROMA_FORMAT_420,
    0,
    2048,
    2048,
    2048,
    2048,
    XVMC_MPEG_1,
    XVMC_OVERLAID_SURFACE | XVMC_BACKEND_SUBPICTURE | XVMC_INTRA_UNSIGNED/* | XVMC_SUBPICTURE_INDEPENDENT_SCALING*/,
    &yv12_subpicture_list
};

static XF86MCSurfaceInfoRec via_mpeg2_surface =
{
    FOURCC_VIA,
    XVMC_CHROMA_FORMAT_420,
    0,
    2048,
    2048,
    2048,
    2048,
    XVMC_MPEG_2 | XVMC_VLD,
    XVMC_OVERLAID_SURFACE | XVMC_BACKEND_SUBPICTURE | XVMC_INTRA_UNSIGNED/* | XVMC_SUBPICTURE_INDEPENDENT_SCALING*/,
    &yv12_subpicture_list
};

static XF86MCSurfaceInfoRec via_mpeg4_surface =
{
    FOURCC_VIA,
    XVMC_CHROMA_FORMAT_420,
    0,
    2048,
    2048,
    2048,
    2048,
    XVMC_MPEG_4 | XVMC_IDCT,
    XVMC_OVERLAID_SURFACE | XVMC_BACKEND_SUBPICTURE | XVMC_INTRA_UNSIGNED/* | XVMC_SUBPICTURE_INDEPENDENT_SCALING*/,
    &yv12_subpicture_list
};

static XF86MCSurfaceInfoRec via_wmv9_surface =
{
    FOURCC_VIA,
    XVMC_CHROMA_FORMAT_420,
    0,
    2048,
    2048,
    2048,
    2048,
    XVMC_WMV_9 | XVMC_IDCT,
    XVMC_OVERLAID_SURFACE | XVMC_BACKEND_SUBPICTURE | XVMC_INTRA_UNSIGNED/* | XVMC_SUBPICTURE_INDEPENDENT_SCALING*/,
    &yv12_subpicture_list
};

static XF86MCSurfaceInfoRec via_wvc1_surface =
{
    FOURCC_VIA,
    XVMC_CHROMA_FORMAT_420,
    0,
    2048,
    2048,
    2048,
    2048,
    XVMC_WVC_1 | XVMC_IDCT,
    XVMC_OVERLAID_SURFACE | XVMC_BACKEND_SUBPICTURE | XVMC_INTRA_UNSIGNED/* | XVMC_SUBPICTURE_INDEPENDENT_SCALING*/,
    &yv12_subpicture_list
};

static XF86MCSurfaceInfoPtr ppSI[5] =
{
    (XF86MCSurfaceInfoPtr)&via_mpeg1_surface,
    (XF86MCSurfaceInfoPtr)&via_mpeg2_surface,
    (XF86MCSurfaceInfoPtr)&via_mpeg4_surface,
    (XF86MCSurfaceInfoPtr)&via_wmv9_surface,
    (XF86MCSurfaceInfoPtr)&via_wvc1_surface
};

/* List of subpicture types that we support */
static XF86ImageRec ai44_subpicture = XVIMAGE_AI44;

static XF86ImagePtr via_subpicture_list[1] =
{
    (XF86ImagePtr)&ai44_subpicture
};

/* Fill in the device dependent adaptor record. 
 * This is named "via Video Overlay" because this code falls under the
 * XV extenstion, the name must match or it won't be used.
 *
 * Surface and Subpicture - see above
 * Function pointers to functions below
 */
static XF86MCAdaptorRec pAdapt = 
{
    "XV_SWOV",                    /* name */
    5,                            /* num_surfaces */
    ppSI,                         /* surfaces */
    1,                            /* num_subpictures */
    via_subpicture_list,          /* subpictures */
    (xf86XvMCCreateContextProcPtr)viaXvMCCreateContext,
    (xf86XvMCDestroyContextProcPtr)viaXvMCDestroyContext,
    (xf86XvMCCreateSurfaceProcPtr)viaXvMCCreateSurface,
    (xf86XvMCDestroySurfaceProcPtr)viaXvMCDestroySurface,
    (xf86XvMCCreateSubpictureProcPtr)viaXvMCCreateSubpicture,
    (xf86XvMCDestroySubpictureProcPtr)viaXvMCDestroySubpicture
};

static XF86MCAdaptorPtr ppAdapt[1] = 
{
    (XF86MCAdaptorPtr)&pAdapt
};

/**************************************************************************
 *
 *  viaInitMC
 *
 *  Initialize the hardware motion compenstation extention for this 
 *  hardware. The initialization routines want the address of the pointers
 *  to the structures, not the address of the structures. This means we
 *  allocate (or create static?) the pointer memory and pass that 
 *  address. This seems a little convoluted.
 *
 *  We need to allocate memory for the device depended adaptor record. 
 *  This is what holds the pointers to all our device functions.
 *
 *  We need to map the overlay registers into the drm.
 *
 *  We need to map the surfaces into the drm.
 *
 *  Inputs:
 *    Screen pointer
 *
 *  Outputs:
 *    None, this calls the device independent screen initialization 
 *    function.
 *
 *  Revisions:
 *  
 **************************************************************************/
void viaInitMC(ScreenPtr pScreen)
{
  /*ScrnInfoPtr pScreen = xf86Screens[pScrn->myNum];*/

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaInitMC \n");
  #endif

  xf86XvMCScreenInit(pScreen, 1, ppAdapt);
}

/**************************************************************************
 *
 *  viaXvMCCreateContext
 *
 *  Some info about the private data:
 *
 *  Set *num_priv to the number of 32bit words that make up the size of
 *  of the data that priv will point to.
 *
 *  *priv = (long *) xcalloc (elements, sizeof(element))
 *  *num_priv = (elements * sizeof(element)) >> 2;
 *
 **************************************************************************/

int viaXvMCCreateContext (ScrnInfoPtr pScrn, XvMCContextPtr pContext,
                            int *num_priv, long **priv )
{
  VIAPtr  pVia = VIAPTR(pScrn);
  ViaXvMCCreateContextRec *contextRec;
  int fd = -1;

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCCreateContext \n");
  #endif

  *priv = xcalloc(1, sizeof(ViaXvMCCreateContextRec));
  contextRec = (ViaXvMCCreateContextRec *) * priv;
  contextRec->colorKey = (1 << pScrn->offset.red) 
        | (1 << pScrn->offset.green) | (1 << pScrn->offset.blue);
  *num_priv = sizeof(viaXvMCCreateContextRec) >> 2;
  if (!*priv) {
  	*num_priv = 0;
	  return BadAlloc;
  }

  pVia->pVIAEnt->pVidData->viaXvMC.HAVE_XVMC = 1;

  /* Write some value only */
  /*  contextRec->fbBase = pScrn->memPhysBase;
  contextRec->bviaXVMC_ON = 1;*/

  return Success;
}


int viaXvMCCreateSurface (ScrnInfoPtr pScrn, XvMCSurfacePtr pSurf,
                           int *num_priv, long **priv )
{
  /* VIAPtr viaPtr = VIAPTR(pScrn); */ /* For multiple card */
  /*unsigned long dwRet;*/

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCCreateSurface \n");
  #endif

/*  *priv = (long *)xcalloc(2,sizeof(long));

  if(!*priv) {
    xf86DrvMsg(X_ERROR, pScrn->scrnIndex,
        "viaXvMCCreateSurface: Unable to allocate memory!\n");
    *num_priv = 0;
    return (BadAlloc);
  }
  *num_priv = 2;

  ddSurfaceDesc.dwWidth  =  pSurf->context->width;
  ddSurfaceDesc.dwHeight =  pSurf->context->height;
  ddSurfaceDesc.dwBackBufferCount = 3;
  ddSurfaceDesc.dwColorSpaceLowValue=0x00;
  ddSurfaceDesc.dwFourCC = FOURCC_VIA;

  fpCreateSurface(&ddSurfaceDesc);*/

  return Success;
}

int viaXvMCCreateSubpicture (ScrnInfoPtr pScrn, XvMCSubpicturePtr pSubp,
                              int *num_priv, long **priv )
{
  /* VIAPtr viaPtr = VIAPTR(pScrn); */ /* For multiple card */
  /*unsigned long dwRet;*/

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCCreateSubpicture \n");
  #endif

/*  *priv = (long *)xcalloc(1,sizeof(long));

  if(!*priv) {
    xf86DrvMsg(X_ERROR, pScrn->scrnIndex,
        "viaXvMCCreateSubpicture: Unable to allocate memory!\n");
    *num_priv = 0;
    return (BadAlloc);
  }
  *num_priv = 1;

  ddSurfaceDesc.dwWidth  =  pSubp->context->width;
  ddSurfaceDesc.dwHeight =  pSubp->context->height;
  ddSurfaceDesc.dwBackBufferCount = 1;
  ddSurfaceDesc.dwFourCC = FOURCC_SUBP;

  fpCreateSurface(&ddSurfaceDesc);*/

  return Success;
}

void viaXvMCDestroyContext (ScrnInfoPtr pScrn, XvMCContextPtr pContext)
{
  /* VIAPtr viaPtr = VIAPTR(pScrn); */ /* For multiple card */
  VIAPtr  pVia = VIAPTR(pScrn);
  int fd = -1;
  
  pVia->pVIAEnt->pVidData->viaXvMC.HAVE_XVMC = 0;
  
  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCDestroyContext \n");
  #endif
}

void viaXvMCDestroySurface (ScrnInfoPtr pScrn, XvMCSurfacePtr pSurf)
{
  /* VIAPtr viaPtr = VIAPTR(pScrn); */ /* For multiple card */
  /*unsigned long dwRet;*/

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCDestroySurface \n");
  #endif

/*  ddSurfaceDesc.dwFourCC = FOURCC_VIA;
  dwRet = DestroySurface(&ddSurfaceDesc);*/

  return;
}

void viaXvMCDestroySubpicture (ScrnInfoPtr pScrn, XvMCSubpicturePtr pSubp)
{
  /* VIAPtr viaPtr = VIAPTR(pScrn); */ /* For multiple card */
  /*unsigned long dwRet;*/

  #ifdef XVMC_DEBUG
  ErrorF(" via_xvmc.c : viaXvMCDestroySubpicture \n");
  #endif

/*  ddSurfaceDesc.dwFourCC = FOURCC_SUBP;
  dwRet = DestroySurface(&ddSurfaceDesc);*/
  return;
}

#endif /* CS_FLAG_NO_XVMC */
