/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*************************************************************************
 *
 *  File:       via_driver.c
 *  Content:    XFree86 4.0 for VIA/S3G UniChrome
 *
 ************************************************************************/
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <linux/version.h>

#include "via_driver.h"
#ifndef XSERVER_LIBPCIACCESS
#include "xf86RAC.h"
#endif
#include "xf86Priv.h"
#include "shadowfb.h"
#include "mipointer.h"
#include "mipointrst.h"
#include "inputstr.h"

#include "globals.h"

#define DPMS_SERVER

#include "via_video.h"
#include "via_vt1636.h"
#include "via_hdmi.h"
#include "via_serial.h"
#include "debug.h"      /* for DBG_DD */
#include "via_eng_regs.h"
#include "via_Hw3DScaling.h"

#ifdef XF86DRI
#include "dri.h"
#endif

#include "via_rotate.h" /* for rotate feature.*/
#include "via_dmabuffer.h"
#include "via_exa.h"

#include "via_dvi.h"
#include "via_tv.h"
#include "via_dp.h"

#include "xf86Crtc.h"
#include "xf86RandR12.h"
#include "via_common.h"
#include "via_modedata.h"
#include "via_display.h"
#include "via_output.h"
#include "via_displcd.h"

/*
  *    fix the "Undefined Symbol:vgaHWddc1SetSpeed" problem on FC5(32/64) when starting X,
  *    replace the function vgaHWddc1SetSpeed as VIAvgaHWddc1SetSpeed.
 */
#ifdef XORG_VERSION_CURRENT
 /*
  *    fix the "Undefined Symbol:vgaHWddc1SetSpeed" problem on FC7(32/64) when starting X,
  *    assign the function Xorg 1.3.0 to Xorg 7.2.0.
  */


extern Bool    VIAnoPanoramiXExtension;
#ifdef DRM_SAMM_VIA
extern drm_context_t samm_ctx_hdl;
#endif

/*
 * prototypes
 */

static void VIAIdentify(int flags);
#if XSERVER_LIBPCIACCESS
static Bool VIAPciProbe(DriverPtr drv, int entity_num, struct pci_device *dev, intptr_t data);
#else
static Bool VIAProbe(DriverPtr drv, int flags);
#endif
static Bool VIAPreInit(ScrnInfoPtr pScrn, int flags);
static Bool VIAEnterVT(VT_FUNC_ARGS_DECL);
static void VIALeaveVT(VT_FUNC_ARGS_DECL);
static void VIASave(ScrnInfoPtr pScrn);
static void VIARestore(ScrnInfoPtr, vgaRegPtr, VIARegPtr, int flags);
static Bool VIACloseScreen(CLOSE_SCREEN_ARGS_DECL);
static Bool VIASaveScreen(ScreenPtr pScreen, int mode);
static Bool VIAPMEvent(int scrnIndex, pmEvent event, Bool undo);
static void VIALoadPalette(ScrnInfoPtr pScrn, int numColors, int *indicies,
                           LOCO *colors, VisualPtr pVisual);
static Bool VIAScreenInit(SCREEN_INIT_ARGS_DECL);
static int VIAInternalScreenInit(int scrnIndex, ScreenPtr pScreen);
static void VIAFreeScreen(int scrnIndex, int flags);
static ModeStatus VIAValidMode(int index, DisplayModePtr mode,
                               Bool verbose, int flags);
static const OptionInfoRec * VIAAvailableOptions(int chipid, int busid);
static void VIAProbeChipset(ScrnInfoPtr pScrn, EntityInfoPtr pEnt);

static void VIAEnableMMIO(ScrnInfoPtr pScrn);
static Bool VIAMapMMIO(ScrnInfoPtr pScrn);
static Bool VIAMapFB(ScrnInfoPtr pScrn);
static void VIAUnmapMem(ScrnInfoPtr pScrn);
void VIAFindMaxResFromEstablishedTiming(xf86MonPtr MonitorConf, int* MaxH, int* MaxV);

void viaInitDispEngine(VIAPtr pVia);
void viaInitOutputRegSet(ScrnInfoPtr pScrn);
Bool viaInitHwInfo(VIAPtr pVia);
extern Bool viaGetTvCapsSupported(xf86OutputPtr output, CARD32 hDisplay, CARD32 vDisplay);
extern void viaSetEmbTvBufferAddress(unsigned int MMIOBase, int mapBase);

void VIAInitialize3DEngine(ScrnInfoPtr pScrn);

/* E X T E R N   F U N C T I O N S ------------------------------------------*/
/*fix video can't get enough space from Xserver fb manager*/
extern Bool VIADRIFBInit(ScreenPtr pScreen, VIAPtr pVia);
extern Bool VIADRIRingBufferInit(ScrnInfoPtr pScrn);
extern void VIADRIRingBufferCleanup(ScrnInfoPtr pScrn);
extern Bool VIACheckDrmAlreadyInit(ScrnInfoPtr pScrn);
#ifndef CS_FLAG_NO_XVMC
extern void viaInitMC(ScreenPtr pScreen);
#endif /* CS_FLAG_NO_XVMC */
extern void VIAARGBCursorInit(ScrnInfoPtr pScrn);
extern void VIAEnableVQ(ScrnInfoPtr pScrn);
extern void VIADisableVQ(ScrnInfoPtr pScrn);
extern void viaFinishInitAccel(ScreenPtr pScreen);

BOOL IsNeedReserve3DScalingMem(VIAPtr pVia);

/*---------------------------------------------------------------------------*/

#if XSERVER_LIBPCIACCESS
static const struct pci_id_match VIAPciIdMatch[] = {
    { 0x1106, PCI_CHIP_VT3314, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3324, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3336, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3327, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3364, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3353, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3409, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
    { 0x1106, PCI_CHIP_VT3410, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
#ifdef VIA_VT3293_SUPPORT    
    { 0x1106, PCI_CHIP_VT3293, PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, 0 },
#endif    
    { 0, 0, 0 }
};
#endif

DriverRec VIA =
{
    VIA_VERSION,
    DRIVER_NAME,
    VIAIdentify,
#if XSERVER_LIBPCIACCESS
	NULL,
#else
	VIAProbe,
#endif
    VIAAvailableOptions,
    NULL,
    0,
    NULL,
#if XSERVER_LIBPCIACCESS
	VIAPciIdMatch,
	VIAPciProbe
#endif
};

/* Supported chipsets */

static SymTabRec VIAChipsets[] = {
    {VIA_P4M800PRO,    "P4M800PRO"},
    {VIA_CX700,    "CX700"},
    {VIA_K8M890,   "K8M890"},
    {VIA_P4M890,   "P4M890"},
    {VIA_P4M900,   "P4M900"}, 
    {VIA_VX800,    "VX800"}, 
    {VIA_VX855,    "VX855"}, 
    {VIA_VX900,    "VX900"}, 
#ifdef VIA_VT3293_SUPPORT    
    {VIA_CN750,    "CN750"},
#endif    
    {-1,            NULL }
};


/* This table maps a PCI device ID to a chipset family identifier. */

static PciChipsets VIAPciChipsets[] = {
    {VIA_P4M800PRO, PCI_CHIP_VT3314,    RES_SHARED_VGA}, 
    {VIA_CX700,    PCI_CHIP_VT3324,    RES_SHARED_VGA},
    {VIA_K8M890,   PCI_CHIP_VT3336,    RES_SHARED_VGA},
    {VIA_P4M890,   PCI_CHIP_VT3327,    RES_SHARED_VGA},
    {VIA_P4M900,   PCI_CHIP_VT3364,    RES_SHARED_VGA},
    {VIA_VX800,    PCI_CHIP_VT3353,    RES_SHARED_VGA},
    {VIA_VX855,    PCI_CHIP_VT3409,    RES_SHARED_VGA },
    {VIA_VX900,    PCI_CHIP_VT3410,    RES_SHARED_VGA },
#ifdef VIA_VT3293_SUPPORT    
    {VIA_CN750,    PCI_CHIP_VT3293,    RES_SHARED_VGA},
#endif    
    {-1,            -1,                RES_UNDEFINED}
};

int gVIAEntityIndex = -1;

typedef enum {
    OPTION_NOACCEL,
    OPTION_DRI,
    OPTION_NOCOMPOSITE,
    OPTION_SWCURSOR,
    OPTION_SHADOW_FB,
    OPTION_ROTATE,
    OPTION_ROTATE_TYPE,
    OPTION_VIDEORAM,
    OPTION_VIDEORAM1,
    OPTION_CRT_PORT,
    OPTION_ACTIVEDEVICE,
    OPTION_DVI_PORT,
    OPTION_LCD_PORT,
    OPTION_LCD_BUS_WIDTH,
    OPTION_CENTER,
    OPTION_TV_PORT,
    OPTION_TVDOTCRAWL,
    OPTION_TV2DOTCRAWL,
    OPTION_TVTYPE,
    OPTION_TV2TYPE,
    OPTION_TVOUTPUT,
    OPTION_TV2OUTPUT,
    OPTION_TVVSCAN,
#ifndef CS_FLAG_NO_TV_SCAL
    OPTION_TVHSCALE,
    OPTION_TVVSCALE,
    OPTION_TV2HSCALE,
    OPTION_TV2VSCALE,
#endif /* CS_FLAG_NO_TV_SCAL */
    OPTION_REFRESH,
    OPTION_DISABLEVQ,
    OPTION_NODDCVALUE,
    OPTION_SCRN2POS,
    OPTION_VideoOnDevice, /*set video to work on which device for 3314,3204 etc.in duo view mode */
    OPTION_LCD_PANEL_SIZE, /* to simply LCD panel option. */
    OPTION_DPA_SETTING_DVP0,
    OPTION_DPA_SETTING_DVP1,
    OPTION_DPA_SETTING_DFPHIGH,
    OPTION_DPA_SETTING_DFPLOW,
    OPTION_DPA_SETTING_VT1636,
    OPTION_DPA_SETTING_VT1625,   
#ifndef CS_FLAG_NO_TV_SCAL
    OPTION_TVDISP3DScaling,
#endif /* CS_FLAG_NO_TV_SCAL */
#ifndef CS_FLAG_NO_HDMI_SCAL
    OPTION_HDMIDISP3DScaling,
#endif /* CS_FLAG_NO_HDMI_SCAL */
#ifndef CS_FLAG_NO_LCD_SCAL
    OPTION_LCDDISP3DScaling,  
#endif /* CS_FLAG_NO_LCD_SCAL */
    OPTION_VIDEO_COLOR_KEY,
    OPTION_GOOD_PERFORMANCE,
    OPTION_LCD_MSB_ENABLE,
    OPTION_LCD_NO_DITHERING,
    OPTION_LCD_DUAL_CHANNEL,
    OPTION_INTERLACE_MODE,
    OPTION_FN_HOTKEY,
    OPTION_TEXTUREVIDEOSUPPORT,
    OPTION_TEXTUREVIDEO3DPATH,
    OPTION_HDMI_PORT
} VIAOpts;


static OptionInfoRec VIAOptions[] =
{
    {OPTION_NOACCEL,    "NoAccel",      OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_DRI,    "DRI",   OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_NOCOMPOSITE,     "NoComposite",   OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_SWCURSOR,   "SWCursor",     OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_ROTATE,     "Rotate",     OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_ROTATE_TYPE,"RotateType",     OPTV_ANYSTR,  {0}, FALSE},   
    {OPTION_VIDEORAM,   "VideoRAM",     OPTV_INTEGER, {0}, FALSE},
    {OPTION_VIDEORAM1,  "VideoRAM1",    OPTV_INTEGER, {0}, FALSE},
    {OPTION_CRT_PORT,   "CRTPort",      OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_ACTIVEDEVICE,   "ActiveDevice", OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_DVI_PORT,       "DVIPort",      OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_LCD_PORT,       "LCDPort",      OPTV_ANYSTR, {0}, FALSE},
    {OPTION_LCD_BUS_WIDTH,  "LCDBusWidth",  OPTV_ANYSTR, {0}, FALSE},
    {OPTION_CENTER,         "Center",       OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_LCD_PANEL_SIZE, "PanelSize",    OPTV_STRING, {0}, FALSE},
    {OPTION_TV_PORT,        "TVPort",       OPTV_ANYSTR, {0}, FALSE},
    {OPTION_TVDOTCRAWL,     "TVDotCrawl",   OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_TV2DOTCRAWL,    "TV2DotCrawl",   OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_TVTYPE,     "TVType",       OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_TV2TYPE,    "TV2Type",      OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_TVOUTPUT,   "TVOutput",     OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_TV2OUTPUT,  "TV2Output",    OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_TVVSCAN,    "TVVScan",      OPTV_ANYSTR,  {0}, FALSE},
#ifndef CS_FLAG_NO_TV_SCAL
    {OPTION_TVHSCALE,   "TVHScale",     OPTV_INTEGER, {0}, FALSE},
    {OPTION_TVVSCALE,   "TVVScale",     OPTV_INTEGER, {0}, FALSE},
    {OPTION_TV2HSCALE,  "TV2HScale",    OPTV_INTEGER, {0}, FALSE},
    {OPTION_TV2VSCALE,  "TV2VScale",    OPTV_INTEGER, {0}, FALSE},
#endif /* CS_FLAG_NO_TV_SCAL */
    {OPTION_REFRESH,    "Refresh",      OPTV_INTEGER, {0}, FALSE},
    {OPTION_DISABLEVQ,  "DisableVQ",    OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_NODDCVALUE, "NoDDCValue",   OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_SCRN2POS,   "Scrn2Pos", OPTV_ANYSTR, {0}, FALSE},   
    {OPTION_VideoOnDevice,  "VideoOnDevice",       OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_DPA_SETTING_DVP0,       "DPASetting_DVP0",      OPTV_ANYSTR, {0}, FALSE},
    {OPTION_DPA_SETTING_DVP1,       "DPASetting_DVP1",      OPTV_ANYSTR, {0}, FALSE},
    {OPTION_DPA_SETTING_DFPHIGH,    "DPASetting_DFPHIGH",   OPTV_ANYSTR, {0}, FALSE},
    {OPTION_DPA_SETTING_DFPLOW,     "DPASetting_DFPLOW",    OPTV_ANYSTR, {0}, FALSE},
    {OPTION_DPA_SETTING_VT1636,     "DPASetting_VT1636",    OPTV_ANYSTR, {0}, FALSE},
    {OPTION_DPA_SETTING_VT1625,     "DPASetting_VT1625",    OPTV_ANYSTR, {0}, FALSE},
#ifndef CS_FLAG_NO_TV_SCAL
    {OPTION_TVDISP3DScaling,          "DISPLAY_3DScaling_TV",    OPTV_BOOLEAN, {0}, FALSE},
#endif /* CS_FLAG_NO_TV_SCAL */
#ifndef CS_FLAG_NO_HDMI_SCAL
    {OPTION_HDMIDISP3DScaling,      "DISPLAY_3DScaling_HDMI",    OPTV_BOOLEAN, {0}, FALSE},    
#endif /* CS_FLAG_NO_HDMI_SCAL */
#ifndef CS_FLAG_NO_LCD_SCAL
    {OPTION_LCDDISP3DScaling,       "DISPLAY_3DScaling_LCD",    OPTV_BOOLEAN, {0}, FALSE},    
#endif /* CS_FLAG_NO_LCD_SCAL */
    {OPTION_VIDEO_COLOR_KEY,        "VideoColorKey",  OPTV_INTEGER, {0}, FALSE},
    {OPTION_GOOD_PERFORMANCE,       "GoodPerformance",  OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_LCD_MSB_ENABLE,         "LCD_MSB_ENABLE",  OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_LCD_NO_DITHERING,       "NoDithering",  OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_LCD_DUAL_CHANNEL,       "DualChannel",  OPTV_STRING, {0}, FALSE},
    {OPTION_INTERLACE_MODE,         "Interlace",            OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_FN_HOTKEY,              "FnHotkey",             OPTV_ANYSTR, {0}, FALSE},
    {OPTION_HDMI_PORT,               "HDMIPort",      OPTV_ANYSTR,  {0}, FALSE},    
    {-1,                NULL,           OPTV_NONE,    {0}, FALSE}
};

Bool vt1625_module_loaded = FALSE;
Bool integratedtv_module_loaded = FALSE;
Bool ad9389_module_loaded = FALSE;
Bool sil164_module_loaded = FALSE;

#ifndef XSERVER_LIBPCIACCESS

static int via_mem_en = FALSE;
static int via_io_en = FALSE;
static int via_io_mem_en = FALSE;
static xf86SetAccessFuncRec  viaAccessFunctionSaved;

static void via_mem_enable (void *arg)
{
    if (!via_mem_en) {
        (viaAccessFunctionSaved.mem->AccessEnable)(
            viaAccessFunctionSaved.mem->arg);
        via_mem_en = TRUE;
    }
}
static void via_mem_disable (void *arg)
{
    if (via_mem_en) {
        (viaAccessFunctionSaved.mem->AccessDisable)(
            viaAccessFunctionSaved.mem->arg);
        via_mem_en = FALSE;    
    }
}
static void via_io_enable (void *arg)
{
   if (!via_io_en) {
        (viaAccessFunctionSaved.io->AccessEnable)(
            viaAccessFunctionSaved.io->arg);
        via_io_en = TRUE;
    }
}

static void via_io_disable (void *arg)
{
    if (via_io_en) {
        (viaAccessFunctionSaved.io->AccessDisable)(
            viaAccessFunctionSaved.io->arg);
        via_io_en = FALSE;
    }
}

static void via_io_mem_enable (void *arg)
{
    if (!via_io_mem_en) {
        (viaAccessFunctionSaved.io_mem->AccessEnable)(
            viaAccessFunctionSaved.io_mem->arg);
        via_io_mem_en = TRUE;
    }
}

static void via_io_mem_disable (void *arg)
{    
    if (via_io_mem_en) {
        (viaAccessFunctionSaved.io_mem->AccessDisable)(
            viaAccessFunctionSaved.io_mem->arg);
        via_io_mem_en = FALSE;    
    }
}

static xf86AccessRec via_mem = {
    &via_mem_disable,
    &via_mem_enable,    
    (void *)0
};

static xf86AccessRec via_io = {
    &via_io_disable,
    &via_io_enable,    
    (void *)0
};

static xf86AccessRec via_io_mem = {
    &via_io_mem_disable,
    &via_io_mem_enable,    
    (void *)0
};

static xf86SetAccessFuncRec via_rac_function = {
    &via_mem,
    &via_io,
    &via_io_mem
};
#endif


/* for Via DVD and Video support */
/*
 * No use now, for old Xserver module code use
 */

#ifdef XFree86LOADER

static MODULESETUPPROTO(VIASetup);

static XF86ModuleVersionInfo VIAVersRec = {
    "chrome",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    VERSION_MAJOR, VERSION_MINOR, PATCHLEVEL,
    ABI_CLASS_NONE,
    ABI_VIDEODRV_VERSION,
    MOD_CLASS_VIDEODRV,
    {0, 0, 0, 0}
};

_X_EXPORT XF86ModuleData chromeModuleData = {&VIAVersRec, VIASetup, NULL};

unsigned int via_xf86_version_current;


static Bool
via_xf86crtc_resize (ScrnInfoPtr pScrn, int width, int height)
{
    pScrn->virtualX = width;
    pScrn->virtualY = height;       
    
    return TRUE;
}

static const xf86CrtcConfigFuncsRec via_xf86crtc_config_funcs = {
    via_xf86crtc_resize
};

void FillGraphicInfo(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;

	/* Added to pass DRM info to V4L */
#ifdef XF86DRI
	viaGfxInfo->drmEnabled = pVia->directRenderingEnabled;
#else
    viaGfxInfo->drmEnabled = 0;
#endif
    	
    return;
}


void FillGraphicMemInfo(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;    
	
    /*fill generally info, just do once in ScreenInit, because won't changed*/
    if(!pVia->IsSecondary)
    {
        viaGfxInfo->vheapbase = (unsigned long) pVia->FBFreeStart;
        viaGfxInfo->vheapend = (unsigned long) pVia->FBFreeEnd;
        pVia->FBFreeStart = pVia->FBFreeEnd;

        xf86DrvMsg(pScrn->scrnIndex,X_INFO,"[Memory_layout] video managed Fb size = %ld k\n",
                (viaGfxInfo->vheapend-viaGfxInfo->vheapbase)/1024);

        viaGfxInfo->vramsize = pVia->videoRambytes;
        viaGfxInfo->revisionId = pVia->ChipRev;
        viaGfxInfo->chipId = pVia->ChipId;
        viaGfxInfo->mmioSize = VIA_MMIO_REGSIZE;
        viaGfxInfo->fbSize = pVia->videoRambytes;
    }
}

static pointer VIASetup(pointer module, pointer opts, int *errmaj, int *errmin)
{
    static Bool setupDone = FALSE;

	/*Only be loaded once*/
    if (!setupDone) {
        /*print addtional build string*/
        char *subversion=VIA_SUBVERSION;
        if (subversion)
           xf86Msg(X_INFO,"chrome driver subversion: %s\n",subversion);
        setupDone = TRUE;
		xf86AddDriver(&VIA, module, 
#if XSERVER_LIBPCIACCESS
		HaveDriverFuncs
#else
		0
#endif
		);

        return (pointer) 1;
    }
    else {
        if (errmaj)
            *errmaj = LDR_ONCEONLY;

        return NULL;
    }
} /* VIASetup */

#endif /* XFree86LOADER */


static int
WaitIdle_H2Chips(VIAPtr pVia)
{
    int loop = 0;

    mem_barrier();

    while (!(VIAGETREG(VIA_REG_STATUS) & VIA_VR_QUEUE_EMPTY) && (loop++ < MAXLOOP))
        ;

    while ((VIAGETREG(VIA_REG_STATUS) &
          (VIA_CMD_RGTR_BUSY | VIA_2D_ENG_BUSY | VIA_3D_ENG_BUSY)) &&
          (loop++ < MAXLOOP));

    return loop >= MAXLOOP;
}

static int
WaitIdle_H5Chips(VIAPtr pVia)
{
    int loop = 0;

    mem_barrier();

    while ((VIAGETREG(VIA_REG_STATUS) &
          (VIA_CMD_RGTR_BUSY_H5| VIA_2D_ENG_BUSY_H5|VIA_3D_ENG_BUSY_H5)) &&
          (loop++ < MAXLOOP));

    return loop >= MAXLOOP;
}

static int
WaitIdle_H6Chips(VIAPtr pVia)
{
    int loop = 0;

    mem_barrier();

    while ((VIAGETREG(VIA_REG_STATUS) & 
           (VIA_CMD_RGTR_BUSY_M1 | VIA_2D_ENG_BUSY_M1 | VIA_3D_ENG_BUSY_M1)) &&
           (loop++ < MAXLOOP));

    return loop >= MAXLOOP;
}

static Bool VIAGetRec(ScrnInfoPtr pScrn)
{
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAGetRec\n"));
    
    if (pScrn->driverPrivate)
        return TRUE;

    /* allocate VIARec */
    pScrn->driverPrivate = xnfcalloc(sizeof(VIARec), 1);
    
    ((VIARec *)(pScrn->driverPrivate))->pBIOSInfo =
        xnfcalloc(sizeof(VIABIOSInfoRec), 1);
        
    ((VIARec *)(pScrn->driverPrivate))->pChipInfo =
        xnfcalloc(sizeof(VIAChipInfoRec), 1);

    /* initial value in VIARec */
    ((VIARec *)(pScrn->driverPrivate))->SavedReg.mode = 0xFF;
    ((VIARec *)(pScrn->driverPrivate))->pBIOSInfo->FirstInit = TRUE;
    ((VIARec *)(pScrn->driverPrivate))->HwCursorImage = NULL;
    ((VIARec *)(pScrn->driverPrivate))->HwIconImage = NULL;
    ((VIARec *)(pScrn->driverPrivate))->HIScalingPrimImage = NULL;
    ((VIARec *)(pScrn->driverPrivate))->HIScalingSecImage = NULL;
    
    return TRUE;

} /* VIAGetRec */


static void VIAFreeRec(ScrnInfoPtr pScrn)
{
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAFreeRec\n"));
    if (!pScrn->driverPrivate)
        return;

    xfree(((VIARec *)(pScrn->driverPrivate))->pBIOSInfo);    
    xfree(((VIARec *)(pScrn->driverPrivate))->pChipInfo);    

    xfree(pScrn->driverPrivate);
    pScrn->driverPrivate = NULL;

} /* VIAFreeRec */


static const OptionInfoRec * VIAAvailableOptions(int chipid, int busid)
{

    return VIAOptions;

} /* VIAAvailableOptions */


static void VIAIdentify(int flags)
{
    xf86PrintChipsets(DRIVER_NAME,
                      "driver for VIA chipsets",
                      VIAChipsets);
} /* VIAIdentify */

#if XSERVER_LIBPCIACCESS
struct pci_device * VIAGetHostBridge(uint32_t Bus, uint32_t Device, uint32_t Function)
{
    struct pci_device_iterator	*SlotIterator = NULL;
    struct pci_device		*MatchBridge = NULL;
	
	struct pci_slot_match slot_match = {
		.domain = 0,
		.bus = Bus,
		.dev = Device,
		.func = Function,
		.match_data = 0
	};

    SlotIterator = pci_slot_match_iterator_create (&slot_match);
    MatchBridge = pci_device_next (SlotIterator);
    pci_iterator_destroy (SlotIterator);
    return MatchBridge;
}

static Bool VIAPciProbe(DriverPtr drv, int entity_num, struct pci_device *dev, intptr_t data)
{
    ScrnInfoPtr	    pScrn = NULL;
    EntityInfoPtr   pEnt;
    DevUnion	    *pPriv;
    static int      instance = 0;
    VIAEntPtr       pVIAEnt;
    
    pScrn = xf86ConfigPciEntity(pScrn, 0, entity_num, VIAPciChipsets, 
                NULL, NULL, NULL, NULL, NULL);
    
    if (!pScrn) return FALSE;
    
    pScrn->driverVersion    = VIA_VERSION;
    pScrn->driverName       = DRIVER_NAME;
    pScrn->name             = "VIA";
    pScrn->Probe            = NULL;

    pScrn->PreInit          = VIAPreInit;
    pScrn->ScreenInit       = VIAScreenInit;
    pScrn->SwitchMode       = VIASwitchMode;
    pScrn->AdjustFrame      = VIAAdjustFrame;
    pScrn->EnterVT          = VIAEnterVT;
    pScrn->LeaveVT          = VIALeaveVT;
    pScrn->FreeScreen       = VIAFreeScreen;
    pScrn->ValidMode        = VIAValidMode;
    pScrn->PMEvent          = VIAPMEvent;

    pEnt = xf86GetEntityInfo(entity_num);

    xf86SetEntitySharable(entity_num);
    xf86SetEntityInstanceForScreen(pScrn,
                    pScrn->entityList[0], instance);

    if(gVIAEntityIndex < 0)
    {
        gVIAEntityIndex = xf86AllocateEntityPrivateIndex();
        pPriv = xf86GetEntityPrivate(pScrn->entityList[0],
                gVIAEntityIndex);

        if (!pPriv->ptr)
        {            
            pPriv->ptr = xnfcalloc(sizeof(VIAEntRec), 1);
            pVIAEnt = pPriv->ptr;
            pVIAEnt->HasSecondary = FALSE;
            pVIAEnt->pPrimaryScrn = NULL;
            pVIAEnt->pSecondaryScrn = NULL;
            pVIAEnt->pScreensPublicInfo = xnfcalloc(sizeof(VIAScreensPublicInfo), 1);
            pVIAEnt->pVidData = (PVIDDATA) xalloc(sizeof(VIDDATA));
            memset(pVIAEnt->pVidData, 0, sizeof(VIDDATA));
            
            pVIAEnt->pVidData->viaGfxInfo = (viaGfxInfoPtr) xalloc(sizeof(viaGfxInfoRec));
            memset(pVIAEnt->pVidData->viaGfxInfo, 0, sizeof(viaGfxInfoRec));             
        }
    }
    instance++;

    if (instance >1){
        pPriv = xf86GetEntityPrivate(pScrn->entityList[0],
                gVIAEntityIndex);

        if (pPriv->ptr)
        {
            pVIAEnt = pPriv->ptr;
            pVIAEnt->HasSecondary = TRUE;
        }
    }

    return TRUE;
}
#else /* XSERVER_LIBPCIACCESS */

static Bool VIAProbe(DriverPtr drv, int flags)
{
    GDevPtr *devSections;
    int     *usedChips;
    int     numDevSections;
    int     numUsed;
    Bool    foundScreen = FALSE;
    int     i;

    /* sanity checks */
    if ((numDevSections = xf86MatchDevice(DRIVER_NAME, &devSections)) <= 0)
        return FALSE;

    if (xf86GetPciVideoInfo() == NULL)
        return FALSE;

    numUsed = xf86MatchPciInstances(DRIVER_NAME,
                                    PCI_VIA_VENDOR_ID,
                                    VIAChipsets,
                                    VIAPciChipsets,
                                    devSections,
                                    numDevSections,
                                    drv,
                                    &usedChips);
    xfree(devSections);

    if (numUsed <= 0)
        return FALSE;

    if (flags & PROBE_DETECT) {
        foundScreen = TRUE;
    }
    else {
        /* SAMM on Xorg 1.4.x series, Xorg treats the two screens independent
         * of each other and do lots of screen dependent resources(io*mem)
         * enable*disable when XServer states(SETUP*OPERATING) change, it
         * was this too-frequent-change that cause the GFX won't response to
         * cycle from io or mem range.
         * SO: hook our own enable*disable function, our own enable*disable
         * function will call the xorg's original enable*disable function
         * when it is really needs.
         */
        if (numUsed > 1) {
            EntityInfoRec via_entityinfo;
                    via_entityinfo.index = usedChips[0];
                    xf86SetAccessFuncs(&via_entityinfo,
                        &via_rac_function, &viaAccessFunctionSaved);
        }
        for (i = 0; i < numUsed; i++) {
            ScrnInfoPtr pScrn = NULL;
            static int instance = 0;
            DevUnion* pPriv;
            VIAEntPtr pVIAEnt;

            if ((pScrn = xf86ConfigPciEntity(pScrn, 0, usedChips[i],
                 VIAPciChipsets, 0, 0, 0, 0, 0)))
            {
                pScrn->driverVersion = VIA_VERSION;
                pScrn->driverName = DRIVER_NAME;
                pScrn->name = "VIA";
                pScrn->Probe = VIAProbe;
                pScrn->PreInit = VIAPreInit;
                pScrn->ScreenInit = VIAScreenInit;
                pScrn->SwitchMode = VIASwitchMode;
                pScrn->AdjustFrame = VIAAdjustFrame;
                pScrn->EnterVT = VIAEnterVT;
                pScrn->LeaveVT = VIALeaveVT;
                pScrn->FreeScreen = VIAFreeScreen;
                pScrn->ValidMode = VIAValidMode;
                pScrn->PMEvent=VIAPMEvent;
                foundScreen = TRUE;

                xf86SetEntitySharable(usedChips[i]);
                xf86SetEntityInstanceForScreen(pScrn,
                    pScrn->entityList[0], instance);

                if(gVIAEntityIndex < 0)
                {
                    gVIAEntityIndex = xf86AllocateEntityPrivateIndex();
                    pPriv = xf86GetEntityPrivate(pScrn->entityList[0],
                            gVIAEntityIndex);

                    if (!pPriv->ptr)
                    {
                        pPriv->ptr = xnfcalloc(sizeof(VIAEntRec), 1);
                        pVIAEnt = pPriv->ptr;
                        pVIAEnt->HasSecondary = FALSE;
                        pVIAEnt->pPrimaryScrn = NULL;
                        pVIAEnt->pSecondaryScrn = NULL;
                        pVIAEnt->pScreensPublicInfo = xnfcalloc(sizeof(VIAScreensPublicInfo), 1);
                        pVIAEnt->pVidData = (PVIDDATA) xalloc(sizeof(VIDDATA));
                        memset(pVIAEnt->pVidData, 0, sizeof(VIDDATA));
                        
                        pVIAEnt->pVidData->viaGfxInfo = (viaGfxInfoPtr) xalloc(sizeof(viaGfxInfoRec));
                        memset(pVIAEnt->pVidData->viaGfxInfo, 0, sizeof(viaGfxInfoRec));             
                    }
                }
                instance++;
                if (instance >1){
                    pPriv = xf86GetEntityPrivate(pScrn->entityList[0],
                            gVIAEntityIndex);

                    if (pPriv->ptr)
                    {
                        pVIAEnt = pPriv->ptr;
                        pVIAEnt->HasSecondary = TRUE;
                    }
                }
	        }
        }
    }

    xfree(usedChips);

    return foundScreen;

} /* VIAProbe */
#endif /* XSERVER_LIBPCIACCESS */

static int LookupChipID(PciChipsets* pset, int ChipID)
{
    /* Is there a function to do this for me? */
    while (pset->numChipset >= 0)
    {
        if (pset->PCIid == ChipID)
            return pset->numChipset;

        pset++;
    }

    return -1;

} /* LookupChipID */


static unsigned int
VIAddc1Read(ScrnInfoPtr pScrn)
{
    register vgaHWPtr hwp = VGAHWPTR(pScrn);
    VIAPtr pVia = VIAPTR(pScrn);
    register CARD8 tmp;

    while (hwp->readST01(hwp)&0x8) {};
    while (!(hwp->readST01(hwp)&0x8)) {};

    VGAOUT8(0x3c4, 0x26);
    tmp = VGAIN8(0x3c5);
    return ((unsigned int) ((tmp & 0x08) >> 3));
}



void
VIAvgaHWddc1SetSpeed(ScrnInfoPtr pScrn, xf86ddcSpeed speed)
{
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    unsigned char tmp;
    struct _vgaDdcSave* save;
    switch (speed) {
    case DDC_FAST:
        
        if (hwp->ddc != NULL) break;
        hwp->ddc = xnfcalloc(sizeof(struct _vgaDdcSave),1);
    save = (struct _vgaDdcSave *)hwp->ddc;
      /* Lightpen register disable - allow access to cr10 & 11; just in case */
    save->cr03 = hwp->readCrtc(hwp, 0x03); 
    hwp->writeCrtc(hwp,0x03,(save->cr03 |0x80));
    save->cr12 = hwp->readCrtc(hwp, 0x12); 
    hwp->writeCrtc(hwp,0x12,DISPLAY_END);
    save->cr15 = hwp->readCrtc(hwp, 0x15); 
    hwp->writeCrtc(hwp,0x15,BLANK_START);
    save->cr10 = hwp->readCrtc(hwp, 0x10); 
    hwp->writeCrtc(hwp,0x10,SYNC_START);
    save->cr11 = hwp->readCrtc(hwp, 0x11); 
    /* unprotect group 1 registers; just in case ...*/
    hwp->writeCrtc(hwp,0x11,((save->cr11 & 0x70) | SYNC_END));
    save->cr16 = hwp->readCrtc(hwp, 0x16); 
    hwp->writeCrtc(hwp,0x16,BLANK_END);
    save->cr06 = hwp->readCrtc(hwp, 0x06); 
    hwp->writeCrtc(hwp,0x06,V_TOTAL);
    /* all values have less than 8 bit - mask out 9th and 10th bits */
    save->cr09 = hwp->readCrtc(hwp, 0x09); 
    hwp->writeCrtc(hwp,0x09,(save->cr09 &0xDF));
    save->cr07 = hwp->readCrtc(hwp, 0x07); 
    hwp->writeCrtc(hwp,0x07,(save->cr07 &0x10));
    /* vsync polarity negativ & ensure a 25MHz clock */
    save->msr = hwp->readMiscOut(hwp); 
    hwp->writeMiscOut(hwp,((save->msr & 0xF3) | 0x80));
    break;
    case DDC_SLOW:
        if (hwp->ddc == NULL) break;
    save = (struct _vgaDdcSave *)hwp->ddc;
    hwp->writeMiscOut(hwp,save->msr);
    hwp->writeCrtc(hwp,0x07,save->cr07);
    tmp = hwp->readCrtc(hwp, 0x09);
    hwp->writeCrtc(hwp,0x09,((save->cr09 & 0x20) | (tmp & 0xDF)));
    hwp->writeCrtc(hwp,0x06,save->cr06);
    hwp->writeCrtc(hwp,0x16,save->cr16);
    hwp->writeCrtc(hwp,0x11,save->cr11);
    hwp->writeCrtc(hwp,0x10,save->cr10);
    hwp->writeCrtc(hwp,0x15,save->cr15);
    hwp->writeCrtc(hwp,0x12,save->cr12);
    hwp->writeCrtc(hwp,0x03,save->cr03);
    xfree(save);
    hwp->ddc = NULL;
    break;
    default:
    break;
    }
}

#else 
    #define VIAvgaHWddc1SetSpeed vgaHWddc1SetSpeed
#endif


static Bool
VIAddc1(int scrnIndex)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);
    xf86MonPtr pMon;
    CARD8 tmp;
    Bool success = FALSE;

    /* initialize chipset */
    VGAOUT8(0x3c4, 0x26);
    tmp = VGAIN8(0x3c5);
    VGAOUT8(0x3c4, 0x26);
    VGAOUT8(0x3c5, (tmp | 0x11));

    if ((pMon = xf86PrintEDID(
        xf86DoEDID_DDC1(scrnIndex,VIAvgaHWddc1SetSpeed,VIAddc1Read))) != NULL)
        success = TRUE;
    xf86SetDDCproperties(pScrn,pMon);

    /* undo initialization */
    VGAOUT8(0x3c4, 0x26);
    VGAOUT8(0x3c5, tmp);
    return success;
}

static Bool
VIAddc2(int scrnIndex)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);
    xf86MonPtr pMon;
    CARD8 tmp;
    Bool success = FALSE;

    VGAOUT8(0x3c4, 0x26);
    tmp = VGAIN8(0x3c5);
    pMon = xf86DoEDID_DDC2(pScrn->scrnIndex, pVia->I2C_Port1);
    if (pMon)
        success = TRUE;
    pVia->DDC1 =  pMon;
    xf86PrintEDID(pMon);
    xf86SetDDCproperties(pScrn, pMon);
    VGAOUT8(0x3c4, 0x26);
    VGAOUT8(0x3c5, tmp);

    return success;
}


void 
VIACheckCursorTypeToUse(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;

    if (pVia->ForceSWCursor) {
        /* Use SW Cursor */
        pVia->pVIAEnt->pScreensPublicInfo->UseHwCursor = FALSE;
    } else {
        /* Use HW Cursor */
        pVia->pVIAEnt->pScreensPublicInfo->UseHwCursor = TRUE;
    }

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Using %s cursor\n", pVia->pVIAEnt->pScreensPublicInfo->UseHwCursor ? "HW" : "SW"));
}


/* For Hotkey Timer Use. */
CARD32 
VIACheckHotkeyStatus(ScrnInfoPtr pScrn)
{            
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;           
    CARD16  deviceCount = 0;        
    CARD16  i;
    CARD16  deviceSeq[4];
    
    /* Check the change of register CR3C[0]. */
    if (read_reg(VIACR, CR3C) & BIT0) {
        /* Clear */                
        write_reg_mask(CR3C, VIACR, 0x00, BIT0);                         

        /* To count how many prefered and valid devices can be actived.*/
        for (i = 0; i < 4; i++) {
            if (pVia->deviceSequence[i]) {
                if (pVia->deviceSequence[i] == VIA_DEVICE_CRT1) {
                    if (pVia->IsForceHotkeySwitch) {
                        deviceSeq[deviceCount] = pVia->deviceSequence[i];
                        deviceCount++;                                        
                    } else if (VIASenseCRT(pScrn)) {
                        deviceSeq[deviceCount] = pVia->deviceSequence[i];
                        deviceCount++;                                        
                    }
                }                
                
                if ((pVia->deviceSequence[i] == VIA_DEVICE_DFP) &&
                    (VIASenseDVI(pBIOSInfo,&(pBIOSInfo->TMDSSettingInfo)))) {
                    
                    deviceSeq[deviceCount] = pVia->deviceSequence[i];
                    deviceCount++;                      
                }
                
                if (pVia->deviceSequence[i] == VIA_DEVICE_LCD) {
                    deviceSeq[deviceCount] = pVia->deviceSequence[i];
                    deviceCount++;                      
                }
            }
        }                
        
        /* Only one device was conneted. */                                
        if (deviceCount == 1) {
            DeviceCounter = 0;
        /* 2 devices were input. ex: CRT, LCD */    
        } else if (deviceCount == 2) {
            if (pBIOSInfo->ActiveDevice == deviceSeq[0]) {
                DeviceCounter = 0;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[1]) {
                DeviceCounter = 1;
            }
            else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[1])) {
                DeviceCounter = 2;            
            } else {
                DeviceCounter = 0;
            }
        /* 3 devices were input. ex: CRT, LCD, DVI */    
        } else if (deviceCount == 3) {
            /* Configure current state */        
            if (pBIOSInfo->ActiveDevice == deviceSeq[0]) {
                DeviceCounter = 0;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[1]) {
                DeviceCounter = 1;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[2]) {
                DeviceCounter = 2;
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[1])) {
                DeviceCounter = 3;            
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[2])) {
                DeviceCounter = 4;            
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[1] | deviceSeq[2])) {
                DeviceCounter = 5;
            } else {
                DeviceCounter = 0;
            }
        /* 4 devices were input. ex: CRT, LCD, DVI, TV */
        } else if (deviceCount == 4) {
            /* Configure current state */        
            if (pBIOSInfo->ActiveDevice == deviceSeq[0]) {
                DeviceCounter = 0;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[1]) {
                DeviceCounter = 1;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[2]) {
                DeviceCounter = 2;
            } else if (pBIOSInfo->ActiveDevice == deviceSeq[3]) {
                DeviceCounter = 3;
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[1])) {
                DeviceCounter = 4;            
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[2])) {
                DeviceCounter = 5;            
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[0] | deviceSeq[3])) {
                DeviceCounter = 6;
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[1] | deviceSeq[2])) {
                DeviceCounter = 7;
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[1] | deviceSeq[3])) {
                DeviceCounter = 8;
            } else if (pBIOSInfo->ActiveDevice == (deviceSeq[2] | deviceSeq[3])) {
                DeviceCounter = 9;
            } else {
                DeviceCounter = 0;
            }
        }               
                                                  
        
        /* Loop to switch device */
        DeviceCounter++;        

        if (deviceCount == 1) {
            DeviceCounter = 0;   
            pBIOSInfo->ActiveDevice = deviceSeq[0];
        } else if (deviceCount == 2) {
            DeviceCounter = DeviceCounter%3;   
            switch(DeviceCounter) {
                case 0:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0];
                    break;                    
                case 1:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1];
                    break;                    
                case 2:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[1];
                    break;                       
            } 
        } else if (deviceCount == 3) {
            DeviceCounter = DeviceCounter%6; 
            switch(DeviceCounter)
            {
                case 0:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0];
                    break;
                case 1:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1];
                    break;
                case 2:                
                    pBIOSInfo->ActiveDevice = deviceSeq[2];
                    break;    
                case 3:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[1];
                    break;    
                case 4:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[2];
                    break;    
                case 5:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1] | deviceSeq[2];
                    break;                
            } 
        } else if (deviceCount == 4) {
            DeviceCounter = DeviceCounter%10;    
            switch(DeviceCounter) {
                case 0:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0];
                    break;                    
                case 1:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1];
                    break;                    
                case 2:                
                    pBIOSInfo->ActiveDevice = deviceSeq[2];
                    break;   
                case 3:                
                    pBIOSInfo->ActiveDevice = deviceSeq[3];
                    break;
                case 4:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[1];
                    break;    
                case 5:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[2];
                    break;    
                case 6:                
                    pBIOSInfo->ActiveDevice = deviceSeq[0] | deviceSeq[3];
                    break;   
                case 7:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1] | deviceSeq[2];
                    break;  
                case 8:                
                    pBIOSInfo->ActiveDevice = deviceSeq[1] | deviceSeq[3];
                    break;  
                case 9:                
                    pBIOSInfo->ActiveDevice = deviceSeq[2] | deviceSeq[3];
                    break;  
            } 
        } else {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD;
        }
                
        /* If there is no any device connected. Use the first device as default */
        if (!pBIOSInfo->ActiveDevice) {
            pBIOSInfo->ActiveDevice = deviceSeq[0];
        }
        VIASwitchMode(SWITCH_MODE_ARGS(pScrn, pScrn->currentMode));
               
    }        

    return 1000;
}


CARD32 
VIASyncTimer(OsTimerPtr timer, CARD32 now, pointer arg)
{        
    ScrnInfoPtr pScrn = (ScrnInfoPtr) arg;
    VIAPtr pVia = VIAPTR(pScrn);

    if(pVia->pBIOSInfo->IsStableStatus) {        
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "need to do VIADISP3DScalingImageProcessing\n"));
        VIADISP3DScalingImageProcessing(pVia);
    }

    if (pVia->IsHotkeyTimerEnabled) {
        VIACheckHotkeyStatus(pScrn);
    }

    return 30;
}

static int 
VIATranslateTVOutputNameToID(char *pName)
{
    if (!xf86NameCmp(pName, "S-Video")) {
        return TV_OUTPUT_SVIDEO;
    } else if(!xf86NameCmp(pName, "Composite")) {
        return TV_OUTPUT_COMPOSITE;
    } else if(!xf86NameCmp(pName, "SC")) {
        return TV_OUTPUT_COMPOSITE_SVIDEO;
    } else if(!xf86NameCmp(pName, "RGB")) {
        return TV_OUTPUT_RGB;
    } else if(!xf86NameCmp(pName, "YCbCr") || !xf86NameCmp(pName, "YPbPr")) {
        return TV_OUTPUT_YPBPR;
    } else if(!xf86NameCmp(pName, "RGBSC")) {
        return TV_OUTPUT_RGB_COMPOSITE_SVIDEO;
    } else if(!xf86NameCmp(pName, "YCbCrSC") || !xf86NameCmp(pName, "YPbPrSC")) {
        return TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO;
    } else {
        ErrorF("Invalid TV Output.\n");
        return TV_OUTPUT_NONE;
    }
}

static void 
VIAGetOptions_Cursor(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);

    /*
     * The default if neither is specified is HW.
     */
    pVia->ForceSWCursor = FALSE;
    
    if (xf86ReturnOptValBool(VIAOptions, OPTION_SWCURSOR, FALSE)) {
        pVia->ForceSWCursor = TRUE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option SWCursor enabled.\n");
    }
}

static void 
VIAGetOptions_ActiveDevice(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;
    
    pBIOSInfo->TwoCRT = FALSE;    
    
    /* ActiveDevice Option for device selection */
    pBIOSInfo->ActiveDevice = 0x00;
    
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_ACTIVEDEVICE))) {
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Active Device is %s.\n", pStr);
        
        if (!xf86NameCmp(pStr, "CRT,TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_TV;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "TV,CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_TV;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "CRT,CRT2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_CRT2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
            pBIOSInfo->TwoCRT = TRUE;
        } else if(!xf86NameCmp(pStr, "CRT2,CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_CRT2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT2;
            pBIOSInfo->TwoCRT = TRUE;
        } else if(!xf86NameCmp(pStr, "CRT,LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "LCD,CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "LCD,LCD2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_LCD2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "LCD2,LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_LCD2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD2;
        } else if(!xf86NameCmp(pStr, "TV,LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "LCD,TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "TV,TV2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_TV2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "TV2,TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_TV2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV2;
        } else if(!xf86NameCmp(pStr, "DP,DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "DP2,DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "CRT,DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DFP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "DVI,CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DFP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "DVI,TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DFP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "TV,DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DFP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "DVI,LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "LCD,DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "LCD")) { 
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "DVI2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP2;
        } else if(!xf86NameCmp(pStr, "HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        }  else if(!xf86NameCmp(pStr, "DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        }  else if(!xf86NameCmp(pStr, "DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "CRT, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "HDMI, CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "TV, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "HDMI, TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "LCD, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "HDMI, LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "DVI, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "HDMI, DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "DP, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "HDMI, DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "DP2, HDMI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP2 | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "HDMI, DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DP2 | VIA_DEVICE_HDMI;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "HDMI, HDMI2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_HDMI | VIA_DEVICE_HDMI2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_HDMI;
        } else if(!xf86NameCmp(pStr, "CRT, DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "DP, CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "TV, DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "DP, TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "LCD, DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "DP, LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "DVI, DP")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "DP, DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DP;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP;
        } else if(!xf86NameCmp(pStr, "CRT, DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "DP2, CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "TV, DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_TV;
        } else if(!xf86NameCmp(pStr, "DP2, TV")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_TV | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "LCD, DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_LCD;
        } else if(!xf86NameCmp(pStr, "DP2, LCD")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_LCD | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "DVI, DP2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "DP2, DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DP2;
        } else if(!xf86NameCmp(pStr, "DVI, DVI2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DFP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP;
        } else if(!xf86NameCmp(pStr, "DVI2, DVI")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_DFP | VIA_DEVICE_DFP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP2;
        } else if(!xf86NameCmp(pStr, "CRT, DVI2")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DFP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_CRT1;
        } else if(!xf86NameCmp(pStr, "DVI2, CRT")) {
            pBIOSInfo->ActiveDevice = VIA_DEVICE_CRT1 | VIA_DEVICE_DFP2;
            pBIOSInfo->PrimaryDevice = VIA_DEVICE_DFP2;
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                "Option \"%s\" can't recognize!, Active Device by default.\n", pStr);
        }
    }

    pBIOSInfo->OptRefresh = 0;
    if (xf86GetOptValInteger(VIAOptions, OPTION_REFRESH,
                             &(pBIOSInfo->OptRefresh))) {
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Refresh is %d\n",
                   pBIOSInfo->OptRefresh);
    }

    /* Initialize CRT2DIport value even if not support two crt. */
    pBIOSInfo->CRTDIPort = VIA_DI_NONE;
    if(pBIOSInfo->TwoCRT) {
        if ((pStr = xf86GetOptValString(VIAOptions, OPTION_CRT_PORT))) {
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "CRT port set to %s.\n", pStr);
            
            if (!xf86NameCmp(pStr, "DFP_HIGH")) {                
                pBIOSInfo->CRTDIPort = VIA_DI_DFPHIGH;
            } else if (!xf86NameCmp(pStr, "DFP_LOW")) {
                pBIOSInfo->CRTDIPort = VIA_DI_DFPLOW;
            } else if (!xf86NameCmp(pStr, "DVP0")) {
                pBIOSInfo->CRTDIPort = VIA_DI_DVP0;
            } else if (!xf86NameCmp(pStr, "DVP1")) {
                pBIOSInfo->CRTDIPort = VIA_DI_DVP1;
            } 
        }
    }
}


int 
TranslateDIPortNameToID(char *pStr)
{
    if (!xf86NameCmp(pStr, "DFP_HIGH"))
        return VIA_DI_DFPHIGH;
    else if (!xf86NameCmp(pStr, "DFP_LOW"))
        return VIA_DI_DFPLOW;
    else if (!xf86NameCmp(pStr, "DFP_HIGHLOW"))
        return VIA_DI_DFPHIGHLOW;
    else if (!xf86NameCmp(pStr, "DVP0"))
        return VIA_DI_DVP0;
    else if (!xf86NameCmp(pStr, "DVP1"))
        return VIA_DI_DVP1;
    else if (!xf86NameCmp(pStr, "LVDS0"))
        return VIA_DI_LVDS0;
    else if (!xf86NameCmp(pStr, "LVDS1"))
        return VIA_DI_LVDS1;
    else if (!xf86NameCmp(pStr, "LVDS0LVDS1"))
        return VIA_DI_LVDS0LVDS1;    
    else if (!xf86NameCmp(pStr, "TMDS"))
        return VIA_DI_TMDS;        
    else if (!xf86NameCmp(pStr, "TTL"))
        return VIA_DI_TTL;        
    else if (!xf86NameCmp(pStr, "LVDS1DVP1"))
        return VIA_DI_LVDS1DVP1;
    else        
        return VIA_DI_NONE; /* Invalid port. */
}


static void
viaSetDefaultLCDPanelChannelInfo( VIABIOSInfoPtr pBIOSInfo)
{
    switch(pBIOSInfo->LVDSSettingInfo.PanelSizeID){
    case VIA_1280X1024:
    case VIA_1400X1050:
    case VIA_1440X900:
    case VIA_1600X1200:
        pBIOSInfo->LVDSSettingInfo.IsDualEdge=TRUE;
		pBIOSInfo->LVDSSettingInfo2.IsDualEdge=TRUE;
        break;
    default:
        pBIOSInfo->LVDSSettingInfo.IsDualEdge=FALSE;
		pBIOSInfo->LVDSSettingInfo2.IsDualEdge=FALSE;
        break;
    }
}


static void 
VIAGetOptions_LCDInfo(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;    
    char    *pStr = NULL;
    
    /* Initialize LCD panel size: */
    pBIOSInfo->LVDSSettingInfo.PanelSizeID = VIA_INVALID;
    
    switch (pVia->Chipset) {
    case VIA_CX700:
    case VIA_VX800:
    case VIA_VX855:
    case VIA_VX900:
        pBIOSInfo->LVDSSettingInfo.ChipID = VIA_INTEGRATED_LVDS;
        break;
    default:
        pBIOSInfo->LVDSSettingInfo.ChipID = VIA_Hardwired;
        break;
    }

    /* LCD MSB/LSB Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_LCD_MSB_ENABLE, FALSE)) {
        pBIOSInfo->LVDSSettingInfo.MsbEnable=TRUE;
		pBIOSInfo->LVDSSettingInfo2.MsbEnable=TRUE;
    } else {
        pBIOSInfo->LVDSSettingInfo.MsbEnable=FALSE;
		pBIOSInfo->LVDSSettingInfo2.MsbEnable=FALSE;
    }

    /* Set LCD Panel Info. */
    if (pStr = xf86GetOptValString(VIAOptions, OPTION_LCD_PANEL_SIZE)) {
		char * b = strdup(pStr);
		pStr = strtok(b, "x");
        CARD32 panelWidth = (CARD32)atoi(pStr);
		pStr = strtok(NULL, "x");
		CARD32 panelHeight = (CARD32)atoi(pStr);
		/*Check if the panel size is valid*/
		if (isPanelSizeValid(panelWidth, panelHeight)) {
			pBIOSInfo->LVDSSettingInfo.PanelSizeID = 
				VIA_MAKE_ID(panelWidth, panelHeight);

            viaSetDefaultLCDPanelChannelInfo(pBIOSInfo);

			xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "LCD Panel Size is %lux%lu.\n", 
					   panelWidth, panelHeight);
		} else {
			xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "LCD Panel size %lux%lu is invalid!\n", 
        		panelWidth, panelHeight);	
		}
    }
    
    /* LCD Dithering Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_LCD_NO_DITHERING, FALSE)) {
        pBIOSInfo->LVDSSettingInfo.IsDithering=FALSE;
		pBIOSInfo->LVDSSettingInfo2.IsDithering=FALSE;
    } else {
        pBIOSInfo->LVDSSettingInfo.IsDithering=TRUE;
		pBIOSInfo->LVDSSettingInfo2.IsDithering=TRUE;
    }

    /* LCD DualChannel Option */
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_LCD_DUAL_CHANNEL))) {
        if ((*pStr == '\0') ||
            !xf86NameCmp(pStr, "TRUE") ||
            !xf86NameCmp(pStr, "True") ||
            !xf86NameCmp(pStr, "true")) {
            pBIOSInfo->LVDSSettingInfo.IsDualEdge = TRUE;
		    pBIOSInfo->LVDSSettingInfo2.IsDualEdge = TRUE;
        } else if (!xf86NameCmp(pStr, "FALSE") ||
                   !xf86NameCmp(pStr, "False") ||
                   !xf86NameCmp(pStr, "false")) {
            pBIOSInfo->LVDSSettingInfo.IsDualEdge = FALSE;
		    pBIOSInfo->LVDSSettingInfo2.IsDualEdge = FALSE;
        }    
    } 
    
    /* LCD Center/Expand Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_CENTER, FALSE)) {
        pBIOSInfo->LVDSSettingInfo.Center = TRUE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "(OPTION) LCD Center is On\n");
    } else {
        pBIOSInfo->LVDSSettingInfo.Center = FALSE;
    }

    /* Bus Width Option */
    pBIOSInfo->LVDSSettingInfo.BusWidth = VIA_DI_12BIT;

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_LCD_BUS_WIDTH))) {
        if (!xf86NameCmp(pStr, "12BIT")) {
            pBIOSInfo->LVDSSettingInfo.BusWidth = VIA_DI_12BIT;
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "LCD Bus Width is 12BIT\n");
        } else if (!xf86NameCmp(pStr, "24BIT")) {
            pBIOSInfo->LVDSSettingInfo.BusWidth = VIA_DI_24BIT;
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "LCD Bus Width is 24BIT\n");
        }
    }

    /* Support LCD port selection */
    pBIOSInfo->LVDSSettingInfo.DIPort = VIA_DI_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_LCD_PORT))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "LCD port set to %s.\n", pStr));
        pBIOSInfo->LVDSSettingInfo.DIPort = TranslateDIPortNameToID(pStr); 

        if (pBIOSInfo->LVDSSettingInfo.DIPort == VIA_DI_TTL) {
            if (pBIOSInfo->Chipset == VIA_VX855)           
                pBIOSInfo->LVDSSettingInfo.DIPort = (VIA_DI_LVDS0LVDS1DVP1 | VIA_DI_TTL);
            else
                pBIOSInfo->LVDSSettingInfo.DIPort = (VIA_DI_LVDS0DVP1 | VIA_DI_TTL);
        }
        
        /* For dual LCD using DVP1+LVDS1 */
        if (pBIOSInfo->ActiveDevice == (VIA_DEVICE_LCD|VIA_DEVICE_LCD2)) {
            if (pBIOSInfo->LVDSSettingInfo.DIPort == VIA_DI_LVDS1DVP1) {
                pBIOSInfo->LVDSSettingInfo.DIPort = VIA_DI_DVP1;
                pBIOSInfo->LVDSSettingInfo2.DIPort = VIA_DI_LVDS1;                
            }
        }            
    }
}

static void 
VIAGetOptions_TVInfo(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;
    
    /* TV DotCrawl Enable Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_TVDOTCRAWL, FALSE)) {
        pBIOSInfo->TVSettingInfo.TVDotCrawl = TRUE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DotCrawl is Enable\n");
    } else {
        pBIOSInfo->TVSettingInfo.TVDotCrawl = FALSE;
    }

    /* TV Type: */
    pBIOSInfo->TVSettingInfo.TVType = TVTYPE_NONE;
    pBIOSInfo->TVSettingInfo.TVTypeIsNTSCJ = FALSE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TVTYPE))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "TV Type is %s.\n", pStr));
        
        if (!xf86NameCmp(pStr, "NTSC"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_NTSC;
        else if (!xf86NameCmp(pStr, "NTSC_J")) {
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_NTSC;
            pBIOSInfo->TVSettingInfo.TVTypeIsNTSCJ = TRUE;
        } else if(!xf86NameCmp(pStr, "PAL"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_PAL;
        else if(!xf86NameCmp(pStr, "480P"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_480P;
        else if(!xf86NameCmp(pStr, "576P"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_576P;
        else if(!xf86NameCmp(pStr, "720P"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_720P;                
        else if(!xf86NameCmp(pStr, "1080I"))
            pBIOSInfo->TVSettingInfo.TVType = TVTYPE_1080I;                
    }

    /* TV out put signal Option */
    pBIOSInfo->TVSettingInfo.TVOutput = TV_OUTPUT_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TVOUTPUT))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "TV Output Signal is %s\n", pStr));
        pBIOSInfo->TVSettingInfo.TVOutput = VIATranslateTVOutputNameToID(pStr);
    }

    if(pVia->DISP3DScalingCaps & VIA_DEVICE_TV) {
        /* TV Shrink Option */
        if ((TVTYPE_1080I==pBIOSInfo->TVSettingInfo.TVType )||
            (TVTYPE_720P==pBIOSInfo->TVSettingInfo.TVType))
            pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVNORMAL;
        else {
            pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVOVER;
            if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TVVSCAN)))
            {
                if (!xf86NameCmp(pStr, "under") || !xf86NameCmp(pStr, "normal"))
                    pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVNORMAL;
                else if (!xf86NameCmp(pStr, "fit"))
                    pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVFIT;
                else if (!xf86NameCmp(pStr, "over"))
                    pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVOVER;
            }
        }
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "TV Shrink scan mode is %d\n", 
                         pBIOSInfo->TVSettingInfo.TVVScan));
    }
    else
    {
         /* TV Standard Option */
        pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVNORMAL;
	    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TVVSCAN))) {
    		if (!xf86NameCmp(pStr, "under") || !xf86NameCmp(pStr, "normal"))
    			pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVNORMAL;
    		else if (!xf86NameCmp(pStr, "fit"))
    			pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVFIT;
    		else if (!xf86NameCmp(pStr, "over"))
    			pBIOSInfo->TVSettingInfo.TVVScan = VIA_TVOVER;
    	}
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "TV scan mode is %d\n", 
               pBIOSInfo->TVSettingInfo.TVVScan));
    }

#ifndef CS_FLAG_NO_TV_SCAL
    /* For H-Scaling adjustment: */
    /* We can use this option to specify the number of pixels to adjust each time. */
    pBIOSInfo->TVSettingInfo.TVHScale = VIA_TVSCALE5;
    if (xf86GetOptValInteger(VIAOptions, OPTION_TVHSCALE, 
                    &(pBIOSInfo->TVSettingInfo.TVHScale))) {
        if ((pBIOSInfo->TVSettingInfo.TVHScale < VIA_TVSCALE1) ||
            (pBIOSInfo->TVSettingInfo.TVHScale > VIA_TVSCALE5))
            pBIOSInfo->TVSettingInfo.TVHScale = VIA_TVSCALE5;
    }

    /* For V-Scaling adjustment: */
    /* We can use this option to specify the number of lines to adjust each time. */
    pBIOSInfo->TVSettingInfo.TVVScale = VIA_TVSCALE5;
    if (xf86GetOptValInteger(VIAOptions, OPTION_TVVSCALE, 
        &(pBIOSInfo->TVSettingInfo.TVVScale))) {
        if ((pBIOSInfo->TVSettingInfo.TVVScale < VIA_TVSCALE1) ||
            (pBIOSInfo->TVSettingInfo.TVVScale > VIA_TVSCALE5))
            pBIOSInfo->TVSettingInfo.TVVScale = VIA_TVSCALE5;
    }
#endif /* CS_FLAG_NO_TV_SCAL */

    pBIOSInfo->TVDIPort = VIA_DI_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TV_PORT))) {
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "TV port set to %s.\n", pStr);
        pBIOSInfo->TVDIPort = TranslateDIPortNameToID(pStr);        
    }
}


static void 
VIAGetOptions_TV2Info(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;    


    /* TV2 DotCrawl Enable Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_TV2DOTCRAWL, FALSE)) {
        pBIOSInfo->TVSettingInfo2.TVDotCrawl = TRUE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: TV2 DotCrawl Enable\n");
    } else {
        pBIOSInfo->TVSettingInfo2.TVDotCrawl = FALSE;
    }

    /* TV2 Type: */
    pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TV2TYPE))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: TV2 Type is %s.\n", pStr));
        
        if (!xf86NameCmp(pStr, "NTSC")) 
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_NTSC;
        else if(!xf86NameCmp(pStr, "PAL"))
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_PAL;
        else if(!xf86NameCmp(pStr, "480P"))
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_480P;
        else if(!xf86NameCmp(pStr, "576P"))
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_576P;
        else if(!xf86NameCmp(pStr, "720P"))
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_720P;                            
        else if(!xf86NameCmp(pStr, "1080I"))
            pBIOSInfo->TVSettingInfo2.TVType = TVTYPE_1080I;                            
    }
    
    /* TV2 out put signal Option */
    pBIOSInfo->TVSettingInfo2.TVOutput = TV_OUTPUT_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_TV2OUTPUT))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
              "Option: TV2 Output Signal is %s\n", pStr));
        pBIOSInfo->TVSettingInfo2.TVOutput = VIATranslateTVOutputNameToID(pStr);
    }

    /* TV2 Standard Option: TV2 only could be internal TV, so it only use normal scan. */
    pBIOSInfo->TVSettingInfo2.TVVScan = VIA_TVNORMAL;

#ifndef CS_FLAG_NO_TV_SCAL
    /* For TV2 H-Scaling adjustment: */
    /* We can use this option to specify the number of pixels to adjust each time. */
    pBIOSInfo->TVSettingInfo2.TVHScale = VIA_TVSCALE5;
    if (xf86GetOptValInteger(VIAOptions, OPTION_TV2HSCALE, 
        &(pBIOSInfo->TVSettingInfo2.TVHScale))) {
        if ((pBIOSInfo->TVSettingInfo2.TVHScale < VIA_TVSCALE1) ||
            (pBIOSInfo->TVSettingInfo2.TVHScale > VIA_TVSCALE5)) {
            pBIOSInfo->TVSettingInfo2.TVHScale = VIA_TVSCALE5;
        }
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Option: TV2 HScale is %d\n", pBIOSInfo->TVSettingInfo2.TVHScale));
    }
    
    /* For TV2 V-Scaling adjustment: */
    /* We can use this option to specify the number of lines to adjust each time. */
    pBIOSInfo->TVSettingInfo2.TVVScale = VIA_TVSCALE5;
    if (xf86GetOptValInteger(VIAOptions, OPTION_TV2VSCALE, 
            &(pBIOSInfo->TVSettingInfo2.TVVScale))) {
        if ((pBIOSInfo->TVSettingInfo2.TVVScale < VIA_TVSCALE1) ||
            (pBIOSInfo->TVSettingInfo2.TVVScale > VIA_TVSCALE5)) {
            pBIOSInfo->TVSettingInfo2.TVVScale = VIA_TVSCALE5;
        }
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Option: TV2 VScale is %d\n", pBIOSInfo->TVSettingInfo2.TVVScale));
    }
#endif /* CS_FLAG_NO_TV_SCAL */

}



static void 
VIAGetOptions_DVIInfo(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;

    /* Initialize DVI panel size: */
    pBIOSInfo->TMDSSettingInfo.DFPSize = M_INVALID;
    pBIOSInfo->TMDSSettingInfo2.DFPSize = M_INVALID;

    /* Support DVI port selection */
    pBIOSInfo->TMDSSettingInfo.DFPDIPort = VIA_DI_NONE;
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DVI_PORT))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DVI port set to %s.\n", pStr));
        
        if (!xf86NameCmp(pStr, "DFP_HIGH")) {
            pBIOSInfo->TMDSSettingInfo.DFPDIPort = VIA_DI_DFPHIGH;
        } else if (!xf86NameCmp(pStr, "DFP_LOW"))
            pBIOSInfo->TMDSSettingInfo.DFPDIPort = VIA_DI_DFPLOW;
        else if (!xf86NameCmp(pStr, "DVP0"))
            pBIOSInfo->TMDSSettingInfo.DFPDIPort = VIA_DI_DVP0;
        else if (!xf86NameCmp(pStr, "DVP1"))
            pBIOSInfo->TMDSSettingInfo.DFPDIPort = VIA_DI_DVP1;
    }
}


static void 
VIAParseDPAValues(char *pStr, int *pReturnedBuffer, int BufSize)
{
    char *pToken;
    int i = 0;

    pToken = strtok(pStr, ",");
    
    if (pToken) {
        *pReturnedBuffer = atoi(pToken);
        pReturnedBuffer++;
        i++;
    }

    for (; i < BufSize; i++) {
        pToken = strtok('\0', ",");

        if (pToken) {
            *pReturnedBuffer = atoi(pToken);
            pReturnedBuffer++;
        }
    }
}

static void 
VIAGetOptions_DPASettingInfo(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;
    int     TmpBuffer[3];

    pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx = FALSE;

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_DVP0))) {
        /* The data we get should be "v1,v2,v3", they are for */
        /* "Clock Polarity/Adjust,Data Driving,Clock Driving". */
        VIAParseDPAValues(pStr, TmpBuffer, 3);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DVP0 DPA=%d, DataDriving=%d, ClockDriving=%d.\n", 
              TmpBuffer[0], TmpBuffer[1], TmpBuffer[2]));
        
        pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx = TRUE;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP0 = TmpBuffer[0] & 0x0F;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP0DataDri_S = (TmpBuffer[1] & BIT0) << 1;     /* for SR1B[1] */
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP0DataDri_S1 = (TmpBuffer[1] & BIT1) << 4;    /* for SR2A[5] */
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP0ClockDri_S = (TmpBuffer[2] & BIT0) << 2;    /* for SR1E[2] */
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP0ClockDri_S1 = (TmpBuffer[2] & BIT1) << 3;   /* for SR2A[4] */
    }

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_DVP1))) {
        /* The data we get should be "v1,v2,v3", they are for */
        /* "Clock Polarity/Adjust,Data Driving,Clock Driving". */
        VIAParseDPAValues(pStr, TmpBuffer, 3);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DVP1 DPA=%d, DataDriving=%d, ClockDriving=%d.\n", 
              TmpBuffer[0], TmpBuffer[1], TmpBuffer[2]));

        pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx = TRUE;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP1 = TmpBuffer[0] & 0x0F;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DVP1Driving = 
            ((TmpBuffer[2] & 0x03) << 2) | (TmpBuffer[1] & 0x03);
    }

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_DFPHIGH))) {
        /* The data we get is for "Clock Polarity/Adjust" */
        VIAParseDPAValues(pStr, TmpBuffer, 1);

        pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx = TRUE;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DFPHigh = TmpBuffer[0] & 0x0F;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DFPHIGH DPA = %d.\n", 
            pBIOSInfo->UserDPASettingInfo.GfxDPA.DFPHigh));
    }

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_DFPLOW))) {
        /* The data we get is for "Clock Polarity/Adjust" */
        VIAParseDPAValues(pStr, TmpBuffer, 1);

        pBIOSInfo->UserDPASettingInfo.HasUserSetting_Gfx = TRUE;
        pBIOSInfo->UserDPASettingInfo.GfxDPA.DFPLow = TmpBuffer[0] & 0x0F;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "DFPLOW DPA = %d.\n", pBIOSInfo->UserDPASettingInfo.GfxDPA.DFPLow));
    }

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_VT1636))) {
        /* The data we get should be "v1,v2", they are for "CLK_SEL_ST1, CLK_SEL_ST2" */
        VIAParseDPAValues(pStr, TmpBuffer, 2);

        pBIOSInfo->UserDPASettingInfo.HasUserSetting_VT1636 = TRUE;        
        pBIOSInfo->UserDPASettingInfo.VT1636DPA.CLK_SEL_ST1 = TmpBuffer[0] & 0x1F;
        pBIOSInfo->UserDPASettingInfo.VT1636DPA.CLK_SEL_ST2 = TmpBuffer[1] & 0x0F;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "VT1636 DPA: CLK_SEL_ST1=%d, CLK_SEL_ST2=%d.\n", 
            pBIOSInfo->UserDPASettingInfo.VT1636DPA.CLK_SEL_ST1,
            pBIOSInfo->UserDPASettingInfo.VT1636DPA.CLK_SEL_ST2));
    } else {
        pBIOSInfo->UserDPASettingInfo.HasUserSetting_VT1636 = FALSE;
    }

    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_DPA_SETTING_VT1625))) {
        /* The data we get is for VT1625 DPA. */
        VIAParseDPAValues(pStr, TmpBuffer, 1);
        
        pBIOSInfo->UserDPASettingInfo.HasUserSetting_VT1625 = TRUE;        
        pBIOSInfo->UserDPASettingInfo.VT1625DPA = TmpBuffer[0] & 0x1F;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "VT1625 DPA = %d.\n", 
            pBIOSInfo->UserDPASettingInfo.VT1625DPA));
    } else {
        pBIOSInfo->UserDPASettingInfo.HasUserSetting_VT1625 = FALSE;
    }
}


static void 
VIAGetOptions_RotateInfo(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    viaGfxInfoPtr viaGfxInfo = pVia->pVIAEnt->pVidData->viaGfxInfo;       
    CARD32 RotateType = VIA_ROTATE_TYPE_HW;
    char* s = NULL;

    viaGfxInfo->hwIconEnabled = FALSE;
    
    /* Initilize */    
    pVia->RotateDegree = VIA_ROTATE_DEGREE_0;
    pVia->IsHWRotateEnabled = FALSE;
    
    if ((s = xf86GetOptValString(VIAOptions, OPTION_ROTATE))) {                    
        if (!xf86NameCmp(s, "CW")) {                        
            pVia->RotateDegree = VIA_ROTATE_DEGREE_90;                        
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Rotating screen clockwise.\n"));
        } else if(!xf86NameCmp(s, "CCW")) {            
            pVia->RotateDegree = VIA_ROTATE_DEGREE_270;                        
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,  "Rotating screen counter-clockwise. \n"));
        } else if(!xf86NameCmp(s, "UD")) {            
            pVia->RotateDegree = VIA_ROTATE_DEGREE_180;                                    
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,  "Rotating screen upside-down. \n"));
        } else {
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "\"%s\" is not a valid"
                       "value for Option \"Rotate\"\n", s));
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                       "Valid options are \"CW\" or \"CCW\" or \"UD\" \n"));
        }        
    }    
       
    if ((s = xf86GetOptValString(VIAOptions, OPTION_ROTATE_TYPE))) {        
        if (!xf86NameCmp(s, "SW")) {
            RotateType = VIA_ROTATE_TYPE_SW;
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Use Software Rotation - Acceleration disabled\n");
        } else if(!xf86NameCmp(s, "HW")) {            
            RotateType = VIA_ROTATE_TYPE_HW;          
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Use Hardware Rotation.\n");
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "\"%s\" is not a valid"
                       "value for Option \"RotateType\"\n", s);
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                       "Valid options are \"SW\" or \"HW\"  \n");
        }
    }

    if (pVia->IsSecondary)
        viaGfxInfo->screenInfo[1].rotatetype = RotateType;
    else
        viaGfxInfo->screenInfo[0].rotatetype = RotateType;
    
    if (pVia->RotateDegree != VIA_ROTATE_DEGREE_0) {            
        switch (RotateType) {
            /* accel is disabled below for shadowFB */
            case VIA_ROTATE_TYPE_SW:            
                pVia->shadowFB = TRUE;
                break;
                
            case VIA_ROTATE_TYPE_HW:
                pVia->IsHWRotateEnabled = TRUE;
                break;                
        }

        viaGfxInfo->hwIconEnabled = TRUE;
    }
}

static void 
VIAGetOptions_Video(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    viaGfxInfoPtr viaGfxInfo = pVia->pVIAEnt->pVidData->viaGfxInfo;
    int VideoColorKey;
    
    if (xf86GetOptValInteger(VIAOptions, OPTION_VIDEO_COLOR_KEY, &VideoColorKey)) {
        viaGfxInfo->preferedColorKey = VideoColorKey;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "video color key set to 0x%lx\n", 
            viaGfxInfo->preferedColorKey);
    } else {
        viaGfxInfo->preferedColorKey = (1 << pScrn->offset.red) |
                                          (1 << pScrn->offset.green) |
                                          (1 << pScrn->offset.blue);
    }
}

static void 
VIAGetOptions_general(ScrnInfoPtr pScrn)
{
    VIAPtr          pVia;
    VIABIOSInfoPtr  pBIOSInfo;
    MessageType     from = X_DEFAULT;
    char            *s = NULL;
    pVia = VIAPTR(pScrn);
    pBIOSInfo = pVia->pBIOSInfo;    
    
    if (!xf86ReturnOptValBool(VIAOptions, OPTION_DRI, TRUE)) {
        pVia->directRenderingEnabled = FALSE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: DRI is disabled\n");
    } else {
        pVia->directRenderingEnabled = TRUE;
    }

#ifndef CS_FLAG_NO_TV_SCAL
    /* Initialize the DISPLAY 3D scaling CAP */
    /*modified the variable type,and change the function of this variable*/
    pVia->DISP3DScalingCaps = FALSE;
    /* the max number should be 2, so when you add more device need 
       this function, please check the number*/
    pVia->NumOfDevNeed3dScl = 0;
    if (xf86IsOptionSet(VIAOptions, OPTION_TVDISP3DScaling)) {
        pVia->DISP3DScalingCaps |= DISPLAY_3D_SCALING_SUPPORT  + VIA_DEVICE_TV; 
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Display SW scaling %s.\n",
                  ( pVia->DISP3DScalingCaps&VIA_DEVICE_TV )? "enabled" : "disabled");
    }
#endif /* CS_FLAG_NO_TV_SCAL */

#ifndef CS_FLAG_NO_HDMI_SCAL
    if (xf86IsOptionSet(VIAOptions, OPTION_HDMIDISP3DScaling)) {
        pVia->DISP3DScalingCaps |= DISPLAY_3D_SCALING_SUPPORT + VIA_DEVICE_HDMI; 

        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
                   "Option: HDMI Display SW scaling %s.\n",
                   (pVia->DISP3DScalingCaps&VIA_DEVICE_HDMI) ? "enabled" : "disabled");
    }
#endif /* CS_FLAG_NO_HDMI_SCAL */

#ifndef CS_FLAG_NO_LCD_SCAL
    if (xf86IsOptionSet(VIAOptions, OPTION_LCDDISP3DScaling)) {
        pVia->DISP3DScalingCaps |= DISPLAY_3D_SCALING_SUPPORT  + VIA_DEVICE_LCD; 
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: LCD Display SW scaling %s.\n",
                   (pVia->DISP3DScalingCaps&VIA_DEVICE_LCD) ? "enabled" : "disabled");
    }
#endif /* CS_FLAG_NO_LCD_SCAL */

    if (xf86IsOptionSet(VIAOptions, OPTION_GOOD_PERFORMANCE)) {
        pVia->GoodPerformance = DISPLAY_3D_SCALING_GOOD_PERFORMANCE; 
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: 3D scaling use good performance mode.\n");
    }

    from = X_DEFAULT;

    pScrn->videoRam = 0;
    if(xf86GetOptValInteger(VIAOptions, OPTION_VIDEORAM, &pScrn->videoRam)) {
        DEBUG(xf86DrvMsg( pScrn->scrnIndex, X_CONFIG,
                    "Option: VideoRAM %dkB\n", pScrn->videoRam));
    }

    if (xf86ReturnOptValBool(VIAOptions, OPTION_DISABLEVQ, FALSE)) {
        pVia->VQEnable = FALSE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Option: DisableVQ -VQ Disabled\n"));
    } else {
        pVia->VQEnable = TRUE;
    }
    
    /* Get rotate option. */
    VIAGetOptions_RotateInfo(pScrn);   

    if (xf86ReturnOptValBool(VIAOptions, OPTION_NOACCEL, FALSE)) {
        pVia->NoAccel = TRUE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Option: NoAccel -Acceleration Disabled\n"));
    } else {
        pVia->NoAccel = FALSE;
    }

    if (xf86ReturnOptValBool(VIAOptions, OPTION_NOCOMPOSITE, FALSE)) {
        pVia->NoComposite = TRUE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Option: NoComposite - Accelerated Compositing Disabled\n"));
    } else {
        pVia->NoComposite = FALSE;
    }

    /* NoDDCValue Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_NODDCVALUE, FALSE)) {        
        pBIOSInfo->NoDDCValue = TRUE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
            "Option: Not using DDC probed value to set HorizSync & VertRefresh\n"));
    } else {       
        pBIOSInfo->NoDDCValue = FALSE;
    }

    /* Interlace Mode Option */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_INTERLACE_MODE, FALSE)) {        
        pBIOSInfo->IsEnableInterlaceMode = TRUE;
       DEBUG( xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Enable Interlace Mode. \n"));
    } else {       
        pBIOSInfo->IsEnableInterlaceMode = FALSE;
    }

    pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_NONE;
    if ((s = xf86GetOptValString(VIAOptions, OPTION_HDMI_PORT))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "HDMI port set to %s.\n", s));
        
        if (!xf86NameCmp(s, "DFP_HIGH")) 
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DFPHIGH;
        else if (!xf86NameCmp(s, "DFP_LOW"))
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DFPLOW;
        else if (!xf86NameCmp(s, "DVP0"))
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DVP0;
        else if (!xf86NameCmp(s, "DVP1"))
            pBIOSInfo->HDMISettingInfo.HDMIDIPort = VIA_DI_DVP1;
    }

    if (pVia->shadowFB && !pVia->NoAccel) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "HW acceleration not supported with \"shadowFB\".\n"));
            pVia->NoAccel = TRUE;
    }    
    
    return ;
}


static void 
VIAGetOptions_Hotkey(ScrnInfoPtr pScrn)
{
    VIAPtr          pVia;
    VIABIOSInfoPtr  pBIOSInfo;    
    pVia = VIAPTR(pScrn);
    pBIOSInfo = pVia->pBIOSInfo;
    char    *pStr = NULL;
    char    *pTmpStr = NULL;       
    CARD32  i = 0;

    pVia->IsForceHotkeySwitch = FALSE;
    /* Hotkey triggered by a timer. */
    if ((pStr = xf86GetOptValString(VIAOptions, OPTION_FN_HOTKEY))) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
            "Option: Enable a timer to active hotkey.\n"));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Device Sequence = %s\n", pStr));
 
        pTmpStr = strtok(pStr, ",");
        while (pTmpStr != NULL) {
            if (!xf86NameCmp(pTmpStr, "CRT"))
                pVia->deviceSequence[i] = VIA_DEVICE_CRT1;
            else if (!xf86NameCmp(pTmpStr, "LCD"))
                pVia->deviceSequence[i] = VIA_DEVICE_LCD;
            else if (!xf86NameCmp(pTmpStr, "DVI"))
                pVia->deviceSequence[i] = VIA_DEVICE_DFP;
            else if (!xf86NameCmp(pTmpStr, "TV"))
                pVia->deviceSequence[i] = VIA_DEVICE_TV;
            else if (!xf86NameCmp(pTmpStr, "FORCE"))
                pVia->IsForceHotkeySwitch = TRUE;
            else
                ErrorF("Unknow Option %s for hotkey used. \n", pTmpStr);
                        
            pTmpStr = strtok(NULL, ",");
            i++;
        }

        /* It should has more than one prefered device at least. */
        if (i>1)
            pVia->IsHotkeyTimerEnabled = TRUE;
        else
            pVia->IsHotkeyTimerEnabled = FALSE;
    }
}


static void 
VIAProcessCfgOptions(ScrnInfoPtr pScrn)
{
    VIAPtr          pVia;
    VIABIOSInfoPtr  pBIOSInfo;

    pVia = VIAPTR(pScrn);
    pBIOSInfo = pVia->pBIOSInfo;

    xf86CollectOptions(pScrn, NULL);
    xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, VIAOptions);  
    
    VIAGetOptions_general(pScrn);

    VIAGetOptions_Hotkey(pScrn);
    
     /* Probe the capabilities about rotation. */
    ProbeRotateCaps(pScrn);   

    VIAGetOptions_ActiveDevice(pScrn);
    VIAGetOptions_LCDInfo(pScrn);
    VIAGetOptions_TVInfo(pScrn);    
    VIAGetOptions_DVIInfo(pScrn);
    VIAGetOptions_DPASettingInfo(pScrn);

    /* For CX700 InTV + ExtTV */
    if (pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2)
        VIAGetOptions_TV2Info(pScrn);
    
    /*here add this function because  chipset id is identified just above*/ 
    VIAGetOptions_Cursor(pScrn);
    VIAGetOptions_Video(pScrn);

    return ;    
}



static void 
VIAProbeChipset(ScrnInfoPtr pScrn, EntityInfoPtr pEnt)
{
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAProbeChipSet!\n"));
    
    VIAPtr pVia;
    VIABIOSInfoPtr pBIOSInfo;
    MessageType     from = X_DEFAULT;

    pVia = VIAPTR(pScrn);
    pBIOSInfo = pVia->pBIOSInfo;

    if (pEnt->device->chipset && *pEnt->device->chipset) {
        pScrn->chipset = pEnt->device->chipset;
        pVia->ChipId = pEnt->device->chipID;
        pVia->Chipset = xf86StringToToken(VIAChipsets, pScrn->chipset);
        from = X_CONFIG;
    } else if (pEnt->device->chipID >= 0) {
        pVia->ChipId = pEnt->device->chipID;
        pVia->Chipset = LookupChipID(VIAPciChipsets, pVia->ChipId);
        pScrn->chipset = (char *)xf86TokenToString(VIAChipsets,
                                                   pVia->Chipset);
        from = X_CONFIG;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipID override: 0x%04X\n",
                   pEnt->device->chipID));
    } else {
        from = X_PROBED;
        pVia->ChipId = DEVICE_ID(pVia->PciInfo);
        pVia->Chipset = LookupChipID(VIAPciChipsets, pVia->ChipId);
        pScrn->chipset = (char *)xf86TokenToString(VIAChipsets,
                                                   pVia->Chipset);
    }

    if (pEnt->device->chipRev >= 0) {
        pVia->ChipRev = pEnt->device->chipRev;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipRev override: %d\n",
                   pVia->ChipRev));
    } else {
        /* Read PCI bus 0, dev 0, function 0, index 0xF6 to get chip rev. */
		uint8_t ChipRev;
#if XSERVER_LIBPCIACCESS
		pci_device_cfg_read_u8(VIAGetHostBridge(0, 0, 0), &ChipRev, 0xF6);
#else
		ChipRev = pciReadByte(pciTag(0, 0, 0), 0xF6);
#endif
		pVia->ChipRev = ChipRev;
    }
    
    xf86DrvMsg(pScrn->scrnIndex, from, "Chipset: \"%s\"\n", pScrn->chipset);
    xf86DrvMsg(pScrn->scrnIndex, from, "Chipset Rev.: %d\n", pVia->ChipRev);
	
    pBIOSInfo->Chipset = pVia->Chipset;
    pBIOSInfo->ChipRev = pVia->ChipRev;
}

void VIAFindMaxResFromEstablishedTiming(xf86MonPtr MonitorConf, int* MaxH, int* MaxV)
{               
    if (MonitorConf->timings1.t1 & 0xD0) {
        *MaxH = 720;
        *MaxV = 400;
    }
    
    if (MonitorConf->timings1.t1 & 0x3D) {
        *MaxH = 640;
        *MaxV = 480;
    }

    if (MonitorConf->timings1.t1 & 0x03) {
        *MaxH = 800;
        *MaxV = 600;
    }

    if (MonitorConf->timings1.t2 & 0xD0) {
        *MaxH = 800;
        *MaxV = 600;
    }

    if (MonitorConf->timings1.t2 & 0x20) {
        *MaxH = 832;
        *MaxV = 624;
    }

    if (MonitorConf->timings1.t2 & 0x1E) {
        *MaxH = 1024;
        *MaxV = 768;
    }

    if (MonitorConf->timings1.t2 & 0x01) {
        *MaxH = 1280;
        *MaxV = 1024;
    }        
}

xf86MonPtr 
VIADoEDID(ScrnInfoPtr pScrn, pointer pDDCModule)
{
    xf86MonPtr      pMonitor = NULL;
    unsigned char   *edidBuffer = NULL; 
    VIAPtr          pVia= VIAPTR(pScrn); ;
    VIABIOSInfoPtr  pBIOSInfo=pVia->pBIOSInfo;     
    vbeInfoPtr pVbe = pVia->pVbe;
    int edidSource = 0;

    if(xf86IsEntityShared(pScrn->entityList[0])) {
        if (pVia->IsSecondary){
            edidSource = pBIOSInfo->ActiveDevice - pBIOSInfo->PrimaryDevice;
            if(!edidSource){ 
                edidSource = pBIOSInfo->ActiveDevice;
            }
        }else
            edidSource = pBIOSInfo->PrimaryDevice;    
    } else if (pBIOSInfo->PrimaryDevice)
        edidSource = pBIOSInfo->PrimaryDevice;    
    else
        edidSource = pBIOSInfo->ActiveDevice;
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "edidSource = 0x%x\n", edidSource));       

    edidBuffer = xcalloc(1,sizeof(unsigned char) * (128));

    if (edidSource & VIA_DEVICE_LCD) {
        /* Read EDID v1.1 Block */  
        if (VIALVDSIdentify(pBIOSInfo)) {
            switch (pBIOSInfo->LVDSSettingInfo.ChipID) {
                case VIA_Hardwired:          
                    pBIOSInfo->LVDSSettingInfo.I2CPort = 0x31;
                    break;

                case VIA_INTEGRATED_LVDS: 
                    if (pBIOSInfo->Chipset == VIA_VX800)
                        pBIOSInfo->LVDSSettingInfo.I2CPort = 0x31;
                    else
                        pBIOSInfo->LVDSSettingInfo.I2CPort = 0x2C;                    
                    break;

                case VIA_VT1636:
                case VIA_VT1631: 
                    /* use the serial port which is used to sense LVDS */
                    break;      

                default:
                    pBIOSInfo->LVDSSettingInfo.I2CPort = 0x31;                    
                    break;
            }

            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                  "Use Port %#x to probe LCD EDID\n", pBIOSInfo->LVDSSettingInfo.I2CPort));

            viaSerialReadBytes(pBIOSInfo->LVDSSettingInfo.I2CPort, 0xA0, 0, edidBuffer, 128);
        }

        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex,edidBuffer);    

        if (pMonitor)
            pBIOSInfo->EDIDType = VIA_DEVICE_LCD;
                
    } else if (edidSource & VIA_DEVICE_DFP) {
        /* Read EDID v1.1 Block */
        if(VIATMDSIdentify_External(pBIOSInfo, &pBIOSInfo->TMDSSettingInfo)) {
            viaSerialReadBytes(pBIOSInfo->TMDSSettingInfo.I2CPort, 0xA0, 0, edidBuffer, 128);
        }
        
        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex,edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = edidSource;
        
    }else if (edidSource & VIA_DEVICE_DFP2) {
        if(VIATMDSIdentify_IntegratedTMDS(pBIOSInfo, &pBIOSInfo->TMDSSettingInfo2)) {
			if (pBIOSInfo->TMDSSettingInfo2.ChipID == VIA_INTEGRATED_HDMI)
                viaSerialReadBytesByHDCP(0xA0, 0, edidBuffer, 128);    
            else
                viaSerialReadBytes(pBIOSInfo->TMDSSettingInfo2.I2CPort, 0xA0, 0, edidBuffer, 128);
        }
		
		/* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex,edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = edidSource;
		
    } else if (edidSource & VIA_DEVICE_HDMI) {
        /* Read EDID v1.1 Block */
        if(viaIdentifyHDMI(pBIOSInfo)) {
            if (pBIOSInfo->HDMISettingInfo.ChipID == VIA_INTEGRATED_HDMI)
                viaSerialReadBytesByHDCP(0xA0, 0, edidBuffer, 128); 
            else
                viaGetAD9389EDIDBlockData(pBIOSInfo, edidBuffer);
        }
        
        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex,edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = VIA_DEVICE_HDMI;        
    } else if (edidSource & VIA_DEVICE_HDMI2) {
        /* Read EDID v1.1 Block */
        if(viaIdentifyHDMI(pBIOSInfo)) {
            if (pBIOSInfo->HDMISettingInfo2.ChipID == VIA_INTEGRATED_HDMI)
                viaSerialReadBytesByHDCP(0xA0, 0, edidBuffer, 128); 
        }
        
        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex,edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = VIA_DEVICE_HDMI2;
    } else if (edidSource & VIA_DEVICE_DP) {
        if (VIAIdentifyDP(pBIOSInfo)) {
            /* Hardware issue, we can't read EDID on VT3410 A0 A1, may remove it future */
            if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1)
                VIAGetDPEDIDBlockData(pBIOSInfo, pBIOSInfo->DPSettingInfo.ChipID, edidBuffer);
        }
        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex, edidBuffer);
		
        if (pMonitor) 
            pBIOSInfo->EDIDType = VIA_DEVICE_DP;
    } else if (edidSource & VIA_DEVICE_DP2) {
        if (VIAIdentifyDP(pBIOSInfo)) {
            /* Hardware issue, we can't read EDID on VT3410 A0 A1, may remove it future */
            if (VIAGetRevisionOfVX900(pVia) > REVISION_VX900_A1)
                VIAGetDPEDIDBlockData(pBIOSInfo, pBIOSInfo->DPSettingInfo2.ChipID, edidBuffer);
        }
        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex, edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = VIA_DEVICE_DP2;
    } else if (edidSource & VIA_DEVICE_CRT2) {
        /* Read EDID v1.1 Block */
        if (vt1625_module_loaded) {
        
            if (TVEncoderIdentify_VT1625(pBIOSInfo, &(pBIOSInfo->TVSettingInfo))) {
                viaSerialReadBytes(pBIOSInfo->TVI2CPort, 0xA0, 0, edidBuffer, 128);
            }
        }    

        /* Parse the EDID block */
        pMonitor = xf86InterpretEDID(pScrn->scrnIndex, edidBuffer);

        if (pMonitor) 
            pBIOSInfo->EDIDType = VIA_DEVICE_CRT2;
    }
   
   xfree(edidBuffer);
   if (pMonitor){
       DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
             "EDID data is from device %#x.\n", pBIOSInfo->EDIDType));
   } else{
       DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
         "VIADoEDID fail.\n"));
   }    
    
    return pMonitor;
}    


static Bool
VIAIsMAMM(ScrnInfoPtr pScrn)
{
    char *	viaDriverName; 
    Bool retValue;

    viaDriverName = pScrn->confScreen->device->driver;
    xf86DrvMsg( pScrn->scrnIndex, X_INFO, "the driver module for this screen 0 is \"%s\" driver.\n", viaDriverName);
    if (!memcmp(viaDriverName, "VIA", 3))
        retValue= TRUE;
    else 
        retValue= FALSE; 
    return retValue;        
}


/* Get the CRTC timing from xorg.conf */
void 
VIAGetModeLineTiming(ScrnInfoPtr pScrn, DisplayModePtr mode)
{
    VIAPtr          pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;

    /* 
    case 1: user specified timing in xorg.conf, we'll use it,
        and the option "Refresh" doesn't take effect
    */
    if(mode->type == M_T_USERDEF) {
        pBIOSInfo->ModeLineStatus = MODE_LINE_TYPE_USER;
        /* 
        The secondary device will use hard code timing
        */
        if(!pVia->pVIAEnt->HasSecondary) {
            if (pBIOSInfo->PrimaryDevice) /* Duoview*/
                pBIOSInfo->EDIDType = pBIOSInfo->PrimaryDevice;
            else /* Single */
                pBIOSInfo->EDIDType = pBIOSInfo->ActiveDevice;
        }
    /* 
    case 2: monitor preferred timing, we'll use it, 
        and the option "Refresh" doesn't take effect
    */    
    } else if(mode->type & M_T_PREFERRED)
        pBIOSInfo->ModeLineStatus = MODE_LINE_TYPE_EDID;
    /* 
    other case: we'll use hard code timing,
        and the option "Refresh" take effect
    */
    else
        pBIOSInfo->ModeLineStatus = 0;

    if (pBIOSInfo->ModeLineStatus & (MODE_LINE_TYPE_USER | MODE_LINE_TYPE_EDID)) {
        pBIOSInfo->UserSpecifiedMode.clk = mode->Clock * 1000;
        if (mode->Flags & V_PHSYNC) 
            pBIOSInfo->UserSpecifiedMode.h_sync_polarity = POSITIVE;
        if (mode->Flags & V_NHSYNC) 
            pBIOSInfo->UserSpecifiedMode.h_sync_polarity = NEGATIVE;
        if (mode->Flags & V_PVSYNC)
            pBIOSInfo->UserSpecifiedMode.v_sync_polarity = POSITIVE;
        if (mode->Flags & V_NVSYNC) 
            pBIOSInfo->UserSpecifiedMode.v_sync_polarity = NEGATIVE;
        
        pBIOSInfo->UserSpecifiedMode.crtc.hor_total = mode->CrtcHTotal;
        pBIOSInfo->UserSpecifiedMode.crtc.hor_addr = mode->CrtcHDisplay;
        pBIOSInfo->UserSpecifiedMode.crtc.hor_blank_start = mode->CrtcHDisplay; /* mode->CrtcHBlankStart is incorrect! */
        pBIOSInfo->UserSpecifiedMode.crtc.hor_blank_end = mode->CrtcHBlankEnd - mode->CrtcHDisplay;
        pBIOSInfo->UserSpecifiedMode.crtc.hor_sync_start = mode->CrtcHSyncStart;
        pBIOSInfo->UserSpecifiedMode.crtc.hor_sync_end = mode->CrtcHSyncEnd - mode->CrtcHSyncStart;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_total = mode->CrtcVTotal;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_addr = mode->CrtcVDisplay;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_blank_start = mode->CrtcVBlankStart;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_blank_end = mode->CrtcVBlankEnd - mode->CrtcVBlankStart;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_sync_start = mode->CrtcVSyncStart;
        pBIOSInfo->UserSpecifiedMode.crtc.ver_sync_end = mode->CrtcVSyncEnd - mode->CrtcVSyncStart;    
    }

}
    

/* for CLE AX/CX differential behavior.                         *
 * Define UniChrome Family Project ID/Revision, for simplify    *
 * Device ID + Revision combination.                            *
 * TODO: Use uni-variable capability flag to replace this.      */
static void VIAGetdwProjectIDRevision(ScrnInfoPtr pScrn)
{
    CARD8  BondingOption;
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    
    switch (pVia->Chipset) {
        case VIA_P4M800PRO:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3314;
            break;
        case VIA_CX700:
            /*add for get 3324A3 revision*/
            VGAOUT8(VIASR, SR43);
            BondingOption = VGAIN8(VIASR + 1);
            
            if(BondingOption & 0x02)
                pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3324M2;
            else if(BondingOption & 0x40)
                pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3324M;
            else
                pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3324;
            break;
        case VIA_K8M890:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3336;
            break;
        case VIA_P4M890:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3327;            
            break;
        case VIA_P4M900:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3364;
            break;
        case VIA_VX800:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3353;
            break;
        case VIA_VX855:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3409;
            break;
        case VIA_VX900:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3410;
            break;
#ifdef VIA_VT3293_SUPPORT
        case VIA_CN750:
            pBIOSInfo->dwProjectIDRevision = UNICHROME_VT3293;
            break;
#endif            
        default:
            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "VIABIOSInit: Unknown Device ID\n"));
            break;
    }
}

static Bool 
VIAPreInit(ScrnInfoPtr pScrn, int flags)
{
    EntityInfoPtr   pEnt;
    VIAPtr          pVia;
    VIABIOSInfoPtr  pBIOSInfo;
    int             i;
    int             pitch;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAPreInit\n"));
    DEBUG(ErrorF("XORG_VERSION_CURRENT = %d\n",XORG_VERSION_CURRENT));
#if XORG_VERSION_CURRENT < XF86_VERSION_NUMERIC(1,5,99,0,0)||XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(7,0,0,0,0)
    via_xf86_version_current = xf86GetVersion();    
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "XF86_VER=%d\n,Driver Not depending on this anymore\n", via_xf86_version_current));    
#endif

    if (pScrn->numEntities > 1)
        return FALSE;

    if (flags & PROBE_DETECT)
        return FALSE;
   
    if (!xf86LoadSubModule(pScrn, "vgahw"))
        return FALSE;

    if (!vgaHWGetHWRec(pScrn))
        return FALSE;

    if (!VIAGetRec(pScrn)) 
        return FALSE;

    pVia = VIAPTR(pScrn);

#define VIAFB_QUEUE_CMD_AVAILABLE 0x56494139
    pVia->cmd_queue_fd = open("/dev/fb0", O_RDWR);
    if (pVia->cmd_queue_fd >= 0) {
        int r = ioctl(pVia->cmd_queue_fd, VIAFB_QUEUE_CMD_AVAILABLE, NULL);
        if (r < 0) {
            close(pVia->cmd_queue_fd);
            pVia->cmd_queue_fd = -1;
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Kernel command queueing support not available\n");
        } else {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Detected kernel command queueing support\n");
        }
    }

    DevUnion* pPriv;
    pPriv = xf86GetEntityPrivate(pScrn->entityList[0],
            gVIAEntityIndex);
    pVia->pVIAEnt = pPriv->ptr; 
    
    pVia->IsSuspending = FALSE;
    
    pBIOSInfo = pVia->pBIOSInfo;
    pBIOSInfo->s3utility=FALSE;
    pBIOSInfo->FirstTimeToSenseIntegratedTV = TRUE;
    pBIOSInfo->scrnIndex = pScrn->scrnIndex;


    if(VIAIsMAMM(pScrn)) {              
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "MAMM mode enable\n");
        pVia->IsMAMMEnable = TRUE;   
    } else {               
        pVia->IsMAMMEnable = FALSE;   
    }     

   
    pVia->IsSecondary = FALSE;
    pEnt = xf86GetEntityInfo(pScrn->entityList[0]);
#ifndef XSERVER_LIBPCIACCESS
    if (pEnt->resources) {
        xfree(pEnt);
        VIAFreeRec(pScrn);
        return FALSE;
    }
#endif

    pVia->PciInfo = xf86GetPciInfoForEntity(pEnt->index);

#ifndef XSERVER_LIBPCIACCESS
    pVia->PciTag = pciTag(pVia->PciInfo->bus, pVia->PciInfo->device,
              pVia->PciInfo->func);

    if(xf86RegisterResources(pEnt->index, NULL, ResNone)) {
        xfree(pEnt);
        VIAFreeRec(pScrn);
        return FALSE;
    }
#endif	

    if(xf86IsEntityShared(pScrn->entityList[0])) {
        if(xf86IsPrimInitDone(pScrn->entityList[0])) {

            pVia->pVIAEnt->pSecondaryScrn = pScrn;
            pVia->IsSecondary = TRUE;

            /* If SAMM, Hot key zoom locked
             * pScrn->zoomLocked = TRUE;*/
            pVia->pVIAEnt->pSecondaryScrn->zoomLocked = TRUE;
            pVia->pVIAEnt->pPrimaryScrn->zoomLocked = TRUE;
        } else {
            xf86SetPrimInitDone(pScrn->entityList[0]);
            pVia->pVIAEnt->pPrimaryScrn = pScrn;
            if(pVia->pVIAEnt->HasSecondary)
                pVia->IsSecondary = FALSE;
        }
    }

    /* Detect which chipset we used. */
    VIAProbeChipset(pScrn, pEnt);

    pScrn->monitor = pScrn->confScreen->monitor;

    
    /*
     * We support depths of 8, 16 and 24.
     * We support bpp of 8, 16, and 32.
     */

    if (!xf86SetDepthBpp(pScrn, 0, 0, 0, Support32bppFb)) {
        xfree(pEnt);
        VIAFreeRec(pScrn);
        return FALSE;
    } else {
        switch (pScrn->depth)
        {
            case 8:
            case 16:
            case 24:
            case 32:
                /* OK */
                break;
            default:
                xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                       "Given depth (%d) is not supported by this driver\n",
                       pScrn->depth);
                xfree(pEnt);
                VIAFreeRec(pScrn);
                return FALSE;
        }
    }

    xf86PrintDepthBpp(pScrn);

    if (pScrn->depth == 32) 
        pScrn->depth = 24;

    if (pScrn->depth > 8) {
        rgb zeros = {0, 0, 0};

        if (!xf86SetWeight(pScrn, zeros, zeros)) {
            xfree(pEnt);
            VIAFreeRec(pScrn);
            return FALSE;
        } else {
            /* TODO check weight returned is supported */
            ;
        }
    }

    if (!xf86SetDefaultVisual(pScrn, -1)) {
        return FALSE;
    } else {
        /* We don't currently support DirectColor at > 8bpp */
        if (pScrn->depth > 8 && pScrn->defaultVisual != TrueColor) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Given default visual"
                       " (%s) is not supported at depth %d\n",
                       xf86GetVisualName(pScrn->defaultVisual), pScrn->depth);
            xfree(pEnt);
            VIAFreeRec(pScrn);
            return FALSE;
        }
    }

    /* Set the bits per RGB for 8bpp mode */
    if (pScrn->depth == 8)
        pScrn->rgbBits = 8;

    /* We use a programmable clock */
    pScrn->progClock = TRUE;

    /* If MAMM mode, we can't use int10,vbe */
    if (!pVia->IsMAMMEnable) {
        if (xf86LoadSubModule(pScrn, "int10")) {
            pVia->pInt10 = xf86InitInt10(pEnt->index);
        }

        if (pVia->pInt10 && xf86LoadSubModule(pScrn, "vbe")) {
            pVia->pVbe = VBEExtendedInit(pVia->pInt10, pEnt->index,0);
        }    
    }

    /* Set status word positions based on chip type. */
    switch (pVia->Chipset) {
        case VIA_K8M890:
        case VIA_P4M900: 
#ifdef VIA_VT3293_SUPPORT        
        case VIA_CN750:
#endif        
            pVia->myWaitIdle = WaitIdle_H5Chips;
            break;
        case VIA_VX800:
        case VIA_VX855:
        case VIA_VX900:
            pVia->myWaitIdle = WaitIdle_H6Chips;
            break;
        default:
            pVia->myWaitIdle = WaitIdle_H2Chips;
            break;
    }

    /* *_* We process all config file's options in one function *_* */
    VIAProcessCfgOptions(pScrn);
    
    if (pEnt->device->videoRam != 0) {
        if (!pScrn->videoRam)
            pScrn->videoRam = pEnt->device->videoRam;
        else
            xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                       "Video Memory Size in Option is %d KB, Detect is %d KB!",
                       pScrn->videoRam, pEnt->device->videoRam);
    }

    xfree(pEnt);    

    /* FrameBufferBase always represets physical start address for the 
       whole video memory.*/
    if (pVia->pBIOSInfo->Chipset == VIA_VX900)
    {
        pVia->FrameBufferBase = MEMBASE(pVia->PciInfo, 2);
    }
    else 
    {
        pVia->FrameBufferBase = MEMBASE(pVia->PciInfo, 0);
    }  

    /* ScreenBase represents physical start address for each screen*/
    if (pVia->pBIOSInfo->Chipset == VIA_VX900)
    {
        pVia->ScreenBase = MEMBASE(pVia->PciInfo, 2);
    }
    else 
    {
        pVia->ScreenBase = MEMBASE(pVia->PciInfo, 0); 
    }  
    
    if (!VIAMapMMIO(pScrn)) {
        if (pVia->pVbe!=NULL)
            vbeFree(pVia->pVbe);
            return FALSE;
    }

    /*Get dwProjectIDRevision for capture and TV
      This function must call after VIAMapMMIO*/
    VIAGetdwProjectIDRevision(pScrn);

    /* load vt1625 sub module */
    /* using LoadSubModule instead of xf86LoadSubModule 
       for if module does not exsit don't print error*/
    if (!LoadSubModule(pScrn->module, "vt1625", NULL, NULL, NULL, NULL,
			NULL, NULL)) {
        vt1625_module_loaded = FALSE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   " Couldn't load vt1625 submodule\n"));
    } else
        vt1625_module_loaded = TRUE;
    
    /* load integratedtv sub module */
    if (!LoadSubModule(pScrn->module, "ViaIntegratedTV", NULL, NULL, NULL, NULL,
			NULL, NULL)) {
        integratedtv_module_loaded = FALSE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   " Couldn't load ViaIntegratedTV submodule\n"));
    } else
        integratedtv_module_loaded = TRUE;

    /* load ad9389 sub module */
    if (!LoadSubModule(pScrn->module, "ad9389", NULL, NULL, NULL, NULL,
			NULL, NULL)) {
        ad9389_module_loaded = FALSE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
            "Failed to load AD9389 module\n"));
    }
    else
        ad9389_module_loaded = TRUE;

    if (!LoadSubModule(pScrn->module, "sil164", NULL, NULL, NULL, NULL,
			NULL, NULL)) {
        sil164_module_loaded = FALSE;
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
            "Failed to load sil164 module\n"));
    } else {
        sil164_module_loaded = TRUE;
    }
        
    /* Collect some information about Chip or CMOS setting */
    if (!VIACollectBIOSInfo(pScrn)) {
        VIAFreeRec(pScrn);
        return FALSE;
    }

    /* Initialize the colormap and this step must be done before calling xf86HandleColormaps. */
    Gamma zeros = {0.0, 0.0, 0.0};
    if (!xf86SetGamma(pScrn, zeros)) {
        VIAFreeRec(pScrn);
        vbeFree(pVia->pVbe);
        return FALSE;
    }

    /* Split FB for SAMM */
    /* FIXME: For now, split FB into two equal sections. This should
     *        be able to be adjusted by user with a config option. */
    if (pVia->IsSecondary) {
        VIAPtr    pVia0;
        int videoram1=0;

        if(xf86GetOptValInteger(VIAOptions, OPTION_VIDEORAM1, &videoram1))
            xf86DrvMsg( pScrn->scrnIndex, X_CONFIG,
                        "Option: VideoRAM1 %dkB\n", videoram1 );
        
        if(videoram1 > 0)
            videoram1 = (videoram1 >> 10) << 10;
  
        if(videoram1 >= pScrn->videoRam - 16384)
            videoram1 = pScrn->videoRam - 16384;

        if((videoram1 <= 16384) && (videoram1 > 0))
            videoram1 = 16384;

        if(videoram1)
            pScrn->videoRam = videoram1;
        else
            pScrn->videoRam = 11 * 1024; /* Reserve 11M for second scrn, because 1920x1440x4/2^20 = 10.5M  */
        
        pVia->pVIAEnt->pPrimaryScrn->videoRam -= pScrn->videoRam;
        pVia0 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);
        /*Rewrite pVia0->videoRambytes here*/
        pVia0->videoRambytes = pVia->pVIAEnt->pPrimaryScrn->videoRam << 10;
        pVia->ScreenBase += (pVia->pVIAEnt->pPrimaryScrn->videoRam << 10);
                   
    }
    
    if (!pVia->IsSecondary) {
        pVia->pVIAEnt->pScreensPublicInfo->VRamSize = pScrn->videoRam << 10;
    }

    /*Init videoRambytes to All Video Memory Size, pVia->videoRambytes may rewrite in SAMM mode*/
    pVia->videoRambytes = pScrn->videoRam << 10;

    xf86DrvMsg(pScrn->scrnIndex, X_PROBED, "videoram =  %dk\n",
               pScrn->videoRam);

    /* here load any modules that we need */    
    /* Before using DDC, we must initialize the I2C. */
    if (!xf86LoadSubModule(pScrn, "i2c")) {
        VIAFreeRec(pScrn);
        return FALSE;
    } else {        
        VIAI2CInit(pScrn);
    }
    
    if (xf86LoadSubModule(pScrn, "fb") == NULL) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Failed to load fb module");
        VIAFreeRec(pScrn);
        return FALSE;
    }

    if (!pVia->NoAccel) {
#if (EXA_VERSION_MAJOR >= 2)
        XF86ModReqInfo req;
        int errmaj, errmin;

        memset(&req, 0, sizeof(req));
        req.majorversion = 2;
        req.minorversion = 0;
        if (!LoadSubModule(pScrn->module, "exa", NULL, NULL, NULL, &req,
                           &errmaj, &errmin)) {
            LoaderErrorMsg(NULL, "exa", errmaj, errmin);
            VIAFreeRec(pScrn);
            return FALSE;
        }
#else
        if (!xf86LoadSubModule(pScrn, "exa")) {
            VIAFreeRec(pScrn);
            return FALSE;
        }
#endif /* EXA_VERSION */
    }

    if (!pVia->ForceSWCursor) {
        if (!xf86LoadSubModule(pScrn, "ramdac")) {
            VIAFreeRec(pScrn);
            return FALSE;
        }
    }

    if (pVia->shadowFB) {
        if (!xf86LoadSubModule(pScrn, "shadowfb")) {
            VIAFreeRec(pScrn);
            return FALSE;
        }
    }

	viaInitHwInfo(pVia);
  
	/* Check Spread Spectrum. Or it will cause garbage when we sense the external encoder.
	   Actual project: On Qunta's IL1 case, this will cause the TTL LCD appears garbage.
	*/
	if (pVia->pChipInfo->biosIsSpreadSpectrum) {		 
		/* Disable spread spectrum temporarily, we will turn it on when set mode. */
		if ((pVia->Chipset == VIA_CX700) ||
			(pVia->Chipset == VIA_VX800) ||
			(pVia->Chipset == VIA_VX855)) {
			VGAOUT8(0x3c4, SR3D);
			VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xFE);
		} else {
			VGAOUT8(0x3c4, SR2C);
			VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xFE);
		}

		VGAOUT8(0x3c4, SR1E);
		VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xF7);
	}
	
    /* Load DDC module for EDID. */
    xf86LoadSubModule(pScrn, "ddc");           
    
    xf86CrtcConfigInit (pScrn, &via_xf86crtc_config_funcs);
    // xf86CrtcSetSizeRange (pScrn, 320, 200, 8192, 8192);     
    xf86CrtcSetSizeRange (pScrn, 320, 200, 4096, 4096);     

    via_crtc_init(pScrn, 2);    
    via_output_init(pScrn);        

    if (!xf86InitialConfiguration(pScrn, FALSE)) {        
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
            "No valid initial configuration found\n");   
        return FALSE;
    }     

    pitch = CMDISP_ALIGN_TO(pScrn->virtualX * (pScrn->bitsPerPixel >> 3), 32);
    pScrn->displayWidth = pitch / (pScrn->bitsPerPixel >> 3);

    xf86CrtcConfigPtr   via_crtc_config = XF86_CRTC_CONFIG_PTR (pScrn);
    
    xf86SetDpi(pScrn, 0, 0);   

    /* If TV output create,get the TV Caps*/
    for (i = 0; i < via_crtc_config->num_output; i++) {
        if ((!xf86NameCmp(via_crtc_config->output[i]->name, "TV"))||
                           (!xf86NameCmp(via_crtc_config->output[i]->name, "TV-2"))) {
            int HDisplay = 1024;
            int VDisplay = 768;
            if (via_crtc_config->output[i]->crtc) {
                HDisplay = via_crtc_config->output[i]->crtc->desiredMode.HDisplay;
                VDisplay = via_crtc_config->output[i]->crtc->desiredMode.VDisplay;
            }
            viaGetTvCapsSupported(via_crtc_config->output[i], HDisplay, VDisplay);		
        }
    }

    /* Set up screen parameters. */
    pVia->Bpp = pScrn->bitsPerPixel >> 3;
    pVia->Bpl = pScrn->displayWidth * pVia->Bpp;
 
    /* Check what type of cursor (SW/HW) to use. */
    VIACheckCursorTypeToUse(pScrn);            
    
    VIAUnmapMem(pScrn);

    return TRUE;
}


static Bool 
VIAEnterVT(VT_FUNC_ARGS_DECL)
{
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIAEnterVT\n"));

    SCRN_INFO_PTR(arg);
    VIAPtr      pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    vgaRegPtr   vgaSavePtr = &hwp->SavedReg;
    VIARegPtr   viaSavePtr = &pVia->SavedReg;
    Bool        ret;
    int i;
    miPointerScreenPtr  PointPriv;
    int x, y;

    /* 410 chipset some register setting are lost when suspend from
     * S3, the extended memory access enable/disable bit is the one
     * so enable the bit for we need to access the extended MMIO space
     */
    if (pVia->Chipset == VIA_VX900 && !pVia->IsSecondary) {       
        VIAEnableMMIO(pScrn);
    }

    if(pBIOSInfo->Chipset == VIA_CX700) {
        /* In SAMM mode, X and VT switch, with random large times, 
        * the Screen will covered by Vertical Line Garbages.
        * Called VIASave here to reduced it's probability.
        */
        pVia->myWaitIdle(pVia);
        VIASave(pScrn); 

        /*Fix on 3324 graphic can't resume
        *Theorically, we neend't to touch 0x3c3(switch of graphic system)
        */
        if(1 != VGAIN8(0x3c3))
            VGAOUT8(0x3c3,1);
    }

    if(pVia->VQEnable && pVia->VQStart)
        VIAEnableVQ(pScrn);

    /* FIXME: Rebind AGP memory here */
    /* FIXME: Unlock DRI here */

    vgaHWUnlock(hwp);

#if XORG_VERSION_CURRENT < XF86_VERSION_NUMERIC(1,5,99,0,0)||XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(7,0,0,0,0)
    /* This is really a patch. patch for ubuntu8.04 os has garbage when resume from S3
    it seems the ubuntu os's own behavior. When os behavior right, the patch can
    be removed.And this patch take no effect on other OS like FCX whose xorg version is > 7.0.
    but this patch has side effect to xorg<7.0,like suse10.1 suse10.2, when switch console,
    the cursor disppear. so we take the condision into consideration*/
    if (pVia->pVIAEnt->pScreensPublicInfo->UseHwCursor) {
#if XORG_VERSION_CURRENT <= XF86_VERSION_NUMERIC(2,0,0,0,0) && XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(1,4,99,0,0)
        PointPriv = (miPointerScreenPtr)dixLookupPrivate(&pScrn->pScreen->devPrivates, miPointerScreenKey);
#else
        extern int miPointerScreenIndex;
        PointPriv = pScrn->pScreen->devPrivates[miPointerScreenIndex].ptr;
#endif

#if XORG_VERSION_CURRENT <= XF86_VERSION_NUMERIC(2,0,0,0,0) && XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(1,4,0,0,0)
        miPointerGetPosition(inputInfo.pointer,&x, &y);
#else
     	miPointerPosition(&x, &y);
#endif
        PointPriv->spriteFuncs->SetCursor(pScrn->pScreen, 0, x, y);
    } 
#endif

    /* PCIE series new Registers */
    if (pVia->Chipset == VIA_VX800 || pVia->Chipset == VIA_VX855 ||
        pVia->Chipset == VIA_VX900) {
        for (i = 0; i < 10; i++) {
             VGAOUT8(0x3c4, 0x66 + i);
             VGAOUT8(0x3c5, viaSavePtr->SRRegs[0x66 + i]);
        }

        for (i = 0; i < 3; i++) {
            VGAOUT8(0x3c4, 0x79 + i);
            VGAOUT8(0x3c5, viaSavePtr->SRRegs[0x79 + i]);
        }
 
        for (i = 0; i < 6; i++) {
            VGAOUT8(0x3c4, 0x70 + i);
            VGAOUT8(0x3c5, viaSavePtr->SRRegs[0x70 + i]);
        }
    }

    if (!pVia->IsSecondary) {
        /* 1. Restore first or some of BIOS settings will lose. */
        /* 2. Flags=TRUE implies decrease flick. When the system resume from
              suspend mode. LCD is switch too many times, so it cause eyes
              feel flick. But We don't need to enable LCD so many times.
              LCD panel will be enabled at the following setmode function.
        */
        VIARestore(pScrn, vgaSavePtr, viaSavePtr, TRUE);

        if((pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI) &&
           (pBIOSInfo->HDMISettingInfo.ChipID == VIA_AD9389)) {
            viaIdentifyHDMI(pBIOSInfo);
        }

        /*Clear FB with black color to avoid garbage caused by VIARestore*/
        memset(pVia->FBBase, 0x00, ((pScrn->displayWidth * pScrn->bitsPerPixel>>3) * pScrn->virtualY));

        /*Clear 3Dscaling buffer to avoid garbage*/
        if (pVia->DISP3DScalingBufferStart)
           memset(pVia->FBBase + pVia->DISP3DScalingBufferStart, 0x00,
           DISP_3D_SCALING_BUFFER_SIZE*DISP_3D_SCALING_BUFNUM_ONESET*pVia->NumOfDevNeed3dScl);
        } else {
            /*Clear FB with black color to avoid garbage caused by VIARestore*/
            memset(pVia->FBBase, 0x00, ((pScrn->displayWidth * pScrn->bitsPerPixel>>3) * pScrn->virtualY));
        }

    /* Initialize display engine*/
    viaInitDispEngine(pVia);
	viaInitOutputRegSet(pScrn);
    ret = xf86SetDesiredModes(pScrn);

	/* Enable the graphics engine. */
	if (!pVia->NoAccel) {
	    VIAInitialize2DEngine(pScrn);

        VIASETREG(0x43C, (0x00020000 | 0xfe000000));
        int j;
        for ( j = 0 ; j <= 0x13 ; j++ )
            VIASETREG(0x440, (CARD32) j << 24);
    }

#ifdef XF86DRI
    if (pVia->directRenderingEnabled 
#ifdef DRM_SAMM_VIA
        || pVia->drmSammEnabled 
#endif
        ) {
        VIADRIRingBufferInit(pScrn);
    }
#endif

    /* We need to initialize related registers about hardware icon again */
    /* when system resumes, or the cursor will become garbage. */
    VIAARGBCursorInit(pScrn);
    
    /* restore video status */
    if (!pVia->IsSecondary) {
        viaRestoreHqv(pScrn);
        viaRestoreVideo(pScrn);
    }
    
    VIAAdjustFrame(ADJUST_FRAME_ARGS(pScrn, pScrn->frameX0, pScrn->frameY0));

    /* Insert a timer */
    if  ((pBIOSInfo->Is3DScalingEnable & DISP_3D_SCALING_ENABLE) ||
        ((pVia->IsHotkeyTimerEnabled) && (!pVia->pVIAEnt->HasSecondary))) {
        if(!pVia->devicesTimer) {
            pVia->devicesTimer = TimerSet(NULL, 0, 30, VIASyncTimer, pScrn);
        }
    } else {
        if(pVia->devicesTimer) {
            TimerCancel(pVia->devicesTimer);
            pVia->devicesTimer=NULL;
        }
    }

    
    pVia->IsSuspending = FALSE;    /* System has resumed. */

    /* Release the DRM lock, or the DRI APs will be pending forever.*/
#ifdef XF86DRI
    if (pVia->directRenderingEnabled) {
        DRIUnlock(pScrn->pScreen);
    }    
#endif

    return ret;
}


static void 
VIALeaveVT(VT_FUNC_ARGS_DECL)
{
    SCRN_INFO_PTR(arg);
    vgaHWPtr    hwp = VGAHWPTR(pScrn);
    VIAPtr      pVia = VIAPTR(pScrn);
    vgaRegPtr   vgaSavePtr = &hwp->SavedReg;
    VIARegPtr   viaSavePtr = &pVia->SavedReg;
    int i, rotate_degree;
    
	DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIALeaveVT\n"));

    /* FIXME: take the DRI lock here to avoid accidents */
    /* FIXME: unbind the AGP memory ? */
    /* Get DRM Lock to avoid DRI APs run unexpected under non-X situation */
#ifdef XF86DRI
    CHROMESAREAPriv *saPriv;
    if (pVia->directRenderingEnabled) {
        DRILock(pScrn->pScreen, 0);
		if (H5_UMA_CHIPID || H6_UMA_CHIPID) {
			saPriv = DRIGetSAREAPrivate(pScrn->pScreen);
			saPriv->tex_age[S3G_TEX_FB]++;
		}
    }
#endif

    /* Unload Timer */
    if (pVia->devicesTimer)
        TimerCancel(pVia->devicesTimer);

    pVia->devicesTimer = NULL; 

    /* Wait Hardware Engine idle to exit graphicd mode */
    WaitIdle();

    if (pVia->VQEnable) {
     /* if we use VQ, disable it before we exit */
        switch( pVia->Chipset ) { 
            case VIA_P4M890: 
            case VIA_K8M890:
                VIASETREG(0x41c, 0x00100000);
                VIASETREG(0x420, 0x74301000);
                break;
            default:	 
                VIASETREG(0x43c, 0x00fe0000);
                VIASETREG(0x440, 0x00000004);
                VIASETREG(0x440, 0x40008c0f);
                VIASETREG(0x440, 0x44000000);
                VIASETREG(0x440, 0x45080c04);
                VIASETREG(0x440, 0x46800408);
                break;
        }  
    }

    /* Save video status and turn off all video activities */
    if (!pVia->IsSecondary) {
        viaSaveHqv(pScrn);
        viaSaveVideo(pScrn);
    }

    if (pVia->IsHWRotateEnabled) {
        DisableCBUrotate(pScrn);

        /* Disable CBU HW Rotate.
         * VX855 and VX900 rotation 90 and 270 degree has issue
         * but 180 degree doesn't have.
         * So here mimic the behavior of CBU 180
         * The root cause is still unknown.
         */
        if ((pVia->Chipset == VIA_VX855 || pVia->Chipset == VIA_VX900) &&
            (pVia->RotateDegree == VIA_ROTATE_DEGREE_90 ||
            pVia->RotateDegree == VIA_ROTATE_DEGREE_270)) {
            /* Mimic the behavior of rotate 180 */
            rotate_degree = pVia->RotateDegree;
            pVia->RotateDegree = VIA_ROTATE_DEGREE_180;
            EnableCBUrotate(pScrn);
            pVia->RotateDegree = rotate_degree;
            /* Let CBU 180 rotate take effect */
            memset(pVia->FBBase, 0x00,
            (pScrn->displayWidth * pScrn->bitsPerPixel >> 3));
            
            DisableCBUrotate(pScrn);
        }
     }

    if(pVia->RotateDegree == VIA_ROTATE_DEGREE_90
        ||pVia->RotateDegree == VIA_ROTATE_DEGREE_270){
        memset(pVia->FBBase, 0x00, ((pScrn->displayWidth * pScrn->bitsPerPixel>>3) * pScrn->virtualX));
    } else {
        memset(pVia->FBBase, 0x00, ((pScrn->displayWidth * pScrn->bitsPerPixel>>3) * pScrn->virtualY));
    }
    
    /* FIX ME: For vt3353, some registers can't be restore in DRM module, it seems that
     * these registers be modified between DRM resume function and VIAEnterVT function
     * when using "suspend" button shipped with ubuntu 8.04.
     * So we save these registers here and restore them in VIAEnterVT.
     * These registers including: SR66~SR6F, SR70~SR75, SR79~SR7B.
     */
    if (pVia->Chipset == VIA_VX800 || pVia->Chipset == VIA_VX855 ||
        pVia->Chipset == VIA_VX900) {
        for (i = 0; i < 10; i++) {
            VGAOUT8(0x3c4, 0x66 + i);
            viaSavePtr->SRRegs[0x66 + i] = VGAIN8(0x3c5);
        }

        for (i = 0; i < 6; i++) {
            VGAOUT8(0x3c4, 0x70 + i);
            viaSavePtr->SRRegs[0x70 + i] = VGAIN8(0x3c5);
        }

        for (i = 0; i < 3; i++) {
            VGAOUT8(0x3c4, 0x79 + i);
            viaSavePtr->SRRegs[0x79 + i] = VGAIN8(0x3c5);
        }
    }

#ifdef XF86DRI
    if (pVia->directRenderingEnabled 
#ifdef DRM_SAMM_VIA
        || pVia->drmSammEnabled
#endif
        )
        VIADRIRingBufferCleanup(pScrn);
#endif

    VIARestore(pScrn, vgaSavePtr, viaSavePtr, FALSE);
        
    pVia->IsSuspending = TRUE;     /* System will suspend. */

    /*To fix hang issue when switch VT using 3D desktop*/
    VGAOUT8(0x3c4, 0x32);
    VGAOUT8(0x3c5, 0x00);

    vgaHWLock(hwp);
}


static void 
VIASave(ScrnInfoPtr pScrn)
{
    vgaHWPtr        hwp = VGAHWPTR(pScrn);
    vgaRegPtr       vgaSavePtr = &hwp->SavedReg;
    VIAPtr          pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    VIARegPtr       save = &pVia->SavedReg;
    int             vgaCRIndex, vgaCRReg, vgaIOBase;
    int             i;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIASave\n"));

    vgaHWProtect(pScrn, TRUE);
    
    if (!pVia->IsMAMMEnable) {
        vgaHWSave(pScrn, vgaSavePtr, VGA_SR_ALL);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "Primary Adapter! saving VGA_SR_ALL !!\n"));            
    } else {
        vgaHWSave(pScrn, vgaSavePtr, VGA_SR_MODE);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
              "Non-Primary Adapter! saving VGA_SR_MODE only !!\n"));
    }

    /* Unlock Extended Regs */
    viaWriteVgaIoBits(REG_SR01, 0x01, BIT0);        
    /* Unlock CRTC register protect */
    viaWriteVgaIoBits(REG_CR47, 0x00, BIT0);

    /* Save Sequencer regs: from SR00 to SRFF */
    i = 0;
    while (1) {
        /* We have saved SR1A in VIAPreInit(). */
        if (i != 0x1A)
            save->SRRegs[i] = viaReadVgaIo(i + REG_SR00);

        if (i == 0xFF)
            break;
        else
            i++;
    }

    /* Save CRTC regs: from CR00 to CRFF */
    i = 0;
    while (1) {
        /* We have saved CR08 and CR09 in VIAPreInit(). */
        if ((i != 0x08) && (i != 0x09))
            save->CRTCRegs[i] = viaReadVgaIo(i + REG_CR00);

        if (i == 0xFF)
            break;
        else
            i++;
    }
           
    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
    for (i = 0; i < xf86_config->num_output; i++) {
        xf86OutputPtr output = xf86_config->output[i];
        if (output->funcs->save)
            (*output->funcs->save)(output);
    }

    /* Saving HW Cursor state */
    save->dwCursorMode = VIAGETREG(VIA_REG_CURSOR_MODE);
    save->dwCursorPOS = VIAGETREG(VIA_REG_CURSOR_POS);
    save->dwCursorORG = VIAGETREG(VIA_REG_CURSOR_ORG);
    save->dwCursorBG = VIAGETREG(VIA_REG_CURSOR_BG);
    save->dwCursorFG = VIAGETREG(VIA_REG_CURSOR_FG);

    vgaHWProtect(pScrn, FALSE);
        
    return;
}

static void 
VIARestore(ScrnInfoPtr pScrn, vgaRegPtr vgaSavePtr, VIARegPtr restore, int flags)
{
	vgaHWPtr        hwp = VGAHWPTR(pScrn);
	VIAPtr          pVia = VIAPTR(pScrn);
	VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
	int             vgaCRIndex, vgaCRReg, vgaIOBase;
	int             i;
    Bool            deFlick = flags;
    
	DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIARestore\n"));
	vgaIOBase = hwp->IOBase;
	vgaCRIndex = vgaIOBase + 4;
	vgaCRReg = vgaIOBase + 5;

    xf86CrtcConfigPtr	xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
    
    /* Disable crtc */
    for (i = 0; i < xf86_config->num_crtc; i++) {
       xf86CrtcPtr crtc = xf86_config->crtc[i];
       crtc->funcs->dpms(crtc, DPMSModeOff);
    }
    
    vgaHWProtect(pScrn, TRUE);
    
    /* Unlock Extended Regs */
    viaWriteVgaIoBits(REG_SR10, 0x01, BIT0);
    /* Unlock CRTC register protect */
    viaWriteVgaIoBits(REG_CR47, 0x00, 0xFE);

    /*=* CR6A, CR6B, CR6C must be reset before restore
        standard vga regs, or system will be hang. *=*/
    /*=* Reset IGA2 channel before disable IGA2 channel
        or it may cause some line garbage. *=*/
    /* Modify for the sequence issue of enable and disable */
    /* second display channel. */
    viaWriteVgaIoBits(REG_CR6A, 0x00, BIT6); /* Set CR6A[6] to 0. */
    viaWriteVgaIoBits(REG_CR6A, 0x00, BIT7); /* Set CR6A[7] to 0. */
    viaWriteVgaIoBits(REG_CR6A, 0x40, BIT6); /* Set CR6A[6] to 1. */
    viaWriteVgaIoBits(REG_CR6A, 0x00, 0x3F); /* Set CR6A[5:0] to 0. */

    viaWriteVgaIoBits(REG_CR6B, 0x00, 0xFF); /* Set CR6B[7:0] to 0. */
    viaWriteVgaIoBits(REG_CR6C, 0x00, 0xFF); /* Set CR6C[7:0] to 0. */

    /* Gamma must disable before restore pallette */
    viaWriteVgaIoBits(REG_CR33, 0x00, BIT7); /* Set CR33[7] to 0. */

    /* restore the standard vga regs */
    if (!pVia->IsMAMMEnable)
       vgaHWRestore(pScrn, vgaSavePtr, VGA_SR_ALL);
    else
       vgaHWRestore(pScrn, vgaSavePtr, VGA_SR_MODE);

    /* restore Sequencer regs: From SR14 to SR3F  */		 
    for (i=0x14; i <= 0x40; i++) {
        if ((i == 0x1a) || (i == 0x1e) || (i == 0x2c) || (i == 0x3d))
            continue;
        viaWriteVgaIoBits(i+REG_SR00, restore->SRRegs[i], 0xFF);
    } 

    viaWriteVgaIoBits(REG_SR42, restore->SRRegs[0x42], 0xFF);
    viaWriteVgaIoBits(REG_SR4F, restore->SRRegs[0x4F], 0xFF);
    viaWriteVgaIoBits(REG_SR50, restore->SRRegs[0x50], 0xFF);
    viaWriteVgaIoBits(REG_SR57, restore->SRRegs[0x57], 0xFF);
    viaWriteVgaIoBits(REG_SR5D, restore->SRRegs[0x5D], 0xFF);

    /* DVP1 Clock and Data Pads Driving: */
    viaWriteVgaIoBits(REG_SR65, restore->SRRegs[0x65], 0xFF);

    /*=* restore VCK, LCDCK and ECK *=*/
    /* VCK: */
    viaWriteVgaIoBits(REG_SR44, restore->SRRegs[0x44], 0xFF);
    viaWriteVgaIoBits(REG_SR45, restore->SRRegs[0x45], 0xFF);
    viaWriteVgaIoBits(REG_SR46, restore->SRRegs[0x46], 0xFF);
    /* Reset VCK PLL */
    viaWriteVgaIoBits(REG_SR40, 0x02, BIT1);	/* Set SR40[1] to 1 */
    viaWriteVgaIoBits(REG_SR40, 0x00, BIT1);	/* Set SR40[1] to 0 */

    /* LCK: */
    viaWriteVgaIoBits(REG_SR4A, restore->SRRegs[0x4A], 0xFF);
    viaWriteVgaIoBits(REG_SR4B, restore->SRRegs[0x4B], 0xFF);
    viaWriteVgaIoBits(REG_SR4C, restore->SRRegs[0x4C], 0xFF);
    /* Reset LCK PLL */
    viaWriteVgaIoBits(REG_SR40, 0x04, BIT2);	/* Set SR40[2] to 1 */
    viaWriteVgaIoBits(REG_SR40, 0x00, BIT2);	/* Set SR40[2] to 0 */

    /* ECK: */
    viaWriteVgaIoBits(REG_SR47, restore->SRRegs[0x47], 0xFF);
    viaWriteVgaIoBits(REG_SR48, restore->SRRegs[0x48], 0xFF);
    viaWriteVgaIoBits(REG_SR49, restore->SRRegs[0x49], 0xFF);
    /* Reset ECK PLL */
    viaWriteVgaIoBits(REG_SR40, 0x01, BIT0);	/* Set SR40[0] to 1 */
    viaWriteVgaIoBits(REG_SR40, 0x00, BIT0);	/* Set SR40[0] to 0 */

    /* Restore Preemptive Arbiter Control Register */
    if (pVia->Chipset == VIA_VX900) {
        viaWriteVgaIoBits(REG_SR4D, restore->SRRegs[0x4D], 0xFF);
    }        
      
    /* Restore CRTC controller extended regs: */
    viaWriteVgaIoBits(REG_CR08, restore->CRTCRegs[0x08], 0xFF);
    viaWriteVgaIoBits(REG_CR09, restore->CRTCRegs[0x09], 0xFF);

    viaWriteVgaIoBits(REG_CR13, restore->CRTCRegs[0x13], 0xFF);
    viaWriteVgaIoBits(REG_CR30, restore->CRTCRegs[0x30], 0xFF);
    viaWriteVgaIoBits(REG_CR31, restore->CRTCRegs[0x31], 0xFF);
    viaWriteVgaIoBits(REG_CR32, restore->CRTCRegs[0x32], 0xFF);
    viaWriteVgaIoBits(REG_CR33, restore->CRTCRegs[0x33], 0xFF);

    /* From CR35 to CR47 */
    i = 0x35;
    while (1) {
        viaWriteVgaIoBits(i+REG_CR00, restore->CRTCRegs[i], 0xFF);
        if (i == 0x47)
           break;
        else
           i++;
    }

    /* Starting Address: */
    viaWriteVgaIoBits(REG_CR0C, restore->CRTCRegs[0x0C], 0xFF);
    viaWriteVgaIoBits(REG_CR0D, restore->CRTCRegs[0x0D], 0xFF);
    viaWriteVgaIoBits(REG_CR48, restore->CRTCRegs[0x48], 0xFF);
    viaWriteVgaIoBits(REG_CR34, restore->CRTCRegs[0x34], 0xFF);

    /* From CR49 to CRFF */
    for (i = 0x49; i < 0xFF; i++) {
        if (i == 0x91) {
            /* 
                LCD Panel could use software power sequence, so we should resotore it 
                individually to avoid incorrect enable LCD.
            */
        } else {
            viaWriteVgaIoBits(i+REG_CR00, restore->CRTCRegs[i], 0xFF);
        }
    }
    
    /* Restore outputs */
    for (i = 0; i < xf86_config->num_output; i++) {
        xf86OutputPtr   output = xf86_config->output[i];
        if (output->funcs->restore)
            output->funcs->restore(output);
    }
    
    /* Restore CR91 after lvds restore*/
    viaWriteVgaIoBits(REG_CR91, restore->CRTCRegs[0x91], 0xFF);

    /* Reset clock */
    viaWriteMiscIo(viaReadMiscIo());
    
    /* Restore HW Cursor state */
    VIASETREG(VIA_REG_CURSOR_MODE, restore->dwCursorMode);
    VIASETREG(VIA_REG_CURSOR_POS, restore->dwCursorPOS);
    VIASETREG(VIA_REG_CURSOR_ORG, restore->dwCursorORG);
    VIASETREG(VIA_REG_CURSOR_BG, restore->dwCursorBG);
    VIASETREG(VIA_REG_CURSOR_FG, restore->dwCursorFG);

    vgaHWProtect(pScrn, FALSE);
    return;
}


static Bool VIAMapMMIO(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    vgaHWPtr hwp;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAMapMMIO\n"));

    pVia->MmioBase = MEMBASE(pVia->PciInfo, 1);

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping MMIO @ 0x%lx with size 0x%x\n",
               pVia->MmioBase, VIA_MMIO_REGSIZE));

    /*Notice: pci_device_map_range will return error if there is already a map region
                which base and size are the same as what you want to map. So in
                SAMM case, just map these regions only once.*/
    /*Map MMIO*/                
#if XSERVER_LIBPCIACCESS
    pci_device_map_range (pVia->PciInfo,
                          pVia->MmioBase,
                          VIA_MMIO_REGSIZE,
                          PCI_DEV_MAP_FLAG_WRITABLE,
                          (void **) &pVia->MapBase);
#else
    pVia->MapBase = xf86MapPciMem(pScrn->scrnIndex,
                                  VIDMEM_MMIO,
                                  pVia->PciTag,
                                  pVia->MmioBase,
                                  VIA_MMIO_REGSIZE);
#endif
    pBIOSInfo->MapBase = pVia->MapBase;

    MMIO_MB1 = pVia->MapBase;

    /*Map Blit space*/
    MMIOMapBase = pVia->MapBase + 0x8000;
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping BitBlt MMIO @ 0x%lx with size 0x%x\n",
               pVia->MmioBase + VIA_MMIO_BLTBASE, VIA_MMIO_BLTSIZE));

#if XSERVER_LIBPCIACCESS
    pci_device_map_range (pVia->PciInfo,
                   pVia->MmioBase + VIA_MMIO_BLTBASE,
                   VIA_MMIO_BLTSIZE,
                   PCI_DEV_MAP_FLAG_WRITABLE,
                   (void **) &pVia->BltBase);
#else
    pVia->BltBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                                  pVia->MmioBase + VIA_MMIO_BLTBASE,
                                  VIA_MMIO_BLTSIZE);
#endif

    if (!pVia->MapBase || !pVia->BltBase) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Internal error: cound not map registers\n");
        return FALSE;
    }

    /* Memory mapped IO for Video Engine */
    pVia->VidMapBase = pVia->MapBase + 0x200;

    /* Memory mapped IO for Integrated TV. */
#ifdef VIA_VT3293_SUPPORT    
    if ((pVia->pBIOSInfo->Chipset == VIA_CX700) ||
        (pVia->pBIOSInfo->Chipset == VIA_CN750)) {
#else
    if (pVia->pBIOSInfo->Chipset == VIA_CX700) {
#endif
        xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
                   "mapping Integrated TV MMIO @ 0x%lx with size 0x%x\n",
                   pVia->MmioBase + VIA_MMIO_INTEGRATED_TV_BASE, VIA_MMIO_INTEGRATED_TV_SIZE);

#if XSERVER_LIBPCIACCESS
        pci_device_map_range (pVia->PciInfo,
                              pVia->MmioBase + VIA_MMIO_INTEGRATED_TV_BASE,
                              VIA_MMIO_INTEGRATED_TV_SIZE,
                              PCI_DEV_MAP_FLAG_WRITABLE,
                              (void **) &pVia->IntegratedTVMapBase);
#else
        pVia->IntegratedTVMapBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                                                  pVia->MmioBase + VIA_MMIO_INTEGRATED_TV_BASE, 
                                                  VIA_MMIO_INTEGRATED_TV_SIZE);
#endif
        
        if (!pVia->IntegratedTVMapBase) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Internal error: cound not map integrated TV registers\n");
            return FALSE;
        }

        pBIOSInfo->IntegratedTVMapBase = pVia->IntegratedTVMapBase;
    }
    
    VIAEnableMMIO(pScrn);
    hwp = VGAHWPTR(pScrn);
    vgaHWGetIOBase(hwp);

    if (!vgaHWMapMem(pScrn))
        return FALSE;

    return TRUE;
}


static Bool 
VIAMapFB(ScrnInfoPtr pScrn)
{

    VIAPtr    pVia;
    pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAMapFB\n"));
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping framebuffer @ 0x%lx with size 0x%lx\n",
               pVia->ScreenBase, pVia->videoRambytes));

    if (!pVia->IsSecondary) {
        if (pVia->videoRambytes) {
#if XSERVER_LIBPCIACCESS
            /*Call PCI Map for Graphic Only ONE Time*/
		    pci_device_map_range(pVia->PciInfo, pVia->ScreenBase, pVia->pVIAEnt->pScreensPublicInfo->VRamSize,
			                     PCI_DEV_MAP_FLAG_WRITABLE | PCI_DEV_MAP_FLAG_WRITE_COMBINE,
					             (void **)&pVia->FBBase);
#else
		    pVia->FBBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_FRAMEBUFFER,
			                             pVia->PciTag, pVia->ScreenBase,
									     pVia->pVIAEnt->pScreensPublicInfo->VRamSize);
#endif
		    pBIOSInfo->FBBase = pVia->FBBase;
		    pBIOSInfo->videoRambytes = pVia->videoRambytes;

            if (!pVia->FBBase) {
                xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                           "Internal error: could not map framebuffer\n");
                return FALSE;
            }

            pVia->FBFreeStart = (pScrn->displayWidth * pScrn->bitsPerPixel >> 3) * pScrn->virtualY;        
            pVia->FBFreeEnd = pVia->videoRambytes;
            if (pVia->FBFreeStart >= pVia->FBFreeEnd ) {
	            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                           "NOT ENOUGH Memory For OnScreen !!\n");	
                    return FALSE;
    	    }
            pScrn->fbOffset = 0;

            DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
                   "Frame buffer start: %p, free start: 0x%x end: 0x%x\n",
                   pVia->FBBase, pVia->FBFreeStart, pVia->FBFreeEnd));
        }
    }
    else {   /*pVia->IsSecondary*/
        /*Get PCI Map correct Offset for Screen 1*/
        VIAPtr    pVia0;
        if(pVia->pVIAEnt)
            pVia0 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);
            
        pVia->FBBase = pVia0->FBBase + pVia0->videoRambytes;
        pBIOSInfo->FBBase = pVia0->FBBase + pVia0->videoRambytes;
        pBIOSInfo->videoRambytes = pVia->videoRambytes;

        pVia->FBFreeStart = (pScrn->displayWidth * pScrn->bitsPerPixel >> 3) * pScrn->virtualY;
        pVia->FBFreeEnd = pVia->videoRambytes;         
            
        pScrn->fbOffset = pVia->pVIAEnt->pPrimaryScrn->videoRam << 10;
    }

    if (pVia->pBIOSInfo->Chipset == VIA_VX900)
    {
        pScrn->memPhysBase = MEMBASE(pVia->PciInfo, 2);
    }
    else 
    {
        pScrn->memPhysBase = MEMBASE(pVia->PciInfo, 0);
    }

    return TRUE;
}


static void 
VIAUnmapMem(ScrnInfoPtr pScrn)
{
    VIAPtr pVia;

    pVia = VIAPTR(pScrn);

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAUnmapMem\n"));

    if (pVia->MapBase) {
#if XSERVER_LIBPCIACCESS
		pci_device_unmap_range (pVia->PciInfo, pVia->MapBase, 
						VIA_MMIO_REGSIZE);
#else
		xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->MapBase,
						VIA_MMIO_REGSIZE);
#endif
		pVia->MapBase = NULL;
    }

    if (pVia->BltBase) {
#if XSERVER_LIBPCIACCESS
		pci_device_unmap_range (pVia->PciInfo, pVia->BltBase, 
						VIA_MMIO_BLTSIZE);
#else
		xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->BltBase,
						VIA_MMIO_BLTSIZE);
#endif
		pVia->BltBase = NULL;
    }

    if (pVia->FBBase) {
#if XSERVER_LIBPCIACCESS
		pci_device_unmap_range (pVia->PciInfo, pVia->FBBase, 
						pVia->pVIAEnt->pScreensPublicInfo->VRamSize);
#else
		xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->FBBase,
						pVia->pVIAEnt->pScreensPublicInfo->VRamSize);
#endif
		pVia->FBBase = NULL;
    }

#ifdef VIA_VT3293_SUPPORT
    if ((pVia->pBIOSInfo->Chipset == VIA_CX700) ||
        (pVia->pBIOSInfo->Chipset == VIA_CN750)) {
#else
    if (pVia->pBIOSInfo->Chipset == VIA_CX700) {
#endif
        if (pVia->IntegratedTVMapBase) {
#if XSERVER_LIBPCIACCESS
			pci_device_unmap_range (pVia->PciInfo, pVia->IntegratedTVMapBase, 
							VIA_MMIO_INTEGRATED_TV_SIZE);
#else
			xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->IntegratedTVMapBase,
							VIA_MMIO_INTEGRATED_TV_SIZE);
#endif
			pVia->IntegratedTVMapBase = NULL;
        }
    }
    return;
}

 
int 
VIAHWCursorSpaceAllocate(ScreenPtr pScreen)
{	
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    unsigned int addr = 0;

    if(!pVia->IsSecondary) {
        if (!pVia->ForceSWCursor) {
            int i;
            pVia->FBFreeStart = (pVia->FBFreeStart+ 0x3FF) & (~0x3FF);
            for (i = 0; i < 2; i++) {
                addr = viaFBAlloctor(pScrn, VIA_CURSOR_SIZE, SEQUENCE, "RandR HW Icon Buf");
                if (addr) {
                    pVia->HiStartBuf[i] = addr;
                } else {
                    return -1;
                }
            }
        }
	}
    return 0;
}
#ifdef DRM_SAMM_VIA

void VIADrmUnlock(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    VIAPtr pVia0 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);

    if (pVia0->LockRefCount > 0) {
        pVia0->LockRefCount--;
    } else {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
            "In SAMM+DRM case, driver unlock too many times.\n");
    }

    if (!pVia0->LockRefCount) {
        drmUnlock(pVia0->drmFD, samm_ctx_hdl);
    }
}

void VIADrmLock(ScreenPtr pScreen, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    VIAPtr pVia0 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);

    if (!pVia0->LockRefCount) {
        drmGetLock(pVia0->drmFD, samm_ctx_hdl, 0);
    }
    pVia0->LockRefCount++;
}

/* These functions will be called When X reponse to something */
void
VIABlockHandler(BLOCKHANDLER_ARGS_DECL)
{
    SCREEN_PTR(arg);
    SCRN_INFO_PTR(arg);
    VIAPtr pVia = VIAPTR(pScrn);

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
        "enter VIABlockHandler function\n"));

    pScreen->BlockHandler = pVia->viaBlockHandlerback;
    (*pScreen->BlockHandler) (BLOCKHANDLER_ARGS);

    pVia->viaBlockHandlerback = pScreen->BlockHandler;
    pScreen->BlockHandler = VIABlockHandler;

    /* do usefull work here */
    /* We can take it for granted that 
     * the DRM has been initialized properly
     */
    VIADrmUnlock(pScreen);
}

void
VIAWakeupHandler(BLOCKHANDLER_ARGS_DECL)
{
    SCREEN_PTR(arg);
    SCRN_INFO_PTR(arg);
    VIAPtr pVia = VIAPTR(pScrn);

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO,
        "enter VIAWakeupHandler function\n"));

    pScreen->WakeupHandler = pVia->viaWakeupHandlerback;
    (*pScreen->WakeupHandler) (BLOCKHANDLER_ARGS);

    pVia->viaWakeupHandlerback = pScreen->WakeupHandler;
    pScreen->WakeupHandler = VIAWakeupHandler;

    /* do usefull work here */
    /* We can take it for granted that 
     * the DRM has been initialized properly
     */
    VIADrmLock(pScreen, 0);
}
#endif

static Bool 
VIAScreenInit(SCREEN_INIT_ARGS_DECL)
{
    ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
    VIAPtr pVia = VIAPTR(pScrn);
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    int ret;
    unsigned int TmpSwapValue;
	unsigned int addr;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;

    
    
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAScreenInit\n"));    
/*version usingdef */
#ifdef XORG_VERSION_CURRENT
	DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "XF86_VER=%d\n, Not using in driver Now.\n", via_xf86_version_current));
#if XORG_VERSION_CURRENT <= XF86_VERSION_NUMERIC(2,0,0,0,0)&& XORG_VERSION_CURRENT >= XF86_VERSION_NUMERIC(1,0,0,0,0)
	DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "XORG_VER=%d\n, Now driver depending only on this.\n", XORG_VERSION_CURRENT));    
#else
	DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "WARNING!!XORG_VER Out of (2,0,0,0,0)and(1,0,0,0,0),if =7.0~2,will ok,or Old version driver no surpport.\n"));    
#endif
#else
	DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "WARNING!!XORG_VER=%d\n,Not Exist,flow may wrong  Now.\n", XORG_VERSION_CURRENT));    
#endif
/*version using define*/

    /* For MAMM*/
    if (!pVia->IsMAMMEnable)
        vgaHWUnlock(hwp);
      
    if (!VIAMapFB(pScrn))
        return FALSE;

	/*Notice: SAMM, both two screens share the same MMIO space, so just need to map once.*/
	if (!pVia->IsSecondary)	{
    	if (!VIAMapMMIO(pScrn))
        	return FALSE;
	} else {
		/*Copy the first screen's MMIO map address to the second screen*/
		VIAPtr pVia0 = VIAPTR(pVia->pVIAEnt->pPrimaryScrn);	
		/*MMIO*/
		pVia->MapBase = pVia0->MapBase;
		pBIOSInfo->MapBase = pVia->MapBase;
		/*BitBlt MMIO*/
		pVia->BltBase = pVia0->BltBase;
		/*Video MMIO*/
		pVia->VidMapBase = pVia0->VidMapBase;
		/*Integrated TV MMIO*/
		if (pVia->pBIOSInfo->Chipset == VIA_CX700) {
			pVia->IntegratedTVMapBase = pVia0->IntegratedTVMapBase;
			pBIOSInfo->IntegratedTVMapBase = pVia->IntegratedTVMapBase;
		}
		VIAEnableMMIO(pScrn);
		vgaHWGetIOBase(hwp);
		/*For what?*/
		if (!vgaHWMapMem(pScrn))
			return FALSE;
	}
	
    if (pVia->numOfEmbTv) {
        if (integratedtv_module_loaded) {
            addr = viaFBAlloctor(pScrn, INTEGRATED_TV_BUFFER_SIZE, INVERT_SEQ, "Integrated TV buf");
            if(addr) {
                pVia->IntegratedTVBufferStart = addr; 
            }
            else {
                return FALSE;
            }
            viaSetEmbTvBufferAddress((unsigned int)pVia->IntegratedTVMapBase,
                                     pVia->IntegratedTVBufferStart);
        }
	}

    /*modified the variable type,and change the function of this variable*/
    /*in SAMM, maybe have a issue, study in laterly*/
    if(IsNeedReserve3DScalingMem(pVia) )
        Reserve3DScalingMemory(pScrn);

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Mem Mapped\n"));

#ifdef XF86DRI
    if (pVia->directRenderingEnabled == TRUE) {
        pVia->directRenderingEnabled = VIADRIScreenInit(pScreen);
    }
#endif

    VIASave(pScrn);

    /*Clear FB with black color to avoid garbage*/
    memset(pVia->FBBase, 0x00, ((pScrn->displayWidth * pScrn->bitsPerPixel>>3) * pScrn->virtualY));

    /* Initialize display engine, must after VIASave() and before 2D and 3D Engine init() */
    viaInitDispEngine(pVia);
    viaInitOutputRegSet(pScrn);

    /* Even we had disable the spread spectrum at PreInit.
       But we still disable it again for safty.
    */
    if (pVia->pChipInfo->biosIsSpreadSpectrum) {                 
        /* Disable spread spectrum temporarily, we will turn it on when set mode. */
        VGAOUT8(0x3c4, SR3D);
        VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xFE);

        VGAOUT8(0x3c4, SR1E);
        VGAOUT8(0x3c5, VGAIN8(0x3c5) & 0xF7);        
    }

    vgaHWBlankScreen(pScrn, FALSE);   


	/* Enable the graphics engine. */
	if (!pVia->NoAccel) {
	    VIAInitialize2DEngine(pScrn);
	    VIAInitialize3DEngine(pScrn);
    }

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- State saved\n"));

    /* Darken the screen for aesthetic reasons and set the viewport */
    VIASaveScreen(pScreen, SCREEN_SAVER_ON);
    pScrn->AdjustFrame(ADJUST_FRAME_ARGS(pScrn, pScrn->frameX0, pScrn->frameY0));

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Blanked\n"));

    miClearVisualTypes();

    if (pScrn->bitsPerPixel > 8 && !pVia->IsSecondary) {
        if (!miSetVisualTypes(pScrn->depth, TrueColorMask,
                              pScrn->rgbBits, pScrn->defaultVisual))
            return FALSE;
        if (!miSetPixmapDepths())
            return FALSE;
    } else {
        if (!miSetVisualTypes(pScrn->depth, miGetDefaultVisualMask(pScrn->depth),
                              pScrn->rgbBits, pScrn->defaultVisual))
            return FALSE;
        if (!miSetPixmapDepths())
            return FALSE;
    }

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Visuals set up\n"));

    ret = VIAInternalScreenInit(pScrn->scrnIndex, pScreen);

    if (!ret)
        return FALSE;

    xf86SetBlackWhitePixels(pScreen);
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- B & W\n"));

    if (pScrn->bitsPerPixel > 8) {
        VisualPtr visual;

        visual = pScreen->visuals + pScreen->numVisuals;
        while (--visual >= pScreen->visuals) {
            if ((visual->class | DynamicClass) == DirectColor) {
                visual->offsetRed = pScrn->offset.red;
                visual->offsetGreen = pScrn->offset.green;
                visual->offsetBlue = pScrn->offset.blue;
                visual->redMask = pScrn->mask.red;
                visual->greenMask = pScrn->mask.green;
                visual->blueMask = pScrn->mask.blue;
            }
        }
    }

    /* must be after RGB ordering fixed */
    fbPictureInit(pScreen, 0, 0);

    /*for 2D accel fail*/
    if (!pVia->NoAccel){
        if(FALSE==VIAInitAccel(pScreen))
	        return FALSE;
    } else {
        /* No Accel path :
         * Sync marker space. 
         */
         /* No Accel path :
           Init the 3D pipeline API for the uniform 3D texture blting interface */
        viaExa3DInit(pScreen);
        viaExa2DInit(pScreen);
        addr = viaFBAlloctor( pScrn, 32, INVERT_SEQ, "EXA Sync Marker");
        if(addr) { 
            pVia->markerOffset = addr;
        }
        else {
            return FALSE;    
        }
        pVia->markerOffset = ( pVia->markerOffset + 31) & ~31;
        pVia->markerBuf = (volatile CARD32 *) ((char *)pVia->FBBase + pVia->markerOffset);
        *pVia->markerBuf = 0;
        pVia->curMarker = 0;
        pVia->lastMarkerRead = 0;
    }
    if (pVia->NoAccel) {
	    ret = VIAHWCursorSpaceAllocate(pScreen);
	    if(ret != 0) {
	        return FALSE; 
	    }
    }
#ifdef XF86DRI
    /*for VIDEO play memory overlap PIXEL_MAP_CACHE*/
    if (pVia->directRenderingEnabled 
#ifdef DRM_SAMM_VIA
        || pVia->drmSammEnabled
#endif
        ) {
	    if (!(VIADRIFBInit(pScreen, pVia))) {
	        VIADRICloseScreen(pScreen);
	        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "[dri] frame buffer initialize failed .\n" );
	        pVia->directRenderingEnabled = FALSE;
	    }
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] frame buffer initialized done.\n" );
    }
#endif

    miInitializeBackingStore(pScreen);
    xf86SetBackingStore(pScreen);
    xf86SetSilkenMouse(pScreen);
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Backing store set up\n"));

    /* hardware cursor needs to wrap this layer */
    xf86DiDGAInit (pScreen, pVia->FrameBufferBase + pScrn->fbOffset);
        
    miDCInitialize(pScreen, xf86GetPointerScreenFuncs());
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- SW cursor set up\n"));

	
    if (!pVia->ForceSWCursor) {
        /*Disable HW cursor 0x2D0[0]*/
        CARD32 regCursorMode = VIAGETREG(VIA_REG_CURSOR_MODE);
        VIASETREG(VIA_REG_CURSOR_MODE, regCursorMode & 0xFFFFFFFE);                        
        /*Init HW Icon*/
        VIAARGBCursorInit(pScrn);
        if(!xf86_cursors_init(pScreen, MAX_CURS, MAX_CURS,
               (HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_64 |
                HARDWARE_CURSOR_AND_SOURCE_WITH_MASK |
                HARDWARE_CURSOR_TRUECOLOR_AT_8BPP |
                HARDWARE_CURSOR_INVERT_MASK |
                HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |                                              
                HARDWARE_CURSOR_ARGB | 0)))
            DEBUG(ErrorF("Fail to initial xf86_cursors_init()\n"));
    }

    /* For RandR v1.2 */
    if (!xf86CrtcScreenInit(pScreen)) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "xf86CrtcScreenInit Fail\n"));
        return FALSE;
    }    

    if (!miCreateDefColormap(pScreen))
        return FALSE;
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Def Color map set up\n"));

    /* only 3123Ax IGA2 use 6-Bit LUT, others chipset IGA2 use 8-Bit LUT */
    /* So, except 3123Ax, we use 8-Bit LUT as default. */
    pScrn->rgbBits = 8;
    VGAOUT8(0x3C4, SR15);
    VGAOUT8(0x3C5, VGAIN8(0x3C5) | 0x80);
    if (!xf86HandleColormaps(pScreen, 256, 8, VIALoadPalette, NULL,
                             CMAP_RELOAD_ON_MODE_SWITCH | CMAP_PALETTED_TRUECOLOR))
        return FALSE;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Palette loaded\n"));

    vgaHWBlankScreen(pScrn, TRUE);

    pVia->CloseScreen = pScreen->CloseScreen;   
    pScreen->CloseScreen = VIACloseScreen;

    xf86DPMSInit(pScreen, xf86DPMSSet, 0);
    pScreen->SaveScreen = xf86SaveScreen;

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- DPMS set up\n"));

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Color maps etc. set up\n"));

#ifdef XF86DRI
    if (pVia->directRenderingEnabled 
#ifdef DRM_SAMM_VIA
       || pVia->drmSammEnabled
#endif
       ) 
        pVia->directRenderingEnabled = VIADRIFinishScreenInit(pScreen);

    if (pVia->directRenderingEnabled) {
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "direct rendering enabled\n");
        VIASet3DSyncInfoParas(pScreen);
    } else {
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "direct rendering disabled\n");
    }
#endif

    if (!pVia->NoAccel){
        viaFinishInitAccel(pScreen);
    } else {    
        /* Init Command Buffer for the uniform 3D texture blting interface */
        viaSetupCBuffer(pScrn, &pVia->cb, 0);
    }

    /* Initial VIA Video Driver */
    FillGraphicMemInfo(pScrn);
    FillGraphicInfo(pScrn);

    /*added for RANDR12 static rotate , because of the xorg's randr12 has bug, at which point pScrn->pScreen is not set but the 
    xf86SetDesiredModes will use it, so we just manually assign it before the function, we only need to do this manually in 
    xorg-1.4.0.90,in xorg-1.4.99.905 is fix, we do not need to assign it manully*/
    ScrnInfoPtr pTemp = pScrn;
    pScrn->pScreen = screenInfo.screens[pScrn->scrnIndex];
    
    if (!xf86SetDesiredModes(pScrn))
        return FALSE;                

    pScrn->pScreen = pTemp->pScreen;  

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "DRMEnabled: 0x%x\n", 
            viaGfxInfo->drmEnabled));
    if (viaGfxInfo->igaInfo[IGA1-1].actived) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA1 dwWidth: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA1-1].desired_width));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA1 dwHeight: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA1-1].desired_height));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA1 dwRefreshRate: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA1-1].refreshrate));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA1 dwExpand: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA1-1].igaStatus.expanded));
    }
    if (viaGfxInfo->igaInfo[IGA2-1].actived) {
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA2 dwWidth: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA2-1].desired_width));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA2 dwSaveWidth: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA2-1].desired_height));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA2 dwBPP: 0x%lx\n", 
                viaGfxInfo->screenInfo[IGA2-1].bpp));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA2 dwRefreshRate: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA2-1].refreshrate));
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "using IGA2 dwExpand: 0x%lx\n", 
                viaGfxInfo->igaInfo[IGA2-1].igaStatus.expanded));
    }

    /*  Init video structure, video flags & HW difference flags */
    vidInitVideoInfo(pVidData );
    
    viaInitVideo(pScreen);

#ifndef CS_FLAG_NO_XVMC
    /* 
       When rotate in FC3,viaInitMC will result in error.So disable it when rotate */
    if(!pVia->RotateDegree)
        viaInitMC(pScreen);
#endif /* CS_FLAG_NO_XVMC */

    /* For XV Rotate+panning support, set the correct Clipping window. */
    if ((pVia->RotateDegree == VIA_ROTATE_DEGREE_90) ||
        (pVia->RotateDegree == VIA_ROTATE_DEGREE_270)) {
        TmpSwapValue = pScrn->currentMode->HDisplay;
        pScrn->currentMode->HDisplay = pScrn->currentMode->VDisplay;
        pScrn->currentMode->VDisplay = TmpSwapValue;
    }

    if (serverGeneration == 1)
        xf86ShowUnusedOptions(pScrn->scrnIndex, pScrn->options);


/* Only hook this two function when SAMM + DRM enabled.
 * for dealing with DRM lock.
 * Maybe we can do other usefull things in this function.
 */
#ifdef DRM_SAMM_VIA
    if (pVia->drmSammEnabled) {
        pVia->viaBlockHandlerback = pScreen->BlockHandler;
        pScreen->BlockHandler = VIABlockHandler;
        pVia->viaWakeupHandlerback = pScreen->WakeupHandler;
        pScreen->WakeupHandler = VIAWakeupHandler;
    }
#endif

    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "- Done\n"));
    return TRUE;
}

static int 
VIAInternalScreenInit(int scrnIndex, ScreenPtr pScreen)
{
    ScrnInfoPtr     pScrn;
    VIAPtr          pVia;
    VIABIOSInfoPtr  pBIOSInfo;
    int             width, height, displayWidth, shadowHeight;
    unsigned char   *FBStart;
    int             ret = TRUE;

    xf86DrvMsg(scrnIndex, X_INFO, "VIAInternalScreenInit\n");

    pScrn = xf86Screens[pScreen->myNum];
    pVia = VIAPTR(pScrn);
    pBIOSInfo = pVia->pBIOSInfo;

    displayWidth = pScrn->displayWidth;

    if ((pVia->RotateDegree == VIA_ROTATE_DEGREE_90)||
        (pVia->RotateDegree == VIA_ROTATE_DEGREE_270)) {
        height = pScrn->virtualX;
        width = pScrn->virtualY;
		
        /* For XV Rotate+panning support, set the correct Clipping window. */
        pScrn->virtualX = width;
        pScrn->virtualY = height;
    } else {
        width = pScrn->virtualX;
        height = pScrn->virtualY;        
    }

    /* 
    *   Use Direct Access Window Rotate. 
    *   When screen will be rotated, we must report a rotated size to OS
    *   to generate a rotated screen data. 
    */
    if(pVia->IsHWRotateEnabled) {
        switch (pVia->RotateDegree) {
        case VIA_ROTATE_DEGREE_90:
        case VIA_ROTATE_DEGREE_270:
             displayWidth = width;
        break;
        }
    }
   
    ret = fbScreenInit(pScreen, pVia->FBBase, width, height,
                       pScrn->xDpi, pScrn->yDpi, displayWidth,
                       pScrn->bitsPerPixel);

    return ret;
}


static ModeStatus 
VIAValidMode(int scrnIndex, DisplayModePtr mode, Bool verbose, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr      pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    int ModeIndex;
    CARD32 DesiredModePixelClock; 
    
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIAValidMode\n"));

    if (!memcmp(mode->name, "PanelMode", 9)) {
        mode->type = M_T_USERDEF;        
        VIAGetModeLineTiming(pScrn, mode);
        pBIOSInfo->EDIDType = VIA_DEVICE_LCD;
        pBIOSInfo->LVDSSettingInfo.PanelSizeID =
            VIA_MAKE_ID(mode->HDisplay, mode->VDisplay);
        pBIOSInfo->isPanelModeLine = TRUE;
        return MODE_OK;
    }

    if(!pBIOSInfo->NoDDCValue) {
    	/*1.if user don't set modes in config file, we want to use PreferMode, so we use PreferModeH  PreferModeV to valid modes.
    	   2.if user set modes in config file, we use EDID can support maxH, maxV to valid modes*/ 
        /* TODO: Determine the heuristics for figuring this out. */
        /*            if ((mode->CrtcHDisplay > pBIOSInfo->CRTSettingInfo.PreferModeH) || 
         *      		    (mode->CrtcVDisplay > pBIOSInfo->CRTSettingInfo.PreferModeV))    
         *               return MODE_BAD_VVALUE;
         */
    	if ((mode->CrtcHDisplay > pBIOSInfo->CRTSettingInfo.MonitorSizeH) || 
    	    (mode->CrtcVDisplay > pBIOSInfo->CRTSettingInfo.MonitorSizeV))
			return MODE_BAD_VVALUE;
    }
    
    /* DP hardware limitation: TU ratio only 15 bits */ 
    if (pBIOSInfo->ActiveDevice & (VIA_DEVICE_DP | VIA_DEVICE_DP2)) {
        if (!IsValidMode(mode->CrtcHDisplay, mode->CrtcVDisplay))
            return MODE_BAD;
		
        ModeIndex = VIAGetModeIndex(mode->CrtcHDisplay, mode->CrtcVDisplay);
        DesiredModePixelClock = VIAGetPixelClockByMode(ModeIndex);
        if (pBIOSInfo->ActiveDevice & VIA_DEVICE_DP) {
            if (VIADPGetLinkSpeed(pBIOSInfo)) {
                VIACalculate_TU(&(pBIOSInfo->DPSettingInfo), DesiredModePixelClock, pBIOSInfo->bitsPerPixel);
                if (pBIOSInfo->DPSettingInfo.TURatio > 0x7FFF)
                    return MODE_BAD;
            }
        } else {
            if (VIADP2GetLinkSpeed(pBIOSInfo)) {
                VIACalculate_TU(&(pBIOSInfo->DPSettingInfo2), DesiredModePixelClock, pBIOSInfo->bitsPerPixel);
                if (pBIOSInfo->DPSettingInfo2.TURatio > 0x7FFF)
                    return MODE_BAD;
            }
        }    
    }

    /* If the modeline is added by user, pass it. */
    if (( mode->type == M_T_USERDEF) || (mode->type == 0))
        return MODE_OK;

    /* If the modeline is added by monitor, pass it. */
    switch (mode->type) {
        case M_T_DRIVER:                                /* Monitor EDID */
        case M_T_DRIVER | M_T_USERDEF:                  /* Monitor EDID + Auto-detection */
        case M_T_DRIVER | M_T_PREFERRED:                /* Monitor EDID + Monitor Preferred Mode */
        case M_T_DRIVER | M_T_USERDEF | M_T_PREFERRED:  /* Monitor EDID + Auto-detection + Monitor Preferred Mode */
            return MODE_OK;
            break;
    }    
     
    /* check modes */
    if(IsValidMode(mode->CrtcHDisplay, mode->CrtcVDisplay) == 0){
        return MODE_BAD;
    } else {
        return MODE_OK;
    }
}


static void 
VIABIOSInit(VIAPtr pVia, ScrnInfoPtr pScrn, DisplayModePtr pMode)
{
    int             i;
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIABIOSInit begin\n"));

    pBIOSInfo->bitsPerPixel = pScrn->bitsPerPixel;
    pBIOSInfo->displayWidth = pScrn->displayWidth;
    pBIOSInfo->IGA1SettingInfo.IsActive = FALSE;
    pBIOSInfo->IGA2SettingInfo.IsActive = FALSE;

    pBIOSInfo->Clock = pMode->Clock;
    pBIOSInfo->HTotal = pMode->HTotal;
    pBIOSInfo->VTotal = pMode->VTotal;
    pBIOSInfo->HDisplay = pMode->HDisplay;
    pBIOSInfo->VDisplay = pMode->VDisplay;
    pBIOSInfo->CrtcHDisplay = pMode->CrtcHDisplay;
    pBIOSInfo->CrtcVDisplay = pMode->CrtcVDisplay;

    if (pBIOSInfo->FirstInit) {
        pBIOSInfo->SaveHDisplay = pMode->HDisplay;
        pBIOSInfo->SaveVDisplay = pMode->VDisplay;
        pBIOSInfo->SaveVirtualX = pScrn->virtualX;
        pBIOSInfo->SaveVirtualY = pScrn->virtualY;
    }

    for (i = 0; i < 0xFF; i++) {
        pBIOSInfo->TVRegs[i] = pVia->SavedReg.TVRegs[i];
    }
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pBIOSInfo:bitsPerPixel=%d,displayWidth=%d,frameX1=%d,frameY1=%d\n",pBIOSInfo->bitsPerPixel,pBIOSInfo->displayWidth,pBIOSInfo->frameX1,pBIOSInfo->frameY1));     
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pMode:HTotal=%d,VTotal=%d,HDisplay=%d,VDisplay=%d\n",pMode->HTotal,pMode->VTotal,pMode->HDisplay,pMode->VDisplay));     
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "pMode:,CrtcHDisplay=%d,CrtcVDisplay=%d\n",pMode->CrtcHDisplay,pMode->CrtcVDisplay)); 
    DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIABIOSInit exit\n"));    
}


void 
VIACorrectModeList(ScrnInfoPtr pScrn)
{
    DisplayModePtr mode;

    for (mode = pScrn->modes; ; mode = mode->next) {
        VIAGetModeSizeByName(mode->name, &mode->HDisplay, &mode->VDisplay);

        if (mode->next == pScrn->modes)
            break;  /* The end of mode list. */
    }
}


static
Bool VIACloseScreen(CLOSE_SCREEN_ARGS_DECL)
{
    ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
    vgaHWPtr    hwp = VGAHWPTR(pScrn);
    VIAPtr      pVia = VIAPTR(pScrn);
    vgaRegPtr   vgaSavePtr = &hwp->SavedReg;
    VIARegPtr   viaSavePtr = &pVia->SavedReg;
    CARD32      dwCursorMode;
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;
    int rotate_degree;

    if (pVia->devicesTimer)
        TimerCancel(pVia->devicesTimer);

    pVia->devicesTimer = NULL;
    
    /*when log out X, if DRI is enabled, the DRI resource should be cleaned up*/
    if (pVia->directRenderingEnabled
#ifdef DRM_SAMM_VIA
        || pVia->drmSammEnabled
#endif
      ) {
        VIADRICloseScreen(pScreen);
    }

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIACloseScreen\n");

    if (pVia->IsHWRotateEnabled) {
        DisableCBUrotate(pScrn);

        /* Disable CBU HW Rotate.
         * VX855 and VX900 rotation 90 and 270 degree has issue
         * but 180 degree doesn't have.
         * So here mimic the behavior of CBU 180
         * The root cause is still unknown.
         */
        if ((pVia->Chipset == VIA_VX855 || pVia->Chipset == VIA_VX900) &&
            (pVia->RotateDegree == VIA_ROTATE_DEGREE_90 ||
            pVia->RotateDegree == VIA_ROTATE_DEGREE_270)) {
            /* Mimic the behavior of rotate 180 */
            rotate_degree = pVia->RotateDegree;
            pVia->RotateDegree = VIA_ROTATE_DEGREE_180;
            EnableCBUrotate(pScrn);
            pVia->RotateDegree = rotate_degree;
            /* Let CBU 180 rotate take effect */
            memset(pVia->FBBase, 0x00,
            (pScrn->displayWidth * pScrn->bitsPerPixel >> 3));
            
            DisableCBUrotate(pScrn);
        }
     }
    
    /*Turn off color cursor here*/
    VIASETREG(PRIM_HI_CTRL, 0);
    VIASETREG(HI_CONTROL, 0);

    if (pScrn->vtSema) {
        /* Move exit video here if exitvideo after restore SR1A (PCI reset)
         * the V3 command fire may stick and caused system hang when playing video
         * Turn off all video activities */
        viaExitVideo(pScrn);
    }

    if (!pVia->IsSecondary) {
        /* clean viddata structure*/
        int i;

        /* Reset panning mode parameter */
        /*Modified for SAMM, Panning for Video XV rotate*/
        for (i = 0; i < 2; i ++) {
            pVidData->panning_x[i] = 0;
            pVidData->panning_y[i] = 0;
        }

        viaGfxInfo->drmEnabled = FALSE;

        /* Diable Hardware Cursor */
        dwCursorMode = VIAGETREG(VIA_REG_CURSOR_MODE);
        VIASETREG(VIA_REG_CURSOR_MODE, dwCursorMode & 0xFFFFFFFE);
    }

    /* Wait Hardware Engine idle to exit graphic mode */
    WaitIdle();

    if ((pVia->VQEnable)&&(pScrn->vtSema)) {
        /* if we use VQ, disable it before we exit */
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIACloseScreen, VQ, disable \n");
        VIADisableVQ(pScrn);
    }

    if(pVia->HwCursorImage) {
        xfree(pVia->HwCursorImage);
        pVia->HwCursorImage = NULL;
    }
    
    if(pVia->HwIconImage) {
        xfree(pVia->HwIconImage);
        pVia->HwIconImage = NULL;
    }
    
    if(pVia->HIScalingPrimImage) {
        xfree(pVia->HIScalingPrimImage);
        pVia->HIScalingPrimImage = NULL;
    }
    
    if(pVia->HIScalingSecImage) {
        xfree(pVia->HIScalingSecImage);
        pVia->HIScalingSecImage = NULL;
    }
    
    if (pVia->CursorInfoRec) {
        xf86DestroyCursorInfoRec(pVia->CursorInfoRec);
        pVia->CursorInfoRec = NULL;
    }
    
    viaExitAccel(pScreen);
    
    if (pScrn->vtSema) {    
        DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIACloseScreen, pScrn->vtSema is true \n"));            
        /* Clear Memory to avoid garbage */
        memset(pVia->pBIOSInfo->FBBase, 0x00, pVia->pBIOSInfo->videoRambytes - VIA_FB_CURSOR_SIZE);                
        
        if (!pVia->IsSecondary)
            VIARestore(pScrn, vgaSavePtr, viaSavePtr, FALSE);
		
        if (pVia->Chipset == VIA_VX900) {
            if (MMIO_RD(DP_EPHY_PLL_REG) & BIT0)
                viaDPDisableEPHY();
            if (MMIO_RD(DP2_ENABLE_IF_REG) & BIT0)
                viaDP2DisableEPHY();
        }
		
        vgaHWLock(hwp);
        /*SAMM case, make sure MMIO only unmap once!*/
        if (!pVia->IsSecondary)
            VIAUnmapMem(pScrn);
        else {
            if (pVia->FBBase) {
                pVia->FBBase = NULL;
            }
        }
        vgaHWUnmapMem(pScrn);
    }

    /* For MAMM*/
    if (!pVia->IsMAMMEnable) {
        if (pVia->pInt10) {
            xf86FreeInt10(pVia->pInt10);
            pVia->pInt10 = NULL;
        }
    }

    if (!pVia->IsSecondary) {
        /* delete the memory of VidData */
        if(pVia->pVIAEnt->pVidData) {
            if(pVia->pVIAEnt->pVidData->viaGfxInfo){
                xfree((void *)pVia->pVIAEnt->pVidData->viaGfxInfo);
                pVia->pVIAEnt->pVidData->viaGfxInfo = NULL;
            }
            xfree((void *)pVia->pVIAEnt->pVidData);
            pVia->pVIAEnt->pVidData = NULL;
        }
    }
    
    pScrn->vtSema = FALSE;
    pScreen->CloseScreen = pVia->CloseScreen;

    return (*pScreen->CloseScreen)(CLOSE_SCREEN_ARGS);
}

/*
 * This only gets called when a screen is being deleted.  It does not
 * get called routinely at the end of a server generation.
 */
static void 
VIAFreeScreen(int scrnIndex, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];    
    VIAPtr      pVia = VIAPTR(pScrn);
    
    xf86DrvMsg(scrnIndex, X_INFO, "VIAFreeScreen\n");
    
    if (xf86LoaderCheckSymbol("vgaHWFreeHWRec"))
    vgaHWFreeHWRec(pScrn);
	
	/*SAMM case, make sure MMIO only unmap once!*/
    if (!pVia->IsSecondary) {
        /* delete the memory of VidData */
        if (pVia->pVIAEnt->pVidData) {
            if(pVia->pVIAEnt->pVidData->viaGfxInfo){
                xfree((void *)pVia->pVIAEnt->pVidData->viaGfxInfo);
                pVia->pVIAEnt->pVidData->viaGfxInfo = NULL;
            }
            xfree((void *)pVia->pVIAEnt->pVidData);
            pVia->pVIAEnt->pVidData = NULL;
        }
        VIAUnmapMem(pScrn);
    } else {
        if (pVia->FBBase) {
			pVia->FBBase = NULL;
		}
	}
	
    VIAFreeRec(pScrn);
}

static Bool 
VIASaveScreen(ScreenPtr pScreen, int mode)
{
	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr      pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo = pVia->pBIOSInfo;

    Bool on;

    on = xf86IsUnblank(mode);

    if (on)
    {
        EnableSecondDisplayChannel();
        /*Enable IGA2 input data*/
        switch (pVia->Chipset)
        {
            case VIA_P4M800PRO:
            case VIA_K8M890:
            case VIA_P4M890:
            case VIA_P4M900:
                write_reg_mask(CR91, VIACR, 0x00, BIT5);
                break;
            case VIA_CX700:  
            case VIA_VX800:
            case VIA_VX855:
            case VIA_VX900:
                write_reg_mask(CR6B, VIACR, 0x00, BIT2);
                break;
            default:
                write_reg_mask(CR91, VIACR, 0x00, BIT5);
                break;
        }
    }
    else
    {
        /*Disable IGA2 input data*/
        switch (pVia->Chipset)
        {
            case VIA_P4M800PRO:
            case VIA_K8M890:
            case VIA_P4M890:
            case VIA_P4M900:
                write_reg_mask(CR91, VIACR, BIT5, BIT5);
                break;
            case VIA_CX700:  
            case VIA_VX800:
            case VIA_VX855:
            case VIA_VX900:
                write_reg_mask(CR6B, VIACR, BIT2, BIT2);
                break;
            default:
                write_reg_mask(CR91, VIACR, BIT5, BIT5);
                break;
        }        
        DisableSecondDisplayChannel();
    }

    return vgaHWSaveScreen(pScreen, mode);
}

int 
VIAGetIGAInUse(VIABIOSInfoPtr pBIOSInfo)
{
    switch (pBIOSInfo->ActiveDevice) {
        case VIA_DEVICE_CRT1:
            return pBIOSInfo->CRTSettingInfo.IGAPath;

        case VIA_DEVICE_LCD:
            return pBIOSInfo->LVDSSettingInfo.IGAPath;

        case VIA_DEVICE_DFP:
            return pBIOSInfo->TMDSSettingInfo.IGAPath;

        case VIA_DEVICE_TV:
            return pBIOSInfo->TVSettingInfo.IGAPath;

        case VIA_DEVICE_TV2:
            return pBIOSInfo->TVSettingInfo2.IGAPath;

        case VIA_DEVICE_CRT2:
            return pBIOSInfo->CRTSettingInfo2.IGAPath;

        case VIA_DEVICE_DP:
            return pBIOSInfo->DPSettingInfo.IGAPath;  

        case VIA_DEVICE_DP2:
            return pBIOSInfo->DPSettingInfo2.IGAPath;  

        default:
            /* Wrong Device ID! */
            return 0;
    }
}


void 
VIARandR12UpdateOverlay(int scrnIndex, int x, int y, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr  pVia = VIAPTR(pScrn);
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;    
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;
 
    int i = 0;
    int ret;
    viaPortPrivPtr pPriv = NULL;
    DBG_DD(("Enter Function : %s\n",__FUNCTION__));    

    for (i = XV_PORT_NUM_OVERLAY -1 ; i >= 0; i--) {        
        if (pVia->pPriv[i] != NULL) {
            if(pVia->pPriv[i]->curIGA != 0){
                pPriv = pVia->pPriv[i];
                
                if(pPriv->scrnAttr != viaGfxInfo->screenAttr.uint){
                    pPriv->forceEngReAlloc = TRUE;
                    if(pPriv->videoFlag & VIDEO_ACTIVE) {
                        viaPutImageG(pScrn, pPriv->src_x, pPriv->src_y, 
                                    pPriv->drw_x, pPriv->drw_y, pPriv->src_w,
                                    pPriv->src_h, pPriv->drw_w, pPriv->drw_h,
                                    pPriv->fourCC, pPriv->buf, pPriv->width, pPriv->height,
                                    pPriv->sync, &pPriv->clip, pPriv, pPriv->pDraw);
                    }
                }        
            }
        }
    } 

    return;
}

void 
VIAAdjustFrame(ADJUST_FRAME_ARGS_DECL)
{
    SCRN_INFO_PTR(arg);
    VIAPtr      pVia = VIAPTR(pScrn);
    PVIDDATA pVidData = pVia->pVIAEnt->pVidData;
    viaGfxInfoPtr viaGfxInfo = pVidData->viaGfxInfo;
    viaPortPrivPtr pPriv;
    int i = 0;
    int ret;
    unsigned long curIGA = 0;

    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIAAdjustFrame\n"));

    if(x < 0 || y <0)
        return;

    /* Patch for Duoview panning cases:
       The panning_x[], panning_y[] are indexed by Iga Index. */
    /* Unifiy the process of panning_x[], panning_y[] for all case: single, duoview and samm. */
    if (viaGfxInfo->screenInfo[pScrn->scrnIndex].igaInuse & IGA1) {
        if (viaGfxInfo->igaInfo[0].igaStatus.panning) {
            viaGfxInfo->igaInfo[0].start_x = x;
            viaGfxInfo->igaInfo[0].start_y = y;
            pVidData->panning_x[0] = x;
            pVidData->panning_y[0] = y;
        }
    }

    if (viaGfxInfo->screenInfo[pScrn->scrnIndex].igaInuse & IGA2) {
        if(viaGfxInfo->igaInfo[1].igaStatus.panning){
            viaGfxInfo->igaInfo[1].start_x = x;
            viaGfxInfo->igaInfo[1].start_y = y;
            pVidData->panning_x[1] = x;
            pVidData->panning_y[1] = y;
        }
    }

    for (i = XV_PORT_NUM_OVERLAY -1 ; i >= 0; i--) {    
        
        if (pVia->pPriv[i] != NULL) {
            
            if(pVia->pPriv[i]->curIGA != 0){
                pPriv = pVia->pPriv[i];
                
                if(pPriv->videoFlag & VIDEO_ACTIVE) {                    
                    determineCurrentIGAInUse( pScrn, &curIGA,x,y);                    
                    
                    if(curIGA == pPriv->curIGA){
                        viaPutImageG(pScrn, pPriv->src_x, pPriv->src_y, 
                                    pPriv->drw_x, pPriv->drw_y, pPriv->src_w,
                                    pPriv->src_h, pPriv->drw_w, pPriv->drw_h,
                                    pPriv->fourCC, pPriv->buf, pPriv->width, pPriv->height,
                                    pPriv->sync, &pPriv->clip, pPriv, pPriv->pDraw);
                    }
                }
            }
        }
    } 

    return;
}


Bool 
VIASwitchMode(SWITCH_MODE_ARGS_DECL)
{   
    SCRN_INFO_PTR(arg);
    VIAPtr      pVia = VIAPTR(pScrn);
    ServerInfoPtr pServerInfo = pVia->pServerInfo;
    
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "VIASwitchMode\n"));
    DEBUG(xf86DrvMsg(scrnIndex, X_INFO, "xf86SetSingleMode\n"));

    return xf86SetSingleMode(pScrn, mode, RR_Rotate_0);
}



static void 
VIAPreSaveReg(ScrnInfoPtr pScrn)
{
    static int  IsSaved = FALSE;
    VIAPtr      pVia = VIAPTR(pScrn);
    VIARegPtr   save = &pVia->SavedReg; 

    /* To fix the problem: In SAMM, the system will hang when suspends.
       (On VT3364) In VIAEnableMMIO(), it will enable software reset of 
       SR1A, so the value of SR1A we saved is not the original value set 
       by BIOS, and we found that it will cause system hang when restore
       registers in VIALeaveVT(). In order to keep the original value set
       by BIOS, we should save SR1A before doing VIAEnableMMIO() to avoid
       this problem. So, we move the saving code from VIASave() to here.
    */
    if (!IsSaved) {
        VGAOUT8(0x3d4, CR08);
        save->CRTCRegs[0x08] = VGAIN8(0x3d5);
        VGAOUT8(0x3d4, CR09);
        save->CRTCRegs[0x09] = VGAIN8(0x3d5);
        VGAOUT8(0x3C4, SR1A);
        save->SRRegs[0x1A] = VGAIN8(0x3C5);   

        IsSaved  = TRUE;
    }
}


void 
VIAEnableMMIO(ScrnInfoPtr pScrn)
{
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    VIAPtr pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr pBIOSInfo = pVia->pBIOSInfo;
    unsigned char val;

    /* If we are primary card, we still use std vga port. If we use
     * MMIO, system will hang in vgaHWSave when our card used in
     * PLE and KLE (integrated Trident MVP4)
     */
    if (!pVia->IsMAMMEnable) {
        vgaHWSetStdFuncs(hwp);
    }
    else {
        vgaHWSetMmioFuncs(hwp, pVia->MapBase, 0x8000);
    }

    val = VGAIN8(0x3c3);
    VGAOUT8(0x3c3, val | 0x01);
    val = VGAIN8(VGA_MISC_OUT_R);
    VGAOUT8(VGA_MISC_OUT_W, val | 0x01);

    /* Unlock Extended IO Space */
    VGAOUT8(0x3c4, 0x10);
    VGAOUT8(0x3c5, 0x01);
    /* Unlock CRTC register protect */
    VGAOUT8(0x3d4, 0x47);
    VGAOUT8(0x3d5, VGAIN8(0x3d5) & ~0x01);

    /* Save some important register before we change its value */
    VIAPreSaveReg(pScrn);

    /* Enable MMIO */
    if (pVia->IsSecondary) {
        VGAOUT8(0x3c4, 0x1a);
        val = VGAIN8(0x3c5);
        DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "secondary val = %x\n", val));
        VGAOUT8(0x3c5, val | 0x38);
    } 
    return;
}


void 
VIALoadPalette(ScrnInfoPtr pScrn, int numColors, int *indicies,
                    LOCO *colors, VisualPtr pVisual)
{
    VIAPtr pVia = VIAPTR(pScrn);
    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
    int i,j,index,k;
    unsigned short lut_r[256], lut_g[256], lut_b[256];
    
    for (k = 0; k < xf86_config->num_crtc; k++) {
        xf86CrtcPtr    crtc = xf86_config->crtc[k];
        VIACrtcPrivatePtr viaCrtc = VIACrtcPrivate(crtc);

        /* Initialize to the old lookup table values. */
        for (i = 0; i < 256; i++) {
                lut_r[i] = viaCrtc->colors[i].red<< 8;
                lut_g[i] = viaCrtc->colors[i].green << 8;
                lut_b[i] = viaCrtc->colors[i].blue<< 8;
        }

        switch (pScrn->depth) {
            case 15:
                for (i = 0; i < numColors; i++) {
                    index = indicies[i];
                    for (j = 0; j < 8; j++) {
                        lut_r[index * 8 + j] = colors[index].red << 8;
                        lut_g[index * 8 + j] = colors[index].green << 8;
                        lut_b[index * 8 + j] = colors[index].blue << 8;
                    }
                }
                break;
            case 16:
                for (i = 0; i < numColors; i++) {
                    index = indicies[i];
    
                    if (index <= 31) {
                        for (j = 0; j < 8; j++) {
                            lut_r[index * 8 + j] = colors[index].red << 8;
                            lut_b[index * 8 + j] = colors[index].blue << 8;
                        }
                    }
    
                    for (j = 0; j < 4; j++) {
                        lut_g[index * 4 + j] = colors[index].green << 8;
                    }
                }
                break;
            default:
                for (i = 0; i < numColors; i++) {
                    index = indicies[i];
                    lut_r[index] = colors[index].red << 8;
                    lut_g[index] = colors[index].green << 8;
                    lut_b[index] = colors[index].blue << 8;
                }
                break;
        }
        
        RRCrtcGammaSet(crtc->randr_crtc, lut_r, lut_g, lut_b);
    }
}


void
VIAInitialize3DEngine(ScrnInfoPtr pScrn)
{
#ifndef CS_FLAG_NO_3D_SCAL
    VIAPtr  pVia = VIAPTR(pScrn);
    int i;
    DWORD StageOfTexture;
    WaitIdle();
    /*
     * read CR32 for the suspend or resume information
     * Bit 7:1 work from suspend
     *       0 work normally
     */
    if (H5_UMA_CHIPID) {        
        VIASETREG(0x43C, 0x00010000);
    
        for (i = 0; i <= 0x8A; i++)
            VIASETREG(0x440, (CARD32) i << 24);
        
        /* Initial Texture Stage Setting*/
        for (StageOfTexture = 0; StageOfTexture < 0xf; StageOfTexture++) {
            VIASETREG(0x43C, (0x00020000 | 0x00000000 | (StageOfTexture & 0xf)<<24));
            for (i = 0 ; i <= 0x30 ; i++ )
                VIASETREG(0x440, (CARD32) i << 24);
        }
        
        /* Initial Texture Sampler Setting*/
        for (StageOfTexture = 0; StageOfTexture < 0xf; StageOfTexture++) {
            VIASETREG(0x43C, (0x00020000 | 0x00020000 | (StageOfTexture & 0xf)<<24));
            for (i = 0 ; i <= 0x30 ; i++ )
                VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x43C, (0x00020000 | 0xfe000000));
        for (i = 0 ; i <= 0x13 ; i++ )
            VIASETREG(0x440, (CARD32) i << 24);

        /* Initial Gamma Table Setting*/
        /* Initial Gamma Table Setting*/
        /* 5 + 4 = 9 (12) dwords*/
        /* sRGB texture is not directly support by H3 hardware.*/
        /* We have to set the deGamma table for texture sampling.*/

        /* degamma table*/
        VIASETREG(0x43C, ( 0x00030000 | 0x15000000));
        VIASETREG(0x440, (0x40000000 | (30 << 20) | (15 << 10) | (5)));
        VIASETREG(0x440, ((119 << 20) | (81 << 10) | (52)));
        VIASETREG(0x440, ((283 << 20) | (219 << 10) | (165)));
        VIASETREG(0x440, ((535 << 20) | (441 << 10) | (357)));
        VIASETREG(0x440, ((119 << 20) | (884 << 20) | (757 << 10) | (640)));

        /* gamma table*/
        VIASETREG(0x43C, ( 0x00030000 | 0x17000000));
        VIASETREG(0x440, (0x40000000 | (13 << 20) | (13 << 10) | (13)));
        VIASETREG(0x440, (0x40000000 | (26 << 20) | (26 << 10) | (26)));
        VIASETREG(0x440, (0x40000000 | (39 << 20) | (39 << 10) | (39)));
        VIASETREG(0x440, ((51 << 20) | (51 << 10) | (51)));
        VIASETREG(0x440, ((71 << 20) | (71 << 10) | (71)));
        VIASETREG(0x440, (87 << 20) | (87 << 10) | (87));
        VIASETREG(0x440, (113 << 20) | (113 << 10) | (113));
        VIASETREG(0x440, (135 << 20) | (135 << 10) | (135));
        VIASETREG(0x440, (170 << 20) | (170 << 10) | (170));
        VIASETREG(0x440, (199 << 20) | (199 << 10) | (199));
        VIASETREG(0x440, (246 << 20) | (246 << 10) | (246));
        VIASETREG(0x440, (284 << 20) | (284 << 10) | (284));
        VIASETREG(0x440, (317 << 20) | (317 << 10) | (317));
        VIASETREG(0x440, (347 << 20) | (347 << 10) | (347));
        VIASETREG(0x440, (373 << 20) | (373 << 10) | (373));
        VIASETREG(0x440, (398 << 20) | (398 << 10) | (398));
        VIASETREG(0x440, (442 << 20) | (442 << 10) | (442));
        VIASETREG(0x440, (481 << 20) | (481 << 10) | (481));
        VIASETREG(0x440, (517 << 20) | (517 << 10) | (517));
        VIASETREG(0x440, (550 << 20) | (550 << 10) | (550));
        VIASETREG(0x440, (609 << 20) | (609 << 10) | (609));
        VIASETREG(0x440, (662 << 20) | (662 << 10) | (662));
        VIASETREG(0x440, (709 << 20) | (709 << 10) | (709));
        VIASETREG(0x440, (753 << 20) | (753 << 10) | (753));
        VIASETREG(0x440, (794 << 20) | (794 << 10) | (794));
        VIASETREG(0x440, (832 << 20) | (832 << 10) | (832));
        VIASETREG(0x440, (868 << 20) | (868 << 10) | (868));
        VIASETREG(0x440, (902 << 20) | (902 << 10) | (902));
        VIASETREG(0x440, (934 << 20) | (934 << 10) | (934));
        VIASETREG(0x440, (966 << 20) | (966 << 10) | (966));
        VIASETREG(0x440, (996 << 20) | (996 << 10) | (996));
        
        
        /* For Interrupt Restore only
           All types of write through regsiters should be write header data to
           hardware at least before it can restore. H/W will automatically record
           the header to write through state buffer for resture usage.
           By Jaren:
           HParaType = 8'h03, HParaSubType = 8'h00
                                             8'h11
                                             8'h12
                                             8'h14
                                             8'h15
                                             8'h17
           HParaSubType 8'h12, 8'h15 is initialized.
          [HWLimit]
           1. All these write through registers can't be partial update.
           2. All these write through must be AGP command
           16 entries : 4 128-bit data */

	 /* Initialize INV_ParaSubType_TexPal  	 */
	    VIASETREG(0x43C, ( 0x00030000 | 0x00000000));
        for (i = 0; i < 16; i++)
            VIASETREG(0x440, 0x00000000);

        /* Initialize INV_ParaSubType_4X4Cof */
        /* 32 entries : 8 128-bit data */
        VIASETREG(0x43C, ( 0x00030000 | 0x11000000)); 
        for (i = 0; i < 32; i++)
            VIASETREG(0x440, 0x00000000);

        /* Initialize INV_ParaSubType_StipPal */
        /* 5 entries : 2 128-bit data */
        VIASETREG(0x43C, ( 0x00030000 | 0x14000000)); 
        for (i = 0; i < (5+3); i++)
            VIASETREG(0x440, 0x00000000);

        /* primitive setting & vertex format*/
        VIASETREG(0x43C, ( 0x00040000 |0x14000000));
        for( i = 0; i<52; i++)
            VIASETREG(0x440, ((DWORD) i << 24));
#if 1            
        VIASETREG(0x43C, 0x00fe0000);
        VIASETREG(0x440, 0x4000840f);
        VIASETREG(0x440, 0x47000400);
        VIASETREG(0x440, 0x44000000);
        VIASETREG(0x440, 0x46000000); 
#endif
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
            "H5 3D Engine has been initialized.\n");
        /* setting Misconfig*/
        VIASETREG(0x43C, 0x00fe0000); 
        VIASETREG(0x440, 0x00001004);
        VIASETREG(0x440, 0x0800004b);
        VIASETREG(0x440, 0x0a000049);
        VIASETREG(0x440, 0x0b0000fb);
        VIASETREG(0x440, 0x0c000001);
        VIASETREG(0x440, 0x0d0000cb);
        VIASETREG(0x440, 0x0e000009);
        VIASETREG(0x440, 0x10000000);
        VIASETREG(0x440, 0x110000ff);
        VIASETREG(0x440, 0x12000000);
        VIASETREG(0x440, 0x130000db);
        VIASETREG(0x440, 0x14000000);
        VIASETREG(0x440, 0x15000000);
        VIASETREG(0x440, 0x16000000);
        VIASETREG(0x440, 0x17000000);
        VIASETREG(0x440, 0x18000000);
        VIASETREG(0x440, 0x19000000);
        VIASETREG(0x440, 0x20000000);
    } else if(H6_UMA_CHIPID) {        
        VIASETREG(0x43C, 0x00010000);
        for (i = 0; i <= 0x9A; i++)
            VIASETREG(0x440, (CARD32) i << 24);
        
        /* Initial Texture Stage Setting*/
        for (StageOfTexture = 0; StageOfTexture <= 0xf; StageOfTexture++) {
            VIASETREG(0x43C, (0x00020000 | 0x00000000 | (StageOfTexture & 0xf)<<24));
            for (i = 0 ; i <= 0x30 ; i++ )
                 VIASETREG(0x440, (CARD32) i << 24);
        }
        
        /* Initial Texture Sampler Setting*/
        for (StageOfTexture = 0; StageOfTexture <= 0xf; StageOfTexture++) {
            VIASETREG(0x43C, (0x00020000 | 0x20000000 | (StageOfTexture & 0xf)<<24));
            for ( i = 0 ; i <= 0x36 ; i++ )
                VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x43C, (0x00020000 | 0xfe000000));
        for ( i = 0 ; i <= 0x13 ; i++ )
            VIASETREG(0x440, (CARD32) i << 24);

        /* Initial Gamma Table Setting*/
        /* Initial Gamma Table Setting*/
        /* 5 + 4 = 9 (12) dwords*/
        /* sRGB texture is not directly support by H3 hardware.*/
        /* We have to set the deGamma table for texture sampling.*/

        /* degamma table*/
        VIASETREG(0x43C, ( 0x00030000 | 0x15000000));
        VIASETREG(0x440, (0x40000000 | (30 << 20) | (15 << 10) | (5)));
        VIASETREG(0x440, ((119 << 20) | (81 << 10) | (52)));
        VIASETREG(0x440, ((283 << 20) | (219 << 10) | (165)));
        VIASETREG(0x440, ((535 << 20) | (441 << 10) | (357)));
        VIASETREG(0x440, ((119 << 20) | (884 << 20) | (757 << 10) | (640)));

        /* gamma table*/
        VIASETREG(0x43C, ( 0x00030000 | 0x17000000));
        VIASETREG(0x440, (0x40000000 | (13 << 20) | (13 << 10) | (13)));
        VIASETREG(0x440, (0x40000000 | (26 << 20) | (26 << 10) | (26)));
        VIASETREG(0x440, (0x40000000 | (39 << 20) | (39 << 10) | (39)));
        VIASETREG(0x440, ((51 << 20) | (51 << 10) | (51)));
        VIASETREG(0x440, ((71 << 20) | (71 << 10) | (71)));
        VIASETREG(0x440, (87 << 20) | (87 << 10) | (87));
        VIASETREG(0x440, (113 << 20) | (113 << 10) | (113));
        VIASETREG(0x440, (135 << 20) | (135 << 10) | (135));
        VIASETREG(0x440, (170 << 20) | (170 << 10) | (170));
        VIASETREG(0x440, (199 << 20) | (199 << 10) | (199));
        VIASETREG(0x440, (246 << 20) | (246 << 10) | (246));
        VIASETREG(0x440, (284 << 20) | (284 << 10) | (284));
        VIASETREG(0x440, (317 << 20) | (317 << 10) | (317));
        VIASETREG(0x440, (347 << 20) | (347 << 10) | (347));
        VIASETREG(0x440, (373 << 20) | (373 << 10) | (373));
        VIASETREG(0x440, (398 << 20) | (398 << 10) | (398));
        VIASETREG(0x440, (442 << 20) | (442 << 10) | (442));
        VIASETREG(0x440, (481 << 20) | (481 << 10) | (481));
        VIASETREG(0x440, (517 << 20) | (517 << 10) | (517));
        VIASETREG(0x440, (550 << 20) | (550 << 10) | (550));
        VIASETREG(0x440, (609 << 20) | (609 << 10) | (609));
        VIASETREG(0x440, (662 << 20) | (662 << 10) | (662));
        VIASETREG(0x440, (709 << 20) | (709 << 10) | (709));
        VIASETREG(0x440, (753 << 20) | (753 << 10) | (753));
        VIASETREG(0x440, (794 << 20) | (794 << 10) | (794));
        VIASETREG(0x440, (832 << 20) | (832 << 10) | (832));
        VIASETREG(0x440, (868 << 20) | (868 << 10) | (868));
        VIASETREG(0x440, (902 << 20) | (902 << 10) | (902));
        VIASETREG(0x440, (934 << 20) | (934 << 10) | (934));
        VIASETREG(0x440, (966 << 20) | (966 << 10) | (966));
        VIASETREG(0x440, (996 << 20) | (996 << 10) | (996));
        
        
        /* For Interrupt Restore only
           All types of write through regsiters should be write header data to
           hardware at least before it can restore. H/W will automatically record
           the header to write through state buffer for resture usage.
           By Jaren:
           HParaType = 8'h03, HParaSubType = 8'h00
                                             8'h11
                                             8'h12
                                             8'h14
                                             8'h15
                                             8'h17
           HParaSubType 8'h12, 8'h15 is initialized.
          [HWLimit]
           1. All these write through registers can't be partial update.
           2. All these write through must be AGP command
           16 entries : 4 128-bit data */

	 /* Initialize INV_ParaSubType_TexPal  	 */
	    VIASETREG(0x43C, ( 0x00030000 | 0x00000000));
        for (i = 0; i < 16; i++)
            VIASETREG(0x440, 0x00000000);

        /* Initialize INV_ParaSubType_4X4Cof */
        /* 32 entries : 8 128-bit data */
        VIASETREG(0x43C, ( 0x00030000 | 0x11000000)); 
        for (i = 0; i < 32; i++)
            VIASETREG(0x440, 0x00000000);

        /* Initialize INV_ParaSubType_StipPal */
        /* 5 entries : 2 128-bit data */
        VIASETREG(0x43C, ( 0x00030000 | 0x14000000)); 
        for (i = 0; i < (5+3); i++)
             VIASETREG(0x440, 0x00000000);

        /* primitive setting & vertex format*/
        VIASETREG(0x43C, ( 0x00040000));
        for( i = 0; i<=0x62; i++)
             VIASETREG(0x440, ((DWORD) i << 24));

        /*ParaType 0xFE - Configure and Misc Setting*/
		VIASETREG(0x43C, ( 0x00fe0000));
        for( i = 0; i<=0x47; i++)
             VIASETREG(0x440, ((DWORD) i << 24));

        /*ParaType 0x11 - Frame Buffer Auto-Swapping and 
        Command Regulator Misc*/
		VIASETREG(0x43C, ( 0x00110000));
        for( i = 0; i<=0x20; i++)
             VIASETREG(0x440, ((DWORD) i << 24));

#if 1            
        VIASETREG(0x43C, 0x00fe0000);
        VIASETREG(0x440, 0x4000840f);
        VIASETREG(0x440, 0x47000404);
        VIASETREG(0x440, 0x44000000);
        VIASETREG(0x440, 0x46000005); 
#endif
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
            "H6 3D Engine has been initialized.\n");    
        /* setting Misconfig*/
        VIASETREG(0x43C, 0x00fe0000); 
        VIASETREG(0x440, 0x00001004);
        VIASETREG(0x440, 0x08000249);
        VIASETREG(0x440, 0x0a0002c9);
        VIASETREG(0x440, 0x0b0002fb);
        VIASETREG(0x440, 0x0c000000);
        VIASETREG(0x440, 0x0d0002cb);
        VIASETREG(0x440, 0x0e000009);
        VIASETREG(0x440, 0x10000049);
        VIASETREG(0x440, 0x110002ff);
        VIASETREG(0x440, 0x12000008);
        VIASETREG(0x440, 0x130002db);
    } else {
        if (!VIACheckDrmAlreadyInit(pScrn)) {
            VIASETREG(0x43C, 0x00010000);
            for (i = 0; i <= 0x7D; i++)
                VIASETREG(0x440, (CARD32) i << 24);
        
            VIASETREG(0x43C, 0x00020000);
            for (i = 0; i <= 0x94; i++)
                VIASETREG(0x440, (CARD32) i << 24);
        
            VIASETREG(0x440, 0x82400000);
            VIASETREG(0x43C, 0x01020000);
            for (i = 0; i <= 0x94; i++)
                VIASETREG(0x440, (CARD32) i << 24);
        
            VIASETREG(0x440, 0x82400000);
            VIASETREG(0x43C, 0xfe020000);
            for (i = 0; i <= 0x03; i++)
                VIASETREG(0x440, (CARD32) i << 24);
        
            VIASETREG(0x43C, 0x00030000);
            for (i = 0; i <= 0xff; i++)
                VIASETREG(0x440, 0);

            VIASETREG(0x43C, 0x00100000);
            VIASETREG(0x440, 0x00333004);
            VIASETREG(0x440, 0x10000002);
            VIASETREG(0x440, 0x60000000);
            VIASETREG(0x440, 0x61000000);
            VIASETREG(0x440, 0x62000000);
            VIASETREG(0x440, 0x63000000);
            VIASETREG(0x440, 0x64000000);
            
            VIASETREG(0x43C, 0x00fe0000);
        
            if (pVia->ChipRev >= 3 )
                VIASETREG(0x440,0x40008c0f);
            else
                VIASETREG(0x440,0x4000800f);
        
            VIASETREG(0x440,0x44000000);
            VIASETREG(0x440,0x45080C04);
            VIASETREG(0x440,0x46800408);
            VIASETREG(0x440,0x50000000);
            VIASETREG(0x440,0x51000000);
            VIASETREG(0x440,0x52000000);
            VIASETREG(0x440,0x53000000);
        
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "H2 3D Engine has been initialized.\n");
        }
        
        VIASETREG(0x43C,0x00fe0000);
        VIASETREG(0x440,0x08000001);
        VIASETREG(0x440,0x0A000183);
        VIASETREG(0x440,0x0B00019F);
        VIASETREG(0x440,0x0C00018B);
        VIASETREG(0x440,0x0D00019B);
        VIASETREG(0x440,0x0E000000);
        VIASETREG(0x440,0x0F000000);
        VIASETREG(0x440,0x10000000);
        VIASETREG(0x440,0x11000000);
        VIASETREG(0x440,0x20000000); 
    
    }    
#endif /* CS_FLAG_NO_3D_SCAL */
DEBUG(xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VIAInitialize3DEngine exit\n"));
}


/*
 * This function is only required if we need to do anything related to power managemant
 * and via private event handling.
 */
static Bool 
VIAPMEvent(int scrnIndex, pmEvent event, Bool undo)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);
     
    xf86DrvMsg(X_INFO,pScrn->scrnIndex,"enter pm event\n");
    ErrorF("VIAPMEvent: received APM event %d\n", event);

    switch (event) {
        case XF86_APM_SYS_SUSPEND:
        case XF86_APM_CRITICAL_SUSPEND: /*do we want to delay a critical suspend?*/
        case XF86_APM_USER_SUSPEND:
        case XF86_APM_SYS_STANDBY:
        case XF86_APM_USER_STANDBY:
            if (!undo && !pVia->IsSuspending) {
                pScrn->LeaveVT(VT_FUNC_ARGS);
                pVia->IsSuspending = TRUE;
                sleep(50);
                pScrn->EnterVT(VT_FUNC_ARGS);
            } else if (undo && pVia->IsSuspending) {
                sleep(1);
                pScrn->EnterVT(VT_FUNC_ARGS);
                pVia->IsSuspending = FALSE;
            }
            break;
        case XF86_APM_STANDBY_RESUME:
        case XF86_APM_NORMAL_RESUME:
        case XF86_APM_CRITICAL_RESUME:
            if (pVia->IsSuspending) {
                sleep(1);
                pScrn->EnterVT(VT_FUNC_ARGS);
                pVia->IsSuspending = FALSE;
            }
            break;
        case XF86_APM_CAPABILITY_CHANGED:
            if (pVia->IsSecondary)
                return TRUE;
        default:
            ErrorF("VIAPMEvent: received APM event %d\n", event);
            return TRUE;
    }
    return TRUE;
}


void 
viaInitHwDisplayCaps(VIAPtr pVia)
{
	/*Get graphic chip display caps*/
	switch (pVia->Chipset) {
		case VIA_CX700:		/* VT3324 */
			pVia->GfxDispCaps = INTERNAL_LVDS | INTERNAL_TMDS | INTERNAL_TVEncoder;
			break;
		case VIA_VX800:		/* VT3353 */
		    pVia->GfxDispCaps = INTERNAL_LVDS | INTERNAL_TMDS;
			break;
		case VIA_VX855:		/* VT3409*/
			pVia->GfxDispCaps = INTERNAL_LVDS;
			break;
        case VIA_VX900:		/* VT3410*/
			pVia->GfxDispCaps = INTERNAL_LVDS | INTERNAL_HDMI | INTERNAL_DP;
			break;
		default:
			break;
	}
	return;
}


void 
viaCheckMbStrapping(VIAPtr pVia)
{
    unsigned char sr12Date = 0;
    sr12Date = viaReadVgaIo(REG_SR12);

    pVia->MbDiPortUsedInfo = 0;

    switch (pVia->Chipset) {
        case VIA_CX700:
        case VIA_VX800:
            /* First, remove the DFP port support*/
            /*Note that the port is used, in fact no such port*/
            pVia->MbDiPortUsedInfo |= DISP_DI_DFP; 

            if (MB_DVP0_ENABLE & sr12Date) {
                /*Enable the ports*/
                pVia->MbDiPortUsedInfo &= ~(DISP_DI_DVP0 ); 
                /* IF SR12[6]=1, board supported all devices defualt */
                pVia->MbDvp0Device = DISP_DEV_CRT | DISP_DEV_TV | DISP_DEV_DVI | DISP_DEV_LCD;            
            } else {
                /*Note that the port is used, in fact no such port*/
                pVia->MbDiPortUsedInfo  |= DISP_DI_DVP0; 
                pVia->MbDvp0Device = DISP_DEV_NONE;
            }         
            break;     

        case VIA_VX855:
        case VIA_VX900:    
            /* Remove one 24-bit DFP port support,
               VT3409 doesn't support DVP0 */
            pVia->MbDiPortUsedInfo |= DISP_DI_DFP | DISP_DI_DVP0; 
            break;
            
        case VIA_P4M800PRO:
        case VIA_P4M890:
        case VIA_P4M900:
        case VIA_K8M890:
        default:
            /*BIT4 is "1", DFP is one 24bit DFP port, so no DFPH/DFPL*/
            if (MB_DFP_24BIT & sr12Date) {
                /*Note that the port is used, in fact no such port*/
                pVia->MbDiPortUsedInfo  |= DISP_DI_DFPH | DISP_DI_DFPL; 
            /* two 12 bit port, no DFP port*/
            } else {
                /*Note that the port is used, in fact no such port*/
                pVia->MbDiPortUsedInfo  |= DISP_DI_DFP; 
            }
            
            /* After test, DVP0 support crt,lcd,dvi whatever the value SR12[5] is */
            pVia->MbDiPortUsedInfo &= ~DISP_DI_DVP0;
            pVia->MbDvp0Device = DISP_DEV_CRT | DISP_DEV_DVI | DISP_DEV_LCD;
            
            /* DVP0 support tv only when SR12[5]=1 */
            if (MB_DVP0_TV & sr12Date) {
                pVia->MbDvp0Device |= DISP_DEV_TV;            
            }            
            break;
    }
    
    return;
}

Bool 
viaInitHwInfo(VIAPtr pVia)
{
	/*1. Get graphic chip display caps*/
	viaInitHwDisplayCaps(pVia);

	/*2. Get the MB HW straping info*/
	viaCheckMbStrapping(pVia);
	
	return TRUE;
}


/*
    Function Name:  LoadFixedCrtcRegs
    Description: Load fixed CRTC timing registers
*/
static void 
LoadFixedCrtcRegs(void)
{
    viaUnlockCrtc();
    
    viaWriteVgaIoBits(REG_CR03, 0x80, BIT7);       /* always set to 1 */
    viaWriteVgaIo(REG_CR18, 0xFF);                 /* line compare should set all bits = 1 (extend modes) */
    viaWriteVgaIoBits(REG_CR07, 0x10, BIT4);
    viaWriteVgaIoBits(REG_CR09, 0x40, 0xFF);
    viaWriteVgaIoBits(REG_CR35, 0x10, BIT4);
    viaWriteVgaIoBits(REG_CR33, 0x05, BIT0+BIT1+BIT2);
    viaWriteVgaIo(REG_CR17, 0xE3);                 /* extend mode always set to e3h */
    viaWriteVgaIo(REG_CR08, 0x00);                     /* extend mode always set to 0h */
    viaWriteVgaIo(REG_CR14, 0x00);                     /* extend mode always set to 0h */

    viaLockCrtc(); 
}

/*
    Function Name:  LoadFixedCrtcRegs
    Description: Init TD timing register (power sequence)
*/
static void 
InitTdTimingRegs(VIAPtr pVia)
{
    unsigned int TdTimer[PANEL_TD_TIMING_AMOUNT]; 
    unsigned int i;

    /* Calculate TD Timer, every step is 572.1uSec */
    TdTimer[0] = PanelTdTimer.tdTimer0*10000/5721;
    TdTimer[1] = PanelTdTimer.tdTimer1*10000/5721;
    TdTimer[2] = PanelTdTimer.tdTimer2*10000/5721;
    TdTimer[3] = PanelTdTimer.tdTimer3*10000/5721;

    /* Fill primary power sequence */
    for (i = 0; i < PANEL_TD_TIMING_AMOUNT; i++) {
        viaLoadRegs(TdTimer[i], PANEL_TD_TIMING_REG_AMOUNT, TdTimerRegs[i].tdRegs);
    }

    /*Note: VT3293, VT3324, VT3353 have two hardware power sequences
    other chips only have one hardware power sequence*/
    switch (pVia->Chipset) { 
        case VIA_CX700:
        case VIA_VX800:
            /* set CRD4[0] to "1" to select 2nd LCD power sequence. */ 
            viaWriteVgaIoBits(REG_CRD4, 0x01, 0x01);

        /* Fill secondary power sequence */
        for (i = 0; i < PANEL_TD_TIMING_AMOUNT; i++) {
            viaLoadRegs(TdTimer[i], PANEL_TD_TIMING_REG_AMOUNT, TdTimerRegs[i].tdRegs);
        }
        break;
    default:
        break;
    }
}



void
viaInitDispEngine(VIAPtr pVia)
{
    /*1. Write Common Setting for Video Mode */
    switch(pVia->Chipset) {   
        case VIA_P4M800PRO:
        case VIA_K8M890:
        case VIA_P4M890:
        case VIA_P4M900:
            viaWriteVgaIoMultiBits(InitialRegTable_CN700, NUM_TOTAL_CN700_INIT_REG_TBL);
            break;
        case VIA_CX700:
        case VIA_VX800:
            viaWriteVgaIoMultiBits(InitialRegTable_CX700, NUM_TOTAL_CX700_INIT_REG_TBL);
            break;
        case VIA_VX855:
            viaWriteVgaIoMultiBits(InitialRegTable_VX855, NUM_TOTAL_VX855_INIT_REG_TBL);
            break;
        case VIA_VX900:    
            viaWriteVgaIoMultiBits(InitialRegTable_VX900, NUM_TOTAL_VX900_INIT_REG_TBL);
            break;
    }

    /*2. load fixed CRTC timing registers*/
    LoadFixedCrtcRegs();

    /*3. Init TD timing register (power sequence)*/
    InitTdTimingRegs(pVia);

    /*4. I/O address bit to be 1 */
    /*Enables access to frame buffer at A0000-BFFFFh*/
    viaWriteMiscIo(viaReadMiscIo() | 0x01);

}

/* Set register CRD2 according to internal device setting */
static void
setChannel4InternalDev(VIAPtr pVia)
{
    if (pVia->numOfEmbDvi == 0 && pVia->numOfEmbLcd == 0)
        return;

    /*One Dual LVDS Channel (High Resolution Pannel)*/
    if (pVia->dualChannelEmbLcdExist) {
        viaWriteVgaIoBits(REG_CRD2, BIT5, BIT4+BIT5);
    } else if (pVia->numOfEmbDvi && pVia->numOfEmbLcd) {
    	/*One single channel LVDS and one single channel TMDS*/
        viaWriteVgaIoBits(REG_CRD2, BIT4, BIT4+BIT5);
    } else if (pVia->numOfEmbDvi && !pVia->numOfEmbLcd) {
    	/*Single Channel TMDS*/
        viaWriteVgaIoBits(REG_CRD2, BIT4+BIT5, BIT4+BIT5);
    } else {
		/*LVDS1 Channel + LVDS2 Channel*/ 
        viaWriteVgaIoBits(REG_CRD2, 0x00, BIT4+BIT5);
    }
}

void
viaInitOutputRegSet(ScrnInfoPtr pScrn)
{
	VIAPtr pVia = VIAPTR(pScrn);
	xf86CrtcConfigPtr xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
	int i = 0;

	setChannel4InternalDev(pVia);
	
	for (i = 0; i < xf86_config->num_output; i++) {
		 if ((!xf86NameCmp(xf86_config->output[i]->name, "LCD"))||
		 	 (!xf86NameCmp(xf86_config->output[i]->name, "LCD-2"))) {
		 	ViaLcdPrivateInfoPtr viaLcdInfo = xf86_config->output[i]->driver_private;	
			/*Init subchip*/
			initLVDS(viaLcdInfo);
		 }
	}
}

CARD32  VIAGetRevisionOfVX800(VIAPtr pVia)
{    
    /* Get VT3353 Revision num. */
    VGAOUT8(0x3C4, 0x3B);    
    if (VGAIN8(0x3C5) < 0x10)
    {        
        return REVISION_VX800_A;
    }    
    else if (VGAIN8(0x3C5) < 0x12)
    {        
        return REVISION_VX800_B;
    }
    else
    {        
        return REVISION_VX800_B2;
    }
}

CARD32  VIAGetRevisionOfVX900(VIAPtr pVia)
{    
    /* Get VT3410 Revision num. */
    VGAOUT8(0x3C4, 0x3B);    

    if (VGAIN8(0x3C5) == 0x00)
    {
        return REVISION_VX900_A0;
    }
    else if (VGAIN8(0x3C5) == 0x01)
    {
        return REVISION_VX900_A1;
    }
    else
    {
        return REVISION_VX900_A2;
    }
}

/*
	at SAMM situation, screen0 will allocate 3DScaling memory for screen0 && screen1,
	screen1 only obtain memory address from screen0.
*/
BOOL IsNeedReserve3DScalingMem(VIAPtr pVia)
{
	BOOL ret = FALSE;
	char *sTV3DScaling, *sHDMI3DScaling, *sLCD3DScaling;
	char *sActive, *sMatch;
	BOOL TVOption = FALSE;
	BOOL HDMIOption = FALSE;
	BOOL LCDOption = FALSE;
	ScrnInfoPtr pScrn;
	/*
		1. No matter screen0 or screen1, If Option "DISPLAY_3DScaling_TV"/"DISPLAY_3DScaling_HDMI"/"DISPLAY_3DScaling_LCD"
		is set in Device section, we should check in advance whether device is connected.
	*/
	if((pVia->DISP3DScalingCaps & VIA_DEVICE_TV)&&((pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_TV)||(pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_TV2)))
	{
		pVia->NumOfDevNeed3dScl++;
		ret = TRUE;
	}

	if((pVia->DISP3DScalingCaps & VIA_DEVICE_LCD)&&((pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD)||(pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_LCD2)))
	{	
		pVia->NumOfDevNeed3dScl++;
		ret = TRUE;
	}

	if((pVia->DISP3DScalingCaps & VIA_DEVICE_HDMI)&&((pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI)||(pVia->pBIOSInfo->ActiveDevice & VIA_DEVICE_HDMI2)))
	{
		pVia->NumOfDevNeed3dScl++;
		ret = TRUE;
	}

	/*
		2. In screen0, we should check screen1's Option wether Option "DISPLAY_3DScaling_TV"/"DISPLAY_3DScaling_HDMI"/"DISPLAY_3DScaling_LCD"
		is set in xorg.conf, if so,adds  up the number of devices needed 3D Scaling.
	*/
	if(!pVia->IsSecondary)
	{
		if(pVia->pVIAEnt->HasSecondary)
		{
			pScrn = pVia->pVIAEnt->pSecondaryScrn;
			xf86CollectOptions(pScrn, NULL);
 			sTV3DScaling = xf86findOptionValue(pScrn->options, "DISPLAY_3DScaling_TV");
			sHDMI3DScaling = xf86findOptionValue(pScrn->options, "DISPLAY_3DScaling_HDMI");
			sLCD3DScaling = xf86findOptionValue(pScrn->options, "DISPLAY_3DScaling_LCD");
			sActive = xf86findOptionValue(pScrn->options, "ActiveDevice");
			if(sActive != NULL)
			{
				sMatch = strchr(sActive, ',');
				if(sMatch == NULL)
				{
					sMatch = sActive;
				}
				else
				{
					sMatch ++;
				}
				if((!xf86NameCmp(sMatch, "TV")) || (!xf86NameCmp(sMatch, "TV2")))
				{
					TVOption = TRUE;
				}

				if((!xf86NameCmp(sMatch, "HDMI")) || (!xf86NameCmp(sMatch, "HDMI2")))
				{
					HDMIOption = TRUE;
				}

				if((!xf86NameCmp(sMatch, "LCD")) || (!xf86NameCmp(sMatch, "LCD2")))
				{
					LCDOption = TRUE;
				}
		
			
				if(TVOption && (sTV3DScaling != NULL) && (pVia->pBIOSInfo->ConnectedDevice & VIA_DEVICE_TV))
				{
					pVia->NumOfDevNeed3dScl++;
					ret = TRUE;
				}

				if(HDMIOption && (sHDMI3DScaling != NULL) && (pVia->pBIOSInfo->ConnectedDevice & VIA_DEVICE_HDMI))
				{
					pVia->NumOfDevNeed3dScl++;
					ret = TRUE;
				}

				if(LCDOption && (sLCD3DScaling != NULL)&& (pVia->pBIOSInfo->ConnectedDevice & VIA_DEVICE_LCD))
				{
					pVia->NumOfDevNeed3dScl++;
					ret = TRUE;
				}
			}
		}
	}
	return ret;
}
